Package: MSCMT
Type: Package
Encoding: UTF-8
Version: 1.4.2
Date: 2026-05-19
Title: Multivariate Synthetic Control Method Using Time Series
Authors@R: c(person("Martin","Becker", role = c("aut","cre"),
    email = "martin.becker@mx.uni-saarland.de",
    comment=c(ORCID="0000-0003-2336-9751")),
    person("Stefan","Klößner", role = c("aut"),
    email = "S.Kloessner@mx.uni-saarland.de"),
    person("Karline","Soetaert", role = c("com"),
    email = "karline.soetaert@nioz.nl"),
    person("Jack","Dongarra",role=c("cph"),
    email = "dongarra@icl.utk.edu"),
    person("R.J.","Hanson",role=c("cph")),
    person("K.H.","Haskell",role=c("cph")),
    person("Cleve","Moler",role=c("cph")),
    person("LAPACK authors", role = "cph"))
Depends: R (>= 3.2.0)
Imports: stats, utils, parallel, lpSolve, ggplot2, lpSolveAPI, Rglpk,
        Rdpack
Suggests: Synth, DEoptim, rgenoud, DEoptimR, GenSA, GA, soma, cmaes,
        Rmalschains, NMOF, nloptr, pso, LowRankQP, kernlab, reshape,
        knitr, rmarkdown
Description: Three generalizations of the synthetic control method (which has 
    already an implementation in package 'Synth') are implemented: first, 
    'MSCMT' allows for using multiple outcome variables, second, time series 
    can be supplied as economic predictors, and third, a well-defined 
    cross-validation approach can be used.
    Much effort has been taken to make the implementation as stable as possible 
    (including edge cases) without losing computational efficiency.
    A detailed description of the main algorithms is given in 
    Becker and Klößner (2018) <doi:10.1016/j.ecosta.2017.08.002>.
License: GPL
Copyright: inst/COPYRIGHTS
RoxygenNote: 7.3.1
VignetteBuilder: knitr
RdMacros: Rdpack
BuildVignettes: yes
NeedsCompilation: yes
Packaged: 2026-05-19 13:09:36 UTC; mabe003
Author: Martin Becker [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-2336-9751>),
  Stefan Klößner [aut],
  Karline Soetaert [com],
  Jack Dongarra [cph],
  R.J. Hanson [cph],
  K.H. Haskell [cph],
  Cleve Moler [cph],
  LAPACK authors [cph]
Maintainer: Martin Becker <martin.becker@mx.uni-saarland.de>
Repository: CRAN
Date/Publication: 2026-05-19 13:40:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-19 16:49:35 UTC; unix
Archs: MSCMT.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(ggplot,mscmt)
S3method(plot,mscmt)
S3method(print,mscmt)
export(compare)
export(did)
export(improveSynth)
export(listFromLong)
export(mscmt)
export(ppratio)
export(pvalue)
importFrom(Rdpack,insert_ref)
importFrom(Rdpack,reprompt)
importFrom(Rglpk,Rglpk_solve_LP)
importFrom(ggplot2,aes)
importFrom(ggplot2,aes_string)
importFrom(ggplot2,facet_wrap)
importFrom(ggplot2,geom_hline)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_rect)
importFrom(ggplot2,geom_smooth)
importFrom(ggplot2,geom_vline)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,guide_legend)
importFrom(ggplot2,guides)
importFrom(ggplot2,labs)
importFrom(ggplot2,scale_alpha_manual)
importFrom(ggplot2,scale_colour_manual)
importFrom(ggplot2,scale_linetype_manual)
importFrom(ggplot2,scale_size_manual)
importFrom(ggplot2,scale_x_date)
importFrom(ggplot2,scale_y_continuous)
importFrom(ggplot2,stat_summary)
importFrom(graphics,abline)
importFrom(graphics,plot)
importFrom(lpSolve,lp)
importFrom(lpSolveAPI,get.objective)
importFrom(lpSolveAPI,get.variables)
importFrom(lpSolveAPI,lp.control)
importFrom(lpSolveAPI,make.lp)
importFrom(lpSolveAPI,set.bounds)
importFrom(lpSolveAPI,set.column)
importFrom(lpSolveAPI,set.constr.type)
importFrom(lpSolveAPI,set.objfn)
importFrom(lpSolveAPI,set.rhs)
importFrom(parallel,clusterApplyLB)
importFrom(parallel,clusterEvalQ)
importFrom(parallel,clusterExport)
importFrom(stats,end)
importFrom(stats,frequency)
importFrom(stats,na.omit)
importFrom(stats,sd)
importFrom(stats,start)
importFrom(stats,ts)
importFrom(stats,var)
importFrom(stats,window)
importFrom(utils,capture.output)
importFrom(utils,head)
importFrom(utils,packageVersion)
importFrom(utils,stack)
importFrom(utils,tail)
useDynLib(MSCMT, .registration = TRUE, .fixes = "C_")

## internal function, export later?
## todo (?): @export prepare
## user interface made for compatibility with function dataprep of package Synth
## use listFromLong to convert 'long' format to list (as required by prepare)
#' @importFrom stats var
prepare <- function(dat, predictors, predictors.op = "mean", special.predictors,
                    dependent, treatment.identifier, controls.identifier, 
                    time.predictors.prior, time.optimize.ssr,
                    alpha=NULL, beta=NULL, gamma=NULL, scale.Z=FALSE)
{
  if (!is.list(dat)) stop("dat must be a list, consider using listFromLong")

  # main workhorse
	BuildMatrix <- function(dat,vars,times,donors,treated,aggfuns,alpha=NULL,
                          betagamma=NULL,scale=FALSE) {
		Select <- function(lst,nam,tim,cols,agg=NULL) {
      if (!(nam %in% names(lst))) stop (paste0("variable '",nam,"' is missing ",
                                               "in the supplied data"))
      if (any(!(cols %in% colnames(lst[[nam]])))) 
        stop (paste0("variable '",nam,"' is missing for units ",
                  paste(cols[!(cols %in% colnames(lst[[nam]]))],collapse=", ")))
      tmp <- lst[[nam]][intersect(tim,rownames(lst[[nam]])),cols,drop=FALSE]
            if (is.null(agg)||agg=="id") {                                      # w/o aggregation
                sapply(1:ncol(tmp),function(i) if (any(is.na(tmp[,i]))) 
                  stop(paste0(cols[i]," has some NAs for variable '",nam,"'")))
                rownames(tmp) <- paste(nam,rownames(tmp),sep=".")
            } else {                                                            # with aggregation
                sapply(1:ncol(tmp),function(i) if (all(is.na(tmp[,i]))) 
                  stop(paste0(cols[i]," has only NAs for variable '",nam,"'")))
                sapply(1:ncol(tmp),function(i) if (any(is.na(tmp[,i]))) 
                  warning(paste0(cols[i]," has some NAs for variable '",nam,"'")))
                tmp <- apply(tmp,2,agg,na.rm=TRUE)
                if (!is.matrix(tmp)) tmp <- matrix(tmp,nrow=1)
                rownames(tmp) <- paste(nam,agg,tim[1],tim[length(tim)],sep=".")
            }    
			tmp
		}		
		if (!is.null(alpha)) stopifnot(length(alpha)==length(vars))
		if (!is.null(betagamma)) {
      stopifnot(length(betagamma)==length(vars))
      if (!is.list(betagamma)) betagamma <- as.list(betagamma)
    }  
		X0X1     <- NULL
    len      <- NULL
    names.v  <- NULL
    divisors <- NULL
    if ((length(times)==1)&&(length(vars)>1)) 
      for (i in 2:length(vars)) times[[i]] <- times[[1]]
		for (i in seq_along(vars)) {
      if (is.list(times[[i]])) {
        if (!is.null(betagamma)) 
          if ((length(betagamma[[i]])>1) && 
              (sum(sapply(times[[i]],length))!=length(betagamma[[i]]))) 
            stop("beta/gamma probably has incorrect length")
        tmp <- NULL
        for (j in seq_along(times[[i]])) 
          tmp <- rbind(tmp,Select(dat,vars[i],times[[i]][[j]],c(donors,treated),
                                  if(missing(aggfuns)) NULL else aggfuns[i]))
      } else tmp <- Select(dat,vars[i],times[[i]],c(donors,treated),
                           if(missing(aggfuns)) NULL else aggfuns[i])
      if (scale) {
        divisor  <- sqrt(var(as.vector(tmp))*nrow(tmp))
        divisors <- c(divisors,divisor)
        tmp <- tmp/divisor
      }    
      if ((length(betagamma[[i]])>1)&&(length(betagamma[[i]])!=nrow(tmp)))
        stop("beta/gamma probably has incorrect length")
      if (!is.null(betagamma))   
        tmp <- tmp * if(length(betagamma[[i]])==1) 
                       betagamma[[i]]^seq(nrow(tmp)-1,0,by=-1) else
                       rev(betagamma[[i]])
      if (!is.null(alpha)) tmp <- tmp * sqrt(alpha[[i]])                       
      X0X1    <- rbind(X0X1,tmp)
      len     <- c(len,nrow(tmp))
      names.v <- c(names.v,
                   if (missing(aggfuns)||(aggfuns[i]=="id")) vars[i] else
                     if (nrow(tmp)==1) rownames(tmp) else vars[i])
    }
    colnames(X0X1) <- c(donors,treated)
    list(X0=X0X1[,-ncol(X0X1),drop=FALSE],X1=X0X1[,ncol(X0X1),drop=FALSE],
#         info.v=list(n.v=length(len),names.v=names.v,len.v=len),
         trafo.v=genTrafo(n.v=length(len),names.v=names.v,len.v=len),
         scaled=scale,divisors=divisors)
	}
    
  # check treated unit
  if (length(treatment.identifier) != 1 ) 
    stop("please specify exactly one treatment unit\n")

  # check control units
  if (length(controls.identifier) < 2 ) 
    stop("\n please specify at least two control units\n")
  if (any(duplicated(controls.identifier))) 
    stop("\n duplicate control units in controls.identifier\n")

  # more checks
  if (treatment.identifier %in% controls.identifier) 
    stop("\n treated unit among controls\n")

  allpredictors <- c(predictors,sapply(special.predictors,function(x) x[[1]]))
  allpredtimes  <- vector("list",length(predictors))
  for (i in seq_along(predictors)) 
    allpredtimes[[i]] <- time.predictors.prior
  for (i in seq_along(special.predictors)) 
    allpredtimes <- c(allpredtimes,list(special.predictors[[i]][[2]]))
  allpredaggfun <- rep(predictors.op,length(predictors))
  for (i in seq_along(special.predictors)) 
    allpredaggfun <- c(allpredaggfun,special.predictors[[i]][[3]])

  if (!is.list(time.optimize.ssr)) time.optimize.ssr <- list(time.optimize.ssr)
    
	tmpX <- BuildMatrix(dat,allpredictors,allpredtimes,controls.identifier,
                      treatment.identifier,allpredaggfun,scale=TRUE,
                      betagamma=gamma)
	tmpXu <- BuildMatrix(dat,allpredictors,allpredtimes,controls.identifier,
                       treatment.identifier,allpredaggfun,scale=FALSE,
                       betagamma=gamma)
	tmpZ <- BuildMatrix(dat,dependent,time.optimize.ssr,controls.identifier,
                      treatment.identifier,scale=scale.Z,alpha=alpha,
                      betagamma=beta)
	tmpZu <- BuildMatrix(dat,dependent,time.optimize.ssr,controls.identifier,
                       treatment.identifier,scale=FALSE,alpha=alpha,
                       betagamma=beta)
  storage.mode(tmpX$X0) <- storage.mode(tmpX$X1) <- storage.mode(tmpZ$X0) <- 
    storage.mode(tmpZ$X1) <- storage.mode(tmpXu$X0) <- storage.mode(tmpXu$X1) <- 
    "double"
  res <- list(X0=tmpX$X0,X1=tmpX$X1,Z0=tmpZ$X0,Z1=tmpZ$X1,X0.unscaled=tmpXu$X0,
              X1.unscaled=tmpXu$X1,trafo.v=tmpX$trafo.v,Z.scaled=scale.Z)
  if (scale.Z) {
    storage.mode(tmpZu$X0) <- storage.mode(tmpZu$X1) <- "double"
    res <- c(res,list(Z0.unscaled=tmpZu$X0,Z1.unscaled=tmpZu$X1,
                      Z.len=tmpZu$trafo.v$len.v))
  }
  res
}

#' Post-pre-(r)mspe-ratios for placebo studies
#'
#' \code{ppratio} calculates post-to-pre-(r)mspe-ratios for placebo studies.
#'
#' \code{ppratio} calculates post-to-pre-(r)mspe-ratios for placebo studies based
#' on Synthetic Control Methods.
#'
#' @param x An object of class \code{"mscmt"}, usually obtained as
#' the result of a call to function \code{\link{mscmt}}.
#' @param what A character vector. Name of the variable to be considered. If 
#' missing, the (first) dependent variable will be used.
##' @param range.pre A vector of length 2 defining the range of the pre-treatment
#' period with start and end time given as 
#' \itemize{
#' \item annual dates, if the format of start/end time is "dddd", e.g. "2016",
#' \item quarterly dates, if the format of start/end time is "ddddQd", e.g. 
#' "2016Q1",
#' \item monthly dates, if the format of start/end time is "dddd?dd" with "?"
#' different from "W" (see below), e.g. "2016/03" or "2016-10",
#' \item weekly dates, if the format of start/end time is "ddddWdd", e.g. 
#' "2016W23",
#' \item daily dates, if the format of start/end time is "dddd-dd-dd", e.g. 
#' "2016-08-18",
#' }
#' corresponding to the format of the respective column of the \code{times.dep}
#' argument of \code{\link{mscmt}}.
#' If missing, the corresponding column of \code{times.dep} will be used.
#' @param range.post A vector of length 2 defining the range of the 
#' post-treatment period with start and end time given as 
#' \itemize{
#' \item annual dates, if the format of start/end time is "dddd", e.g. "2016",
#' \item quarterly dates, if the format of start/end time is "ddddQd", e.g. 
#' "2016Q1",
#' \item monthly dates, if the format of start/end time is "dddd?dd" with "?"
#' different from "W" (see below), e.g. "2016/03" or "2016-10",
#' \item weekly dates, if the format of start/end time is "ddddWdd", e.g. 
#' "2016W23",
#' \item daily dates, if the format of start/end time is "dddd-dd-dd", e.g. 
#' "2016-08-18",
#' }
#' corresponding to the format of the respective column of the \code{times.dep}
#' argument of \code{\link{mscmt}}. Will be guessed if missing.
#' @param type A character string. Either \code{rmspe} (default) or \code{mspe}.
#' Selects whether root mean squared errors or mean squared errors are 
#' calculated.
#' @param return.all A logical scalar. If \code{FALSE} (default), only the
#' (named) vector of post-pre-(r)mspe-ratios is returned, if \code{TRUE},
#' a three-column matrix with pre- and post-treatment (r)mspe's as well as the 
#' post-pre-ratios will be returned.
#' @return If \code{return.all} is \code{FALSE}, a (named) vector of 
#' post-pre-(r)mspe-ratios. If \code{return.all} is \code{TRUE}, a matrix with
#' three columns containing the pre-treatment (r)mspe, the post-treatment 
#' (r)mspe, and the post-pre-ratio.
#' @export ppratio
ppratio <- function(x,what,range.pre,range.post,type=c("rmspe","mspe"),
                    return.all=FALSE) {
  type <- match.arg(type)                  
  if (is.null(x$placebo)) stop("results of placebo study are missing")
  if (missing(what)) what  <- x[[1]]$dependent[1]
  gaps <- x$placebo[[what]]$gaps
  if (missing(range.pre))  
    if (what %in% x[[1]]$dependent) range.pre <- x[[1]]$times.dep[,what] else
                                    stop("range.pre is missing")
  gaps.pre   <- AQMwindow(gaps,range.pre)
  gaps.post <- if (missing(range.post)) AQMtail(gaps,range.pre[2]) else
                                        AQMwindow(gaps,range.post)
  rmspe.pre  <- apply(gaps.pre,2,function(x) mean(x^2))
  rmspe.post <- apply(gaps.post,2,function(x) mean(x^2))
  if (type=="rmspe") {
    rmspe.pre  <- sqrt(rmspe.pre)
    rmspe.post <- sqrt(rmspe.post)
  }
  if (return.all) 
    cbind(pre=rmspe.pre,post=rmspe.post,ratio=rmspe.post/rmspe.pre) else
    rmspe.post/rmspe.pre
}

#' @importFrom utils packageVersion
.onAttach <- function(libname, pkgname) {
  version <- utils::packageVersion(pkgname)
  if(isTRUE(interactive())) packageStartupMessage(paste0(
    "# MSCMT version ",version,"\n#\n",
    "# When publishing results obtained using this package,\n",
    "# the original authors are to be cited as:\n",
    "# Martin Becker and Stefan Kl\u00F6\u00DFner (2018). \n",
    "# Fast and Reliable Computation of Generalized Synthetic Controls. \n",
    "# Econometrics and Statistics, 5, 1-19."
  ))
}

.onUnload <- function(libpath) {
  library.dynam.unload(utils::packageName(), libpath)
}

C Karline: removed the write statement; where they also passed an integer value, 
C this no long is the case. Each removed statement is preceded by:
C KARLINE: REMOVED WRITE		 


C*********************************************************************
C LEAST DISTANCE SUBROUTINE
C*********************************************************************

      SUBROUTINE ldp(G,H,NUnknowns,NConstraints,NW,X,XNorm,W,xIndex,           &
     &               Mode,verbose, IsError, Iter)


      INTEGER           :: NUnknowns,NConstraints,NW
      DOUBLE PRECISION  :: G(NConstraints,NUnknowns) 
      DOUBLE PRECISION  :: H(NConstraints)

      DOUBLE PRECISION  :: X(NUnknowns)
      DOUBLE PRECISION  :: XNorm

      LOGICAL           :: IsError, verbose

      DOUBLE PRECISION  :: W(NW)
      INTEGER           :: xIndex(NConstraints)
      INTEGER           :: Mode, Iter
      INTEGER :: xLDPSucces,xLDPNoUnknownsOrEquations,                         &
     &       xLDPToomanyIterations,xLDPIncompatibleConstraints,                &
     &       xLDPUnsolvable


      PARAMETER (xLDPSucces = 1,xLDPNoUnknownsOrEquations     = 2,             &
     &       xLDPToomanyIterations= 3,xLDPIncompatibleConstraints   = 4,       &
     &       xLDPUnsolvable  = -1)


      CALL xLDP(G,NConstraints,NConstraints,NUnknowns,H,X,Xnorm,W,             &
     &          xINdex,Mode,Iter)

      IsError=.TRUE.
      if (mode == xLDPSucces) IsError = .FALSE.

      IF (verbose) THEN 
      SELECT CASE (mode)
      CASE (xLDPNoUnknownsOrEquations)
       CALL xMESSAGE ("No unknowns or equations")
      CASE (xLDPToomanyIterations)
       CALL xMESSAGE ("Too many iterations")
      CASE (xLDPIncompatibleConstraints)
       CALL xMESSAGE ("Incompatible constraints ")
      CASE (xLDPUnsolvable       )
       CALL xMESSAGE ("LDP problem unsolvable")
      END SELECT
      ENDIF

      RETURN
      END SUBROUTINE LDP



C****************************************************************************
C SOLVING LINEAR LEAST SQUARES WITH LINEAR INEQUALITIES and EQUALITIES
C****************************************************************************
  
C----------------------------------------------------------------------------
C  Solves the least squares model with equality and inequality constraints
C
C  Mathematically: minimise SUM(A*x-B)^2 
C  where:          E*x=F
C                  G*x>=h   x> 0
C  x(NumUnknowns), A(NumApproximate,NumUnknowns), B(NumApproximate)
C               E(NumEquations,NumUnknowns)     , F(NumEquations)
C               G(NumInequalities,NumUnknowns)  , H(NumInequalities)
C----------------------------------------------------------------------------

      SUBROUTINE lsei (NUnknowns,NEquations,NConstraints,NApproximate,         &
     &          A,B,E,F,G,H,X,mIP,mdW,mWS,IP,W,WS,lpr,ProgOpt,                 &
     &          verbose,IsError)
 
      IMPLICIT NONE

C The arrays and dimensions for the mass balance inversion

      INTEGER          :: NUnknowns,NEquations,NConstraints,NApproximate
      INTEGER          :: mIP,mdW,mWS,lpr
      LOGICAL          :: IsError, verbose
      DOUBLE PRECISION  ::  A (NApproximate,NUnknowns),                        &
     &                      B (NApproximate)          ,                        &
     &                      E (NEquations,NUnknowns)  ,                        &
     &                      F (NEquations)            ,                        &
     &                      G (NConstraints,NUnknowns),                        &
     &                      H (NConstraints)          ,                        &
     &                      X (NUnknowns)                     
C work arrays
      DOUBLE PRECISION :: W(MDW,NUnknowns+1),WS(mWS)
      INTEGER          :: IP(mip)

      INTEGER          :: I,J,K,MOde,ME, MA, MG,N
      DOUBLE PRECISION :: RNORME, RNORML,ProgOpt(lpr)

C---------------------------------------------------------------------
C a linear Least Squares solver with linear Equality and Inequality constraints: (LSEI)
C
C Find MIN ||A*X=B||
C
C and where
C    E*x=F, G*X>H
C    x> 0
C---------------------------------------------------------------------

      N  = NUnknowns
      ME = NEquations
      MA = NApproximate
      MG = NConstraints

C W is Working array with E,A,G   ,F,B,H 

      RNormE = 0.D0
      RNORML = 0.D0
      MODE   = 0

       DO I = 1, ME
         DO J = 1, N
           W(I,J) = E(I,J)
         ENDDO
         W(I,N+1) = F(I)
       ENDDO        

      K = ME
       DO I = 1, MA
         DO J = 1, N
           W(I+K,J) = A(I,J)
         ENDDO
         W(I+K,N+1) = B(I)
       ENDDO        
      K = ME+MA
      DO I = 1,NConstraints
         DO J = 1, N
           W(I+K,J) = G(I,J)
         ENDDO
         W(I+K,N+1) = H(I)
       ENDDO        
      K = K + NConstraints

c      ProgOpt(1) = 1.D0


    
C CALLING SOLVER!

        CALL xdLSEI(W,                                                   &     
     &              MDW,                                                 &     
     &              ME,                                                  & 
     &              MA,                                                  &
     &              MG,                                                  &
     &              N,                                                   &
     &              ProgOpt,                                             &     
     &              X,                                                   &     
     &              RNORME,                                              &     
     &              RNORML,                                              &     
     &              MODE,                                                &     
     &              WS,                                                  &
     &              IP)


      IF (verbose) THEN
       SELECT CASE (Mode) 
       CASE(1)
           CALL XMESSAGE ("LSEI error: equalities contradictory")

       CASE(2)
           CALL XMESSAGE ("LSEI error: inequalities contradictory")

       CASE(3)
           CALL XMESSAGE                                                  &
     &    ("LSEI error: equalities + inequalities contradictory")

       CASE(4)
           CALL XMESSAGE("LSEI error: wrong input")       
       END SELECT
      ENDIF
      IsError = .FALSE.
      IF (mode > 0) Iserror = .TRUE.
      RETURN

      END SUBROUTINE LSEI


C                c<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<c
C                c<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<c
C                c            ERROR HANDLING          c
C                c>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>c
C                c>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>c


      SUBROUTINE XMESSAGE (String)

      CHARACTER (LEN=*)   :: String

C Check whether it is safe to write
        CALL rwarn(String)

      END SUBROUTINE XMESSAGE



C######################################################################
C 
C Routines for solving quadratic and linear programming problem
C
C SUBROUTINES FROM LINPACK LIBRARY
C
C MAIN ROUTINES ARE xLDP
C xLSEI and  xdbocls
C 
C######################################################################



C<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<C
C<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<C
C    Quadratic programming solver    C
C>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>C
C>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>C


C  The original version of this code was developed by
C  Charles L. Lawson and Richard J. Hanson at Jet Propulsion Laboratory
C  1973 JUN 15, and published in the book
C  "SOLVING LEAST SQUARES PROBLEMS", Prentice-HalL, 1974.
C  Revised FEB 1995 to accompany reprinting of the book by SIAM.
C  made compatible with fortran 95 by karline Soetaert
C  added x-prefix
C  captured writing to screen -> XMESSAGE

C*************************************************************************C
C LEAST DISTANCE SUBROUTINE
C*************************************************************************C


      SUBROUTINE xLDP (G,MDG,M,N,H,X,XNORM,W,xINDEX,MODE,ITER)     
C
C  Algorithm LDP: LEAST DISTANCE PROGRAMMING
C
C  The original version of this code was developed by
C  Charles L. Lawson and Richard J. Hanson at Jet Propulsion Laboratory
C  1974 MAR 1, and published in the book
C  "SOLVING LEAST SQUARES PROBLEMS", Prentice-HalL, 1974.
C  Revised FEB 1995 to accompany reprinting of the book by SIAM.
C Karline: added ITER
C     ------------------------------------------------------------------

      IMPLICIT NONE
      INTEGER  :: xLDPSucces ,xLDPNoUnknownsOrEquations 
      INTEGER  :: xLDPToomanyIterations,xLDPIncompatibleConstraints
      INTEGER  :: xLDPUnsolvable 
      PARAMETER(xLDPSucces= 1,xLDPNoUnknownsOrEquations     = 2,                 &
     &                      xLDPToomanyIterations         = 3,                   &
     &                      xLDPIncompatibleConstraints   = 4,                   &
     &                      xLDPUnsolvable                = -1)
C Number of unknowns
      INTEGER  :: M, MDG,N         
C Succes or failure
      INTEGER  :: MODE, ITER             
      INTEGER  :: xINDEX(*)  
C Constraints G*X>H ; W=workarray; xnorm=residual norm
      DOUBLE PRECISION :: G(MDG,*), H(*), X(*)
      DOUBLE PRECISION :: W(*)                
      DOUBLE PRECISION :: XNORM               

      INTEGER          :: I, IW, IWDUAL, IY, IZ, J, JF, NP1
      DOUBLE PRECISION :: xDIFF, FAC, RNORM
      DOUBLE PRECISION ONE,ZERO
      PARAMETER (ONE = 1.0d0, ZERO = 0.0d0)
C     ------------------------------------------------------------------

      MODE=xLDPSucces

      IF (N.LE.0) THEN
        MODE=xLDPNoUnknownsOrEquations
        RETURN
      ENDIF

      DO J=1,N   
        X(J)=ZERO     
      ENDDO
      XNORM=ZERO

      IF (M.LE.0) THEN
        MODE=xLDPNoUnknownsOrEquations           
        RETURN
      ENDIF
C   
C     THE DECLARED DIMENSION OF W() MUST BE AT LEAST (N+1)*(M+2)+2*M.   
C   
C      FIRST (N+1)*M LOCS OF W()   =  MATRIX E FOR PROBLEM NNLS.
C       NEXT     N+1 LOCS OF W()   =  VECTOR F FOR PROBLEM NNLS.
C       NEXT     N+1 LOCS OF W()   =  VECTOR Z FOR PROBLEM NNLS.
C       NEXT       M LOCS OF W()   =  VECTOR Y FOR PROBLEM NNLS.
C       NEXT       M LOCS OF W()   =  VECTOR WDUAL FOR PROBLEM NNLS.    
C     COPY G**T INTO FIRST N ROWS AND M COLUMNS OF E.   
C     COPY H**T INTO ROW N+1 OF E.  
C   
      IW=0  
      DO J=1,M   
          DO I=1,N   
               IW=IW+1   
               W(IW)=G(J,I)  
          ENDDO
          IW=IW+1   
          W(IW)=H(J) 
      ENDDO   
      JF=IW+1   
C                                STORE N ZEROS FOLLOWED BY A ONE INTO F.
      DO I=1,N   
          IW=IW+1   
          W(IW)=ZERO    
      ENDDO
      W(IW+1)=ONE   
C   
      NP1=N+1   
      IZ=IW+2   
      IY=IZ+NP1 
      IWDUAL=IY+M   
C   
      CALL xNNLS (W,NP1,NP1,M,W(JF),W(IY),RNORM,W(IWDUAL),W(IZ),         &
     &  xINDEX,MODE,ITER)  
C                      USE THE FOLLOWING RETURN IF UNSUCCESSFUL IN NNLS.
      IF (MODE.NE.xLDPSucces) RETURN 

C      IF (RNORM) 130,130,50     KARLINE changed
C  50  
      IF (RNORM .LE. 0) THEN
C karline:remove ????
        MODE=xLDPUnsolvable
        RETURN
      ENDIF

      FAC=ONE   
      IW=IY-1   
      DO I=1,M   
         IW=IW+1   
C                               HERE WE ARE USING THE SOLUTION VECTOR Y.
         FAC=FAC-H(I)*W(IW)
      ENDDO
C   
C      IF (DIFF(ONE+FAC,ONE)) 130,130,70 
C   70 FAC=ONE/FAC   

      IF (xDIFF(ONE+FAC,ONE) .LE. 0) THEN
        MODE=xLDPIncompatibleConstraints
        RETURN
      ENDIF
      FAC=ONE/FAC   

      DO J=1,N   
         IW=IY-1   
         DO I=1,M   
            IW=IW+1   
C                               HERE WE ARE USING THE SOLUTION VECTOR Y.
            X(J)=X(J)+G(I,J)*W(IW)
         ENDDO 
         X(J)=X(J)*FAC 
      ENDDO

      DO J=1,N  
         XNORM=XNORM+X(J)**2   
      ENDDO

      XNORM=sqrt(XNORM) 
      RETURN
      END SUBROUTINE xLDP



C*************************************************************************C

C     SUBROUTINE NNLS  (A,MDA,M,N,B,X,RNORM,W,ZZ,INDEX,MODE,ITER)
C   
C  Algorithm NNLS: NONNEGATIVE LEAST SQUARES
C   
C  The original version of this code was developed by
C  Charles L. Lawson and Richard J. Hanson at Jet Propulsion Laboratory
C  1973 JUN 15, and published in the book
C  "SOLVING LEAST SQUARES PROBLEMS", Prentice-HalL, 1974.
C  Revised FEB 1995 to accompany reprinting of the book by SIAM.
C  Karline:added iter as a request by Judy Wang
C     GIVEN AN M BY N MATRIX, A, AND AN M-VECTOR, B,  COMPUTE AN
C     N-VECTOR, X, THAT SOLVES THE LEAST SQUARES PROBLEM   
C   
C                      A * X = B  SUBJECT TO X .GE. 0   
C     ------------------------------------------------------------------
C                     Subroutine Arguments
C
C     A(),MDA,M,N     MDA IS THE FIRST DIMENSIONING PARAMETER FOR THE   
C                     ARRAY, A().   ON ENTRY A() CONTAINS THE M BY N    
C                     MATRIX, A.           ON EXIT A() CONTAINS 
C                     THE PRODUCT MATRIX, Q*A , WHERE Q IS AN   
C                     M BY M ORTHOGONAL MATRIX GENERATED IMPLICITLY BY  
C                     THIS SUBROUTINE.  
C     B()     ON ENTRY B() CONTAINS THE M-VECTOR, B.   ON EXIT B() CON- 
C             TAINS Q*B.
C     X()     ON ENTRY X() NEED NOT BE INITIALIZED.  ON EXIT X() WILL   
C             CONTAIN THE SOLUTION VECTOR.  
C     RNORM   ON EXIT RNORM CONTAINS THE EUCLIDEAN NORM OF THE  
C             RESIDUAL VECTOR.  
C     W()     AN N-ARRAY OF WORKING SPACE.  ON EXIT W() WILL CONTAIN    
C             THE DUAL SOLUTION VECTOR.   W WILL SATISFY W(I) = 0.  
C             FOR ALL I IN SET P  AND W(I) .LE. 0. FOR ALL I IN SET Z   
C     ZZ()     AN M-ARRAY OF WORKING SPACE.     
C     INDEX()     AN INTEGER WORKING ARRAY OF LENGTH AT LEAST N.
C                 ON EXIT THE CONTENTS OF THIS ARRAY DEFINE THE SETS    
C                 P AND Z AS FOLLOWS..  
C   
C                 INDEX(1)   THRU INDEX(NSETP) = SET P.     
C                 INDEX(IZ1) THRU INDEX(IZ2)   = SET Z.     
C                 IZ1 = NSETP + 1 = NPP1
C                 IZ2 = N   
C     MODE    THIS IS A SUCCESS-FAILURE FLAG WITH THE FOLLOWING 
C             MEANINGS. 
C             1     THE SOLUTION HAS BEEN COMPUTED SUCCESSFULLY.
C             2     THE DIMENSIONS OF THE PROBLEM ARE BAD.  
C                   EITHER M .LE. 0 OR N .LE. 0.
C             3    ITERATION COUNT EXCEEDED.  MORE THAN 3*N ITERATIONS. 
C   
C     ------------------------------------------------------------------
      SUBROUTINE xnnls (A,MDA,M,N,B,X,RNORM,W,ZZ,INDEX,MODE,ITER) 
C     ------------------------------------------------------------------
      integer I, II, IP, ITER, ITMAX, IZ, IZ1, IZ2, IZMAX, J, JJ, JZ, L
      integer M, MDA, MODE,N, NPP1, NSETP, RTNKEY
C     integer INDEX(N)  
C     double precision A(MDA,N), B(M), W(N), X(N), ZZ(M)   
      integer INDEX(*)  
      double precision A(MDA,*), B(*), W(*), X(*), ZZ(*)
C Karline: changed dummy -> dummy(1) to avoid warining line 495         
      double precision ALPHA, ASAVE, CC, xDIFF, DUMMY(1), FACTOR, RNORM
      double precision SM, SS, T, TEMP, TWO, UNORM, UP, WMAX
      double precision ZERO, ZTEST
      parameter(FACTOR = 0.01d0)
      parameter(TWO = 2.0d0, ZERO = 0.0d0)
C     ------------------------------------------------------------------
      MODE=1
      IF (M .le. 0 .or. N .le. 0) then
         MODE=2
         RETURN
      endif
      ITER=0
      ITMAX=3*N 
C   
C                    INITIALIZE THE ARRAYS INDEX() AND X(). 
C   
      DO I=1,N   
         X(I)=ZERO     
         INDEX(I)=I    
      ENDDO
C   
      IZ2=N 
      IZ1=1 
      NSETP=0   
      NPP1=1
C                             &*****  MAIN LOOP BEGINS HERE  ******     
   30 CONTINUE  
C                  QUIT IF ALL COEFFICIENTS ARE ALREADY IN THE SOLUTION.
C                        OR IF M COLS OF A HAVE BEEN TRIANGULARIZED.    
C   
      IF (IZ1 .GT.IZ2.OR.NSETP.GE.M) GO TO 350   
C   
C         COMPUTE COMPONENTS OF THE DUAL (NEGATIVE GRADIENT) VECTOR W().
C   
      DO IZ=IZ1,IZ2  
         J=INDEX(IZ)   
         SM=ZERO   
         DO L=NPP1,M
           SM=SM+A(L,J)*B(L)   
         ENDDO  
         W(J)=SM   
      ENDDO
C                                   FIND LARGEST POSITIVE W(J). 
   60 continue
      WMAX=ZERO 
      DO IZ=IZ1,IZ2  
         J=INDEX(IZ)   
         IF (W(J) .gt. WMAX) then
            WMAX=W(J)     
            IZMAX=IZ  
         endif
      ENDDO
C   
C             IF WMAX .LE. 0. GO TO TERMINATION.
C             THIS INDICATES SATISFACTION OF THE KUHN-TUCKER CONDITIONS.
C   
      IF (WMAX .le. ZERO) go to 350
      IZ=IZMAX  
      J=INDEX(IZ)   
C   
C     THE SIGN OF W(J) IS OK FOR J TO BE MOVED TO SET P.    
C     BEGIN THE TRANSFORMATION AND CHECK NEW DIAGONAL ELEMENT TO AVOID  
C     NEAR LINEAR DEPENDENCE.   
C   
      ASAVE=A(NPP1,J)   
      CALL xH12 (1,NPP1,NPP1+1,M,A(1,J),1,UP,DUMMY,1,1,0)    
      UNORM=ZERO
      IF (NSETP .ne. 0) then
          DO L=1,NSETP   
            UNORM=UNORM+A(L,J)**2     
          ENDDO
      endif
      UNORM=sqrt(UNORM) 
      IF (xDIFF(UNORM+ABS(A(NPP1,J))*FACTOR,UNORM) .gt. ZERO) then
C   
C        COL J IS SUFFICIENTLY INDEPENDENT.  COPY B INTO ZZ, UPDATE ZZ
C        AND SOLVE FOR ZTEST ( = PROPOSED NEW VALUE FOR X(J) ).    
C   
        DO L=1,M  
          ZZ(L)=B(L)    
        ENDDO
        CALL xH12 (2,NPP1,NPP1+1,M,A(1,J),1,UP,ZZ,1,1,1)   
        ZTEST=ZZ(NPP1)/A(NPP1,J)  
C   
C                                     SEE IF ZTEST IS POSITIVE  
C   
         IF (ZTEST .gt. ZERO) go to 140
      endif
C   
C     REJECT J AS A CANDIDATE TO BE MOVED FROM SET Z TO SET P.  
C     RESTORE A(NPP1,J), SET W(J)=0., AND LOOP BACK TO TEST DUAL
C     COEFFS AGAIN.     
C   
      A(NPP1,J)=ASAVE   
      W(J)=ZERO 
      GO TO 60  
C   
C     THE INDEX  J=INDEX(IZ)  HAS BEEN SELECTED TO BE MOVED FROM
C     SET Z TO SET P.    UPDATE B,  UPDATE INDICES,  APPLY HOUSEHOLDER  
C     TRANSFORMATIONS TO COLS IN NEW SET Z,  ZERO SUBDIAGONAL ELTS IN   
C     COL J,  SET W(J)=0.   
C   
  140 continue
      DO L=1,M  
        B(L)=ZZ(L)    
      ENDDO
C   
      INDEX(IZ)=INDEX(IZ1)  
      INDEX(IZ1)=J  
      IZ1=IZ1+1 
      NSETP=NPP1
      NPP1=NPP1+1   
C   
      IF (IZ1 .le. IZ2) then
         DO JZ=IZ1,IZ2 
            JJ=INDEX(JZ)  
            CALL xH12 (2,NSETP,NPP1,M,A(1,J),1,UP,A(1,JJ),1,MDA,1)
         ENDDO
      endif
C   
      IF (NSETP .ne. M) then
         DO L=NPP1,M   
           A(L,J)=ZERO   
         ENDDO
      endif
C   
      W(J)=ZERO 
C                                SOLVE THE TRIANGULAR SYSTEM.   
C                                STORE THE SOLUTION TEMPORARILY IN ZZ().
      RTNKEY = 1
      GO TO 400 
  200 CONTINUE  
C   
C                       &*****  SECONDARY LOOP BEGINS HERE ******   
C   
C                          ITERATION COUNTER.   
C 
  210 continue  
      ITER=ITER+1   
      IF (ITER .gt. ITMAX) then
         MODE=3
C         write (*,'(/a)') ' NNLS quitting on iteration count.'
      CALL XMESSAGE ('error in LDP - NNLS quitting on iteration count.')
         GO TO 350 
      endif
C   
C                    SEE IF ALL NEW CONSTRAINED COEFFS ARE FEASIBLE.    
C                                  IF NOT COMPUTE ALPHA.    
C   
      ALPHA=TWO 
      DO IP=1,NSETP 
         L=INDEX(IP)   
         IF (ZZ(IP) .le. ZERO) then
            T=-X(L)/(ZZ(IP)-X(L))     
            IF (ALPHA .gt. T) then
               ALPHA=T   
               JJ=IP 
            endif
         endif
      ENDDO 
C   
C          IF ALL NEW CONSTRAINED COEFFS ARE FEASIBLE THEN ALPHA WILL   
C          STILL = 2.    IF SO EXIT FROM SECONDARY LOOP TO MAIN LOOP.   
C   
      IF (ALPHA.EQ.TWO) GO TO 330   
C   
C          OTHERWISE USE ALPHA WHICH WILL BE BETWEEN 0. AND 1. TO   
C          INTERPOLATE BETWEEN THE OLD X AND THE NEW ZZ.    
C   
      DO IP=1,NSETP 
         L=INDEX(IP)   
         X(L)=X(L)+ALPHA*(ZZ(IP)-X(L)) 
      ENDDO
C   
C        MODIFY A AND B AND THE INDEX ARRAYS TO MOVE COEFFICIENT I  
C        FROM SET P TO SET Z.   
C   
      I=INDEX(JJ)   
  260 continue
      X(I)=ZERO 
C   
      IF (JJ .ne. NSETP) then
         JJ=JJ+1   
         DO 280 J=JJ,NSETP 
            II=INDEX(J)   
            INDEX(J-1)=II 
            CALL xG1 (A(J-1,II),A(J,II),CC,SS,A(J-1,II))   
            A(J,II)=ZERO  
            DO L=1,N  
               IF (L.NE.II) then
C
C                 Apply procedure G2 (CC,SS,A(J-1,L),A(J,L))  
C
                  TEMP = A(J-1,L)
                  A(J-1,L) = CC*TEMP + SS*A(J,L)
                  A(J,L)   =-SS*TEMP + CC*A(J,L)
               endif
            ENDDO   
C
C                 Apply procedure G2 (CC,SS,B(J-1),B(J))   
C
            TEMP = B(J-1)
            B(J-1) = CC*TEMP + SS*B(J)    
            B(J)   =-SS*TEMP + CC*B(J)    
  280    continue
      endif
C
      NPP1=NSETP
      NSETP=NSETP-1     
      IZ1=IZ1-1 
      INDEX(IZ1)=I  
C   
C        SEE IF THE REMAINING COEFFS IN SET P ARE FEASIBLE.  THEY SHOULD
C        BE BECAUSE OF THE WAY ALPHA WAS DETERMINED.
C        IF ANY ARE INFEASIBLE IT IS DUE TO ROUND-OFF ERROR.  ANY   
C        THAT ARE NONPOSITIVE WILL BE SET TO ZERO   
C        AND MOVED FROM SET P TO SET Z. 
C   
      DO JJ=1,NSETP 
         I=INDEX(JJ)   
         IF (X(I) .le. ZERO) go to 260
      ENDDO 
C   
C         COPY B( ) INTO ZZ( ).  THEN SOLVE AGAIN AND LOOP BACK.
C   
      DO I=1,M  
        ZZ(I)=B(I)    
      ENDDO
      RTNKEY = 2
      GO TO 400 
  320 CONTINUE  
      GO TO 210 
C                      &*****  END OF SECONDARY LOOP  ******
C   
  330 continue
      DO IP=1,NSETP 
        I=INDEX(IP)   
        X(I)=ZZ(IP)   
      ENDDO
C        ALL NEW COEFFS ARE POSITIVE.  LOOP BACK TO BEGINNING.  
      GO TO 30  
C   
C                        &*****  END OF MAIN LOOP  ******   
C   
C                        COME TO HERE FOR TERMINATION.  
C                     COMPUTE THE NORM OF THE FINAL RESIDUAL VECTOR.    
C 
  350 continue  
      SM=ZERO   
      IF (NPP1 .le. M) then
         DO I=NPP1,M   
           SM=SM+B(I)**2 
         ENDDO
      else
         DO J=1,N  
           W(J)=ZERO     
         ENDDO
      endif
      RNORM=sqrt(SM)    
      RETURN
C   
C     THE FOLLOWING BLOCK OF CODE IS USED AS AN INTERNAL SUBROUTINE     
C     TO SOLVE THE TRIANGULAR SYSTEM, PUTTING THE SOLUTION IN ZZ().     
C   
  400 continue
      DO L=1,NSETP  
         IP=NSETP+1-L  
         IF (L .ne. 1) then
            DO II=1,IP
               ZZ(II)=ZZ(II)-A(II,JJ)*ZZ(IP+1)   
            ENDDO
         endif
         JJ=INDEX(IP)  
         ZZ(IP)=ZZ(IP)/A(IP,JJ)    
      ENDDO
C
      IF (RTNKEY .EQ. 1) THEN
        GOTO 200
      ELSE IF (RTNKEY .EQ. 2) THEN
        GOTO 320
      ENDIF    
C      go to (200, 320), RTNKEY
      END SUBROUTINE xNNLS


C*************************************************************************C

      double precision FUNCTION xDIFF(X,Y)
C
C  Function used in tests that depend on machine precision.
C
C  The original version of this code was developed by
C  Charles L. Lawson and Richard J. Hanson at Jet Propulsion Laboratory
C  1973 JUN 7, and published in the book
C  "SOLVING LEAST SQUARES PROBLEMS", Prentice-HalL, 1974.
C  Revised FEB 1995 to accompany reprinting of the book by SIAM.
C
      double precision X, Y
      xDIFF=X-Y  
      RETURN
      END FUNCTION xDIFF  



C*************************************************************************C

      SUBROUTINE xG1 (A,B,CTERM,STERM,SIG)   
C
C     COMPUTE ORTHOGONAL ROTATION MATRIX..  
C
C  The original version of this code was developed by
C  Charles L. Lawson and Richard J. Hanson at Jet Propulsion Laboratory
C  1973 JUN 12, and published in the book
C  "SOLVING LEAST SQUARES PROBLEMS", Prentice-HalL, 1974.
C  Revised FEB 1995 to accompany reprinting of the book by SIAM.
C   
C     COMPUTE.. MATRIX   (C, S) SO THAT (C, S)(A) = (SQRT(A**2+B**2))   
C                        (-S,C)         (-S,C)(B)   (   0          )    
C     COMPUTE SIG = SQRT(A**2+B**2) 
C        SIG IS COMPUTED LAST TO ALLOW FOR THE POSSIBILITY THAT 
C        SIG MAY BE IN THE SAME LOCATION AS A OR B .
C     ------------------------------------------------------------------
      double precision A, B, CTERM, ONE, SIG, STERM, XR, YR, ZERO
      parameter(ONE = 1.0d0, ZERO = 0.0d0)
C     ------------------------------------------------------------------
      if (abs(A) .gt. abs(B)) then
         XR=B/A
         YR=sqrt(ONE+XR**2)
         CTERM=sign(ONE/YR,A)
         STERM=CTERM*XR
         SIG=abs(A)*YR     
         RETURN
      endif

      if (B .ne. ZERO) then
         XR=A/B
         YR=sqrt(ONE+XR**2)
         STERM=sign(ONE/YR,B)
         CTERM=STERM*XR
         SIG=abs(B)*YR     
         RETURN
      endif

      SIG=ZERO  
      CTERM=ZERO  
      STERM=ONE   
      RETURN
      END SUBROUTINE xG1

C*************************************************************************C

C     SUBROUTINE H12 (MODE,LPIVOT,L1,M,U,IUE,UP,C,ICE,ICV,NCV)  
C   
C  CONSTRUCTION AND/OR APPLICATION OF A SINGLE   
C  HOUSEHOLDER TRANSFORMATION..     Q = I + U*(U**T)/B   
C   
C  The original version of this code was developed by
C  Charles L. Lawson and Richard J. Hanson at Jet Propulsion Laboratory
C  1973 JUN 12, and published in the book
C  "SOLVING LEAST SQUARES PROBLEMS", Prentice-HalL, 1974.
C  Revised FEB 1995 to accompany reprinting of the book by SIAM.
C     ------------------------------------------------------------------
C                     Subroutine Arguments
C
C     MODE   = 1 OR 2   Selects Algorithm H1 to construct and apply a
C            Householder transformation, or Algorithm H2 to apply a
C            previously constructed transformation.
C     LPIVOT IS THE INDEX OF THE PIVOT ELEMENT. 
C     L1,M   IF L1 .LE. M   THE TRANSFORMATION WILL BE CONSTRUCTED TO   
C            ZERO ELEMENTS INDEXED FROM L1 THROUGH M.   IF L1 GT. M     
C            THE SUBROUTINE DOES AN IDENTITY TRANSFORMATION.
C     U(),IUE,UP    On entry with MODE = 1, U() contains the pivot
C            vector.  IUE is the storage increment between elements.  
C            On exit when MODE = 1, U() and UP contain quantities
C            defining the vector U of the Householder transformation.
C            on entry with MODE = 2, U() and UP should contain
C            quantities previously computed with MODE = 1.  These will
C            not be modified during the entry with MODE = 2.   
C     C()    ON ENTRY with MODE = 1 or 2, C() CONTAINS A MATRIX WHICH
C            WILL BE REGARDED AS A SET OF VECTORS TO WHICH THE
C            HOUSEHOLDER TRANSFORMATION IS TO BE APPLIED.
C            ON EXIT C() CONTAINS THE SET OF TRANSFORMED VECTORS.
C     ICE    STORAGE INCREMENT BETWEEN ELEMENTS OF VECTORS IN C().  
C     ICV    STORAGE INCREMENT BETWEEN VECTORS IN C().  
C     NCV    NUMBER OF VECTORS IN C() TO BE TRANSFORMED. IF NCV .LE. 0  
C            NO OPERATIONS WILL BE DONE ON C(). 
C     ------------------------------------------------------------------
      SUBROUTINE xH12 (MODE,LPIVOT,L1,M,U,IUE,UP,C,ICE,ICV,NCV)  
C     ------------------------------------------------------------------
      integer I, I2, I3, I4, ICE, ICV, INCR, IUE, J
      integer L1, LPIVOT, M, MODE, NCV
      double precision B, C(*), CL, CLINV, ONE, SM
C     double precision U(IUE,M)
      double precision U(IUE,*)
      double precision UP
      parameter(ONE = 1.0d0)
C     ------------------------------------------------------------------
      IF (0.GE.LPIVOT.OR.LPIVOT.GE.L1.OR.L1.GT.M) RETURN    
      CL=abs(U(1,LPIVOT))   
      IF (MODE.EQ.2) GO TO 60   
C                            &***** CONSTRUCT THE TRANSFORMATION. ******
          DO J=L1,M  
           CL=MAX(abs(U(1,J)),CL)  
          ENDDO
C KARLINE      IF (CL) 130,130,20
      IF (CL .LE. 0) GOTO 130
      CLINV=ONE/CL  
      SM=(U(1,LPIVOT)*CLINV)**2   
      DO J=L1,M  
        SM=SM+(U(1,J)*CLINV)**2 
      ENDDO
      CL=CL*SQRT(SM)   
C KARLINE     IF (U(1,LPIVOT)) 50,50,40     
      IF (U(1,LPIVOT) .LE. 0) GOTO 50
      CL=-CL
   50 UP=U(1,LPIVOT)-CL 
      U(1,LPIVOT)=CL    
      GO TO 70  
C            &***** APPLY THE TRANSFORMATION  I+U*(U**T)/B  TO C. ******
C   
CKARLINE   60 IF (CL) 130,130,70
   60 IF (CL .LE. 0) GOTO 130
   70 IF (NCV.LE.0) RETURN  
      B= UP*U(1,LPIVOT)
C                       B  MUST BE NONPOSITIVE HERE.  IF B = 0., RETURN.
C   
CKARLINE      IF (B) 80,130,130 
      IF (B .GE. 0) GOTO 130 
      B=ONE/B   
      I2=1-ICV+ICE*(LPIVOT-1)   
      INCR=ICE*(L1-LPIVOT)  
          DO 120 J=1,NCV
          I2=I2+ICV     
          I3=I2+INCR    
          I4=I3 
          SM=C(I2)*UP
           DO I=L1,M  
             SM=SM+C(I3)*U(1,I)
             I3=I3+ICE 
           ENDDO
CKARLINE          IF (SM) 100,120,100   
          IF (SM .EQ. 0) GOTO 120
          SM=SM*B   
          C(I2)=C(I2)+SM*UP
            DO I=L1,M 
              C(I4)=C(I4)+SM*U(1,I)
              I4=I4+ICE 
            ENDDO  
  120     CONTINUE  
  130 RETURN
      END SUBROUTINE xH12





C*************************************************************************C

      DOUBLE PRECISION FUNCTION xDNRM2 (N, DX, INCX)
C***BEGIN PROLOGUE  xDNRM2
C***PURPOSE  Compute the Euclidean length (L2 norm) of a vector.
C***LIBRARY   SLATEC (BLAS)
C***CATEGORY  D1A3B
C***TYPE      DOUBLE PRECISION (SNRM2-S, xDNRM2-D, SCNRM2-C)
C***KEYWORDS  BLAS, EUCLIDEAN LENGTH, EUCLIDEAN NORM, L2,
C             LINEAR ALGEBRA, UNITARY, VECTOR
C***AUTHOR  Lawson, C. L., (JPL)
C           Hanson, R. J., (SNLA)
C           Kincaid, D. R., (U. of Texas)
C           Krogh, F. T., (JPL)
C***DESCRIPTION
C
C                B L A S  Subprogram
C    Description of parameters
C
C     --Input--
C        N  number of elements in input vector(s)
C       DX  double precision vector with N elements
C     INCX  storage spacing between elements of DX
C
C     --Output--
C    xDNRM2  double precision result (zero if N .LE. 0)
C
C     Euclidean norm of the N-vector stored in DX with storage
C     increment INCX.
C     If N .LE. 0, return with result = 0.
C     If N .GE. 1, then INCX must be .GE. 1
C
C     Four phase method using two built-in constants that are
C     hopefully applicable to all machines.
C         CUTLO = maximum of  SQRT(U/EPS)  over all known machines.
C         CUTHI = minimum of  SQRT(V)      over all known machines.
C     where
C         EPS = smallest no. such that EPS + 1. .GT. 1.
C         U   = smallest positive no.   (underflow limit)
C         V   = largest  no.            (overflow  limit)
C
C     Brief outline of algorithm.
C
C     Phase 1 scans zero components.
C     move to phase 2 when a component is nonzero and .LE. CUTLO
C     move to phase 3 when a component is .GT. CUTLO
C     move to phase 4 when a component is .GE. CUTHI/M
C     where M = N for X() real and M = 2*N for complex.
C
C     Values for CUTLO and CUTHI.
C     From the environmental parameters listed in the IMSL converter
C     document the limiting values are as follows:
C     CUTLO, S.P.   U/EPS = 2**(-102) for  Honeywell.  Close seconds are
C                   Univac and DEC at 2**(-103)
C                   Thus CUTLO = 2**(-51) = 4.44089E-16
C     CUTHI, S.P.   V = 2**127 for Univac, Honeywell, and DEC.
C                   Thus CUTHI = 2**(63.5) = 1.30438E19
C     CUTLO, D.P.   U/EPS = 2**(-67) for Honeywell and DEC.
C                   Thus CUTLO = 2**(-33.5) = 8.23181D-11
C     CUTHI, D.P.   same as S.P.  CUTHI = 1.30438D19
C     DATA CUTLO, CUTHI /8.232D-11,  1.304D19/
C     DATA CUTLO, CUTHI /4.441E-16,  1.304E19/
C
C***REFERENCES  C. L. Lawson, R. J. Hanson, D. R. Kincaid and F. T.
C                 Krogh, Basic linear algebra subprograms for Fortran
C                 usage, Algorithm No. 539, Transactions on Mathematical
C                 Software 5, 3 (September 1979), pp. 308-323.
C***ROUTINES CALLED  (NONE)
C***REVISION HISTORY  (YYMMDD)
C   791001  DATE WRITTEN
C   890531  Changed all specific intrinsics to generic.  (WRB)
C   890831  Modified array declarations.  (WRB)
C   890831  REVISION DATE from Version 3.2
C   891214  Prologue converted to Version 4.0 format.  (BAB)
C   920501  Reformatted the REFERENCES section.  (WRB)
C***END PROLOGUE  xDNRM2
      INTEGER NEXT, N,NN,INCX,I,J
      DOUBLE PRECISION DX(*), CUTLO, CUTHI, HITEST, SUM, XMAX, ZERO,      &
     &                 ONE
      SAVE CUTLO, CUTHI, ZERO, ONE
      DATA ZERO, ONE /0.0D0, 1.0D0/
C
      DATA CUTLO, CUTHI /8.232D-11,  1.304D19/
C***FIRST EXECUTABLE STATEMENT  xDNRM2
      IF (N .GT. 0) GO TO 10
         XDNRM2  = ZERO
         GO TO 300
C
C   10 ASSIGN 30 TO NEXT               CHANGED INTO:
   10 NEXT = 30      
      SUM = ZERO
      NN = N * INCX
C
C                                                 BEGIN MAIN LOOP
C
      I = 1
C  20    GO TO NEXT,(30, 50, 70, 110)  CHANGED INTO:
   20 SELECT CASE (NEXT)

        CASE (30)
          GOTO 30
        CASE (50)
          GOTO 50
        CASE (70)
          GOTO 70
        CASE (110)
          GOTO 110
                   
      END SELECT
      
      
        
   30 IF (ABS(DX(I)) .GT. CUTLO) GO TO 85
C     ASSIGN 50 TO NEXT               CHANGED INTO:
      NEXT = 50

      XMAX = ZERO
C
C                        PHASE 1.  SUM IS ZERO
C
   50 IF (DX(I) .EQ. ZERO) GO TO 200
      IF (ABS(DX(I)) .GT. CUTLO) GO TO 85
C
C                                PREPARE FOR PHASE 2.
C
C      ASSIGN 70 TO NEXT              CHANGED INTO:
      NEXT = 70
      GO TO 105
C
C                                PREPARE FOR PHASE 4.
C
  100 I = J
C      ASSIGN 110 TO NEXT              CHANGED INTO:
      NEXT = 110
      SUM = (SUM / DX(I)) / DX(I)
  105 XMAX = ABS(DX(I))
      GO TO 115
C
C                   PHASE 2.  SUM IS SMALL.
C                             SCALE TO AVOID DESTRUCTIVE UNDERFLOW.
C
   70 IF (ABS(DX(I)) .GT. CUTLO) GO TO 75
C
C                     COMMON CODE FOR PHASES 2 AND 4.
C                     IN PHASE 4 SUM IS LARGE.  SCALE TO AVOID OVERFLOW.
C
  110 IF (ABS(DX(I)) .LE. XMAX) GO TO 115
         SUM = ONE + SUM * (XMAX / DX(I))**2
         XMAX = ABS(DX(I))
         GO TO 200
C
  115 SUM = SUM + (DX(I)/XMAX)**2
      GO TO 200
C
C                  PREPARE FOR PHASE 3.
C
   75 SUM = (SUM * XMAX) * XMAX
C
C     FOR REAL OR D.P. SET HITEST = CUTHI/N
C     FOR COMPLEX      SET HITEST = CUTHI/(2*N)
C
   85 HITEST = CUTHI / N
C
C                   PHASE 3.  SUM IS MID-RANGE.  NO SCALING.
C
C KARLINE

C      DO 95 J = I,NN,INCX
C      IF (ABS(DX(J)) .GE. HITEST) GO TO 100
C   95    SUM = SUM + DX(J)**2

      DO J = I,NN,INCX
         IF (ABS(DX(J)) .GE. HITEST) GO TO 100
         SUM = SUM + DX(J)**2
      ENDDO 

      XDNRM2 = SQRT(SUM)
      GO TO 300
C
  200 CONTINUE
      I = I + INCX
      IF (I .LE. NN) GO TO 20
C
C              END OF MAIN LOOP.
C
C              COMPUTE SQUARE ROOT AND ADJUST FOR SCALING.
C
      XDNRM2 = XMAX * SQRT(SUM)
  300 CONTINUE
      RETURN
      END

C*************************************************************************C

C<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<C
C<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<C
C     SOLVING SYSTEMS OF EQUATIONS   C
C>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>C
C>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>C



C********************************************************************

      subroutine  xdscal(n,da,dx,incx)

C-------------------------------------------------------------------*
C Subroutine obtained from LINPACK (ftp-site netlib.att.com)        &
C-------------------------------------------------------------------*
C
C     scales a vector by a constant.
C     uses unrolled loops for increment equal to one.
C     jack dongarra, linpack, 3/11/78.
C     modified 3/93 to return if incx .le. 0.
C     modified 12/3/93, array(1) declarations changed to array(*)
C
C***BEGIN PROLOGUE  xDSCAL
C***PURPOSE  Multiply a vector by a constant.
C***LIBRARY   SLATEC (BLAS)
C***CATEGORY  D1A6
C***TYPE      DOUBLE PRECISION (SSCAL-S, xDSCAL-D, CSCAL-C)
C***KEYWORDS  BLAS, LINEAR ALGEBRA, SCALE, VECTOR
C***AUTHOR  Lawson, C. L., (JPL)
C           Hanson, R. J., (SNLA)
C           Kincaid, D. R., (U. of Texas)
C           Krogh, F. T., (JPL)
C***DESCRIPTION
C
C                B L A S  Subprogram
C    Description of Parameters
C
C     --Input--
C        N  number of elements in input vector(s)
C       DA  double precision scale factor
C       DX  double precision vector with N elements
C     INCX  storage spacing between elements of DX
C
C     --Output--
C       DX  double precision result (unchanged if N.LE.0)
C
C     Replace double precision DX by double precision DA*DX.
C     For I = 0 to N-1, replace DX(IX+I*INCX) with  DA * DX(IX+I*INCX),
C     where IX = 1 if INCX .GE. 0, else IX = 1+(1-N)*INCX.
C
C***REFERENCES  C. L. Lawson, R. J. Hanson, D. R. Kincaid and F. T.
C                 Krogh, Basic linear algebra subprograms for Fortran
C                 usage, Algorithm No. 539, Transactions on Mathematical
C                 Software 5, 3 (September 1979), pp. 308-323.
C***ROUTINES CALLED  (NONE)
C***REVISION HISTORY  (YYMMDD)
C   791001  DATE WRITTEN
C   890831  Modified array declarations.  (WRB)
C   890831  REVISION DATE from Version 3.2
C   891214  Prologue converted to Version 4.0 format.  (BAB)
C   900821  Modified to correct problem with a negative increment.
C           (WRB)
C   920501  Reformatted the REFERENCES section.  (WRB)
C***END PROLOGUE  xDSCAL

      double precision da,dx(*)
      integer i,incx,m,mp1,n,nincx
C
      if( n.le.0 .or. incx.le.0 )return
      if(incx.eq.1)go to 20
C
C        code for increment not equal to 1
C
      nincx = n*incx
      do 10 i = 1,nincx,incx
        dx(i) = da*dx(i)
   10 continue
      return
C
C        code for increment equal to 1
C
C
C        clean-up loop
C
   20 m = mod(n,5)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dx(i) = da*dx(i)
   30 continue
      if( n .lt. 5 ) return
   40 mp1 = m + 1
      do 50 i = mp1,n,5
        dx(i) = da*dx(i)
        dx(i + 1) = da*dx(i + 1)
        dx(i + 2) = da*dx(i + 2)
        dx(i + 3) = da*dx(i + 3)
        dx(i + 4) = da*dx(i + 4)
   50 continue
      return
      end                                        
C of xDscal




C********************************************************************
      subroutine xdaxpy(n,da,dx,incx,dy,incy)

C-------------------------------------------------------------------*
C Subroutine obtained from LINPACK (ftp-site netlib.att.com)        &
C-------------------------------------------------------------------*
C***BEGIN PROLOGUE  xDAXPY
C***PURPOSE  Compute a constant times a vector plus a vector.
C***LIBRARY   SLATEC (BLAS)
C***CATEGORY  D1A7
C***TYPE      DOUBLE PRECISION (SAXPY-S, xDAXPY-D, CAXPY-C)
C***KEYWORDS  BLAS, LINEAR ALGEBRA, TRIAD, VECTOR
C***AUTHOR  Lawson, C. L., (JPL)
C           Hanson, R. J., (SNLA)
C           Kincaid, D. R., (U. of Texas)
C           Krogh, F. T., (JPL)
C***DESCRIPTION
C
C                B L A S  Subprogram
C    Description of Parameters
C
C     --Input--
C        N  number of elements in input vector(s)
C       DA  double precision scalar multiplier
C       DX  double precision vector with N elements
C     INCX  storage spacing between elements of DX
C       DY  double precision vector with N elements
C     INCY  storage spacing between elements of DY
C
C     --Output--
C       DY  double precision result (unchanged if N .LE. 0)
C
C     Overwrite double precision DY with double precision DA*DX + DY.
C     For I = 0 to N-1, replace  DY(LY+I*INCY) with DA*DX(LX+I*INCX) +
C       DY(LY+I*INCY),
C     where LX = 1 if INCX .GE. 0, else LX = 1+(1-N)*INCX, and LY is
C     defined in a similar way using INCY.
C
C***REFERENCES  C. L. Lawson, R. J. Hanson, D. R. Kincaid and F. T.
C                 Krogh, Basic linear algebra subprograms for Fortran
C                 usage, Algorithm No. 539, Transactions on Mathematical
C                 Software 5, 3 (September 1979), pp. 308-323.
C***ROUTINES CALLED  (NONE)
C***REVISION HISTORY  (YYMMDD)
C   791001  DATE WRITTEN
C   890831  Modified array declarations.  (WRB)
C   890831  REVISION DATE from Version 3.2
C   891214  Prologue converted to Version 4.0 format.  (BAB)
C   920310  Corrected definition of LX in DESCRIPTION.  (WRB)
C   920501  Reformatted the REFERENCES section.  (WRB)
C***END PROLOGUE  xDAXPY

C
C     constant times a vector plus a vector.
C     uses unrolled loops for increments equal to one.
C     jack dongarra, linpack, 3/11/78.
C     modified 12/3/93, array(1) declarations changed to array(*)
C
      double precision dx(*),dy(*),da
      integer i,incx,incy,ix,iy,m,mp1,n
C
      if(n.le.0)return
      if (da .eq. 0.0d0) return
      if(incx.eq.1.and.incy.eq.1)go to 20
C
C        code for unequal increments or equal increments
C          not equal to 1
C
      ix = 1
      iy = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      if(incy.lt.0)iy = (-n+1)*incy + 1
      do 10 i = 1,n
        dy(iy) = dy(iy) + da*dx(ix)
        ix = ix + incx
        iy = iy + incy
   10 continue
      return
C
C        code for both increments equal to 1
C
C
C        clean-up loop
C
   20 m = mod(n,4)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dy(i) = dy(i) + da*dx(i)
   30 continue
      if( n .lt. 4 ) return
   40 mp1 = m + 1
      do 50 i = mp1,n,4
        dy(i) = dy(i) + da*dx(i)
        dy(i + 1) = dy(i + 1) + da*dx(i + 1)
        dy(i + 2) = dy(i + 2) + da*dx(i + 2)
        dy(i + 3) = dy(i + 3) + da*dx(i + 3)
   50 continue
      return
      end                                        
C of xdaxpy

C********************************************************************

      double precision function xddot(n,dx,incx,dy,incy)

C-------------------------------------------------------------------*
C Subroutine obtained from LINPACK (ftp-site netlib.att.com)        &
C-------------------------------------------------------------------*
C
C     forms the dot product of two vectors.
C     uses unrolled loops for increments equal to one.
C     jack dongarra, linpack, 3/11/78.
C     modified 12/3/93, array(1) declarations changed to array(*)
C
      double precision dx(*),dy(*),dtemp
      integer i,incx,incy,ix,iy,m,mp1,n
C
      xddot = 0.0d0
      dtemp = 0.0d0
      if(n.le.0)return
      if(incx.eq.1.and.incy.eq.1)go to 20
C
C        code for unequal increments or equal increments
C          not equal to 1
C
      ix = 1
      iy = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      if(incy.lt.0)iy = (-n+1)*incy + 1
      do 10 i = 1,n
        dtemp = dtemp + dx(ix)*dy(iy)
        ix = ix + incx
        iy = iy + incy
   10 continue
      xddot = dtemp
      return
C
C        code for both increments equal to 1
C
C
C        clean-up loop
C
   20 m = mod(n,5)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dtemp = dtemp + dx(i)*dy(i)
   30 continue
      if( n .lt. 5 ) go to 60
   40 mp1 = m + 1
      do 50 i = mp1,n,5
        dtemp = dtemp + dx(i)*dy(i) + dx(i + 1)*dy(i + 1) +                &
     &   dx(i + 2)*dy(i + 2) + dx(i + 3)*dy(i + 3) + dx(i + 4)*dy(i + 4)
   50 continue
   60 xddot = dtemp
      return
      end                                        
C of xDDOT



      subroutine xdswap (n,dx,incx,dy,incy)
C
C     interchanges two vectors.
C     uses unrolled loops for increments equal one.
C     jack dongarra, linpack, 3/11/78.
C     modified 12/3/93, array(1) declarations changed to array(*)
C
      double precision dx(*),dy(*),dtemp
      integer i,incx,incy,ix,iy,m,mp1,n
C
      if(n.le.0)return
      if(incx.eq.1.and.incy.eq.1)go to 20
C
C       code for unequal increments or equal increments not equal
C         to 1
C
      ix = 1
      iy = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      if(incy.lt.0)iy = (-n+1)*incy + 1
      do 10 i = 1,n
        dtemp = dx(ix)
        dx(ix) = dy(iy)
        dy(iy) = dtemp
        ix = ix + incx
        iy = iy + incy
   10 continue
      return
C
C       code for both increments equal to 1
C
C
C       clean-up loop
C
   20 m = mod(n,3)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dtemp = dx(i)
        dx(i) = dy(i)
        dy(i) = dtemp
   30 continue
      if( n .lt. 3 ) return
   40 mp1 = m + 1
      do 50 i = mp1,n,3
        dtemp = dx(i)
        dx(i) = dy(i)
        dy(i) = dtemp
        dtemp = dx(i + 1)
        dx(i + 1) = dy(i + 1)
        dy(i + 1) = dtemp
        dtemp = dx(i + 2)
        dx(i + 2) = dy(i + 2)
        dy(i + 2) = dtemp
   50 continue
      return
      end


C***********************************************************************
      SUBROUTINE XDCOPY(N,DX,INCX,DY,INCY)
C
C     COPIES A VECTOR, X, TO A VECTOR, Y.
C     USES UNROLLED LOOPS FOR INCREMENTS EQUAL TO ONE.
C     JACK DONGARRA, LINPACK, 3/11/78.
C
      DOUBLE PRECISION DX(*),DY(*)
      INTEGER I,INCX,INCY,IX,IY,M,MP1,N
C
      IF(N.LE.0)RETURN
      IF(INCX.EQ.1.AND.INCY.EQ.1)GO TO 20
C
C        CODE FOR UNEQUAL INCREMENTS OR EQUAL INCREMENTS
C          NOT EQUAL TO 1
C
      IX = 1
      IY = 1
      IF(INCX.LT.0)IX = (-N+1)*INCX + 1
      IF(INCY.LT.0)IY = (-N+1)*INCY + 1
      DO 10 I = 1,N
        DY(IY) = DX(IX)
        IX = IX + INCX
        IY = IY + INCY
   10 CONTINUE
      RETURN
C
C        CODE FOR BOTH INCREMENTS EQUAL TO 1
C
C
C        CLEAN-UP LOOP
C
   20 M = MOD(N,7)
      IF( M .EQ. 0 ) GO TO 40
      DO 30 I = 1,M
        DY(I) = DX(I)
   30 CONTINUE
      IF( N .LT. 7 ) RETURN
   40 MP1 = M + 1
      DO 50 I = MP1,N,7
        DY(I) = DX(I)
        DY(I + 1) = DX(I + 1)
        DY(I + 2) = DX(I + 2)
        DY(I + 3) = DX(I + 3)
        DY(I + 4) = DX(I + 4)
        DY(I + 5) = DX(I + 5)
        DY(I + 6) = DX(I + 6)
   50 CONTINUE
      RETURN
      END


C********************************************************************
      integer function xidamax(n,dx,incx)

C-------------------------------------------------------------------*
C Subroutine obtained from LINPACK (ftp-site netlib.att.com)        &
C-------------------------------------------------------------------*
C
C     finds the index of element having max. absolute value.
C     jack dongarra, linpack, 3/11/78.
C     modified 3/93 to return if incx .le. 0.
C     modified 12/3/93, array(1) declarations changed to array(*)
C
      double precision dx(*),dmax
      integer i,incx,ix,n
C
      xidamax = 0
      if( n.lt.1 .or. incx.le.0 ) return
      xidamax = 1
      if(n.eq.1)return
      if(incx.eq.1)go to 20
C
C        code for increment not equal to 1
C
      ix = 1
      dmax = dabs(dx(1))
      ix = ix + incx
      do 10 i = 2,n
         if(dabs(dx(ix)).le.dmax) go to 5
         xidamax = i
         dmax = dabs(dx(ix))
    5    ix = ix + incx
   10 continue
      return
C
C        code for increment equal to 1
C
   20 dmax = dabs(dx(1))
      do 30 i = 2,n
         if(dabs(dx(i)).le.dmax) go to 30
         xidamax = i
         dmax = dabs(dx(i))
   30 continue
      return
      end                                        
C of xidamax


C***********************************************************************

      DOUBLE PRECISION FUNCTION D1MACH (IDUM)
      INTEGER  :: IDUM
C-----------------------------------------------------------------------
C This routine computes:
C the unit roundoff of the machine when IDUM = 4
C the largest value                when IDUM = 2
C Unit roundoffis defined as the smallest positive machine number
C u such that  1.0 + u .ne. 1.0
C Largest value is simply imposed...
C Subroutines/functions called by D1MACH.. None
C-----------------------------------------------------------------------
      DOUBLE PRECISION U, COMP
      DOUBLE PRECISION :: Prec(4)  
      LOGICAL          :: First(4) 
      SAVE Prec, FIRST
      DATA FIRST /.TRUE.,.TRUE.,.TRUE.,.TRUE./
      DATA Prec /1.D-8,1.D-8,1.D-8,1.D-8/


      IF (Idum > 4 .OR. Idum < 0) THEN
C         Write (*,*) "Error in function D1MACH"
C         Write (*,*) "NOT DEFINED FOR IDUM = ", Idum
       CALL XMESSAGE("Error in function D1MACH-NOT DEFINED FOR IDUM  ") 
      ENDIF

      IF (First(Idum)) THEN 

       First(Idum) = .FALSE.

       SELECT CASE (IDUM)

        CASE (2)
C Very large number
         D1MACH = 1.D300

        CASE (4)
C Unit roundoff
         U = 1.0D0
 10      U = U*0.5D0
         COMP = 1.0D0 + U
         IF (COMP .NE. 1.0D0) GO TO 10
         D1MACH = U*2.0D0
        CASE Default
C         Write (*,*) "Error in function D1MACH"
C         Write (*,*) "NOT DEFINED FOR IDUM = ", Idum
         CALL XMESSAGE("Error in function D1MACH-NOT DEFINED FOR IDUM ")
       END SELECT

       PREC (Idum) = D1MACH

      ELSE

       D1mach = Prec(IDUM)

      ENDIF

      RETURN

      END FUNCTION d1Mach





C<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<C
C<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<C
C    LEAST SQUARES with constraints  C
C                DLSEI               C                 
C>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>C
C>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>C


C   LINPACK routine





c                !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<!
c                !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<!
c                !    LEAST SQUARES with constraints  !
c                !                DLSEI               !                 
c                !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>!
c                !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>!


c   LINPACK routine

      SUBROUTINE xDLSEI (W, MDW, ME, MA, MG, N, PRGOPT, X, RNORME,       &
     &   RNORML, MODE, WS, IP)
c***BEGIN PROLOGUE  xDLSEI
c***PURPOSE  Solve a linearly constrained least squares problem with
c            equality and inequality constraints, and optionally compute
c            a covariance matrix.
c***LIBRARY   SLATEC
c***CATEGORY  K1A2A, D9
c***TYPE      DOUBLE PRECISION (LSEI-S, xDLSEI-D)
c***KEYWORDS  CONSTRAINED LEAST SQUARES, CURVE FITTING, DATA FITTING,
c             EQUALITY CONSTRAINTS, INEQUALITY CONSTRAINTS,
c             QUADRATIC PROGRAMMING
c***AUTHOR  Hanson, R. J., (SNLA)
c           Haskell, K. H., (SNLA)
c***DESCRIPTION
c
c     Abstract
c
c     This subprogram solves a linearly constrained least squares
c     problem with both equality and inequality constraints, and, if the
c     user requests, obtains a covariance matrix of the solution
c     parameters.
c
c     Suppose there are given matrices E, A and G of respective
c     dimensions ME by N, MA by N and MG by N, and vectors F, B and H of
c     respective lengths ME, MA and MG.  This subroutine solves the
c     linearly constrained least squares problem
c
c                   EX = F, (E ME by N) (equations to be exactly
c                                       satisfied)
c                   AX = B, (A MA by N) (equations to be
c                                       approximately satisfied,
c                                       least squares sense)
c                   GX .GE. H,(G MG by N) (inequality constraints)
c
c     The inequalities GX .GE. H mean that every component of the
c     product GX must be .GE. the corresponding component of H.
c
c     In case the equality constraints cannot be satisfied, a
c     generalized inverse solution residual vector length is obtained
c     for F-EX.  This is the minimal length possible for F-EX.
c
c     Any values ME .GE. 0, MA .GE. 0, or MG .GE. 0 are permitted.  The
c     rank of the matrix E is estimated during the computation.  We call
c     this value KRANKE.  It is an output parameter in IP(1) defined
c     below.  Using a generalized inverse solution of EX=F, a reduced
c     least squares problem with inequality constraints is obtained.
c     The tolerances used in these tests for determining the rank
c     of E and the rank of the reduced least squares problem are
c     given in Sandia Tech. Rept. SAND-78-1290.  They can be
c     modified by the user if new values are provided in
c     the option list of the array PRGOPT(*).
c
c     The user must dimension all arrays appearing in the call list..
c     W(MDW,N+1),PRGOPT(*),X(N),WS(2*(ME+N)+K+(MG+2)*(N+7)),IP(MG+2*N+2)
c     where K=MAX(MA+MG,N).  This allows for a solution of a range of
c     problems in the given working space.  The dimension of WS(*)
c     given is a necessary overestimate.  Once a particular problem
c     has been run, the output parameter IP(3) gives the actual
c     dimension required for that problem.
c
c     The parameters for xDLSEI( ) are
c
c     Input.. All TYPE REAL variables are DOUBLE PRECISION
c
c     W(*,*),MDW,   The array W(*,*) is doubly subscripted with
c     ME,MA,MG,N    first dimensioning parameter equal to MDW.
c                   For this discussion let us call M = ME+MA+MG.  Then
c                   MDW must satisfy MDW .GE. M.  The condition
c                   MDW .LT. M is an error.
c
c                   The array W(*,*) contains the matrices and vectors
c
c                                  (E  F)
c                                  (A  B)
c                                  (G  H)
c
c                   in rows and columns 1,...,M and 1,...,N+1
c                   respectively.
c
c                   The integers ME, MA, and MG are the
c                   respective matrix row dimensions
c                   of E, A and G.  Each matrix has N columns.
c
c     PRGOPT(*)    This real-valued array is the option vector.
c                  If the user is satisfied with the nominal
c                  subprogram features set
c
c                  PRGOPT(1)=1 (or PRGOPT(1)=1.0)
c
c                  Otherwise PRGOPT(*) is a linked list consisting of
c                  groups of data of the following form
c
c                  LINK
c                  KEY
c                  DATA SET
c
c                  The parameters LINK and KEY are each one word.
c                  The DATA SET can be comprised of several words.
c                  The number of items depends on the value of KEY.
c                  The value of LINK points to the first
c                  entry of the next group of data within
c                  PRGOPT(*).  The exception is when there are
c                  no more options to change.  In that
c                  case, LINK=1 and the values KEY and DATA SET
c                  are not referenced.  The general layout of
c                  PRGOPT(*) is as follows.
c
c               ...PRGOPT(1) = LINK1 (link to first entry of next group)
c               .  PRGOPT(2) = KEY1 (key to the option change)
c               .  PRGOPT(3) = data value (data value for this change)
c               .       .
c               .       .
c               .       .
c               ...PRGOPT(LINK1)   = LINK2 (link to the first entry of
c               .                       next group)
c               .  PRGOPT(LINK1+1) = KEY2 (key to the option change)
c               .  PRGOPT(LINK1+2) = data value
c               ...     .
c               .       .
c               .       .
c               ...PRGOPT(LINK) = 1 (no more options to change)
c
c                  Values of LINK that are nonpositive are errors.
c                  A value of LINK .GT. NLINK=100000 is also an error.
c                  This helps prevent using invalid but positive
c                  values of LINK that will probably extend
c                  beyond the program limits of PRGOPT(*).
c                  Unrecognized values of KEY are ignored.  The
c                  order of the options is arbitrary and any number
c                  of options can be changed with the following
c                  restriction.  To prevent cycling in the
c                  processing of the option array, a count of the
c                  number of options changed is maintained.
c                  Whenever this count exceeds NOPT=1000, an error
c                  message is printed and the subprogram returns.
c
c                  Options..
c
c                  KEY=1
c                         Compute in W(*,*) the N by N
c                  covariance matrix of the solution variables
c                  as an output parameter.  Nominally the
c                  covariance matrix will not be computed.
c                  (This requires no user input.)
c                  The data set for this option is a single value.
c                  It must be nonzero when the covariance matrix
c                  is desired.  If it is zero, the covariance
c                  matrix is not computed.  When the covariance matrix
c                  is computed, the first dimensioning parameter
c                  of the array W(*,*) must satisfy MDW .GE. MAX(M,N).
c
c                  KEY=10
c                         Suppress scaling of the inverse of the
c                  normal matrix by the scale factor RNORM**2/
c                  MAX(1, no. of degrees of freedom).  This option
c                  only applies when the option for computing the
c                  covariance matrix (KEY=1) is used.  With KEY=1 and
c                  KEY=10 used as options the unscaled inverse of the
c                  normal matrix is returned in W(*,*).
c                  The data set for this option is a single value.
c                  When it is nonzero no scaling is done.  When it is
c                  zero scaling is done.  The nominal case is to do
c                  scaling so if option (KEY=1) is used alone, the
c                  matrix will be scaled on output.
c
c                  KEY=2
c                         Scale the nonzero columns of the
c                         entire data matrix.
c                  (E)
c                  (A)
c                  (G)
c
c                  to have length one.  The data set for this
c                  option is a single value.  It must be
c                  nonzero if unit length column scaling
c                  is desired.
c
c                  KEY=3
c                         Scale columns of the entire data matrix
c                  (E)
c                  (A)
c                  (G)
c
c                  with a user-provided diagonal matrix.
c                  The data set for this option consists
c                  of the N diagonal scaling factors, one for
c                  each matrix column.
c
c                  KEY=4
c                         Change the rank determination tolerance for
c                  the equality constraint equations from
c                  the nominal value of SQRT(DRELPR).  This quantity can
c                  be no smaller than DRELPR, the arithmetic-
c                  storage precision.  The quantity DRELPR is the
c                  largest positive number such that T=1.+DRELPR
c                  satisfies T .EQ. 1.  The quantity used
c                  here is internally restricted to be at
c                  least DRELPR.  The data set for this option
c                  is the new tolerance.
c
c                  KEY=5
c                         Change the rank determination tolerance for
c                  the reduced least squares equations from
c                  the nominal value of SQRT(DRELPR).  This quantity can
c                  be no smaller than DRELPR, the arithmetic-
c                  storage precision.  The quantity used
c                  here is internally restricted to be at
c                  least DRELPR.  The data set for this option
c                  is the new tolerance.
c
c                  For example, suppose we want to change
c                  the tolerance for the reduced least squares
c                  problem, compute the covariance matrix of
c                  the solution parameters, and provide
c                  column scaling for the data matrix.  For
c                  these options the dimension of PRGOPT(*)
c                  must be at least N+9.  The Fortran statements
c                  defining these options would be as follows:
c
c                  PRGOPT(1)=4 (link to entry 4 in PRGOPT(*))
c                  PRGOPT(2)=1 (covariance matrix key)
c                  PRGOPT(3)=1 (covariance matrix wanted)
c
c                  PRGOPT(4)=7 (link to entry 7 in PRGOPT(*))
c                  PRGOPT(5)=5 (least squares equas.  tolerance key)
c                  PRGOPT(6)=... (new value of the tolerance)
c
c                  PRGOPT(7)=N+9 (link to entry N+9 in PRGOPT(*))
c                  PRGOPT(8)=3 (user-provided column scaling key)
c
c                  CALL xDCOPY (N, D, 1, PRGOPT(9), 1)  (Copy the N
c                    scaling factors from the user array D(*)
c                    to PRGOPT(9)-PRGOPT(N+8))
c
c                  PRGOPT(N+9)=1 (no more options to change)
c
c                  The contents of PRGOPT(*) are not modified
c                  by the subprogram.
c                  The options for WNNLS( ) can also be included
c                  in this array.  The values of KEY recognized
c                  by WNNLS( ) are 6, 7 and 8.  Their functions
c                  are documented in the usage instructions for
c                  subroutine WNNLS( ).  Normally these options
c                  do not need to be modified when using xDLSEI( ).
c
c     IP(1),       The amounts of working storage actually
c     IP(2)        allocated for the working arrays WS(*) and
c                  IP(*), respectively.  These quantities are
c                  compared with the actual amounts of storage
c                  needed by xDLSEI( ).  Insufficient storage
c                  allocated for either WS(*) or IP(*) is an
c                  error.  This feature was included in xDLSEI( )
c                  because miscalculating the storage formulas
c                  for WS(*) and IP(*) might very well lead to
c                  subtle and hard-to-find execution errors.
c
c                  The length of WS(*) must be at least
c
c                  LW = 2*(ME+N)+K+(MG+2)*(N+7)
c
c                  where K = max(MA+MG,N)
c                  This test will not be made if IP(1).LE.0.
c
c                  The length of IP(*) must be at least
c
c                  LIP = MG+2*N+2
c                  This test will not be made if IP(2).LE.0.
c
c     Output.. All TYPE REAL variables are DOUBLE PRECISION
c
c     X(*),RNORME,  The array X(*) contains the solution parameters
c     RNORML        if the integer output flag MODE = 0 or 1.
c                   The definition of MODE is given directly below.
c                   When MODE = 0 or 1, RNORME and RNORML
c                   respectively contain the residual vector
c                   Euclidean lengths of F - EX and B - AX.  When
c                   MODE=1 the equality constraint equations EX=F
c                   are contradictory, so RNORME .NE. 0.  The residual
c                   vector F-EX has minimal Euclidean length.  For
c                   MODE .GE. 2, none of these parameters is defined.
c
c     MODE          Integer flag that indicates the subprogram
c                   status after completion.  If MODE .GE. 2, no
c                   solution has been computed.
c
c                   MODE =
c
c                   0  Both equality and inequality constraints
c                      are compatible and have been satisfied.
c
c                   1  Equality constraints are contradictory.
c                      A generalized inverse solution of EX=F was used
c                      to minimize the residual vector length F-EX.
c                      In this sense, the solution is still meaningful.
c
c                   2  Inequality constraints are contradictory.
c
c                   3  Both equality and inequality constraints
c                      are contradictory.
c
c                   The following interpretation of
c                   MODE=1,2 or 3 must be made.  The
c                   sets consisting of all solutions
c                   of the equality constraints EX=F
c                   and all vectors satisfying GX .GE. H
c                   have no points in common.  (In
c                   particular this does not say that
c                   each individual set has no points
c                   at all, although this could be the
c                   case.)
c
c                   4  Usage error occurred.  The value
c                      of MDW is .LT. ME+MA+MG, MDW is
c                      .LT. N and a covariance matrix is
c                      requested, or the option vector
c                      PRGOPT(*) is not properly defined,
c                      or the lengths of the working arrays
c                      WS(*) and IP(*), when specified in
c                      IP(1) and IP(2) respectively, are not
c                      long enough.
c
c     W(*,*)        The array W(*,*) contains the N by N symmetric
c                   covariance matrix of the solution parameters,
c                   provided this was requested on input with
c                   the option vector PRGOPT(*) and the output
c                   flag is returned with MODE = 0 or 1.
c
c     IP(*)         The integer working array has three entries
c                   that provide rank and working array length
c                   information after completion.
c
c                      IP(1) = rank of equality constraint
c                              matrix.  Define this quantity
c                              as KRANKE.
c
c                      IP(2) = rank of reduced least squares
c                              problem.
c
c                      IP(3) = the amount of storage in the
c                              working array WS(*) that was
c                              actually used by the subprogram.
c                              The formula given above for the length
c                              of WS(*) is a necessary overestimate.
c                              If exactly the same problem matrices
c                              are used in subsequent executions,
c                              the declared dimension of WS(*) can
c                              be reduced to this output value.
c     User Designated
c     Working Arrays..
c
c     WS(*),IP(*)              These are respectively type real
c                              and type integer working arrays.
c                              Their required minimal lengths are
c                              given above.
c
c***REFERENCES  K. H. Haskell and R. J. Hanson, An algorithm for
c                 linear least squares problems with equality and
c                 nonnegativity constraints, Report SAND77-0552, Sandia
c                 Laboratories, June 1978.
c               K. H. Haskell and R. J. Hanson, Selected algorithms for
c                 the linearly constrained least squares problem - a
c                 users guide, Report SAND78-1290, Sandia Laboratories,
c                 August 1979.
c               K. H. Haskell and R. J. Hanson, An algorithm for
c                 linear least squares problems with equality and
c                 nonnegativity constraints, Mathematical Programming
c                 21 (1981), pp. 98-118.
c               R. J. Hanson and K. H. Haskell, Two algorithms for the
c                 linearly constrained least squares problem, ACM
c                 Transactions on Mathematical Software, September 1982.
c***ROUTINES CALLED  D1MACH, xDASUM, xDAXPY, xDCOPY, xDDOT, xDH12, DLSI,
c                    xDNRM2, xDSCAL, xDSWAP, xXERMSG
c***REVISION HISTORY  (YYMMDD)
c   790701  DATE WRITTEN
c   890531  Changed all specific intrinsics to generic.  (WRB)
c   890618  Completely restructured and extensively revised (WRB & RWC)
c   890831  REVISION DATE from Version 3.2
c   891214  Prologue converted to Version 4.0 format.  (BAB)
c   900315  CALLs to XERROR changed to CALLs to xXERMSG.  (THJ)
c   900510  Convert XERRWV calls to xXERMSG calls.  (RWC)
c   900604  DP version created from SP version.  (RWC)
c   920501  Reformatted the REFERENCES section.  (WRB)
c***END PROLOGUE  xDLSEI
      INTEGER IP(3), MA, MDW, ME, MG, MODE, N
      DOUBLE PRECISION PRGOPT(*), RNORME, RNORML, W(MDW,*), WS(*), X(*)
c
      EXTERNAL D1MACH, xDASUM, xDAXPY, xDCOPY,xDDOT,xDH12,DLSI,xDNRM2,               &
     &   xDSCAL, xDSWAP, xXERMSG
      DOUBLE PRECISION D1MACH, xDASUM, xDDOT, xDNRM2
c KARLINE: 
      DOUBLE PRECISION DRELPR, ENORM, FNORM, GAM, RB, RN, RNMAX, SIZE,             &
     &   SN, SNMAX, T, TAU, UJ, UP, VJ, XNORM, XNRME
      INTEGER I, IMAX, J, JP1, K, KEY, KRANKE, LAST, LCHK, LINK, M,                &
     &   MAPKE1, MDEQC, MEND, MEP1, N1, N2, NEXT, NLINK, NOPT, NP1,                &
     &   NTIMES
      LOGICAL COV, FIRST
c      CHARACTER(LEN=8) XERN1, XERN2, XERN3, XERN4
      SAVE FIRST, DRELPR
c
      DATA FIRST /.TRUE./
c***FIRST EXECUTABLE STATEMENT  xDLSEI
c
c     Set the nominal tolerance used in the code for the equality
c     constraint equations.
c
      IF (FIRST) DRELPR = D1MACH(4)
      FIRST = .FALSE.
      TAU = SQRT(DRELPR)
c
c     Check that enough storage was allocated in WS(*) and IP(*).
c
      MODE = 4
      IF (MIN(N,ME,MA,MG) .LT. 0) THEN
c         WRITE (XERN1, '(I8)') N
c         WRITE (XERN2, '(I8)') ME
c         WRITE (XERN3, '(I8)') MA
c         WRITE (XERN4, '(I8)') MG
C KARLINE: REMOVED WRITE		 
         CALL rwarn ('LSEI: THE VARIABLES N, ME,MA, MG MUST BE>0')
c         CALL xXERMSG ('SLATEC', 'LSEI', 'ALL OF THE VARIABLES N, ME,'//            &
c     &      ' MA, MG MUST BE .GE. 0 ENTERED ROUTINE WITH' //                        &
c     &      ' N  = ' // XERN1 //                                                    &
c     &      ' ME = ' // XERN2 //                                                    &
c     &      ' MA = ' // XERN3 //                                                    &
c     &      ' MG = ' // XERN4, 2, 1)
         RETURN
      ENDIF
c
      IF (IP(1).GT.0) THEN
         LCHK = 2*(ME+N) + MAX(MA+MG,N) + (MG+2)*(N+7)
         IF (IP(1).LT.LCHK) THEN
C KARLINE: REMOVED WRITE		 
         CALL rwarn ('LSEI: insufficient storage')
c            WRITE (XERN1, '(I8)') LCHK
c            CALL xXERMSG ('SLATEC', 'xDLSEI', 'INSUFFICIENT STORAGE ' //             &
c     &         'ALLOCATED FOR WS(*), NEED LW = ' // XERN1, 2, 1)
            RETURN
         ENDIF
      ENDIF
c
      IF (IP(2).GT.0) THEN
         LCHK = MG + 2*N + 2
         IF (IP(2).LT.LCHK) THEN
C KARLINE: REMOVED WRITE		 
         CALL rwarn ('LSEI: insufficient storage')
c            WRITE (XERN1, '(I8)') LCHK
c            CALL xXERMSG ('SLATEC', 'xDLSEI', 'INSUFFICIENT STORAGE ' //             &
c     &         'ALLOCATED FOR IP(*), NEED LIP = ' // XERN1, 2, 1)
            RETURN
         ENDIF
      ENDIF
c
c     Compute number of possible right multiplying Householder
c     transformations.
c
      M = ME + MA + MG
      IF (N.LE.0 .OR. M.LE.0) THEN
         MODE = 0
         RNORME = 0
         RNORML = 0
         RETURN
      ENDIF
c
      IF (MDW.LT.M) THEN
        CALL xXERMSG ('SLATEC', 'xDLSEI', 'MDW.LT.ME+MA+MG IS AN ERROR',            &
     &      2, 1)
         RETURN
      ENDIF
c
      NP1 = N + 1
      KRANKE = MIN(ME,N)
      N1 = 2*KRANKE + 1
      N2 = N1 + N
c
c     Set nominal values.
c
c     The nominal column scaling used in the code is
c     the identity scaling.
c
      CALL XDCOPYSC (N, 1.D0, WS(N1), 1)
c
c     No covariance matrix is nominally computed.
c
      COV = .FALSE.
c
c     Process option vector.
c     Define bound for number of options to change.
c
      NOPT = 1000
      NTIMES = 0
c
c     Define bound for positive values of LINK.
c
      NLINK = 100000
      LAST = 1
      LINK = INT(PRGOPT(1),4)
      IF (LINK.EQ.0 .OR. LINK.GT.NLINK) THEN
         CALL xXERMSG('SLATEC','xDLSEI','THE OPTION VECTOR IS UNDEFINED'            &
     &   ,2,1)
         RETURN
      ENDIF
c
  100 IF (LINK.GT.1) THEN
         NTIMES = NTIMES + 1
         IF (NTIMES.GT.NOPT) THEN
            CALL xXERMSG ('SLATEC','xDLSEI',                                         &
     &         'THE LINKS IN THE OPTION VECTOR ARE CYCLING.', 2, 1)
            RETURN
         ENDIF
c
         KEY = INT(PRGOPT(LAST+1),4)
         IF (KEY.EQ.1) THEN
            COV = PRGOPT(LAST+2) .NE. 0.D0
         ELSEIF (KEY.EQ.2 .AND. PRGOPT(LAST+2).NE.0.D0) THEN
            DO 110 J = 1,N
               T = xDNRM2(M,W(1,J),1)
               IF (T.NE.0.D0) T = 1.D0/T
               WS(J+N1-1) = T
  110       CONTINUE
         ELSEIF (KEY.EQ.3) THEN
            CALL xDCOPY (N, PRGOPT(LAST+2), 1, WS(N1), 1)
         ELSEIF (KEY.EQ.4) THEN
            TAU = MAX(DRELPR,PRGOPT(LAST+2))
         ENDIF
c
         NEXT = INT(PRGOPT(LINK),4)
         IF (NEXT.LE.0 .OR. NEXT.GT.NLINK) THEN
         CALL xXERMSG ('SLATEC', 'xDLSEI',                                           &
     &      'THE OPTION VECTOR IS UNDEFINED', 2, 1)
            RETURN
         ENDIF
c
         LAST = LINK
         LINK = NEXT
         GO TO 100
      ENDIF
c
      DO 120 J = 1,N
         CALL xDSCAL (M, WS(N1+J-1), W(1,J), 1)
  120 CONTINUE
c
      IF (COV .AND. MDW.LT.N) THEN
         CALL xXERMSG ('SLATEC', 'xDLSEI',                                           &
     &      'MDW .LT. N WHEN COV MATRIX NEEDED, IS AN ERROR', 2, 1)
         RETURN
      ENDIF
c
c     Problem definition and option vector OK.
c
      MODE = 0
c
c     Compute norm of equality constraint matrix and right side.
c
      ENORM = 0.D0
      DO 130 J = 1,N
         ENORM = MAX(ENORM,xDASUM(ME,W(1,J),1))
  130 CONTINUE
c
      FNORM = xDASUM(ME,W(1,NP1),1)
      SNMAX = 0.D0
      RNMAX = 0.D0
      DO 150 I = 1,KRANKE
c
c        Compute maximum ratio of vector lengths. Partition is at
c        column I.
c
         DO 140 K = I,ME
            SN = xDDOT(N-I+1,W(K,I),MDW,W(K,I),MDW)
            RN = xDDOT(I-1,W(K,1),MDW,W(K,1),MDW)
            IF (RN.EQ.0.D0 .AND. SN.GT.SNMAX) THEN
               SNMAX = SN
               IMAX = K
            ELSEIF (K.EQ.I .OR. SN*RNMAX.GT.RN*SNMAX) THEN
               SNMAX = SN
               RNMAX = RN
               IMAX = K
            ENDIF
  140    CONTINUE
c
c        Interchange rows if necessary.
c
         IF (I.NE.IMAX) CALL xDSWAP (NP1, W(I,1), MDW, W(IMAX,1), MDW)
         IF (SNMAX.GT.RNMAX*TAU**2) THEN
c
c        Eliminate elements I+1,...,N in row I.
c
            CALL xDH12 (1, I, I+1, N, W(I,1), MDW, WS(I), W(I+1,1), MDW,             &    
     &                1, M-I)
         ELSE
            KRANKE = I - 1
            GO TO 160
         ENDIF
  150 CONTINUE
c
c     Save diagonal terms of lower trapezoidal matrix.
c
  160 CALL xDCOPY (KRANKE, W, MDW+1, WS(KRANKE+1), 1)
c
c     Use Householder transformation from left to achieve
c     KRANKE by KRANKE upper triangular form.
c
      IF (KRANKE.LT.ME) THEN
         DO 170 K = KRANKE,1,-1
c
c           Apply transformation to matrix cols. 1,...,K-1.
c
         CALL xDH12 (1, K, KRANKE+1, ME, W(1,K), 1, UP, W, 1, MDW,K-1)
c
c           Apply to rt side vector.
c
        CALL xDH12 (2, K, KRANKE+1, ME, W(1,K), 1, UP, W(1,NP1),1,1, 1)
  170    CONTINUE
      ENDIF
c
c     Solve for variables 1,...,KRANKE in new coordinates.
c
      CALL xDCOPY (KRANKE, W(1, NP1), 1, X, 1)
      DO 180 I = 1,KRANKE
         X(I) = (X(I)-xDDOT(I-1,W(I,1),MDW,X,1))/W(I,I)
  180 CONTINUE
c
c     Compute residuals for reduced problem.
c
      MEP1 = ME + 1
      RNORML = 0.D0
      DO 190 I = MEP1,M
         W(I,NP1) = W(I,NP1) - xDDOT(KRANKE,W(I,1),MDW,X,1)
         SN = xDDOT(KRANKE,W(I,1),MDW,W(I,1),MDW)
         RN = xDDOT(N-KRANKE,W(I,KRANKE+1),MDW,W(I,KRANKE+1),MDW)
         IF (RN.LE.SN*TAU**2 .AND. KRANKE.LT.N)                                     &
     &      CALL XDCOPYSC (N-KRANKE, 0.D0, W(I,KRANKE+1), MDW)
  190 CONTINUE
c
c     Compute equality constraint equations residual length.
c
      RNORME = xDNRM2(ME-KRANKE,W(KRANKE+1,NP1),1)
c
c     Move reduced problem data upward if KRANKE.LT.ME.
c
      IF (KRANKE.LT.ME) THEN
         DO 200 J = 1,NP1
            CALL xDCOPY (M-ME, W(ME+1,J), 1, W(KRANKE+1,J), 1)
  200    CONTINUE
      ENDIF
c
c     Compute solution of reduced problem.
c
      CALL DLSI(W(KRANKE+1, KRANKE+1), MDW, MA, MG, N-KRANKE, PRGOPT,               &
     &         X(KRANKE+1), RNORML, MODE, WS(N2), IP(2))
c
c     Test for consistency of equality constraints.
c
      IF (ME.GT.0) THEN
         MDEQC = 0
         XNRME = xDASUM(KRANKE,W(1,NP1),1)
         IF (RNORME.GT.TAU*(ENORM*XNRME+FNORM)) MDEQC = 1
         MODE = MODE + MDEQC
c
c        Check if solution to equality constraints satisfies inequality
c        constraints when there are no degrees of freedom left.
c
         IF (KRANKE.EQ.N .AND. MG.GT.0) THEN
            XNORM = xDASUM(N,X,1)
            MAPKE1 = MA + KRANKE + 1
            MEND = MA + KRANKE + MG
            DO 210 I = MAPKE1,MEND
               SIZE = xDASUM(N,W(I,1),MDW)*XNORM + ABS(W(I,NP1))
               IF (W(I,NP1).GT.TAU*SIZE) THEN
                  MODE = MODE + 2
                  GO TO 290
               ENDIF
  210       CONTINUE
         ENDIF
      ENDIF
c
c     Replace diagonal terms of lower trapezoidal matrix.
c
      IF (KRANKE.GT.0) THEN
         CALL xDCOPY (KRANKE, WS(KRANKE+1), 1, W, MDW+1)
c
c        Reapply transformation to put solution in original coordinates.
c
         DO 220 I = KRANKE,1,-1
            CALL xDH12 (2, I, I+1, N, W(I,1), MDW, WS(I), X, 1, 1, 1)
  220    CONTINUE
c
c        Compute covariance matrix of equality constrained problem.
c
         IF (COV) THEN
            DO 270 J = MIN(KRANKE,N-1),1,-1
               RB = WS(J)*W(J,J)
               IF (RB.NE.0.D0) RB = 1.D0/RB
               JP1 = J + 1
               DO 230 I = JP1,N
                  W(I,J) = RB*xDDOT(N-J,W(I,JP1),MDW,W(J,JP1),MDW)
  230          CONTINUE
c
               GAM = 0.5D0*RB*xDDOT(N-J,W(JP1,J),1,W(J,JP1),MDW)
               CALL xDAXPY (N-J, GAM, W(J,JP1), MDW, W(JP1,J), 1)
               DO 250 I = JP1,N
                  DO 240 K = I,N
                     W(I,K) = W(I,K) + W(J,I)*W(K,J) + W(I,J)*W(J,K)
                     W(K,I) = W(I,K)
  240             CONTINUE
  250          CONTINUE
               UJ = WS(J)
               VJ = GAM*UJ
               W(J,J) = UJ*VJ + UJ*VJ
               DO 260 I = JP1,N
                  W(J,I) = UJ*W(I,J) + VJ*W(J,I)
  260          CONTINUE
               CALL xDCOPY (N-J, W(J, JP1), MDW, W(JP1,J), 1)
  270       CONTINUE
         ENDIF
      ENDIF
c
c     Apply the scaling to the covariance matrix.
c
      IF (COV) THEN
         DO 280 I = 1,N
            CALL xDSCAL (N, WS(I+N1-1), W(I,1), MDW)
            CALL xDSCAL (N, WS(I+N1-1), W(1,I), 1)
  280    CONTINUE
      ENDIF
c
c     Rescale solution vector.
c
  290 IF (MODE.LE.1) THEN
         DO 300 J = 1,N
            X(J) = X(J)*WS(N1+J-1)
  300    CONTINUE
      ENDIF
c
      IP(1) = KRANKE
      IP(3) = IP(3) + 2*KRANKE + N
      RETURN
      END





      SUBROUTINE DLSI (W, MDW, MA, MG, N, PRGOPT, X, RNORM, MODE, WS,               &
     &   IP)
c***BEGIN PROLOGUE  DLSI
c***SUBSIDIARY
c***PURPOSE  Subsidiary to xDLSEI
c***LIBRARY   SLATEC
c***TYPE      DOUBLE PRECISION (LSI-S, DLSI-D)
c***AUTHOR  Hanson, R. J., (SNLA)
c***DESCRIPTION
c
c     This is a companion subprogram to xDLSEI.  The documentation for
c     xDLSEI has complete usage instructions.
c
c     Solve..
c              AX = B,  A  MA by N  (least squares equations)
c     subject to..
c
c              GX.GE.H, G  MG by N  (inequality constraints)
c
c     Input..
c
c      W(*,*) contains  (A B) in rows 1,...,MA+MG, cols 1,...,N+1.
c                       (G H)
c
c     MDW,MA,MG,N
c              contain (resp) var. dimension of W(*,*),
c              and matrix dimensions.
c
c     PRGOPT(*),
c              Program option vector.
c
c     OUTPUT..
c
c      X(*),RNORM
c
c              Solution vector(unless MODE=2), length of AX-B.
c
c      MODE
c              =0   Inequality constraints are compatible.
c              =2   Inequality constraints contradictory.
c
c      WS(*),
c              Working storage of dimension K+N+(MG+2)*(N+7),
c              where K=MAX(MA+MG,N).
c      IP(MG+2*N+1)
c              Integer working storage
c
c***ROUTINES CALLED  D1MACH, xDASUM, xDAXPY, xDCOPY, xDDOT, xDH12, xDHFTI,
c                    DLPDP, xDSCAL, xDSWAP
c***REVISION HISTORY  (YYMMDD)
c   790701  DATE WRITTEN
c   890531  Changed all specific intrinsics to generic.  (WRB)
c   890618  Completely restructured and extensively revised (WRB & RWC)
c   891214  Prologue converted to Version 4.0 format.  (BAB)
c   900328  Added TYPE section.  (WRB)
c   900604  DP version created from SP version.  (RWC)
c   920422  Changed CALL to xDHFTI to include variable MA.  (WRB)
c***END PROLOGUE  DLSI  - karline: added RNORMV(1) to avoid warning when calling xdhfti     
      INTEGER IP(*), MA, MDW, MG, MODE, N
      DOUBLE PRECISION PRGOPT(*), RNORM, W(MDW,*),WS(*),X(*),RNORMV(1)
c
      EXTERNAL D1MACH, xDASUM,xDAXPY,xDCOPY,xDDOT,xDH12,xDHFTI,DLPDP,               &
     &   xDSCAL, xDSWAP
      DOUBLE PRECISION D1MACH, xDASUM, xDDOT
c
      DOUBLE PRECISION ANORM, DRELPR, FAC, GAM, RB, TAU, TOL, XNORM
      INTEGER I, J, K, KEY, KRANK, KRM1, KRP1, L, LAST, LINK, M, MAP1,              &
     &   MDLPDP, MINMAN, N1, N2, N3, NEXT, NP1, MDB 
      LOGICAL COV, FIRST, SCLCOV
c
      SAVE DRELPR, FIRST
      DATA FIRST /.TRUE./
c
c***FIRST EXECUTABLE STATEMENT  DLSI
c
c     Set the nominal tolerance used in the code.
c
      IF (FIRST) DRELPR = D1MACH(4)
      FIRST = .FALSE.
      TOL = SQRT(DRELPR)
c
      MODE = 0
      RNORM = 0.D0
      M = MA + MG
      NP1 = N + 1
      KRANK = 0
      IF (N.LE.0 .OR. M.LE.0) GO TO 370
c
c     To process option vector.
c
      COV = .FALSE.
      SCLCOV = .TRUE.
      LAST = 1
      LINK = INT(PRGOPT(1),4)
c
  100 IF (LINK.GT.1) THEN
         KEY = INT(PRGOPT(LAST+1),4)
         IF (KEY.EQ.1) COV = PRGOPT(LAST+2) .NE. 0.D0
         IF (KEY.EQ.10) SCLCOV = PRGOPT(LAST+2) .EQ. 0.D0
         IF (KEY.EQ.5) TOL = MAX(DRELPR,PRGOPT(LAST+2))
         NEXT = INT(PRGOPT(LINK),4)
         LAST = LINK
         LINK = NEXT
         GO TO 100
      ENDIF
c
c     Compute matrix norm of least squares equations.
c
      ANORM = 0.D0
      DO 110 J = 1,N
         ANORM = MAX(ANORM,xDASUM(MA,W(1,J),1))
  110 CONTINUE
c
c     Set tolerance for xDHFTI( ) rank test.
c
      TAU = TOL*ANORM
c
c     Compute Householder orthogonal decomposition of matrix.
c
      CALL XDCOPYSC (N, 0.D0, WS, 1)
      CALL xDCOPY (MA, W(1, NP1), 1, WS, 1)
      K = MAX(M,N)
      MINMAN = MIN(MA,N)
      N1 = K + 1
      N2 = N1 + N
c      CALL xDHFTI (W, MDW, MA, N, WS, MA, 1, TAU, KRANK, RNORM, WS(N2),  WS(N1), IP)
c KARLINE:ADDED both next sentences ...
        RNORMV(1) = RNORM       
        MDB = MAX(MA,N)
      CALL xDHFTI (W, MDW, MA, N, WS, MDB, 1,TAU,KRANK, RNORMV, WS(N2),             &
     &           WS(N1), IP)   
      RNORM = RNORMV(1)   ! and this one added as well
c and changed that...
      FAC = 1.D0
      GAM = MA - KRANK
      IF (KRANK.LT.MA .AND. SCLCOV) FAC = RNORM**2/GAM
c
c     Reduce to DLPDP and solve.
c
      MAP1 = MA + 1
c
c     Compute inequality rt-hand side for DLPDP.
c
      IF (MA.LT.M) THEN
         IF (MINMAN.GT.0) THEN
            DO 120 I = MAP1,M
               W(I,NP1) = W(I,NP1) - xDDOT(N,W(I,1),MDW,WS,1)
  120       CONTINUE
c
c           Apply permutations to col. of inequality constraint matrix.
c
            DO 130 I = 1,MINMAN
               CALL xDSWAP (MG, W(MAP1,I), 1, W(MAP1,IP(I)), 1)
  130       CONTINUE
c
c           Apply Householder transformations to constraint matrix.
c
            IF (KRANK.GT.0 .AND. KRANK.LT.N) THEN
               DO 140 I = KRANK,1,-1
                  CALL xDH12 (2, I, KRANK+1, N, W(I,1), MDW, WS(N1+I-1),             &
     &                      W(MAP1,1), MDW, 1, MG)
  140          CONTINUE
            ENDIF
c
c           Compute permuted inequality constraint matrix times r-inv.
c
            DO 160 I = MAP1,M
               DO 150 J = 1,KRANK
                W(I,J) = (W(I,J)-xDDOT(J-1,W(1,J),1,W(I,1),MDW))/W(J,J)
  150          CONTINUE
  160       CONTINUE
         ENDIF
c
c        Solve the reduced problem with DLPDP algorithm,
c        the least projected distance problem.
c
         CALL DLPDP(W(MAP1,1), MDW, MG, KRANK, N-KRANK, PRGOPT, X,                  &
     &             XNORM, MDLPDP, WS(N2), IP(N+1))
c
c        Compute solution in original coordinates.
c
         IF (MDLPDP.EQ.1) THEN
            DO 170 I = KRANK,1,-1
               X(I) = (X(I)-xDDOT(KRANK-I,W(I,I+1),MDW,X(I+1),1))/W(I,I)
  170       CONTINUE
c
c           Apply Householder transformation to solution vector.
c
            IF (KRANK.LT.N) THEN
               DO 180 I = 1,KRANK
                  CALL xDH12 (2, I, KRANK+1, N, W(I,1), MDW, WS(N1+I-1),             &
     &                      X, 1, 1, 1)
  180          CONTINUE
            ENDIF
c
c           Repermute variables to their input order.
c
            IF (MINMAN.GT.0) THEN
               DO 190 I = MINMAN,1,-1
                  CALL xDSWAP (1, X(I), 1, X(IP(I)), 1)
  190          CONTINUE
c
c              Variables are now in original coordinates.
c              Add solution of unconstrained problem.
c
               DO 200 I = 1,N
                  X(I) = X(I) + WS(I)
  200          CONTINUE
c
c              Compute the residual vector norm.
c
               RNORM = SQRT(RNORM**2+XNORM**2)
            ENDIF
         ELSE
            MODE = 2
         ENDIF
      ELSE
         CALL xDCOPY (N, WS, 1, X, 1)
      ENDIF
c
c     Compute covariance matrix based on the orthogonal decomposition
c     from xDHFTI( ).
c
      IF (.NOT.COV .OR. KRANK.LE.0) GO TO 370
      KRM1 = KRANK - 1
      KRP1 = KRANK + 1
c
c     Copy diagonal terms to working array.
c
      CALL xDCOPY (KRANK, W, MDW+1, WS(N2), 1)
c
c     Reciprocate diagonal terms.
c
      DO 210 J = 1,KRANK
         W(J,J) = 1.D0/W(J,J)
  210 CONTINUE
c
c     Invert the upper triangular QR factor on itself.
c
      IF (KRANK.GT.1) THEN
         DO 230 I = 1,KRM1
            DO 220 J = I+1,KRANK
               W(I,J) = -xDDOT(J-I,W(I,I),MDW,W(I,J),1)*W(J,J)
  220       CONTINUE
  230    CONTINUE
      ENDIF
c
c     Compute the inverted factor times its transpose.
c
      DO 250 I = 1,KRANK
         DO 240 J = I,KRANK
            W(I,J) = xDDOT(KRANK+1-J,W(I,J),MDW,W(J,J),MDW)
  240    CONTINUE
  250 CONTINUE
c
c     Zero out lower trapezoidal part.
c     Copy upper triangular to lower triangular part.
c
      IF (KRANK.LT.N) THEN
         DO 260 J = 1,KRANK
            CALL xDCOPY (J, W(1,J), 1, W(J,1), MDW)
  260    CONTINUE
c
         DO 270 I = KRP1,N
            CALL XDCOPYSC (I, 0.D0, W(I,1), MDW)
  270    CONTINUE
c
c        Apply right side transformations to lower triangle.
c
         N3 = N2 + KRP1
         DO 330 I = 1,KRANK
            L = N1 + I
            K = N2 + I
            RB = WS(L-1)*WS(K-1)
c
c           If RB.GE.0.D0, transformation can be regarded as zero.
c
            IF (RB.LT.0.D0) THEN
               RB = 1.D0/RB
c
c              Store unscaled rank one Householder update in work array.
c
               CALL XDCOPYSC (N, 0.D0, WS(N3), 1)
               L = N1 + I
               K = N3 + I
               WS(K-1) = WS(L-1)
c
               DO 280 J = KRP1,N
                  WS(N3+J-1) = W(I,J)
  280          CONTINUE
c
               DO 290 J = 1,N
                  WS(J) = RB*(xDDOT(J-I,W(J,I),MDW,WS(N3+I-1),1)+                    &
     &                    xDDOT(N-J+1,W(J,J),1,WS(N3+J-1),1))
  290          CONTINUE
c
               L = N3 + I
               GAM = 0.5D0*RB*xDDOT(N-I+1,WS(L-1),1,WS(I),1)
               CALL xDAXPY (N-I+1, GAM, WS(L-1), 1, WS(I), 1)
               DO 320 J = I,N
                  DO 300 L = 1,I-1
                     W(J,L) = W(J,L) + WS(N3+J-1)*WS(L)
  300             CONTINUE
c
                  DO 310 L = I,J
                     W(J,L) = W(J,L) + WS(J)*WS(N3+L-1)+WS(L)*WS(N3+J-1)
  310             CONTINUE
  320          CONTINUE
            ENDIF
  330    CONTINUE
c
c        Copy lower triangle to upper triangle to symmetrize the
c        covariance matrix.
c
         DO 340 I = 1,N
            CALL xDCOPY (I, W(I,1), MDW, W(1,I), 1)
  340    CONTINUE
      ENDIF
c
c     Repermute rows and columns.
c
      DO 350 I = MINMAN,1,-1
         K = IP(I)
         IF (I.NE.K) THEN
            CALL xDSWAP (1, W(I,I), 1, W(K,K), 1)
            CALL xDSWAP (I-1, W(1,I), 1, W(1,K), 1)
            CALL xDSWAP (K-I-1, W(I,I+1), MDW, W(I+1,K), 1)
            CALL xDSWAP (N-K, W(I, K+1), MDW, W(K, K+1), MDW)
         ENDIF
  350 CONTINUE
c
c     Put in normalized residual sum of squares scale factor
c     and symmetrize the resulting covariance matrix.
c
      DO 360 J = 1,N
         CALL xDSCAL (J, FAC, W(1,J), 1)
         CALL xDCOPY (J, W(1,J), 1, W(J,1), MDW)
  360 CONTINUE
c
  370 IP(1) = KRANK
      IP(2) = N + MAX(M,N) + (MG+2)*(N+7)
      RETURN
      END





      SUBROUTINE DWNLIT (W, MDW, M, N, L, IPIVOT, ITYPE, H, SCALE,                  &
     &   RNORM, IDOPE, DOPE, DONE)
c***BEGIN PROLOGUE  DWNLIT
c***SUBSIDIARY
c***PURPOSE  Subsidiary to DWNNLS
c***LIBRARY   SLATEC
c***TYPE      DOUBLE PRECISION (WNLIT-S, DWNLIT-D)
c***AUTHOR  Hanson, R. J., (SNLA)
c           Haskell, K. H., (SNLA)
c***DESCRIPTION
c
c     This is a companion subprogram to DWNNLS( ).
c     The documentation for DWNNLS( ) has complete usage instructions.
c
c     Note  The M by (N+1) matrix W( , ) contains the rt. hand side
c           B as the (N+1)st col.
c
c     Triangularize L1 by L1 subsystem, where L1=MIN(M,L), with
c     col interchanges.
c
c***SEE ALSO  DWNNLS
c***ROUTINES CALLED  xDCOPY, xDH12, xDROTM, xDROTMG, xDSCAL, xDSWAP, DWNLT1,
c                    DWNLT2, DWNLT3, xIDAMAX
c***REVISION HISTORY  (YYMMDD)
c   790701  DATE WRITTEN
c   890531  Changed all specific intrinsics to generic.  (WRB)
c   890618  Completely restructured and revised.  (WRB & RWC)
c   890620  Revised to make WNLT1, WNLT2, and WNLT3 subroutines.  (RWC)
c   891214  Prologue converted to Version 4.0 format.  (BAB)
c   900328  Added TYPE section.  (WRB)
c   900604  DP version created from SP version. .  (RWC)
c***END PROLOGUE  DWNLIT
      INTEGER IDOPE(*), IPIVOT(*), ITYPE(*), L, M, MDW, N
      DOUBLE PRECISION DOPE(*), H(*), RNORM, SCALE(*), W(MDW,*)
      LOGICAL DONE
c
      EXTERNAL xDCOPY, xDH12, xDROTM, xDROTMG, xDSCAL, xDSWAP, DWNLT1,              &
     &   DWNLT2, DWNLT3, xIDAMAX
      INTEGER xIDAMAX
      LOGICAL DWNLT2
c
      DOUBLE PRECISION ALSQ, AMAX, EANORM, FACTOR, HBAR, RN, SPARAM(5),             &
     &   T, TAU
      INTEGER I, I1, IMAX, IR, J, J1, JJ, JP, KRANK, L1, LB, LEND, ME,              &
     &   MEND, NIV, NSOLN
      LOGICAL INDEP, RECALC
c
c***FIRST EXECUTABLE STATEMENT  DWNLIT
      ME    = IDOPE(1)
      NSOLN = IDOPE(2)
      L1    = IDOPE(3)
c
      ALSQ   = DOPE(1)
      EANORM = DOPE(2)
      TAU    = DOPE(3)
c
      LB     = MIN(M-1,L)
      RECALC = .TRUE.
      RNORM  = 0.D0
      KRANK  = 0
c
c     We set FACTOR=1.0 so that the heavy weight ALAMDA will be
c     included in the test for column independence.
c
      FACTOR = 1.D0
      LEND = L
      DO 180 I=1,LB
c
c        Set IR to point to the I-th row.
c
         IR = I
         MEND = M
         CALL DWNLT1 (I, LEND, M, IR, MDW, RECALC, IMAX, HBAR, H, SCALE,            &
     &                W)
c
c        Update column SS and find pivot column.
c
         CALL DWNLT3 (I, IMAX, M, MDW, IPIVOT, H, W)
c
c        Perform column interchange.
c        Test independence of incoming column.
c
  130    IF (DWNLT2(ME, MEND, IR, FACTOR, TAU, SCALE, W(1,I))) THEN
c
c           Eliminate I-th column below diagonal using modified Givens
c           transformations applied to (A B).
c
c           When operating near the ME line, use the largest element
c           above it as the pivot.
c
            DO 160 J=M,I+1,-1
               JP = J-1
               IF (J.EQ.ME+1) THEN
                  IMAX = ME
                  AMAX = SCALE(ME)*W(ME,I)**2
                  DO 150 JP=J-1,I,-1
                     T = SCALE(JP)*W(JP,I)**2
                     IF (T.GT.AMAX) THEN
                        IMAX = JP
                        AMAX = T
                     ENDIF
  150             CONTINUE
                  JP = IMAX
               ENDIF
c
               IF (W(J,I).NE.0.D0) THEN
                  CALL xDROTMG (SCALE(JP), SCALE(J), W(JP,I), W(J,I),                &
     &                         SPARAM)
                  W(J,I) = 0.D0
                  CALL xDROTM (N+1-I, W(JP,I+1), MDW, W(J,I+1), MDW,                 &
     &                        SPARAM)
               ENDIF
  160       CONTINUE
         ELSE IF (LEND.GT.I) THEN
c
c           Column I is dependent.  Swap with column LEND.
c           Perform column interchange,
c           and find column in remaining set with largest SS.
c
            CALL DWNLT3 (I, LEND, M, MDW, IPIVOT, H, W)
            LEND = LEND - 1
            IMAX = xIDAMAX(LEND-I+1, H(I), 1) + I - 1
            HBAR = H(IMAX)
            GO TO 130
         ELSE
            KRANK = I - 1
            GO TO 190
         ENDIF
  180 CONTINUE
      KRANK = L1
c
  190 IF (KRANK.LT.ME) THEN
         FACTOR = ALSQ
         DO 200 I=KRANK+1,ME
            CALL XDCOPYSC (L, 0.D0, W(I,1), MDW)
  200    CONTINUE
c
c        Determine the rank of the remaining equality constraint
c        equations by eliminating within the block of constrained
c        variables.  Remove any redundant constraints.
c
         RECALC = .TRUE.
         LB = MIN(L+ME-KRANK, N)
         DO 270 I=L+1,LB
            IR = KRANK + I - L
            LEND = N
            MEND = ME
            CALL DWNLT1 (I, LEND, ME, IR, MDW, RECALC, IMAX, HBAR, H,               &
     &                   SCALE, W)
c
c           Update col ss and find pivot col
c
            CALL DWNLT3 (I, IMAX, M, MDW, IPIVOT, H, W)
c
c           Perform column interchange
c           Eliminate elements in the I-th col.
c
            DO 240 J=ME,IR+1,-1
               IF (W(J,I).NE.0.D0) THEN
                 CALL xDROTMG (SCALE(J-1), SCALE(J), W(J-1,I), W(J,I),               &
     &                        SPARAM)
                  W(J,I) = 0.D0
                  CALL xDROTM (N+1-I, W(J-1,I+1), MDW,W(J,I+1), MDW,                 &
     &                        SPARAM)
               ENDIF
  240       CONTINUE
c
c           I=column being eliminated.
c           Test independence of incoming column.
c           Remove any redundant or dependent equality constraints.
c
            IF (.NOT.DWNLT2(ME, MEND, IR, FACTOR,TAU,SCALE,W(1,I))) THEN
               JJ = IR
               DO 260 IR=JJ,ME
                  CALL XDCOPYSC (N, 0.D0, W(IR,1), MDW)
                  RNORM = RNORM + (SCALE(IR)*W(IR,N+1)/ALSQ)*W(IR,N+1)
                  W(IR,N+1) = 0.D0
                  SCALE(IR) = 1.D0
c
c                 Reclassify the zeroed row as a least squares equation.
c
                  ITYPE(IR) = 1
  260          CONTINUE
c
c              Reduce ME to reflect any discovered dependent equality
c              constraints.
c
               ME = JJ - 1
               GO TO 280
            ENDIF
  270    CONTINUE
      ENDIF
c
c     Try to determine the variables KRANK+1 through L1 from the
c     least squares equations.  Continue the triangularization with
c     pivot element W(ME+1,I).
c
  280 IF (KRANK.LT.L1) THEN
         RECALC = .TRUE.
c
c        Set FACTOR=ALSQ to remove effect of heavy weight from
c        test for column independence.
c
         FACTOR = ALSQ
         DO 350 I=KRANK+1,L1
c
c           Set IR to point to the ME+1-st row.
c
            IR = ME+1
            LEND = L
            MEND = M
            CALL DWNLT1 (I, L, M, IR, MDW, RECALC, IMAX, HBAR, H, SCALE,            &
     &                   W)
c
c           Update column SS and find pivot column.
c
            CALL DWNLT3 (I, IMAX, M, MDW, IPIVOT, H, W)
c
c           Perform column interchange.
c           Eliminate I-th column below the IR-th element.
c
            DO 320 J=M,IR+1,-1
               IF (W(J,I).NE.0.D0) THEN
                 CALL xDROTMG (SCALE(J-1), SCALE(J), W(J-1,I), W(J,I),               &
     &                        SPARAM)
                  W(J,I) = 0.D0
                  CALL xDROTM (N+1-I, W(J-1,I+1),  MDW, W(J,I+1), MDW,               &
     &                        SPARAM)
               ENDIF
  320       CONTINUE
c
c           Test if new pivot element is near zero.
c           If so, the column is dependent.
c           Then check row norm test to be classified as independent.
c
            T = SCALE(IR)*W(IR,I)**2
            INDEP = T .GT. (TAU*EANORM)**2
            IF (INDEP) THEN
               RN = 0.D0
               DO 340 I1=IR,M
                  DO 330 J1=I+1,N
                     RN = MAX(RN, SCALE(I1)*W(I1,J1)**2)
  330             CONTINUE
  340          CONTINUE
               INDEP = T .GT. RN*TAU**2
            ENDIF
c
c           If independent, swap the IR-th and KRANK+1-th rows to
c           maintain the triangular form.  Update the rank indicator
c           KRANK and the equality constraint pointer ME.
c
            IF (.NOT.INDEP) GO TO 360
            CALL xDSWAP(N+1, W(KRANK+1,1), MDW, W(IR,1), MDW)
            CALL xDSWAP(1, SCALE(KRANK+1), 1, SCALE(IR), 1)
c
c           Reclassify the least square equation as an equality
c           constraint and rescale it.
c
            ITYPE(IR) = 0
            T = SQRT(SCALE(KRANK+1))
            CALL xDSCAL(N+1, T, W(KRANK+1,1), MDW)
            SCALE(KRANK+1) = ALSQ
            ME = ME+1
            KRANK = KRANK+1
  350    CONTINUE
      ENDIF
c
c     If pseudorank is less than L, apply Householder transformation.
c     from right.
c
  360 IF (KRANK.LT.L) THEN
         DO 370 J=KRANK,1,-1
            CALL xDH12 (1, J, KRANK+1, L, W(J,1), MDW, H(J), W, MDW, 1,              &
     &                J-1)
  370    CONTINUE
      ENDIF
c
      NIV = KRANK + NSOLN - L
      IF (L.EQ.N) DONE = .TRUE.
c
c     End of initial triangularization.
c
      IDOPE(1) = ME
      IDOPE(2) = KRANK
      IDOPE(3) = NIV
      RETURN
      END





      SUBROUTINE DWNLSM (W, MDW, MME, MA, N, L, PRGOPT, X, RNORM, MODE,             &
     &   IPIVOT, ITYPE, WD, H, SCALE, Z, TEMP, D)
c***BEGIN PROLOGUE  DWNLSM
c***SUBSIDIARY
c***PURPOSE  Subsidiary to DWNNLS
c***LIBRARY   SLATEC
c***TYPE      DOUBLE PRECISION (WNLSM-S, DWNLSM-D)
c***AUTHOR  Hanson, R. J., (SNLA)
c           Haskell, K. H., (SNLA)
c***DESCRIPTION
c
c     This is a companion subprogram to DWNNLS.
c     The documentation for DWNNLS has complete usage instructions.
c
c     In addition to the parameters discussed in the prologue to
c     subroutine DWNNLS, the following work arrays are used in
c     subroutine DWNLSM  (they are passed through the calling
c     sequence from DWNNLS for purposes of variable dimensioning).
c     Their contents will in general be of no interest to the user.
c
c     Variables of type REAL are DOUBLE PRECISION.
c
c         IPIVOT(*)
c            An array of length N.  Upon completion it contains the
c         pivoting information for the cols of W(*,*).
c
c         ITYPE(*)
c            An array of length M which is used to keep track
c         of the classification of the equations.  ITYPE(I)=0
c         denotes equation I as an equality constraint.
c         ITYPE(I)=1 denotes equation I as a least squares
c         equation.
c
c         WD(*)
c            An array of length N.  Upon completion it contains the
c         dual solution vector.
c
c         H(*)
c            An array of length N.  Upon completion it contains the
c         pivot scalars of the Householder transformations performed
c         in the case KRANK.LT.L.
c
c         SCALE(*)
c            An array of length M which is used by the subroutine
c         to store the diagonal matrix of weights.
c         These are used to apply the modified Givens
c         transformations.
c
c         Z(*),TEMP(*)
c            Working arrays of length N.
c
c         D(*)
c            An array of length N that contains the
c         column scaling for the matrix (E).
c                                       (A)
c
c***SEE ALSO  DWNNLS
c***ROUTINES CALLED  D1MACH, xDASUM, xDAXPY, xDCOPY, xDH12, xDNRM2, xDROTM,
c                    xDROTMG, xDSCAL, xDSWAP, DWNLIT, xIDAMAX, xXERMSG
c***REVISION HISTORY  (YYMMDD)
c   790701  DATE WRITTEN
c   890531  Changed all specific intrinsics to generic.  (WRB)
c   890618  Completely restructured and revised.  (WRB & RWC)
c   891214  Prologue converted to Version 4.0 format.  (BAB)
c   900315  CALLs to XERROR changed to CALLs to xXERMSG.  (THJ)
c   900328  Added TYPE section.  (WRB)
c   900510  Fixed an error message.  (RWC)
c   900604  DP version created from SP version.  (RWC)
c   900911  Restriction on value of ALAMDA included.  (WRB)
c***END PROLOGUE  DWNLSM
      INTEGER IPIVOT(*), ITYPE(*), L, MA, MDW, MME, MODE, N
      DOUBLE PRECISION D(*), H(*), PRGOPT(*), RNORM, SCALE(*), TEMP(*),             &
     &   W(MDW,*), WD(*), X(*), Z(*)
c
      EXTERNAL D1MACH,xDASUM,xDAXPY,xDCOPY,xDH12,xDNRM2,xDROTM,xDROTMG,             &
     &   xDSCAL, xDSWAP, DWNLIT, xIDAMAX, xXERMSG
      DOUBLE PRECISION D1MACH, xDASUM, xDNRM2
      INTEGER xIDAMAX
c
      DOUBLE PRECISION ALAMDA, ALPHA, ALSQ, AMAX, BLOWUP, BNORM,                    &
     &   DOPE(3), DRELPR, EANORM, FAC, SM, SPARAM(5), T, TAU, WMAX, Z2,             &
     &   ZZ
      INTEGER I, IDOPE(3), IMAX, ISOL, ITEMP, ITER, ITMAX, IWMAX, J,                &
     &   JCON, JP, KEY, KRANK, L1, LAST, LINK, M, ME, NEXT, NIV, NLINK,             &  
     &   NOPT, NSOLN, NTIMES
      LOGICAL DONE, FEASBL, FIRST, HITCON, POS
c
      SAVE DRELPR, FIRST
      DATA FIRST /.TRUE./
c***FIRST EXECUTABLE STATEMENT  DWNLSM
c
c     Initialize variables.
c     DRELPR is the precision for the particular machine
c     being used.  This logic avoids resetting it every entry.
c
      IF (FIRST) DRELPR = D1MACH(4)
      FIRST = .FALSE.
c
c     Set the nominal tolerance used in the code.
c
      TAU = SQRT(DRELPR)
c
      M = MA + MME
      ME = MME
      MODE = 2
c
c     To process option vector
c
      FAC = 1.D-4
c
c     Set the nominal blow up factor used in the code.
c
      BLOWUP = TAU
c
c     The nominal column scaling used in the code is
c     the identity scaling.
c
      CALL XDCOPYSC (N, 1.D0, D, 1)
c
c     Define bound for number of options to change.
c
      NOPT = 1000
c
c     Define bound for positive value of LINK.
c
      NLINK = 100000
      NTIMES = 0
      LAST = 1
      LINK = INT(PRGOPT(1),4)
      IF (LINK.LE.0 .OR. LINK.GT.NLINK) THEN
         CALL xXERMSG ('SLATEC', 'DWNLSM',                                          &
     &      'IN DWNNLS, THE OPTION VECTOR IS UNDEFINED', 3, 1)
         RETURN
      ENDIF
c
  100 IF (LINK.GT.1) THEN
         NTIMES = NTIMES + 1
         IF (NTIMES.GT.NOPT) THEN
         CALL xXERMSG ('SLATEC', 'DWNLSM',                                          &
     &      'IN DWNNLS, THE LINKS IN THE OPTION VECTOR ARE CYCLING.',               &
     &      3, 1)
            RETURN
         ENDIF
c
         KEY = INT(PRGOPT(LAST+1),4)
         IF (KEY.EQ.6 .AND. PRGOPT(LAST+2).NE.0.D0) THEN
            DO 110 J = 1,N
               T = xDNRM2(M,W(1,J),1)
               IF (T.NE.0.D0) T = 1.D0/T
               D(J) = T
  110       CONTINUE
         ENDIF
c
         IF (KEY.EQ.7) CALL xDCOPY (N, PRGOPT(LAST+2), 1, D, 1)
         IF (KEY.EQ.8) TAU = MAX(DRELPR,PRGOPT(LAST+2))
         IF (KEY.EQ.9) BLOWUP = MAX(DRELPR,PRGOPT(LAST+2))
c
         NEXT = INT(PRGOPT(LINK),4)
         IF (NEXT.LE.0 .OR. NEXT.GT.NLINK) THEN
            CALL xXERMSG ('SLATEC', 'DWNLSM',                                       &
     &         'IN DWNNLS, THE OPTION VECTOR IS UNDEFINED', 3, 1)
            RETURN
         ENDIF
c
         LAST = LINK
         LINK = NEXT
         GO TO 100
      ENDIF
c
      DO 120 J = 1,N
         CALL xDSCAL (M, D(J), W(1,J), 1)
  120 CONTINUE
c
c     Process option vector
c
      DONE = .FALSE.
      ITER = 0
      ITMAX = 3*(N-L)
      MODE = 0
      NSOLN = L
      L1 = MIN(M,L)
c
c     Compute scale factor to apply to equality constraint equations.
c
      DO 130 J = 1,N
         WD(J) = xDASUM(M,W(1,J),1)
  130 CONTINUE
c
      IMAX = xIDAMAX(N,WD,1)
      EANORM = WD(IMAX)
      BNORM = xDASUM(M,W(1,N+1),1)
      ALAMDA = EANORM/(DRELPR*FAC)
c
c     On machines, such as the VAXes using D floating, with a very
c     limited exponent range for double precision values, the previously
c     computed value of ALAMDA may cause an overflow condition.
c     Therefore, this code further limits the value of ALAMDA.
c
      ALAMDA = MIN(ALAMDA,SQRT(D1MACH(2)))
c
c     Define scaling diagonal matrix for modified Givens usage and
c     classify equation types.
c
      ALSQ = ALAMDA**2
      DO 140 I = 1,M
c
c        When equation I is heavily weighted ITYPE(I)=0,
c        else ITYPE(I)=1.
c
         IF (I.LE.ME) THEN
            T = ALSQ
            ITEMP = 0
         ELSE
            T = 1.D0
            ITEMP = 1
         ENDIF
         SCALE(I) = T
         ITYPE(I) = ITEMP
  140 CONTINUE
c
c     Set the solution vector X(*) to zero and the column interchange
c     matrix to the identity.
c
      CALL XDCOPYSC (N, 0.D0, X, 1)
      DO 150 I = 1,N
         IPIVOT(I) = I
  150 CONTINUE
c
c     Perform initial triangularization in the submatrix
c     corresponding to the unconstrained variables.
c     Set first L components of dual vector to zero because
c     these correspond to the unconstrained variables.
c
      CALL XDCOPYSC (L, 0.D0, WD, 1)
c
c     The arrays IDOPE(*) and DOPE(*) are used to pass
c     information to DWNLIT().  This was done to avoid
c     a long calling sequence or the use of COMMON.
c
      IDOPE(1) = ME
      IDOPE(2) = NSOLN
      IDOPE(3) = L1
c
      DOPE(1) = ALSQ
      DOPE(2) = EANORM
      DOPE(3) = TAU
      CALL DWNLIT (W, MDW, M, N, L, IPIVOT, ITYPE, H, SCALE, RNORM,                 &
     &            IDOPE, DOPE, DONE)
      ME    = IDOPE(1)
      KRANK = IDOPE(2)
      NIV   = IDOPE(3)
c
c     Perform WNNLS algorithm using the following steps.
c
c     Until(DONE)
c        compute search direction and feasible point
c        when (HITCON) add constraints
c        else perform multiplier test and drop a constraint
c        fin
c     Compute-Final-Solution
c
c     To compute search direction and feasible point,
c     solve the triangular system of currently non-active
c     variables and store the solution in Z(*).
c
c     To solve system
c     Copy right hand side into TEMP vector to use overwriting method.
c
  160 IF (DONE) GO TO 330
      ISOL = L + 1
      IF (NSOLN.GE.ISOL) THEN
         CALL xDCOPY (NIV, W(1,N+1), 1, TEMP, 1)
         DO 170 J = NSOLN,ISOL,-1
            IF (J.GT.KRANK) THEN
               I = NIV - NSOLN + J
            ELSE
               I = J
            ENDIF
c
            IF (J.GT.KRANK .AND. J.LE.L) THEN
               Z(J) = 0.D0
            ELSE
               Z(J) = TEMP(I)/W(I,J)
               CALL xDAXPY (I-1, -Z(J), W(1,J), 1, TEMP, 1)
            ENDIF
  170    CONTINUE
      ENDIF
c
c     Increment iteration counter and check against maximum number
c     of iterations.
c
      ITER = ITER + 1
      IF (ITER.GT.ITMAX) THEN
         MODE = 1
         DONE = .TRUE.
      ENDIF
c
c     Check to see if any constraints have become active.
c     If so, calculate an interpolation factor so that all
c     active constraints are removed from the basis.
c
      ALPHA = 2.D0
      HITCON = .FALSE.
      DO 180 J = L+1,NSOLN
         ZZ = Z(J)
         IF (ZZ.LE.0.D0) THEN
            T = X(J)/(X(J)-ZZ)
            IF (T.LT.ALPHA) THEN
               ALPHA = T
               JCON = J
            ENDIF
            HITCON = .TRUE.
         ENDIF
  180 CONTINUE
c
c     Compute search direction and feasible point
c
      IF (HITCON) THEN
c
c        To add constraints, use computed ALPHA to interpolate between
c        last feasible solution X(*) and current unconstrained (and
c        infeasible) solution Z(*).
c
         DO 190 J = L+1,NSOLN
            X(J) = X(J) + ALPHA*(Z(J)-X(J))
  190    CONTINUE
         FEASBL = .FALSE.
c
c        Remove column JCON and shift columns JCON+1 through N to the
c        left.  Swap column JCON into the N th position.  This achieves
c        upper Hessenberg form for the nonactive constraints and
c        leaves an upper Hessenberg matrix to retriangularize.
c
  200    DO 210 I = 1,M
            T = W(I,JCON)
            CALL xDCOPY (N-JCON, W(I, JCON+1), MDW, W(I, JCON), MDW)
            W(I,N) = T
  210    CONTINUE
c
c        Update permuted index vector to reflect this shift and swap.
c
         ITEMP = IPIVOT(JCON)
         DO 220 I = JCON,N - 1
            IPIVOT(I) = IPIVOT(I+1)
  220    CONTINUE
         IPIVOT(N) = ITEMP
c
c        Similarly permute X(*) vector.
c
         CALL xDCOPY (N-JCON, X(JCON+1), 1, X(JCON), 1)
         X(N) = 0.D0
         NSOLN = NSOLN - 1
         NIV = NIV - 1
c
c        Retriangularize upper Hessenberg matrix after adding
c        constraints.
c
         I = KRANK + JCON - L
         DO 230 J = JCON,NSOLN
            IF (ITYPE(I).EQ.0 .AND. ITYPE(I+1).EQ.0) THEN
c
c              Zero IP1 to I in column J
c
               IF (W(I+1,J).NE.0.D0) THEN
                  CALL xDROTMG (SCALE(I), SCALE(I+1), W(I,J), W(I+1,J),              &
     &                         SPARAM)
                  W(I+1,J) = 0.D0
                  CALL xDROTM (N+1-J, W(I,J+1), MDW, W(I+1,J+1), MDW,                &
     &                        SPARAM)
               ENDIF
            ELSEIF (ITYPE(I).EQ.1 .AND. ITYPE(I+1).EQ.1) THEN
c
c              Zero IP1 to I in column J
c
               IF (W(I+1,J).NE.0.D0) THEN
                  CALL xDROTMG (SCALE(I), SCALE(I+1), W(I,J), W(I+1,J),              &
     &                         SPARAM)
                  W(I+1,J) = 0.D0
                  CALL xDROTM (N+1-J, W(I,J+1), MDW, W(I+1,J+1), MDW,                &
     &                        SPARAM)
               ENDIF
            ELSEIF (ITYPE(I).EQ.1 .AND. ITYPE(I+1).EQ.0) THEN
               CALL xDSWAP (N+1, W(I,1), MDW, W(I+1,1), MDW)
               CALL xDSWAP (1, SCALE(I), 1, SCALE(I+1), 1)
               ITEMP = ITYPE(I+1)
               ITYPE(I+1) = ITYPE(I)
               ITYPE(I) = ITEMP
c
c              Swapped row was formerly a pivot element, so it will
c              be large enough to perform elimination.
c              Zero IP1 to I in column J.
c
               IF (W(I+1,J).NE.0.D0) THEN
                  CALL xDROTMG (SCALE(I), SCALE(I+1), W(I,J), W(I+1,J),              &
     &                         SPARAM)
                  W(I+1,J) = 0.D0
                  CALL xDROTM (N+1-J, W(I,J+1), MDW, W(I+1,J+1), MDW,                &
     &                        SPARAM)
               ENDIF
            ELSEIF (ITYPE(I).EQ.0 .AND. ITYPE(I+1).EQ.1) THEN
               IF (SCALE(I)*W(I,J)**2/ALSQ.GT.(TAU*EANORM)**2) THEN
c
c                 Zero IP1 to I in column J
c
                  IF (W(I+1,J).NE.0.D0) THEN
                     CALL xDROTMG (SCALE(I), SCALE(I+1), W(I,J),                     &
     &                            W(I+1,J), SPARAM)
                     W(I+1,J) = 0.D0
                     CALL xDROTM (N+1-J, W(I,J+1), MDW, W(I+1,J+1), MDW,             &
     &                           SPARAM)
                  ENDIF
               ELSE
                  CALL xDSWAP (N+1, W(I,1), MDW, W(I+1,1), MDW)
                  CALL xDSWAP (1, SCALE(I), 1, SCALE(I+1), 1)
                  ITEMP = ITYPE(I+1)
                  ITYPE(I+1) = ITYPE(I)
                  ITYPE(I) = ITEMP
                  W(I+1,J) = 0.D0
               ENDIF
            ENDIF
            I = I + 1
  230    CONTINUE
c
c        See if the remaining coefficients in the solution set are
c        feasible.  They should be because of the way ALPHA was
c        determined.  If any are infeasible, it is due to roundoff
c        error.  Any that are non-positive will be set to zero and
c        removed from the solution set.
c
         DO 240 JCON = L+1,NSOLN
            IF (X(JCON).LE.0.D0) GO TO 250
  240    CONTINUE
         FEASBL = .TRUE.
  250    IF (.NOT.FEASBL) GO TO 200
      ELSE
c
c        To perform multiplier test and drop a constraint.
c
         CALL xDCOPY (NSOLN, Z, 1, X, 1)
         IF (NSOLN.LT.N) CALL XDCOPYSC (N-NSOLN, 0.D0, X(NSOLN+1), 1)
c
c        Reclassify least squares equations as equalities as necessary.
c
         I = NIV + 1
  260    IF (I.LE.ME) THEN
            IF (ITYPE(I).EQ.0) THEN
               I = I + 1
            ELSE
               CALL xDSWAP (N+1, W(I,1), MDW, W(ME,1), MDW)
               CALL xDSWAP (1, SCALE(I), 1, SCALE(ME), 1)
               ITEMP = ITYPE(I)
               ITYPE(I) = ITYPE(ME)
               ITYPE(ME) = ITEMP
               ME = ME - 1
            ENDIF
            GO TO 260
         ENDIF
c
c        Form inner product vector WD(*) of dual coefficients.
c
         DO 280 J = NSOLN+1,N
            SM = 0.D0
            DO 270 I = NSOLN+1,M
               SM = SM + SCALE(I)*W(I,J)*W(I,N+1)
  270       CONTINUE
            WD(J) = SM
  280    CONTINUE
c
c        Find J such that WD(J)=WMAX is maximum.  This determines
c        that the incoming column J will reduce the residual vector
c        and be positive.
c
  290    WMAX = 0.D0
         IWMAX = NSOLN + 1
         DO 300 J = NSOLN+1,N
            IF (WD(J).GT.WMAX) THEN
               WMAX = WD(J)
               IWMAX = J
            ENDIF
  300    CONTINUE
         IF (WMAX.LE.0.D0) GO TO 330
c
c        Set dual coefficients to zero for incoming column.
c
         WD(IWMAX) = 0.D0
c
c        WMAX .GT. 0.D0, so okay to move column IWMAX to solution set.
c        Perform transformation to retriangularize, and test for near
c        linear dependence.
c
c        Swap column IWMAX into NSOLN-th position to maintain upper
c        Hessenberg form of adjacent columns, and add new column to
c        triangular decomposition.
c
         NSOLN = NSOLN + 1
         NIV = NIV + 1
         IF (NSOLN.NE.IWMAX) THEN
            CALL xDSWAP (M, W(1,NSOLN), 1, W(1,IWMAX), 1)
            WD(IWMAX) = WD(NSOLN)
            WD(NSOLN) = 0.D0
            ITEMP = IPIVOT(NSOLN)
            IPIVOT(NSOLN) = IPIVOT(IWMAX)
            IPIVOT(IWMAX) = ITEMP
         ENDIF
c
c        Reduce column NSOLN so that the matrix of nonactive constraints
c        variables is triangular.
c
         DO 320 J = M,NIV+1,-1
            JP = J - 1
c
c           When operating near the ME line, test to see if the pivot
c           element is near zero.  If so, use the largest element above
c           it as the pivot.  This is to maintain the sharp interface
c           between weighted and non-weighted rows in all cases.
c
            IF (J.EQ.ME+1) THEN
               IMAX = ME
               AMAX = SCALE(ME)*W(ME,NSOLN)**2
               DO 310 JP = J - 1,NIV,-1
                  T = SCALE(JP)*W(JP,NSOLN)**2
                  IF (T.GT.AMAX) THEN
                     IMAX = JP
                     AMAX = T
                  ENDIF
  310          CONTINUE
               JP = IMAX
            ENDIF
c
            IF (W(J,NSOLN).NE.0.D0) THEN
               CALL xDROTMG (SCALE(JP), SCALE(J), W(JP,NSOLN),                       &
     &                      W(J,NSOLN), SPARAM)
               W(J,NSOLN) = 0.D0
               CALL xDROTM (N+1-NSOLN, W(JP,NSOLN+1), MDW, W(J,NSOLN+1),             &
     &                     MDW, SPARAM)
            ENDIF
  320    CONTINUE
c
c        Solve for Z(NSOLN)=proposed new value for X(NSOLN).  Test if
c        this is nonpositive or too large.  If this was true or if the
c        pivot term was zero, reject the column as dependent.
c
         IF (W(NIV,NSOLN).NE.0.D0) THEN
            ISOL = NIV
            Z2 = W(ISOL,N+1)/W(ISOL,NSOLN)
            Z(NSOLN) = Z2
            POS = Z2 .GT. 0.D0
            IF (Z2*EANORM.GE.BNORM .AND. POS) THEN
               POS = .NOT. (BLOWUP*Z2*EANORM.GE.BNORM)
            ENDIF
c
c           Try to add row ME+1 as an additional equality constraint.
c           Check size of proposed new solution component.
c           Reject it if it is too large.
c
         ELSEIF (NIV.LE.ME .AND. W(ME+1,NSOLN).NE.0.D0) THEN
            ISOL = ME + 1
            IF (POS) THEN
c
c              Swap rows ME+1 and NIV, and scale factors for these rows.
c
               CALL xDSWAP (N+1, W(ME+1,1), MDW, W(NIV,1), MDW)
               CALL xDSWAP (1, SCALE(ME+1), 1, SCALE(NIV), 1)
               ITEMP = ITYPE(ME+1)
               ITYPE(ME+1) = ITYPE(NIV)
               ITYPE(NIV) = ITEMP
               ME = ME + 1
            ENDIF
         ELSE
            POS = .FALSE.
         ENDIF
c
         IF (.NOT.POS) THEN
            NSOLN = NSOLN - 1
            NIV = NIV - 1
         ENDIF
         IF (.NOT.(POS.OR.DONE)) GO TO 290
      ENDIF
      GO TO 160
c
c     Else perform multiplier test and drop a constraint.  To compute
c     final solution.  Solve system, store results in X(*).
c
c     Copy right hand side into TEMP vector to use overwriting method.
c
  330 ISOL = 1
      IF (NSOLN.GE.ISOL) THEN
         CALL xDCOPY (NIV, W(1,N+1), 1, TEMP, 1)
         DO 340 J = NSOLN,ISOL,-1
            IF (J.GT.KRANK) THEN
               I = NIV - NSOLN + J
            ELSE
               I = J
            ENDIF
c
            IF (J.GT.KRANK .AND. J.LE.L) THEN
               Z(J) = 0.D0
            ELSE
               Z(J) = TEMP(I)/W(I,J)
               CALL xDAXPY (I-1, -Z(J), W(1,J), 1, TEMP, 1)
            ENDIF
  340    CONTINUE
      ENDIF
c
c     Solve system.
c
      CALL xDCOPY (NSOLN, Z, 1, X, 1)
c
c     Apply Householder transformations to X(*) if KRANK.LT.L
c
      IF (KRANK.LT.L) THEN
         DO 350 I = 1,KRANK
            CALL xDH12 (2, I, KRANK+1, L, W(I,1), MDW, H(I), X, 1, 1, 1)
  350    CONTINUE
      ENDIF
c
c     Fill in trailing zeroes for constrained variables not in solution.
c
      IF (NSOLN.LT.N) CALL XDCOPYSC (N-NSOLN, 0.D0, X(NSOLN+1), 1)
c
c     Permute solution vector to natural order.
c
      DO 380 I = 1,N
         J = I
  360    IF (IPIVOT(J).EQ.I) GO TO 370
         J = J + 1
         GO TO 360
c
  370    IPIVOT(J) = IPIVOT(I)
         IPIVOT(I) = J
         CALL xDSWAP (1, X(J), 1, X(I), 1)
  380 CONTINUE
c
c     Rescale the solution using the column scaling.
c
      DO 390 J = 1,N
         X(J) = X(J)*D(J)
  390 CONTINUE
c
      DO 400 I = NSOLN+1,M
         T = W(I,N+1)
         IF (I.LE.ME) T = T/ALAMDA
         T = (SCALE(I)*T)*T
         RNORM = RNORM + T
  400 CONTINUE
c
      RNORM = SQRT(RNORM)
      RETURN
      END





      SUBROUTINE DWNLT1 (I, LEND, MEND, IR, MDW, RECALC, IMAX, HBAR, H,             &
     &   SCALE, W)
c***BEGIN PROLOGUE  DWNLT1
c***SUBSIDIARY
c***PURPOSE  Subsidiary to WNLIT
c***LIBRARY   SLATEC
c***TYPE      DOUBLE PRECISION (WNLT1-S, DWNLT1-D)
c***AUTHOR  Hanson, R. J., (SNLA)
c           Haskell, K. H., (SNLA)
c***DESCRIPTION
c
c     To update the column Sum Of Squares and find the pivot column.
c     The column Sum of Squares Vector will be updated at each step.
c     When numerically necessary, these values will be recomputed.
c
c***SEE ALSO  DWNLIT
c***ROUTINES CALLED  xIDAMAX
c***REVISION HISTORY  (YYMMDD)
c   790701  DATE WRITTEN
c   890620  Code extracted from WNLIT and made a subroutine.  (RWC))
c   900604  DP version created from SP version.  (RWC)
c***END PROLOGUE  DWNLT1
      INTEGER I, IMAX, IR, LEND, MDW, MEND
      DOUBLE PRECISION H(*), HBAR, SCALE(*), W(MDW,*)
      LOGICAL RECALC
c
      EXTERNAL xIDAMAX
      INTEGER xIDAMAX
c
      INTEGER J, K
c
c***FIRST EXECUTABLE STATEMENT  DWNLT1
      IF (IR.NE.1 .AND. (.NOT.RECALC)) THEN
c
c        Update column SS=sum of squares.
c
         DO 10 J=I,LEND
            H(J) = H(J) - SCALE(IR-1)*W(IR-1,J)**2
   10    CONTINUE
c
c        Test for numerical accuracy.
c
         IMAX = xIDAMAX(LEND-I+1, H(I), 1) + I - 1
         RECALC = (HBAR+1.E-3*H(IMAX)) .EQ. HBAR
      ENDIF
c
c     If required, recalculate column SS, using rows IR through MEND.
c
      IF (RECALC) THEN
         DO 30 J=I,LEND
            H(J) = 0.D0
            DO 20 K=IR,MEND
               H(J) = H(J) + SCALE(K)*W(K,J)**2
   20       CONTINUE
   30    CONTINUE
c
c        Find column with largest SS.
c
         IMAX = xIDAMAX(LEND-I+1, H(I), 1) + I - 1
         HBAR = H(IMAX)
      ENDIF
      RETURN
      END





      LOGICAL FUNCTION DWNLT2 (ME, MEND, IR, FACTOR, TAU, SCALE, WIC)
c***BEGIN PROLOGUE  DWNLT2
c***SUBSIDIARY
c***PURPOSE  Subsidiary to WNLIT
c***LIBRARY   SLATEC
c***TYPE      DOUBLE PRECISION (WNLT2-S, DWNLT2-D)
c***AUTHOR  Hanson, R. J., (SNLA)
c           Haskell, K. H., (SNLA)
c***DESCRIPTION
c
c     To test independence of incoming column.
c
c     Test the column IC to determine if it is linearly independent
c     of the columns already in the basis.  In the initial tri. step,
c     we usually want the heavy weight ALAMDA to be included in the
c     test for independence.  In this case, the value of FACTOR will
c     have been set to 1.E0 before this procedure is invoked.
c     In the potentially rank deficient problem, the value of FACTOR
c     will have been set to ALSQ=ALAMDA**2 to remove the effect of the
c     heavy weight from the test for independence.
c
c     Write new column as partitioned vector
c           (A1)  number of components in solution so far = NIV
c           (A2)  M-NIV components
c     And compute  SN = inverse weighted length of A1
c                  RN = inverse weighted length of A2
c     Call the column independent when RN .GT. TAU*SN
c
c***SEE ALSO  DWNLIT
c***ROUTINES CALLED  (NONE)
c***REVISION HISTORY  (YYMMDD)
c   790701  DATE WRITTEN
c   890620  Code extracted from WNLIT and made a subroutine.  (RWC))
c   900604  DP version created from SP version.  (RWC)
c***END PROLOGUE  DWNLT2
      DOUBLE PRECISION FACTOR, SCALE(*), TAU, WIC(*)
      INTEGER IR, ME, MEND
c
      DOUBLE PRECISION RN, SN, T
      INTEGER J
c
c***FIRST EXECUTABLE STATEMENT  DWNLT2
      SN = 0.E0
      RN = 0.E0
      DO 10 J=1,MEND
         T = SCALE(J)
         IF (J.LE.ME) T = T/FACTOR
         T = T*WIC(J)**2
c
         IF (J.LT.IR) THEN
            SN = SN + T
         ELSE
            RN = RN + T
         ENDIF
   10 CONTINUE
      DWNLT2 = RN .GT. SN*TAU**2
      RETURN
      END





      SUBROUTINE DWNLT3 (I, IMAX, M, MDW, IPIVOT, H, W)
c***BEGIN PROLOGUE  DWNLT3
c***SUBSIDIARY
c***PURPOSE  Subsidiary to WNLIT
c***LIBRARY   SLATEC
c***TYPE      DOUBLE PRECISION (WNLT3-S, DWNLT3-D)
c***AUTHOR  Hanson, R. J., (SNLA)
c           Haskell, K. H., (SNLA)
c***DESCRIPTION
c
c     Perform column interchange.
c     Exchange elements of permuted index vector and perform column
c     interchanges.
c
c***SEE ALSO  DWNLIT
c***ROUTINES CALLED  xDSWAP
c***REVISION HISTORY  (YYMMDD)
c   790701  DATE WRITTEN
c   890620  Code extracted from WNLIT and made a subroutine.  (RWC))
c   900604  DP version created from SP version.  (RWC)
c***END PROLOGUE  DWNLT3
      INTEGER I, IMAX, IPIVOT(*), M, MDW
      DOUBLE PRECISION H(*), W(MDW,*)
c
      EXTERNAL xDSWAP
c
      DOUBLE PRECISION T
      INTEGER ITEMP
c
c***FIRST EXECUTABLE STATEMENT  DWNLT3
      IF (IMAX.NE.I) THEN
         ITEMP        = IPIVOT(I)
         IPIVOT(I)    = IPIVOT(IMAX)
         IPIVOT(IMAX) = ITEMP
c
         CALL xDSWAP(M, W(1,IMAX), 1, W(1,I), 1)
c
         T       = H(IMAX)
         H(IMAX) = H(I)
         H(I)    = T
      ENDIF
      RETURN
      END





      SUBROUTINE DWNNLS (W, MDW, ME, MA, N, L, PRGOPT, X, RNORM, MODE,              &
     &   IWORK, WORK)
c***BEGIN PROLOGUE  DWNNLS
c***PURPOSE  Solve a linearly constrained least squares problem with
c            equality constraints and nonnegativity constraints on
c            selected variables.
c***LIBRARY   SLATEC
c***CATEGORY  K1A2A
c***TYPE      DOUBLE PRECISION (WNNLS-S, DWNNLS-D)
c***KEYWORDS  CONSTRAINED LEAST SQUARES, CURVE FITTING, DATA FITTING,
c             EQUALITY CONSTRAINTS, INEQUALITY CONSTRAINTS,
c             NONNEGATIVITY CONSTRAINTS, QUADRATIC PROGRAMMING
c***AUTHOR  Hanson, R. J., (SNLA)
c           Haskell, K. H., (SNLA)
c***DESCRIPTION
c
c     Abstract
c
c     This subprogram solves a linearly constrained least squares
c     problem.  Suppose there are given matrices E and A of
c     respective dimensions ME by N and MA by N, and vectors F
c     and B of respective lengths ME and MA.  This subroutine
c     solves the problem
c
c               EX = F, (equations to be exactly satisfied)
c
c               AX = B, (equations to be approximately satisfied,
c                        in the least squares sense)
c
c               subject to components L+1,...,N nonnegative
c
c     Any values ME.GE.0, MA.GE.0 and 0.LE. L .LE.N are permitted.
c
c     The problem is reposed as problem DWNNLS
c
c               (WT*E)X = (WT*F)
c               (   A)    (   B), (least squares)
c               subject to components L+1,...,N nonnegative.
c
c     The subprogram chooses the heavy weight (or penalty parameter) WT.
c
c     The parameters for DWNNLS are
c
c     INPUT.. All TYPE REAL variables are DOUBLE PRECISION
c
c     W(*,*),MDW,  The array W(*,*) is double subscripted with first
c     ME,MA,N,L    dimensioning parameter equal to MDW.  For this
c                  discussion let us call M = ME + MA.  Then MDW
c                  must satisfy MDW.GE.M.  The condition MDW.LT.M
c                  is an error.
c
c                  The array W(*,*) contains the matrices and vectors
c
c                       (E  F)
c                       (A  B)
c
c                  in rows and columns 1,...,M and 1,...,N+1
c                  respectively.  Columns 1,...,L correspond to
c                  unconstrained variables X(1),...,X(L).  The
c                  remaining variables are constrained to be
c                  nonnegative. The condition L.LT.0 or L.GT.N is
c                  an error.
c
c     PRGOPT(*)    This double precision array is the option vector.
c                  If the user is satisfied with the nominal
c                  subprogram features set
c
c                  PRGOPT(1)=1 (or PRGOPT(1)=1.0)
c
c                  Otherwise PRGOPT(*) is a linked list consisting of
c                  groups of data of the following form
c
c                  LINK
c                  KEY
c                  DATA SET
c
c                  The parameters LINK and KEY are each one word.
c                  The DATA SET can be comprised of several words.
c                  The number of items depends on the value of KEY.
c                  The value of LINK points to the first
c                  entry of the next group of data within
c                  PRGOPT(*).  The exception is when there are
c                  no more options to change.  In that
c                  case LINK=1 and the values KEY and DATA SET
c                  are not referenced. The general layout of
c                  PRGOPT(*) is as follows.
c
c               ...PRGOPT(1)=LINK1 (link to first entry of next group)
c               .  PRGOPT(2)=KEY1 (key to the option change)
c               .  PRGOPT(3)=DATA VALUE (data value for this change)
c               .       .
c               .       .
c               .       .
c               ...PRGOPT(LINK1)=LINK2 (link to the first entry of
c               .                       next group)
c               .  PRGOPT(LINK1+1)=KEY2 (key to the option change)
c               .  PRGOPT(LINK1+2)=DATA VALUE
c               ...     .
c               .       .
c               .       .
c               ...PRGOPT(LINK)=1 (no more options to change)
c
c                  Values of LINK that are nonpositive are errors.
c                  A value of LINK.GT.NLINK=100000 is also an error.
c                  This helps prevent using invalid but positive
c                  values of LINK that will probably extend
c                  beyond the program limits of PRGOPT(*).
c                  Unrecognized values of KEY are ignored.  The
c                  order of the options is arbitrary and any number
c                  of options can be changed with the following
c                  restriction.  To prevent cycling in the
c                  processing of the option array a count of the
c                  number of options changed is maintained.
c                  Whenever this count exceeds NOPT=1000 an error
c                  message is printed and the subprogram returns.
c
c                  OPTIONS..
c
c                  KEY=6
c                         Scale the nonzero columns of the
c                  entire data matrix
c                  (E)
c                  (A)
c                  to have length one. The DATA SET for
c                  this option is a single value.  It must
c                  be nonzero if unit length column scaling is
c                  desired.
c
c                  KEY=7
c                         Scale columns of the entire data matrix
c                  (E)
c                  (A)
c                  with a user-provided diagonal matrix.
c                  The DATA SET for this option consists
c                  of the N diagonal scaling factors, one for
c                  each matrix column.
c
c                  KEY=8
c                         Change the rank determination tolerance from
c                  the nominal value of SQRT(SRELPR).  This quantity
c                  can be no smaller than SRELPR, The arithmetic-
c                  storage precision.  The quantity used
c                  here is internally restricted to be at
c                  least SRELPR.  The DATA SET for this option
c                  is the new tolerance.
c
c                  KEY=9
c                         Change the blow-up parameter from the
c                  nominal value of SQRT(SRELPR).  The reciprocal of
c                  this parameter is used in rejecting solution
c                  components as too large when a variable is
c                  first brought into the active set.  Too large
c                  means that the proposed component times the
c                  reciprocal of the parameter is not less than
c                  the ratio of the norms of the right-side
c                  vector and the data matrix.
c                  This parameter can be no smaller than SRELPR,
c                  the arithmetic-storage precision.
c
c                  For example, suppose we want to provide
c                  a diagonal matrix to scale the problem
c                  matrix and change the tolerance used for
c                  determining linear dependence of dropped col
c                  vectors.  For these options the dimensions of
c                  PRGOPT(*) must be at least N+6.  The FORTRAN
c                  statements defining these options would
c                  be as follows.
c
c                  PRGOPT(1)=N+3 (link to entry N+3 in PRGOPT(*))
c                  PRGOPT(2)=7 (user-provided scaling key)
c
c                  CALL xDCOPY(N,D,1,PRGOPT(3),1) (copy the N
c                  scaling factors from a user array called D(*)
c                  into PRGOPT(3)-PRGOPT(N+2))
c
c                  PRGOPT(N+3)=N+6 (link to entry N+6 of PRGOPT(*))
c                  PRGOPT(N+4)=8 (linear dependence tolerance key)
c                  PRGOPT(N+5)=... (new value of the tolerance)
c
c                  PRGOPT(N+6)=1 (no more options to change)
c
c
c     IWORK(1),    The amounts of working storage actually allocated
c     IWORK(2)     for the working arrays WORK(*) and IWORK(*),
c                  respectively.  These quantities are compared with
c                  the actual amounts of storage needed for DWNNLS( ).
c                  Insufficient storage allocated for either WORK(*)
c                  or IWORK(*) is considered an error.  This feature
c                  was included in DWNNLS( ) because miscalculating
c                  the storage formulas for WORK(*) and IWORK(*)
c                  might very well lead to subtle and hard-to-find
c                  execution errors.
c
c                  The length of WORK(*) must be at least
c
c                  LW = ME+MA+5*N
c                  This test will not be made if IWORK(1).LE.0.
c
c                  The length of IWORK(*) must be at least
c
c                  LIW = ME+MA+N
c                  This test will not be made if IWORK(2).LE.0.
c
c     OUTPUT.. All TYPE REAL variables are DOUBLE PRECISION
c
c     X(*)         An array dimensioned at least N, which will
c                  contain the N components of the solution vector
c                  on output.
c
c     RNORM        The residual norm of the solution.  The value of
c                  RNORM contains the residual vector length of the
c                  equality constraints and least squares equations.
c
c     MODE         The value of MODE indicates the success or failure
c                  of the subprogram.
c
c                  MODE = 0  Subprogram completed successfully.
c
c                       = 1  Max. number of iterations (equal to
c                            3*(N-L)) exceeded. Nearly all problems
c                            should complete in fewer than this
c                            number of iterations. An approximate
c                            solution and its corresponding residual
c                            vector length are in X(*) and RNORM.
c
c                       = 2  Usage error occurred.  The offending
c                            condition is noted with the error
c                            processing subprogram, xXERMSG( ).
c
c     User-designated
c     Working arrays..
c
c     WORK(*)      A double precision working array of length at least
c                  M + 5*N.
c
c     IWORK(*)     An integer-valued working array of length at least
c                  M+N.
c
c***REFERENCES  K. H. Haskell and R. J. Hanson, An algorithm for
c                 linear least squares problems with equality and
c                 nonnegativity constraints, Report SAND77-0552, Sandia
c                 Laboratories, June 1978.
c               K. H. Haskell and R. J. Hanson, Selected algorithms for
c                 the linearly constrained least squares problem - a
c                 users guide, Report SAND78-1290, Sandia Laboratories,
c                 August 1979.
c               K. H. Haskell and R. J. Hanson, An algorithm for
c                 linear least squares problems with equality and
c                 nonnegativity constraints, Mathematical Programming
c                 21 (1981), pp. 98-118.
c               R. J. Hanson and K. H. Haskell, Two algorithms for the
c                 linearly constrained least squares problem, ACM
c                 Transactions on Mathematical Software, September 1982.
c               c. L. Lawson and R. J. Hanson, Solving Least Squares
c                 Problems, Prentice-Hall, Inc., 1974.
c***ROUTINES CALLED  DWNLSM, xXERMSG
c***REVISION HISTORY  (YYMMDD)
c   790701  DATE WRITTEN
c   890531  Changed all specific intrinsics to generic.  (WRB)
c   890618  Completely restructured and revised.  (WRB & RWC)
c   891006  REVISION DATE from Version 3.2
c   891214  Prologue converted to Version 4.0 format.  (BAB)
c   900315  CALLs to XERROR changed to CALLs to xXERMSG.  (THJ)
c   900510  Convert XERRWV calls to xXERMSG calls, change Prologue
c           comments to agree with WNNLS.  (RWC)
c   920501  Reformatted the REFERENCES section.  (WRB)
c***END PROLOGUE  DWNNLS
      INTEGER IWORK(*), L, L1, L2, L3, L4, L5, LIW, LW, MA, MDW, ME,                &
     &     MODE, N
      DOUBLE PRECISION  PRGOPT(*), RNORM, W(MDW,*), WORK(*), X(*)
c      CHARACTER(LEN=8) XERN1
c***FIRST EXECUTABLE STATEMENT  DWNNLS
      MODE = 0
      IF (MA+ME.LE.0 .OR. N.LE.0) RETURN
c
      IF (IWORK(1).GT.0) THEN
         LW = ME + MA + 5*N
         IF (IWORK(1).LT.LW) THEN
C KARLINE: REMOVED WRITE		 
         CALL rwarn ('LSEI: insufficient storage')
    
C            WRITE (XERN1, '(I8)') LW
C            CALL xXERMSG ('SLATEC', 'DWNNLS', 'INSUFFICIENT STORAGE ' //            &
C     &         'ALLOCATED FOR WORK(*), NEED LW = ' // XERN1, 2, 1)
            MODE = 2
            RETURN
         ENDIF
      ENDIF
c
      IF (IWORK(2).GT.0) THEN
         LIW = ME + MA + N
         IF (IWORK(2).LT.LIW) THEN
C KARLINE: REMOVED WRITE		 
         CALL rwarn ('LSEI: insufficient storage')
     
C            WRITE (XERN1, '(I8)') LIW
C            CALL xXERMSG ('SLATEC', 'DWNNLS', 'INSUFFICIENT STORAGE ' //            &
C     &         'ALLOCATED FOR IWORK(*), NEED LIW = ' // XERN1, 2, 1)
            MODE = 2
            RETURN
         ENDIF
      ENDIF
c
      IF (MDW.LT.ME+MA) THEN
         CALL xXERMSG ('SLATEC', 'DWNNLS',                                          &
     &      'THE VALUE MDW.LT.ME+MA IS AN ERROR', 2, 1)
         MODE = 2
         RETURN
      ENDIF
c
      IF (L.LT.0 .OR. L.GT.N) THEN
         CALL xXERMSG ('SLATEC', 'DWNNLS',                                          &
     &      'L.GE.0 .AND. L.LE.N IS REQUIRED', 2, 1)
         MODE = 2
         RETURN
      ENDIF
c
c     THE PURPOSE OF THIS SUBROUTINE IS TO BREAK UP THE ARRAYS
c     WORK(*) AND IWORK(*) INTO SEPARATE WORK ARRAYS
c     REQUIRED BY THE MAIN SUBROUTINE DWNLSM( ).
c
      L1 = N + 1
      L2 = L1 + N
      L3 = L2 + ME + MA
      L4 = L3 + N
      L5 = L4 + N
c
      CALL DWNLSM(W, MDW, ME, MA, N, L, PRGOPT, X, RNORM, MODE, IWORK,              &
     &            IWORK(L1), WORK(1), WORK(L1), WORK(L2), WORK(L3),                 &
     &            WORK(L4), WORK(L5))
      RETURN
      END



      SUBROUTINE xDROTM (N,DX,INCX,DY,INCY,DPARAM)
c
c     APPLY THE MODIFIED GIVENS TRANSFORMATION, H, TO THE 2 BY N MATRIX
c
c     (DX**T) , WHERE **T INDICATES TRANSPOSE. THE ELEMENTS OF DX ARE IN
c     (DY**T)
c
c     DX(LX+I*INCX), I = 0 TO N-1, WHERE LX = 1 IF INCX .GE. 0, ELSE
c     LX = (-INCX)*N, AND SIMILARLY FOR SY USING LY AND INCY.
c     WITH DPARAM(1)=DFLAG, H HAS ONE OF THE FOLLOWING FORMS..
c
c     DFLAG=-1.D0     DFLAG=0.D0        DFLAG=1.D0     DFLAG=-2.D0
c
c       (DH11  DH12)    (1.D0  DH12)    (DH11  1.D0)    (1.D0  0.D0)
c     H=(          )    (          )    (          )    (          )
c       (DH21  DH22),   (DH21  1.D0),   (-1.D0 DH22),   (0.D0  1.D0).
c     SEE DROTMG FOR A DESCRIPTION OF DATA STORAGE IN DPARAM.
c
      DOUBLE PRECISION DFLAG,DH12,DH22,DX,TWO,Z,DH11,DH21,                          &
     & DPARAM,DY,W,ZERO
      INTEGER N,INCX,INCY,NSTEPS,I,KX,KY
      DIMENSION DX(*),DY(*),DPARAM(5)
      DATA ZERO,TWO/0.D0,2.D0/
c
      DFLAG=DPARAM(1)
      IF(N .LE. 0 .OR.(DFLAG+TWO.EQ.ZERO)) GO TO 140
          IF(.NOT.(INCX.EQ.INCY.AND. INCX .GT.0)) GO TO 70
c
               NSTEPS=N*INCX
               IF (DFLAG .LT. 0) THEN
                 GOTO 50
               ELSE IF (DFLAG .EQ. 0) THEN
                 GOTO 10
               ELSE
                 GOTO 30
               ENDIF                    
C               IF(DFLAG) 50,10,30
   10          CONTINUE
               DH12=DPARAM(4)
               DH21=DPARAM(3)
                    DO 20 I=1,NSTEPS,INCX
                    W=DX(I)
                    Z=DY(I)
                    DX(I)=W+Z*DH12
                    DY(I)=W*DH21+Z
   20               CONTINUE
               GO TO 140
   30          CONTINUE
               DH11=DPARAM(2)
               DH22=DPARAM(5)
                    DO 40 I=1,NSTEPS,INCX
                    W=DX(I)
                    Z=DY(I)
                    DX(I)=W*DH11+Z
                    DY(I)=-W+DH22*Z
   40               CONTINUE
               GO TO 140
   50          CONTINUE
               DH11=DPARAM(2)
               DH12=DPARAM(4)
               DH21=DPARAM(3)
               DH22=DPARAM(5)
                    DO 60 I=1,NSTEPS,INCX
                    W=DX(I)
                    Z=DY(I)
                    DX(I)=W*DH11+Z*DH12
                    DY(I)=W*DH21+Z*DH22
   60               CONTINUE
               GO TO 140
   70     CONTINUE
          KX=1
          KY=1
          IF(INCX .LT. 0) KX=1+(1-N)*INCX
          IF(INCY .LT. 0) KY=1+(1-N)*INCY
c
          IF (DFLAG .LT. 0) THEN
             GOTO 120
          ELSE IF (DFLAG .EQ. 0) THEN
             GOTO 80
          ELSE
             GOTO 100
          ENDIF                    

C          IF(DFLAG)120,80,100
   80     CONTINUE
          DH12=DPARAM(4)
          DH21=DPARAM(3)
               DO 90 I=1,N
               W=DX(KX)
               Z=DY(KY)
               DX(KX)=W+Z*DH12
               DY(KY)=W*DH21+Z
               KX=KX+INCX
               KY=KY+INCY
   90          CONTINUE
          GO TO 140
  100     CONTINUE
          DH11=DPARAM(2)
          DH22=DPARAM(5)
               DO 110 I=1,N
               W=DX(KX)
               Z=DY(KY)
               DX(KX)=W*DH11+Z
               DY(KY)=-W+DH22*Z
               KX=KX+INCX
               KY=KY+INCY
  110          CONTINUE
          GO TO 140
  120     CONTINUE
          DH11=DPARAM(2)
          DH12=DPARAM(4)
          DH21=DPARAM(3)
          DH22=DPARAM(5)
               DO 130 I=1,N
               W=DX(KX)
               Z=DY(KY)
               DX(KX)=W*DH11+Z*DH12
               DY(KY)=W*DH21+Z*DH22
               KX=KX+INCX
               KY=KY+INCY
  130          CONTINUE
  140     CONTINUE
          RETURN
          END





      SUBROUTINE xDROTMG (DD1,DD2,DX1,DY1,DPARAM)
c
c     CONSTRUCT THE MODIFIED GIVENS TRANSFORMATION MATRIX H WHICH ZEROS
c     THE SECOND COMPONENT OF THE 2-VECTOR  (DSQRT(DD1)*DX1,DSQRT(DD2)*
c     DY2)**T.
c     WITH DPARAM(1)=DFLAG, H HAS ONE OF THE FOLLOWING FORMS..
c
c     DFLAG=-1.D0     DFLAG=0.D0        DFLAG=1.D0     DFLAG=-2.D0
c
c       (DH11  DH12)    (1.D0  DH12)    (DH11  1.D0)    (1.D0  0.D0)
c     H=(          )    (          )    (          )    (          )
c       (DH21  DH22),   (DH21  1.D0),   (-1.D0 DH22),   (0.D0  1.D0).
c     LOCATIONS 2-4 OF DPARAM CONTAIN DH11, DH21, DH12, AND DH22
c     RESPECTIVELY. (VALUES OF 1.D0, -1.D0, OR 0.D0 IMPLIED BY THE
c     VALUE OF DPARAM(1) ARE NOT STORED IN DPARAM.)
c
c     THE VALUES OF GAMSQ AND RGAMSQ SET IN THE DATA STATEMENT MAY BE
c     INEXACT.  THIS IS OK AS THEY ARE ONLY USED FOR TESTING THE SIZE
c     OF DD1 AND DD2.  ALL ACTUAL SCALING OF DATA IS DONE USING GAM.
c
      DOUBLE PRECISION GAM,ONE,RGAMSQ,DD2,DH11,DH21,DPARAM,DP2,                     &
     & DQ2,DU,DY1,ZERO,GAMSQ,DD1,DFLAG,DH12,DH22,DP1,DQ1,                           &
     & DTEMP,DX1,TWO
      INTEGER IGO
      DIMENSION DPARAM(5)
c
      DATA ZERO,ONE,TWO /0.D0,1.D0,2.D0/
      DATA GAM,GAMSQ,RGAMSQ/4096.D0,16777216.D0,5.9604645D-8/
      IF(.NOT. DD1 .LT. ZERO) GO TO 10
c       GO ZERO-H-D-AND-DX1..
          GO TO 60
   10 CONTINUE
c     CASE-DD1-NONNEGATIVE
      DP2=DD2*DY1
      IF(.NOT. DP2 .EQ. ZERO) GO TO 20
          DFLAG=-TWO
          GO TO 260
c     REGULAR-CASE..
   20 CONTINUE
      DP1=DD1*DX1
      DQ2=DP2*DY1
      DQ1=DP1*DX1
c
      IF(.NOT. DABS(DQ1) .GT. DABS(DQ2)) GO TO 40
          DH21=-DY1/DX1
          DH12=DP2/DP1
c
          DU=ONE-DH12*DH21
c
          IF(.NOT. DU .LE. ZERO) GO TO 30
c         GO ZERO-H-D-AND-DX1..
               GO TO 60
   30     CONTINUE
               DFLAG=ZERO
               DD1=DD1/DU
               DD2=DD2/DU
               DX1=DX1*DU
c         GO SCALE-CHECK..
               GO TO 100
   40 CONTINUE
          IF(.NOT. DQ2 .LT. ZERO) GO TO 50
c         GO ZERO-H-D-AND-DX1..
               GO TO 60
   50     CONTINUE
               DFLAG=ONE
               DH11=DP1/DP2
               DH22=DX1/DY1
               DU=ONE+DH11*DH22
               DTEMP=DD2/DU
               DD2=DD1/DU
               DD1=DTEMP
               DX1=DY1*DU
c         GO SCALE-CHECK
               GO TO 100
c     PROCEDURE..ZERO-H-D-AND-DX1..
   60 CONTINUE
          DFLAG=-ONE
          DH11=ZERO
          DH12=ZERO
          DH21=ZERO
          DH22=ZERO
c
          DD1=ZERO
          DD2=ZERO
          DX1=ZERO
c         RETURN..
          GO TO 220
c     PROCEDURE..FIX-H..
   70 CONTINUE
      IF(.NOT. DFLAG .GE. ZERO) GO TO 90
c
          IF(.NOT. DFLAG .EQ. ZERO) GO TO 80
          DH11=ONE
          DH22=ONE
          DFLAG=-ONE
          GO TO 90
   80     CONTINUE
          DH21=-ONE
          DH12=ONE
          DFLAG=-ONE
   90 CONTINUE
C      GO TO IGO,(120,150,180,210)  CHANGED INTO
      SELECT CASE (IGO)

        CASE (120)
          GOTO 120
        CASE (150)
          GOTO 150
        CASE (180)
          GOTO 180
        CASE (210)
          GOTO 210
                   
      END SELECT
      

C karline: IGO DOES NOT HAVE A VALUE ?
c     PROCEDURE..SCALE-CHECK
  100 CONTINUE
  110     CONTINUE
          IF(.NOT. DD1 .LE. RGAMSQ) GO TO 130
               IF(DD1 .EQ. ZERO) GO TO 160
C               ASSIGN 120 TO IGO  CHANGED INTO:
               IGO = 120
c              FIX-H..
               GO TO 70
  120          CONTINUE
               DD1=DD1*GAM**2
               DX1=DX1/GAM
               DH11=DH11/GAM
               DH12=DH12/GAM
          GO TO 110
  130 CONTINUE
  140     CONTINUE
          IF(.NOT. DD1 .GE. GAMSQ) GO TO 160
C               ASSIGN 150 TO IGO   CAHNGED INTO
               IGO = 150
c              FIX-H..
               GO TO 70
  150          CONTINUE
               DD1=DD1/GAM**2
               DX1=DX1*GAM
               DH11=DH11*GAM
               DH12=DH12*GAM
          GO TO 140
  160 CONTINUE
  170     CONTINUE
          IF(.NOT. DABS(DD2) .LE. RGAMSQ) GO TO 190
               IF(DD2 .EQ. ZERO) GO TO 220
C               ASSIGN 180 TO IGO  CHANGED INTO
               IGO = 180
c              FIX-H..
               GO TO 70
  180          CONTINUE
               DD2=DD2*GAM**2
               DH21=DH21/GAM
               DH22=DH22/GAM
          GO TO 170
  190 CONTINUE
  200     CONTINUE
          IF(.NOT. DABS(DD2) .GE. GAMSQ) GO TO 220
C               ASSIGN 210 TO IGO  CHANGED INTO
               IGO = 210

c              FIX-H..
               GO TO 70
  210          CONTINUE
               DD2=DD2/GAM**2
               DH21=DH21*GAM
               DH22=DH22*GAM
          GO TO 200
  220 CONTINUE
          IF (DFLAG .LT. 0) THEN
             GOTO 250
          ELSE IF (DFLAG .EQ. 0) THEN
             GOTO 230
          ELSE
             GOTO 240
          ENDIF                    

C          IF(DFLAG)250,230,240
  230     CONTINUE
               DPARAM(3)=DH21
               DPARAM(4)=DH12
               GO TO 260
  240     CONTINUE
               DPARAM(2)=DH11
               DPARAM(5)=DH22
               GO TO 260
  250     CONTINUE
               DPARAM(2)=DH11
               DPARAM(3)=DH21
               DPARAM(4)=DH12
               DPARAM(5)=DH22
  260 CONTINUE
          DPARAM(1)=DFLAG
          RETURN
      END



      SUBROUTINE DLPDP (A, MDA, M, N1, N2, PRGOPT, X, WNORM, MODE, WS,              &
     &   IS)
c***BEGIN PROLOGUE  DLPDP
c***SUBSIDIARY
c***PURPOSE  Subsidiary to xDLSEI
c***LIBRARY   SLATEC
c***TYPE      DOUBLE PRECISION (LPDP-S, DLPDP-D)
c***AUTHOR  Hanson, R. J., (SNLA)
c           Haskell, K. H., (SNLA)
c***DESCRIPTION
c
c  **** Double Precision version of LPDP ****
c     DIMENSION A(MDA,N+1),PRGOPT(*),X(N),WS((M+2)*(N+7)),IS(M+N+1),
c     where N=N1+N2.  This is a slight overestimate for WS(*).
c
c     Determine an N1-vector W, and
c               an N2-vector Z
c     which minimizes the Euclidean length of W
c     subject to G*W+H*Z .GE. Y.
c     This is the least projected distance problem, LPDP.
c     The matrices G and H are of respective
c     dimensions M by N1 and M by N2.
c
c     Called by subprogram DLSI( ).
c
c     The matrix
c                (G H Y)
c
c     occupies rows 1,...,M and cols 1,...,N1+N2+1 of A(*,*).
c
c     The solution (W) is returned in X(*).
c                  (Z)
c
c     The value of MODE indicates the status of
c     the computation after returning to the user.
c
c          MODE=1  The solution was successfully obtained.
c
c          MODE=2  The inequalities are inconsistent.
c
c***SEE ALSO  xDLSEI
c***ROUTINES CALLED  xDCOPY, xDDOT, xDNRM2, xDSCAL, DWNNLS
c***REVISION HISTORY  (YYMMDD)
c   790701  DATE WRITTEN
c   890531  Changed all specific intrinsics to generic.  (WRB)
c   891214  Prologue converted to Version 4.0 format.  (BAB)
c   900328  Added TYPE section.  (WRB)
c   910408  Updated the AUTHOR section.  (WRB)
c***END PROLOGUE  DLPDP
c
      INTEGER I, IS(*), IW, IX, J, L, M, MDA, MODE, MODEW, N, N1, N2,               &
     &     NP1
      DOUBLE PRECISION A(MDA,*), xDDOT, xDNRM2, FAC, ONE,                             &
     &     PRGOPT(*), RNORM, SC, WNORM, WS(*), X(*), YNORM, ZERO
      SAVE ZERO, ONE, FAC
      DATA ZERO,ONE /0.0D0,1.0D0/, FAC /0.1D0/
c***FIRST EXECUTABLE STATEMENT  DLPDP
      N = N1 + N2
      MODE = 1
      IF (M .GT. 0) GO TO 20
         IF (N .LE. 0) GO TO 10
            X(1) = ZERO
            CALL xDCOPY(N,X,0,X,1)
   10    CONTINUE
         WNORM = ZERO
      GO TO 200
   20 CONTINUE
c        BEGIN BLOCK PERMITTING ...EXITS TO 190
            NP1 = N + 1
c
c           SCALE NONZERO ROWS OF INEQUALITY MATRIX TO HAVE LENGTH ONE.
            DO 40 I = 1, M
               SC = xDNRM2(N,A(I,1),MDA)
               IF (SC .EQ. ZERO) GO TO 30
                  SC = ONE/SC
                  CALL xDSCAL(NP1,SC,A(I,1),MDA)
   30          CONTINUE
   40       CONTINUE
c
c           SCALE RT.-SIDE VECTOR TO HAVE LENGTH ONE (OR ZERO).
            YNORM = xDNRM2(M,A(1,NP1),1)
            IF (YNORM .EQ. ZERO) GO TO 50
               SC = ONE/YNORM
               CALL xDSCAL(M,SC,A(1,NP1),1)
   50       CONTINUE
c
c           SCALE COLS OF MATRIX H.
            J = N1 + 1
   60       IF (J .GT. N) GO TO 70
               SC = xDNRM2(M,A(1,J),1)
               IF (SC .NE. ZERO) SC = ONE/SC
               CALL xDSCAL(M,SC,A(1,J),1)
               X(J) = SC
               J = J + 1
            GO TO 60
   70       CONTINUE
            IF (N1 .LE. 0) GO TO 130
c
c              COPY TRANSPOSE OF (H G Y) TO WORK ARRAY WS(*).
               IW = 0
               DO 80 I = 1, M
c
c                 MOVE COL OF TRANSPOSE OF H INTO WORK ARRAY.
                  CALL xDCOPY(N2,A(I,N1+1),MDA,WS(IW+1),1)
                  IW = IW + N2
c
c                 MOVE COL OF TRANSPOSE OF G INTO WORK ARRAY.
                  CALL xDCOPY(N1,A(I,1),MDA,WS(IW+1),1)
                  IW = IW + N1
c
c                 MOVE COMPONENT OF VECTOR Y INTO WORK ARRAY.
                  WS(IW+1) = A(I,NP1)
                  IW = IW + 1
   80          CONTINUE
               WS(IW+1) = ZERO
               CALL xDCOPY(N,WS(IW+1),0,WS(IW+1),1)
               IW = IW + N
               WS(IW+1) = ONE
               IW = IW + 1
c
c              SOLVE EU=F SUBJECT TO (TRANSPOSE OF H)U=0, U.GE.0.  THE
c              MATRIX E = TRANSPOSE OF (G Y), AND THE (N+1)-VECTOR
c              F = TRANSPOSE OF (0,...,0,1).
               IX = IW + 1
               IW = IW + M
c
c              DO NOT CHECK LENGTHS OF WORK ARRAYS IN THIS USAGE OF
c              DWNNLS( ).
               IS(1) = 0
               IS(2) = 0
               CALL DWNNLS(WS,NP1,N2,NP1-N2,M,0,PRGOPT,WS(IX),RNORM,                &
     &                     MODEW,IS,WS(IW+1))
c
c              COMPUTE THE COMPONENTS OF THE SOLN DENOTED ABOVE BY W.
               SC = ONE - xDDOT(M,A(1,NP1),1,WS(IX),1)
               IF (ONE + FAC*ABS(SC) .EQ. ONE .OR. RNORM .LE. ZERO)                 &
     &            GO TO 110
                  SC = ONE/SC
                  DO 90 J = 1, N1
                     X(J) = SC*xDDOT(M,A(1,J),1,WS(IX),1)
   90             CONTINUE
c
c                 COMPUTE THE VECTOR Q=Y-GW.  OVERWRITE Y WITH THIS
c                 VECTOR.
                  DO 100 I = 1, M
                     A(I,NP1) = A(I,NP1) - xDDOT(N1,A(I,1),MDA,X,1)
  100             CONTINUE
               GO TO 120
  110          CONTINUE
                  MODE = 2
c        .........EXIT
                  GO TO 190
  120          CONTINUE
  130       CONTINUE
            IF (N2 .LE. 0) GO TO 180
c
c              COPY TRANSPOSE OF (H Q) TO WORK ARRAY WS(*).
               IW = 0
               DO 140 I = 1, M
                  CALL xDCOPY(N2,A(I,N1+1),MDA,WS(IW+1),1)
                  IW = IW + N2
                  WS(IW+1) = A(I,NP1)
                  IW = IW + 1
  140          CONTINUE
               WS(IW+1) = ZERO
               CALL xDCOPY(N2,WS(IW+1),0,WS(IW+1),1)
               IW = IW + N2
               WS(IW+1) = ONE
               IW = IW + 1
               IX = IW + 1
               IW = IW + M
c
c              SOLVE RV=S SUBJECT TO V.GE.0.  THE MATRIX R =(TRANSPOSE
c              OF (H Q)), WHERE Q=Y-GW.  THE (N2+1)-VECTOR S =(TRANSPOSE
c              OF (0,...,0,1)).
c
c              DO NOT CHECK LENGTHS OF WORK ARRAYS IN THIS USAGE OF
c              DWNNLS( ).
               IS(1) = 0
               IS(2) = 0
               CALL DWNNLS(WS,N2+1,0,N2+1,M,0,PRGOPT,WS(IX),RNORM,MODEW,            &
     &                     IS,WS(IW+1))
c
c              COMPUTE THE COMPONENTS OF THE SOLN DENOTED ABOVE BY Z.
               SC = ONE - xDDOT(M,A(1,NP1),1,WS(IX),1)
               IF (ONE + FAC*ABS(SC) .EQ. ONE .OR. RNORM .LE. ZERO)                 &
     &            GO TO 160
                  SC = ONE/SC
                  DO 150 J = 1, N2
                     L = N1 + J
                     X(L) = SC*xDDOT(M,A(1,L),1,WS(IX),1)*X(L)
  150             CONTINUE
               GO TO 170
  160          CONTINUE
                  MODE = 2
c        .........EXIT
                  GO TO 190
  170          CONTINUE
  180       CONTINUE
c
c           ACCOUNT FOR SCALING OF RT.-SIDE VECTOR IN SOLUTION.
            CALL xDSCAL(N,YNORM,X,1)
            WNORM = xDNRM2(N1,X,1)
  190    CONTINUE
  200 CONTINUE
      RETURN
      END




      SUBROUTINE xDH12 (MODE, LPIVOT, L1, M, U, IUE, UP, C, ICE, ICV,                &
     &   NCV)
c***BEGIN PROLOGUE  DH12
c***SUBSIDIARY
c***PURPOSE  Subsidiary to xDHFTI, xDLSEI and DWNNLS
c***LIBRARY   SLATEC
c***TYPE      DOUBLE PRECISION (H12-S, DH12-D)
c***AUTHOR  (UNKNOWN)
c***DESCRIPTION
c
c      &** DOUBLE PRECISION VERSION OF H12 ******
c
c     C.L.Lawson and R.J.Hanson, Jet Propulsion Laboratory, 1973 Jun 12
c     to appear in 'Solving Least Squares Problems', Prentice-Hall, 1974
c
c     Construction and/or application of a single
c     Householder transformation..     Q = I + U*(U**T)/B
c
c     MODE    = 1 or 2   to select algorithm  H1  or  H2 .
c     LPIVOT is the index of the pivot element.
c     L1,M   If L1 .LE. M   the transformation will be constructed to
c            zero elements indexed from L1 through M.   If L1 GT. M
c            THE SUBROUTINE DOES AN IDENTITY TRANSFORMATION.
c     U(),IUE,UP    On entry to H1 U() contains the pivot vector.
c                   IUE is the storage increment between elements.
c                                       On exit from H1 U() and UP
c                   contain quantities defining the vector U of the
c                   Householder transformation.   On entry to H2 U()
c                   and UP should contain quantities previously computed
c                   by H1.  These will not be modified by H2.
c     C()    On entry to H1 or H2 C() contains a matrix which will be
c            regarded as a set of vectors to which the Householder
c            transformation is to be applied.  On exit C() contains the
c            set of transformed vectors.
c     ICE    Storage increment between elements of vectors in C().
c     ICV    Storage increment between vectors in C().
c     NCV    Number of vectors in C() to be transformed. If NCV .LE. 0
c            no operations will be done on C().
c
c***SEE ALSO  xDHFTI, xDLSEI, DWNNLS
c***ROUTINES CALLED  xDAXPY, xDDOT, xDSWAP
c***REVISION HISTORY  (YYMMDD)
c   790101  DATE WRITTEN
c   890531  Changed all specific intrinsics to generic.  (WRB)
c   890831  Modified array declarations.  (WRB)
c   891214  Prologue converted to Version 4.0 format.  (BAB)
c   900328  Added TYPE section.  (WRB)
c   900911  Added xDDOT to DOUBLE PRECISION statement.  (WRB)
c***END PROLOGUE  DH12
      INTEGER I, I2, I3, I4, ICE, ICV, INCR, IUE, J, KL1, KL2, KLP,                 &
     &     L1, L1M1, LPIVOT, M, MML1P2, MODE, NCV
      DOUBLE PRECISION B, C, CL, CLINV, ONE, UL1M1, SM, U, UP, xDDOT
      DIMENSION U(IUE,*), C(*)
c     BEGIN BLOCK PERMITTING ...EXITS TO 140
c***FIRST EXECUTABLE STATEMENT  DH12
         ONE = 1.0D0
c
c     ...EXIT
         IF (0 .GE. LPIVOT .OR. LPIVOT .GE. L1 .OR. L1 .GT. M) GO TO 140
         CL = ABS(U(1,LPIVOT))
         IF (MODE .EQ. 2) GO TO 40
c           &***** CONSTRUCT THE TRANSFORMATION. ******
            DO 10 J = L1, M
               CL = MAX(ABS(U(1,J)),CL)
   10       CONTINUE
            IF (CL .GT. 0.0D0) GO TO 20
c     .........EXIT
               GO TO 140
   20       CONTINUE
            CLINV = ONE/CL
            SM = (U(1,LPIVOT)*CLINV)**2
            DO 30 J = L1, M
               SM = SM + (U(1,J)*CLINV)**2
   30       CONTINUE
            CL = CL*SQRT(SM)
            IF (U(1,LPIVOT) .GT. 0.0D0) CL = -CL
            UP = U(1,LPIVOT) - CL
            U(1,LPIVOT) = CL
         GO TO 50
   40    CONTINUE
c        &***** APPLY THE TRANSFORMATION  I+U*(U**T)/B  TO C. ******
c
         IF (CL .GT. 0.0D0) GO TO 50
c     ......EXIT
            GO TO 140
   50    CONTINUE
c     ...EXIT
         IF (NCV .LE. 0) GO TO 140
         B = UP*U(1,LPIVOT)
c        B  MUST BE NONPOSITIVE HERE.  IF B = 0., RETURN.
c
         IF (B .LT. 0.0D0) GO TO 60
c     ......EXIT
            GO TO 140
   60    CONTINUE
         B = ONE/B
         MML1P2 = M - L1 + 2
         IF (MML1P2 .LE. 20) GO TO 80
            L1M1 = L1 - 1
            KL1 = 1 + (L1M1 - 1)*ICE
            KL2 = KL1
            KLP = 1 + (LPIVOT - 1)*ICE
            UL1M1 = U(1,L1M1)
            U(1,L1M1) = UP
            IF (LPIVOT .NE. L1M1) CALL xDSWAP(NCV,C(KL1),ICV,C(KLP),ICV)
            DO 70 J = 1, NCV
               SM = xDDOT(MML1P2,U(1,L1M1),IUE,C(KL1),ICE)
               SM = SM*B
               CALL xDAXPY(MML1P2,SM,U(1,L1M1),IUE,C(KL1),ICE)
               KL1 = KL1 + ICV
   70       CONTINUE
            U(1,L1M1) = UL1M1
c     ......EXIT
            IF (LPIVOT .EQ. L1M1) GO TO 140
            KL1 = KL2
            CALL xDSWAP(NCV,C(KL1),ICV,C(KLP),ICV)
         GO TO 130
   80    CONTINUE
            I2 = 1 - ICV + ICE*(LPIVOT - 1)
            INCR = ICE*(L1 - LPIVOT)
            DO 120 J = 1, NCV
               I2 = I2 + ICV
               I3 = I2 + INCR
               I4 = I3
               SM = C(I2)*UP
               DO 90 I = L1, M
                  SM = SM + C(I3)*U(1,I)
                  I3 = I3 + ICE
   90          CONTINUE
               IF (SM .EQ. 0.0D0) GO TO 110
                  SM = SM*B
                  C(I2) = C(I2) + SM*UP
                  DO 100 I = L1, M
                     C(I4) = C(I4) + SM*U(1,I)
                     I4 = I4 + ICE
  100             CONTINUE
  110          CONTINUE
  120       CONTINUE
  130    CONTINUE
  140 CONTINUE
      RETURN
      END




      SUBROUTINE xDHFTI (A, MDA, M, N, B, MDB, NB, TAU, KRANK, RNORM, H,             &
     &   G, IP)
c***BEGIN PROLOGUE  xDHFTI
c***PURPOSE  Solve a least squares problem for banded matrices using
c            sequential accumulation of rows of the data matrix.
c            Exactly one right-hand side vector is permitted.
c***LIBRARY   SLATEC
c***CATEGORY  D9
c***TYPE      DOUBLE PRECISION (HFTI-S, xDHFTI-D)
c***KEYWORDS  CURVE FITTING, LEAST SQUARES
c***AUTHOR  Lawson, C. L., (JPL)
c           Hanson, R. J., (SNLA)
c***DESCRIPTION
c
c     DIMENSION A(MDA,N),(B(MDB,NB) or B(M)),RNORM(NB),H(N),G(N),IP(N)
c
c     This subroutine solves a linear least squares problem or a set of
c     linear least squares problems having the same matrix but different
c     right-side vectors.  The problem data consists of an M by N matrix
c     A, an M by NB matrix B, and an absolute tolerance parameter TAU
c     whose usage is described below.  The NB column vectors of B
c     represent right-side vectors for NB distinct linear least squares
c     problems.
c
c     This set of problems can also be written as the matrix least
c     squares problem
c
c                       AX = B,
c
c     where X is the N by NB solution matrix.
c
c     Note that if B is the M by M identity matrix, then X will be the
c     pseudo-inverse of A.
c
c     This subroutine first transforms the augmented matrix (A B) to a
c     matrix (R C) using premultiplying Householder transformations with
c     column interchanges.  All subdiagonal elements in the matrix R are
c     zero and its diagonal elements satisfy
c
c                       ABS(R(I,I)).GE.ABS(R(I+1,I+1)),
c
c                       I = 1,...,L-1, where
c
c                       L = MIN(M,N).
c
c     The subroutine will compute an integer, KRANK, equal to the number
c     of diagonal terms of R that exceed TAU in magnitude. Then a
c     solution of minimum Euclidean length is computed using the first
c     KRANK rows of (R C).
c
c     To be specific we suggest that the user consider an easily
c     computable matrix norm, such as, the maximum of all column sums of
c     magnitudes.
c
c     Now if the relative uncertainty of B is EPS, (norm of uncertainty/
c     norm of B), it is suggested that TAU be set approximately equal to
c     EPS*(norm of A).
c
c     The user must dimension all arrays appearing in the call list..
c     A(MDA,N),(B(MDB,NB) or B(M)),RNORM(NB),H(N),G(N),IP(N).  This
c     permits the solution of a range of problems in the same array
c     space.
c
c     The entire set of parameters for xDHFTI are
c
c     INPUT.. All TYPE REAL variables are DOUBLE PRECISION
c
c     A(*,*),MDA,M,N    The array A(*,*) initially contains the M by N
c                       matrix A of the least squares problem AX = B.
c                       The first dimensioning parameter of the array
c                       A(*,*) is MDA, which must satisfy MDA.GE.M
c                       Either M.GE.N or M.LT.N is permitted.  There
c                       is no restriction on the rank of A.  The
c                       condition MDA.LT.M is considered an error.
c
c     B(*),MDB,NB       If NB = 0 the subroutine will perform the
c                       orthogonal decomposition but will make no
c                       references to the array B(*).  If NB.GT.0
c                       the array B(*) must initially contain the M by
c                       NB matrix B of the least squares problem AX =
c                       B.  If NB.GE.2 the array B(*) must be doubly
c                       subscripted with first dimensioning parameter
c                       MDB.GE.MAX(M,N).  If NB = 1 the array B(*) may
c                       be either doubly or singly subscripted.  In
c                       the latter case the value of MDB is arbitrary
c                       but it should be set to some valid integer
c                       value such as MDB = M.
c
c                       The condition of NB.GT.1.AND.MDB.LT. MAX(M,N)
c                       is considered an error.
c
c     TAU               Absolute tolerance parameter provided by user
c                       for pseudorank determination.
c
c     H(*),G(*),IP(*)   Arrays of working space used by xDHFTI.
c
c     OUTPUT.. All TYPE REAL variables are DOUBLE PRECISION
c
c     A(*,*)            The contents of the array A(*,*) will be
c                       modified by the subroutine. These contents
c                       are not generally required by the user.
c
c     B(*)              On return the array B(*) will contain the N by
c                       NB solution matrix X.
c
c     KRANK             Set by the subroutine to indicate the
c                       pseudorank of A.
c
c     RNORM(*)          On return, RNORM(J) will contain the Euclidean
c                       norm of the residual vector for the problem
c                       defined by the J-th column vector of the array
c                       B(*,*) for J = 1,...,NB.
c
c     H(*),G(*)         On return these arrays respectively contain
c                       elements of the pre- and post-multiplying
c                       Householder transformations used to compute
c                       the minimum Euclidean length solution.
c
c     IP(*)             Array in which the subroutine records indices
c                       describing the permutation of column vectors.
c                       The contents of arrays H(*),G(*) and IP(*)
c                       are not generally required by the user.
c
c***REFERENCES  C. L. Lawson and R. J. Hanson, Solving Least Squares
c                 Problems, Prentice-Hall, Inc., 1974, Chapter 14.
c***ROUTINES CALLED  D1MACH, xDH12, xXERMSG
c***REVISION HISTORY  (YYMMDD)
c   790101  DATE WRITTEN
c   890531  Changed all specific intrinsics to generic.  (WRB)
c   891006  Cosmetic changes to prologue.  (WRB)
c   891006  REVISION DATE from Version 3.2
c   891214  Prologue converted to Version 4.0 format.  (BAB)
c   900315  CALLs to XERROR changed to CALLs to xXERMSG.  (THJ)
c   901005  Replace usage of DDIFF with usage of D1MACH.  (RWC)
c   920501  Reformatted the REFERENCES section.  (WRB)
c***END PROLOGUE  xDHFTI
      INTEGER I, II, IOPT, IP(*), IP1, J, JB, JJ, K, KP1, KRANK, L,                 &
     &     LDIAG, LMAX, M, MDA, MDB, N, NB, NERR
      DOUBLE PRECISION A, B, D1MACH, DZERO, FACTOR,                                 &
     &     G, H, HMAX, RELEPS, RNORM, SM, SM1, SZERO, TAU, TMP
      DIMENSION A(MDA,*),B(MDB,*),H(*),G(*),RNORM(*)
      SAVE RELEPS
      DATA RELEPS /0.D0/
c     BEGIN BLOCK PERMITTING ...EXITS TO 360
c***FIRST EXECUTABLE STATEMENT  xDHFTI
         IF (RELEPS.EQ.0.D0) RELEPS = D1MACH(4)
         SZERO = 0.0D0
         DZERO = 0.0D0
         FACTOR = 0.001D0
c
         K = 0
         LDIAG = MIN(M,N)
         IF (LDIAG .LE. 0) GO TO 350
c           BEGIN BLOCK PERMITTING ...EXITS TO 130
c              BEGIN BLOCK PERMITTING ...EXITS TO 120
                  IF (MDA .GE. M) GO TO 10
                     NERR = 1
                     IOPT = 2
                     CALL xXERMSG ('SLATEC', 'xDHFTI',                               &
     &                  'MDA.LT.M, PROBABLE ERROR.',                                &
     &                  NERR, IOPT)
c     ...............EXIT
                     GO TO 360
   10             CONTINUE
c
                  IF (NB .LE. 1 .OR. MAX(M,N) .LE. MDB) GO TO 20
                     NERR = 2
                     IOPT = 2
                     CALL xXERMSG ('SLATEC', 'xDHFTI',                               &
     &                  'MDB.LT.MAX(M,N).AND.NB.GT.1. PROBABLE ERROR.',             &
     &                  NERR, IOPT)
c     ...............EXIT
                     GO TO 360
   20             CONTINUE
c
                  DO 100 J = 1, LDIAG
c                    BEGIN BLOCK PERMITTING ...EXITS TO 70
                        IF (J .EQ. 1) GO TO 40
c
c                           UPDATE SQUARED COLUMN LENGTHS AND FIND LMAX
c                          ..
                           LMAX = J
                           DO 30 L = J, N
                              H(L) = H(L) - A(J-1,L)**2
                              IF (H(L) .GT. H(LMAX)) LMAX = L
   30                      CONTINUE
c                    ......EXIT
                           IF (FACTOR*H(LMAX) .GT. HMAX*RELEPS) GO TO 70
   40                   CONTINUE
c
c                        COMPUTE SQUARED COLUMN LENGTHS AND FIND LMAX
c                       ..
                        LMAX = J
                        DO 60 L = J, N
                           H(L) = 0.0D0
                           DO 50 I = J, M
                              H(L) = H(L) + A(I,L)**2
   50                      CONTINUE
                           IF (H(L) .GT. H(LMAX)) LMAX = L
   60                   CONTINUE
                        HMAX = H(LMAX)
   70                CONTINUE
c                    ..
c                     LMAX HAS BEEN DETERMINED
c
c                     DO COLUMN INTERCHANGES IF NEEDED.
c                    ..
                     IP(J) = LMAX
                     IF (IP(J) .EQ. J) GO TO 90
                        DO 80 I = 1, M
                           TMP = A(I,J)
                           A(I,J) = A(I,LMAX)
                           A(I,LMAX) = TMP
   80                   CONTINUE
                        H(LMAX) = H(J)
   90                CONTINUE
c
c                     COMPUTE THE J-TH TRANSFORMATION AND APPLY IT TO A
c                     AND B.
c                    ..
                     CALL xDH12(1,J,J+1,M,A(1,J),1,H(J),A(1,J+1),1,MDA,              &
     &                         N-J)
                     CALL xDH12(2,J,J+1,M,A(1,J),1,H(J),B,1,MDB,NB)
  100             CONTINUE
c
c                  DETERMINE THE PSEUDORANK, K, USING THE TOLERANCE,
c                  TAU.
c                 ..
                  DO 110 J = 1, LDIAG
c              ......EXIT
                     IF (ABS(A(J,J)) .LE. TAU) GO TO 120
  110             CONTINUE
                  K = LDIAG
c           ......EXIT
                  GO TO 130
  120          CONTINUE
               K = J - 1
  130       CONTINUE
            KP1 = K + 1
c
c           COMPUTE THE NORMS OF THE RESIDUAL VECTORS.
c
            IF (NB .LT. 1) GO TO 170
            DO 160 JB = 1, NB
               TMP = SZERO
               IF (M .LT. KP1) GO TO 150
               DO 140 I = KP1, M
                  TMP = TMP + B(I,JB)**2
  140          CONTINUE
  150          CONTINUE
               RNORM(JB) = SQRT(TMP)
  160       CONTINUE
  170       CONTINUE
c           SPECIAL FOR PSEUDORANK = 0
            IF (K .GT. 0) GO TO 210
               IF (NB .LT. 1) GO TO 200
               DO 190 JB = 1, NB
                  DO 180 I = 1, N
                     B(I,JB) = SZERO
  180             CONTINUE
  190          CONTINUE
  200          CONTINUE
            GO TO 340
  210       CONTINUE
c
c               IF THE PSEUDORANK IS LESS THAN N COMPUTE HOUSEHOLDER
c               DECOMPOSITION OF FIRST K ROWS.
c              ..
               IF (K .EQ. N) GO TO 230
                  DO 220 II = 1, K
                     I = KP1 - II
                     CALL xDH12(1,I,KP1,N,A(I,1),MDA,G(I),A,MDA,1,I-1)
  220             CONTINUE
  230          CONTINUE
c
c
               IF (NB .LT. 1) GO TO 330
               DO 320 JB = 1, NB
c
c                  SOLVE THE K BY K TRIANGULAR SYSTEM.
c                 ..
                  DO 260 L = 1, K
                     SM = DZERO
                     I = KP1 - L
                     IP1 = I + 1
                     IF (K .LT. IP1) GO TO 250
                     DO 240 J = IP1, K
                        SM = SM + A(I,J)*B(J,JB)
  240                CONTINUE
  250                CONTINUE
                     SM1 = SM
                     B(I,JB) = (B(I,JB) - SM1)/A(I,I)
  260             CONTINUE
c
c                  COMPLETE COMPUTATION OF SOLUTION VECTOR.
c                 ..
                  IF (K .EQ. N) GO TO 290
                     DO 270 J = KP1, N
                        B(J,JB) = SZERO
  270                CONTINUE
                     DO 280 I = 1, K
                        CALL xDH12(2,I,KP1,N,A(I,1),MDA,G(I),B(1,JB),1,              &
     &                            MDB,1)
  280                CONTINUE
  290             CONTINUE
c
c                   RE-ORDER THE SOLUTION VECTOR TO COMPENSATE FOR THE
c                   COLUMN INTERCHANGES.
c                 ..
                  DO 310 JJ = 1, LDIAG
                     J = LDIAG + 1 - JJ
                     IF (IP(J) .EQ. J) GO TO 300
                        L = IP(J)
                        TMP = B(L,JB)
                        B(L,JB) = B(J,JB)
                        B(J,JB) = TMP
  300                CONTINUE
  310             CONTINUE
  320          CONTINUE
  330          CONTINUE
  340       CONTINUE
  350    CONTINUE
c        ..
c         THE SOLUTION VECTORS, X, ARE NOW
c         IN THE FIRST  N  ROWS OF THE ARRAY B(,).
c
         KRANK = K
  360 CONTINUE
      RETURN
      END



C***********************************************************************
C Karline: added as adaptation from xdcopy, to avoid warning of rank mismatch
C***********************************************************************
      SUBROUTINE XDCOPYSC(N,DX,DY,INCY)
C
C     COPIES A scalar, X, TO A VECTOR, Y. 
C
      DOUBLE PRECISION DX, DY(*)
      INTEGER I,INCY,IY,N
C
      IF(N.LE.0)RETURN

      IY = 1
      IF(INCY.LT.0)IY = (-N+1)*INCY + 1

      DO 10 I = 1,N
        DY(IY) = DX
        IY = IY + INCY
   10 CONTINUE
      RETURN
      END

#' Printing Results of MSCMT
#'
#' \code{print.mscmt} prints results of \code{mscmt}.
#'
#' A human-readable summary of \code{mscmt}'s results.
#'
#' @param x An object of class \code{"mscmt"}, usually obtained as
#' the result of a call to function \code{\link{mscmt}}.
#' @param ... Further arguments to be passed to or from other methods. 
#' They are ignored in this function.
#' @return Nothing useful (function is called for its side effects).
#' @importFrom utils capture.output
#' @method print mscmt
#' @export
print.mscmt <- function(x,...) {
  ind <- 16
  mwd <- options()$width

  printSingle <- function(x2) {
    has.placebo   <- !is.null(x2$placebo)
    if (has.placebo) x1 <- x2[[1]] else x1 <- x2
    Vmat <- capture.output(print(rbind(x1$v,"----------"=NA,
                                       "pred. loss"=x1$loss.w),
                                 na.print=""))
    paste0(
      "Specification:\n--------------\n\n",
      csl("Model type",
          paste0(if(length(x1$dependent)>1) "M","SCM",
                 if(x1$dataprep.scaled$trafo.v$has.trafo) "T"),
          mwd,ind),
      csl("Treated unit",x1$treated.unit,mwd,ind),
      csl("Control units",x1$control.units,mwd,ind),
      csl("Dependent(s)",align(x1$dependent," with optimization period from ",
                               x1$times.dep[1,]," to ",x1$times.dep[2,]),
          mwd,ind,TRUE),
      csl("Predictors",align(x1$predictor," from ",x1$times.pred[1,]," to ",
                             x1$times.pred[2,],
                             ifelse(x1$agg.fns=="id","",
                                    paste0(", aggregated via '",x1$agg.fns,"'"))
                             ),
          mwd,ind),"\n\n",
      "Results:\n--------\n\n",
      wrap("Result type",switch(x1$solution.type,
        nosunny = paste0("Perfect predictor fit possible, ie. predictor ",
          "weights V are meaningless since only W's with perfect preditor ",
          "fit are considered for optimization of the 'dependent' loss."),
        onesunny = paste0("Predictor weights V are meaningless since one ",
          "single donor always achieves best predictor fit (irrespective of ",
          "particular choice of predictor weights)."),
        fixed = paste0("Vector V was fixed, outer optimization bypassed."),
        regression = paste0("Vector V was determined by regression, ",
          "outer optimization bypassed."),
        artificial = paste0("Artificial solution, vector W was fixed."),
        global = paste0("Unrestricted outer optimum obtained by ignoring all ",
          "predictors is feasible even when respecting the predictors."),
        paste0("Ordinary solution, ie. no perfect preditor fit possible and ",
          "the predictors impose some restrictions on the outer optimization.")
        ),mwd,ind),
      csl("Optimal W",align(names(ess(x1$w)),": ",
                            format(100*ess(x1$w),scientific=FALSE),"%"),
          mwd,ind),
      csl("Dependent loss",align(c("MSPE ('loss V')","RMSPE"),": ",
                                 c(x1$loss.v,x1$rmspe)),mwd,ind,TRUE),
#      if((x1$solution.type!="global")&&((!is.na(x1$outer.rmspe))))
#        csl("Infeas. opt.",align(c("MSPE ('loss V')","RMSPE"),": ",
#            c(x1$outer.rmspe^2,x1$outer.rmspe)),mwd,ind,TRUE),
      if(length(x1$dependent)>1) paste0(
        csl("Unscaled MSPE",x1$mspe.unscaled,mwd,ind),                               
        csl("Individ. MSPEs",align(x1$dependent,": ",x1$mspes),mwd,ind)                               
      ),  
      wrap("(Optimal) V",switch(x1$solution.type,
        nosunny = paste0("Predictor weights V are meaningless."),
        onesunny = paste0("Predictor weights V are meaningless."),
        fixed = paste0("Fixed predictor weights V (not necessarily optimal)."),
        regression = paste0("Regression-based predictor weights V (not ",
                            "necessarily optimal)."),
        artificial = paste0("Artificial solution, predictor weights V were ",
                            "either fixed or artificially constructed ",
                            "based on fixed donor weights."),
        paste0(if (isTRUE(x1$single.v)) "Single predictor weights V requested. ",
          if (!isTRUE(x1$single.v)) "Some optimal " else "The optimal ",
          "weight vector",
          if (!isTRUE(x1$single.v)) "s V are:" else " V is:")),mwd,ind),
      if ((x1$solution.type!="nosunny")&&(x1$solution.type!="onesunny")) paste0(
        paste0(paste0(paste0(rep(" ",ind),collapse=""),Vmat),collapse="\n"),
        "\n",
        wrap("",paste0("(Predictor weights V are standardized by ",x1$std.v,
                       "(V)=1)"),mwd,ind)
      ),
      if (has.placebo) "\nResults of a placebo study are included.\n"
    )
  }

  printComparison <- function(x2) {
    nam   <- x2$names    
    x1    <- x2$results
    Vmat  <- rbind(x1$v,"----------"=NA,"pred. loss"=x1$loss.w)

    mymerge <- function(a,b) {
      nc  <- max(nchar(rownames(a)),nchar(rownames(b)))
      res <- matrix(NA,nrow=nrow(a)+nrow(b),ncol=ncol(a))
      colnames(res) <- colnames(a)
      rownames(res) <- rep("",nrow(res))
      rownames(res)[2*seq_len(nrow(a))-1] <- 
        paste(format(rownames(a),width=nc),"from")
      rownames(res)[2*seq_len(nrow(a))] <- 
        paste(format(rownames(b),width=nc),"  to")
      res[2*seq_len(nrow(a))-1,]  <- a
      res[2*seq_len(nrow(a)),]    <- b
      res
    }
                                 
    paste0(
      "Specification:\n--------------\n\n",
      wrapM("Model types"    = rbind(x1$model.type),
            "Treated unit"   = rbind(x1$treated.unit),
            max.width=mwd,min.indent=ind),
      wrapM("Control units"  = x1$control.units,
#            "Dependent(s)"   = x1$dependent,
            "Dependent(s)"   = mymerge(x1$dependent.start,x1$dependent.end),
#            "Predictors"     = x1$predictor,
            "Predictors"     = mymerge(x1$predictor.start,x1$predictor.end),
#            "Predictors"     = x1$agg.pred,
            max.width=mwd,min.indent=ind),
      "Results:\n--------\n\n",
      wrapM("Result.type"    = rbind(x1$solution.type),
            max.width=mwd,min.indent=ind),
      wrapM("Optimal W (%)"  = 100*x1$w,
            "Dependent loss" = rbind("MSPE"=x1$loss.v,"RMSPE"=x1$rmspe),
            "Optimal V"      = Vmat,
#            "V std. by"      = rbind(x1$std.v),
            max.width=mwd,min.indent=ind)
    )
  }

  if (!is.null(x$comparison)) {
    out <- "Comparison of (M)SCM(T) optimizations\n\n"
    out <- paste0(out,printComparison(x$comparison))
  } else if (is.null(x$placebo) && is.null(x$combined)) {
    out <- "Collection of univariate SCM(T) optimizations\n\n"
    for (i in seq_along(x)) out <- paste0(out,
      "Dependent variable: ",names(x)[i],"\n",
      paste0(rep("=",20+nchar(names(x)[i])),collapse=""),"\n\n",
      paste0(printSingle(x[[i]])),"\n"
    )
  } else out <- printSingle(x)
  cat(out,"\n")      
}

## R interface to the WNNLS Fortran algorithm 
## Fortran code from package limSolve, with kind permission of Karline Soetaert
## Code is an adapted version of function lsei of package limSolve
## PROGOPT not yet supported
## documenting and exporting this function may be useful...
## to do(?): @export wnnls
wnnls <- function(A=NULL, B=NULL, E=NULL, F=NULL, tol=sqrt(.Machine$double.eps),
                  L=0, verbose=TRUE, check.input=TRUE)  {

  # check input
  if (check.input) {
    if (is.vector(E) & length(F)==1)          E <- t(E)
      else if (! is.matrix(E) & ! is.null(E)) E <- as.matrix(E)
    if (is.vector(A) & length(B)==1)          A <- t(A)
      else if (! is.matrix(A) & ! is.null(A)) A <- as.matrix(A)
    if (! is.matrix(F) & ! is.null(F))        F <- as.matrix(F)
    if (! is.matrix(B) & ! is.null(B))        B <- as.matrix(B)
    if (is.null(A) && is.null (E)) {
      stop("there is no problem to solve: A and E are NULL")
    } else if (is.null(A)) {
      A <- E[1,,drop=FALSE]
      B <- F[1]
    }
  }
  
  # dimensions
  ME  <- nrow(E)
  MA  <- nrow(A)
  MDW <- ME + MA
  N   <- if (!is.null(ncol(A))) ncol(A) else ncol(E)

  if (check.input) {
    # No equalities? -> stop (and hint to nnls), check dimensions otherwise
    if (is.null (ME)) {
      stop("no equalities, please use function nnls in package nnls instead")
    } else  {
      if (ncol(E)   != N)  stop("wrong input: A and E are not compatible")
      if (length(F) != ME) stop("wrong input: E and F are not compatible")
    }
    if (length(B) !=  MA) stop("wrong input: A and B are not compatible")
  }
  
  # Initialize workspace and big matrix input for Fortran code
  IWORK      <- integer(ME+MA+N)
  WORK       <- numeric(ME+MA+5*N)
  IWORK[1:2] <- c(length(WORK),length(IWORK))
  W          <- rbind(cbind(E,F),cbind(A,B))

  # obtain solution
  sol <-.Fortran(C_wnnls,W=as.double(W),MDW=as.integer(MDW),ME=as.integer(ME),
                 MA=as.integer(MA),N=as.integer(N),L=as.integer(L),
                 PRGOPT=as.double(1.0),X=double(N),RNORM=double(1),
                 MODE=integer(1),IWORK=IWORK,WORK=WORK)

  Error <- any(is.infinite(sol$X)) || (sol$MODE>0) 

  X        <- sol$X
  names(X) <- if (!is.null(colnames(A))) colnames(A) else colnames(E)

  # prepare residuals/solution
  if (any(is.infinite(X))) {
    residual <- solution <- Inf
  } else  {
    # residual (l1) norm
    residual <- if (ME > 0) sum(abs(E %*% X - F)) else 0
    if (residual>tol) Error <- TRUE
    # solution value (l2 norm)
    solution <- if (MA > 0) sum ((A %*% X - B)^2) else 0
  }

  list(x=X,value=solution,residual=residual,error=Error,status=sol$MODE)
}

## internally used version of wnnls (no input checks!)
wnnlsInt <- function(W, ME, MA, N, tol=sqrt(.Machine$double.eps))  {
  MDW        <- ME + MA
  IWORK      <- integer(MDW+N)
  WORK       <- double(MDW+5*N)
  IWORK[1:2] <- c(length(WORK),length(IWORK))
  sol <- .Fortran(C_wnnls,W=W,MDW=as.integer(MDW),ME=as.integer(ME),
                  MA=as.integer(MA),N=as.integer(N),L=0L,PRGOPT=as.double(1.0),
                  X=double(N),RNORM=double(1),MODE=integer(1),IWORK=IWORK,
                  WORK=WORK)
  Error <- any(is.infinite(sol$X)) || (sol$MODE>0) 
  list(x=sol$X,error=Error,status=sol$MODE)
}

#' Multivariate Synthetic Control Method Using Time Series
#'
#' MSCMT implements the Multivariate Synthetic Control Method Using
#' Time Series.
#'
#' MSCMT implements three generalizations of the synthetic control method (which 
#' has already an implementation in package 'Synth'): \enumerate{
#' \item it allows for using multiple outcome variables, 
#' \item time series can be supplied as economic predictors,
#' \item a well-defined cross-validation approach can be used.
#' } 
#' Much effort has been taken to make the implementation as stable as possible 
#' (including edge cases) without losing computational efficiency.
#'
#' @rdname MSCMTpackage
#' @aliases MSCMT-package
#' @keywords internal 
#' @name MSCMT
#' @useDynLib MSCMT, .registration = TRUE, .fixes = "C_"
#' @importFrom Rdpack reprompt
#' @importFrom Rdpack insert_ref
#' @examples
#' \dontrun{
#' ## for examples, see the package vignettes:
#' browseVignettes(package="MSCMT")
#' }
#' @references
#' \insertRef{Abadie2003}{MSCMT}
#'
#' \insertRef{Abadie2010}{MSCMT}
#'
#' \insertRef{FastReliable}{MSCMT}
#'
#' \insertRef{CV}{MSCMT}
#'
#' \insertRef{KP16}{MSCMT}
#'
"_PACKAGE"

#' Multivariate SCM Using Time Series
#'
#' \code{mscmt} performs the Multivariate Synthetic Control Method Using Time 
#' Series.
#'
#' \code{mscmt} combines, if necessary, the preparation of the raw data (which 
#' is expected to be in "list" format, possibly after conversion from a 
#' \code{\link[base]{data.frame}}
#' with function \code{\link{listFromLong}}) and the call to the appropriate
#' MSCMT optimization procedures (depending on the input parameters).
#' For details on the input parameters \code{alpha}, \code{beta}, and 
#' \code{gamma}, see [1]. For details on cross-validation, see [2].
#'
#' @param data Typically, a list of matrices with rows corresponding to times 
#' and columns corresponding to units for all relevant features (dependent as
#' well as predictor variables, identified by the list elements' names).
#' This might be the result of converting from a 
#' \code{\link[base]{data.frame}}
#' by using function \code{\link{listFromLong}}. 
#' 
#' For convenience, \code{data} may alternatively be the 
#' result of function \code{\link[Synth]{dataprep}} of package 
#' \code{'Synth'}. In this case, the parameters \code{treatment.identifier},
#' \code{controls.identifier}, \code{times.dep}, \code{times.pred}, 
#' and \code{agg.fns} are ignored, as these input parameters are generated
#' automatically from \code{data}. The parameters \code{univariate}, 
#' \code{alpha}, \code{beta}, and \code{gamma} are ignored by fixing them to 
#' their defaults.
#' Using results of \code{\link[Synth]{dataprep}} is experimental, because
#' the automatic generation of input parameters may fail due to lack of 
#' information contained in results of \code{\link[Synth]{dataprep}}.
#' 
#' @param treatment.identifier A character scalar containing the name of the 
#' treated unit. 
#' Must be contained in the column names of the matrices in \code{data}.
#' @param controls.identifier A character vector containing the names of at 
#' least two control units.
#' Entries must be contained in the column names of the matrices in \code{data}.
#' @param times.dep A matrix with two rows (containing start times in
#' the first and end times in the second row) and one column for each dependent
#' variable, where the column names must exactly match the names of the
#' corresponding dependent variables. 
#' A sequence of dates with the given start and end times of
#' \itemize{
#' \item annual dates, if the format of start/end time is "dddd", e.g. "2016",
#' \item quarterly dates, if the format of start/end time is "ddddQd", e.g. 
#' "2016Q1",
#' \item monthly dates, if the format of start/end time is "dddd?dd" with "?"
#' different from "W" (see below), e.g. "2016/03" or "2016-10",
#' \item weekly dates, if the format of start/end time is "ddddWdd", e.g. 
#' "2016W23",
#' \item daily dates, if the format of start/end time is "dddd-dd-dd", e.g. 
#' "2016-08-18",
#' }
#' will be constructed; these dates are looked for in the row names of
#' the respective matrices in \code{data}. In applications with 
#' cross-validation, \code{times.dep} belongs to the main period.
#' @param times.pred A matrix with two rows (containing start times in
#' the first and end times in the second row) and one column for each predictor
#' variable, where the column names must exactly match the names of the
#' corresponding predictor variables.
#' A sequence of dates with the given start and end times of
#' \itemize{
#' \item annual dates, if the format of start/end time is "dddd", e.g. "2016",
#' \item quarterly dates, if the format of start/end time is "ddddQd", e.g. 
#' "2016Q1",
#' \item monthly dates, if the format of start/end time is "dddd?dd" with "?"
#' different from "W" (see below), e.g. "2016/03" or "2016-10",
#' \item weekly dates, if the format of start/end time is "ddddWdd", e.g. 
#' "2016W23",
#' \item daily dates, if the format of start/end time is "dddd-dd-dd", e.g. 
#' "2016-08-18",
#' }
#' will be constructed; these dates are looked for in the row names of
#' the respective matrices in \code{data}. In applications with 
#' cross-validation, \code{times.pred} belongs to the main period.
#' @param agg.fns Either \code{NULL} (default) or a character vector containing
#' one name of an aggregation function for each predictor variable (i.e., each
#' column of \code{times.pred}). The character string "id" may be used as a
#' "no-op" aggregation. Each aggregation function must accept a numeric vector
#' and return either a numeric scalar ("classical" MSCM) or a numeric vector 
#' (leading to MSCM*T* if length of vector is at least two).
#' @param placebo A logical scalar. If \code{TRUE}, a placebo study is 
#' performed where, apart from the treated unit, each control unit is considered
#' as treated unit in separate optimizations. Defaults to \code{FALSE}. 
#' Depending on the number of control units and the complexity of the problem, 
#' placebo studies may take a long time to finish.
#' @param placebo.with.treated A logical scalar. If \code{TRUE}, the treated
#' unit is included as control unit (for other treated units in placebo 
#' studies). Defaults to \code{FALSE}.
#' @param univariate A logical scalar. If \code{TRUE}, a series of univariate
#' SCMT optimizations is done (instead of one MSCMT optimization) even if
#' there is more than one dependent variable. Defaults to \code{FALSE}.
#' @param univariate.with.dependent A logical scalar. If \code{TRUE} (and if
#' \code{univariate} is also \code{TRUE}), all dependent variables (contained
#' in the column names of \code{times.dep}) apart from the current (real) 
#' dependent variable are included as predictors in the series of univariate
#' SCMT optimizations. Defaults to \code{FALSE}.
#' @param check.global A logical scalar. If \code{TRUE} (default), a check for
#' the feasibility of the unrestricted outer optimum (where actually no 
#' restrictions are imposed by the predictor variables) is made before 
#' starting the actual optimization procedure.
#' @param inner.optim A character scalar containing the name of the optimization
#' method for the inner optimization. Defaults to \code{"wnnlsOpt"}, which
#' (currently) is the only supported implementation, because it outperforms
#' all other inner optimizers we are aware of. 
#' \code{"ipopOpt"}, which uses \code{\link[kernlab]{ipop}}, and 
#' \code{LowRankQPOpt}, which uses \code{\link[LowRankQP]{LowRankQP}} as inner
#' optimizer, have experimental support for benchmark purposes.
#' @param inner.opar A list containing further parameters for the inner 
#' optimizer. Defaults to the empty list. (For \code{"wnnlsOpt"}, there are no
#' meaningful further parameters.)
#' @param outer.optim A character vector containing the name(s) of the 
#' optimization method(s) for the outer optimization. Defaults to 
#' \code{"DEoptC"}, which (currently) is the recommended global optimizer. 
#' The optimizers currently supported can be found in the documentation of
#' parameter \code{outer.opar}, where the default control parameters for
#' the various optimizers are listed.
#' If \code{outer.optim} has length greater
#' than 1, one optimization is invoked for each outer optimizer (and, 
#' potentially, each random seed, see below), and the best result is used.
#' @param outer.par A list containing further parameters for the outer 
#' optimization procedure. Defaults to the empty list. Entries in this list may 
#' override the following hard-coded general defaults:
#' \itemize{
#' \item \code{lb=1e-8}, corresponding to the lower bound for the ratio of
#' predictor weights,
#' \item \code{opt.separate=TRUE}, corresponding
#' to an improved outer optimization where each predictor is treated as the 
#' (potentially) most important predictor (i.e. with maximal weight) in 
#' separate optimizations (one for each predictor), see [1].
#' }
#' @param outer.opar A list (or a list of lists, if \code{outer.optim} has
#' length greater than 1) containing further parameters for the outer 
#' optimizer(s). Defaults to the empty list. Entries in this list may override
#' the following hard-coded defaults for the individual optimizers, which
#' are quite modest concerning the computing time. 
#' \code{dim} is a variable holding the problem dimension, 
#' typically the number of predictors minus one.
#' \tabular{lll}{
#' \bold{Optimizer}  \tab \bold{Package}     \tab \bold{Default parameters} \cr 
#' \code{DEoptC}     \tab \code{MSCMT}       \tab \code{nG=500}, \code{nP=20*dim}, \code{waitgen=100}, \cr
#'                   \tab                    \tab \code{minimpr=1e-14}, \code{F=0.5}, \code{CR=0.9} \cr
#' \code{cma_es}     \tab \code{cmaes}       \tab \code{maxit=2500} \cr
#' \code{crs}        \tab \code{nloptr}      \tab \code{maxeval=2.5e4}, \code{xtol_rel=1e-14}, \cr
#'                   \tab                    \tab \code{population=20*dim}, \code{algorithm="NLOPT_GN_CRS2_LM"} \cr
#' \code{DEopt}      \tab \code{NMOF}        \tab \code{nG=100}, \code{nP=20*dim} \cr
#' \code{DEoptim}    \tab \code{DEoptim}     \tab \code{nP=20*dim} \cr
#' \code{ga}         \tab \code{GA}          \tab \code{maxiter=50}, \code{monitor=FALSE}, \cr
#'                   \tab                    \tab \code{popSize=20*dim} \cr
#' \code{genoud}     \tab \code{rgenoud}     \tab \code{print.level=0}, \code{max.generations=70}, \cr
#'                   \tab                    \tab \code{solution.tolerance=1e-12}, \code{pop.size=20*dim}, \cr
#'                   \tab                    \tab \code{wait.generations=dim}, \code{boundary.enforcement=2}, \cr
#'                   \tab                    \tab \code{gradient.check=FALSE}, \code{MemoryMatrix=FALSE} \cr
#' \code{GenSA}      \tab \code{GenSA}       \tab \code{max.call=1e7}, \code{max.time=25/dim},  \cr
#'                   \tab                    \tab \code{trace.mat=FALSE} \cr
#' \code{isres}      \tab \code{nloptr}      \tab \code{maxeval=2e4}, \code{xtol_rel=1e-14}, \cr
#'                   \tab                    \tab \code{population=20*dim}, \code{algorithm="NLOPT_GN_ISRES"} \cr
#' \code{malschains} \tab \code{Rmalschains} \tab \code{popsize=20*dim}, \code{maxEvals=25000} \cr
#' \code{nlminbOpt}  \tab \code{MSCMT/stats} \tab \code{nrandom=30} \cr
#' \code{optimOpt}   \tab \code{MSCMT/stats} \tab \code{nrandom=25} \cr
#' \code{PSopt}      \tab \code{NMOF}        \tab \code{nG=100}, \code{nP=20*dim} \cr
#' \code{psoptim}    \tab \code{pso}         \tab \code{maxit=700} \cr
#' \code{soma}       \tab \code{soma}        \tab \code{nMigrations=100} 
#' }
#' If \code{outer.opar} is a list of lists, its names must correspond to (a 
#' subset of) the outer optimizers chosen in \code{outer.optim}.
#' @param std.v A character scalar containing one of the function names
#' "sum", "mean", "min", or "max" for the standardization of the predictor 
#' weights (weights are divided by \code{std.v(weights)} before reporting). 
#' Defaults to "sum", partial matching allowed.
#' @param alpha A numerical vector with weights for the dependent variables
#' in an MSCMT optimization or \code{NULL} (default). If not \code{NULL},
#' the length of \code{alpha} must agree with the number of dependent
#' variables, \code{NULL} is equivalent to weight 1 for all dependent 
#' variables.
#' @param beta Either \code{NULL} (default), a numerical vector, or a list.
#' If \code{beta} is a numerical vector or a list, its length must agree
#' with the number of dependent variables. 
#' \itemize{
#' \item If \code{beta} is a numerical vector,
#' the \code{i}th dependent variable is discounted with discount factor 
#' \code{beta[i]} (the observations of the dependent variables must thus be 
#' in chronological order!). 
#' \item If \code{beta} is a list, the components of \code{beta} must be 
#' numerical vectors with lengths corresponding to the numbers of observations 
#' for the individual dependent variables. These observations are then 
#' multiplied with the corresponding component of \code{beta}.
#' }
#' @param gamma Either \code{NULL} (default), a numerical vector, or a list.
#' If \code{gamma} is a numerical vector or a list, its length must agree
#' with the number of predictor variables. 
#' \itemize{
#' \item If \code{gamma} is a numerical vector,
#' the output of \code{agg.fns[i]} applied to the \code{i}th predictor variable
#' is discounted with discount factor \code{gamma[i]} (the output of 
#' \code{agg.fns[i]} must therefore be in chronological order!). 
#' \item If \code{gamma} is a list, the components of \code{gamma} must be 
#' numerical vectors with lengths corresponding to the lengths of the output of 
#' \code{agg.fns} for the individual predictor variables. The output of 
#' \code{agg.fns} is then multiplied with the corresponding component of 
#' \code{gamma}.
#' }
#' @param return.ts A logical scalar. If \code{TRUE} (default), most results are
#' converted to time series.
#' @param single.v A logical scalar. If \code{FALSE} (default), a selection
#' of feasible (optimal!) predictor weight vectors is generated. If \code{TRUE}, 
#' the one optimal weight vector which has maximal order statistics is generated 
#' to facilitate cross validation studies.
#' @param verbose A logical scalar. If \code{TRUE} (default), output is verbose.
#' @param debug A logical scalar. If \code{TRUE}, output is very verbose. 
#' Defaults to \code{FALSE}.
#' @param seed A numerical vector or \code{NULL}. If not \code{NULL}, the
#' random number generator is initialized with the elements of \code{seed} via
#' \code{set.seed(seed)} (see \link[base]{Random}) before
#' calling the optimizer, performing repeated optimizations (and staying with 
#' the best) if \code{seed} has length greater than 1. Defaults to \code{NULL}. 
#' If not \code{NULL}, the seeds \code{int.seed} (default: 53058) and 
#' \code{unif.seed} (default: 812821) for \code{\link[rgenoud]{genoud}} are 
#' also initialized to the corresponding element of \code{seed}, but this can 
#' be overridden with the list elements \code{int.seed} and \code{unif.seed} 
#' of (the corresponding element of) \code{outer.opar}.
#' @param cl \code{NULL} (default) or an object of class \code{cluster}
#' obtained by \code{\link[parallel]{makeCluster}} of package \code{parallel}. 
#' Repeated estimations (see \code{outer.optim} and \code{seed}) and
#' placebo studies will make use of the cluster \code{cl} (if not \code{NULL}).
#' @param times.pred.training A matrix with two rows (containing start times in
#' the first and end times in the second row) and one column for each predictor
#' variable, where the column names must exactly match the names of the
#' corresponding predictor variables (or \code{NULL} by default).
#' If not \code{NULL}, \code{times.pred.training} defines training periods
#' for cross-validation applications. For the format of the start and end times,
#' see the documentation of parameter \code{times.pred}.
#' @param times.dep.validation A matrix with two rows (containing start times in
#' the first and end times in the second row) and one column for each dependent
#' variable, where the column names must exactly match the names of the
#' corresponding dependent variables (or \code{NULL} by default). 
#' If not \code{NULL}, \code{times.dep.validation} defines validation period(s)
#' for cross-validation applications. For the format of the start and end times,
#' see the documentation of parameter \code{times.dep}.
#' @param v.special integer vector containing indices of important predictors
#' with special treatment (see below). Defaults to the empty set.
#' @param cv.alpha numeric scalar containing the minimal proportion (of the
#' maximal feasible weight) for the weights of the predictors selected by 
#' \code{v.special}. Defaults to \code{0}.
#' @param spec.search.treated A logical scalar. If \code{TRUE}, a specification 
#' search (for the optimal set of included predictors) is done for the treated unit. Defaults to \code{FALSE}.
#' @param spec.search.placebos A logical scalar. If \code{TRUE}, a specification 
#' search (for the optimal set of included predictors) is done for the control unit. Defaults to \code{FALSE}.
#' @return An object of class \code{"mscmt"}, which is essentially a list
#' containing the results of the estimation and, if applicable, the placebo
#' study.
#' The most important list elements are 
#' \itemize{
#' \item the weight vector \code{w} for the control units,
#' \item a matrix \code{v} with weight vectors for the predictors in its 
#' columns,
#' \item scalars \code{loss.v} and \code{rmspe} with the dependent loss and its 
#' square root,
#' \item a vector \code{loss.w} with the predictor losses corresponding to the
#' various weight vectors in the columns of \code{v},
#' \item a matrix \code{predictor.table} containing aggregated statistics of
#' predictor values (similar to list element \code{tab.pred} of 
#' function \code{\link[Synth]{synth.tab}} of package \code{'Synth'}),
#' \item a list of multivariate time series \code{combined} containing, 
#' for each dependent and predictor variable, a multivariate time series 
#' with elements \code{treated} for the actual values of the treated unit,
#' \code{synth} for the synthesized values, and \code{gaps} for the differences.
#' }
#' Placebo studies produce a list containing individual results for each 
#' unit (as treated unit), starting with the original treated unit, as well
#' as a list element named \code{placebo} with aggregated results for each
#' dependent and predictor variable.
#'
#' If \code{times.pred.training} and \code{times.dep.validation} are not 
#' \code{NULL}, a cross-validation is done and a list of elements \code{cv}
#' with the results of the cross-validation period and \code{main} with
#' the results of the main period is returned.
#' @references
#' [1] \insertRef{FastReliable}{MSCMT} 
#'
#' [2] \insertRef{CV}{MSCMT}
#'
#' @importFrom stats sd
#' @importFrom parallel clusterExport clusterApplyLB clusterEvalQ
#' @rdname MSCMTfunction
#' @export mscmt
#' @examples
#' \dontrun{
#' ## for examples, see the package vignettes:
#' browseVignettes(package="MSCMT")
#' }
mscmt <- function(data,treatment.identifier=NULL, controls.identifier=NULL, 
                  times.dep=NULL, times.pred=NULL, agg.fns=NULL,
                  placebo=FALSE, placebo.with.treated=FALSE, univariate=FALSE,
                  univariate.with.dependent=FALSE,
                  check.global=TRUE, inner.optim="wnnlsOpt",inner.opar=list(),
                  outer.optim="DEoptC",outer.par=list(),
                  outer.opar=list(), std.v=c("sum","mean","min","max"),
                  alpha=NULL,beta=NULL,gamma=NULL,return.ts=TRUE,single.v=FALSE,
                  verbose=TRUE, debug=FALSE, seed=NULL, cl=NULL,
                  times.pred.training=NULL, times.dep.validation=NULL,
                  v.special=integer(),cv.alpha=0,
                  spec.search.treated=FALSE,spec.search.placebos=FALSE) {
  # check input                  
  std.v <- match.arg(std.v)

  # shall we do cross validation?
  is.cv <- !(is.null(times.pred.training)||is.null(times.dep.validation))
  
  # is data a dataprep object?
  is.dataprep <- isTRUE(all.equal(names(data),c("X0","X1","Z0","Z1","Y0plot",
                                  "Y1plot","names.and.numbers","tag")))
  # initialize v
  v <- NULL                                  
  
  if (is.dataprep&&is.cv) {
    warning("cross validation not supported for dataprep input")                # cross validation not supported for dataprep input
    is.cv <- FALSE
  }
  
  if (placebo&&(!is.null(cl))) { 
    n.placebo <- length(controls.identifier)+1
    n.multi   <- length(outer.optim) * (if (is.null(seed)) 1 else length(seed))
    placebo.on.cluster <- (n.placebo>n.multi)
  } else placebo.on.cluster <- FALSE
    
  if (is.dataprep) {
    storage.mode(data$X0) <- storage.mode(data$X1) <- 
      storage.mode(data$Z0) <- storage.mode(data$Z1) <- "double"
    univariate <- FALSE
    unit.names <- data$names.and.numbers$unit.names
    if (is.null(colnames(data$X1))) colnames(data$X1) <- colnames(data$Z1)      # correct potentially missing names
    if (is.null(colnames(data$Z1))) colnames(data$Z1) <- colnames(data$X1)
    names(unit.names) <- as.character(data$names.and.number$unit.numbers)
    treatment.identifier <- as.character(unit.names[colnames(data$X1)])
    controls.identifier  <- as.character(unit.names[colnames(data$X0)])
    if (any(is.na(treatment.identifier))) 
      treatment.identifier <- colnames(data$X1)
    if (any(is.na(controls.identifier))) 
      controls.identifier <- colnames(data$X0)
    times.dep  <- cbind("Y"=range(data$tag$time.optimize.ssr))
    times.pred <- matrix(rep(range(data$tag$time.predictors.prior),
                             times=length(data$tag$predictors)),nrow=2)
    for (i in seq_along(data$tag$special.predictors))
      times.pred <- cbind(times.pred,
                          range(data$tag$special.predictors[[i]][[2]]))
    colnames(times.pred) <- c(rownames(data$X1)[seq_along(data$tag$predictors)],
                              sapply(data$tag$special.predictors,
                                     function(x) x[[1]]))
    agg.fns <- c(rep(data$tag$predictors.op,length(data$tag$predictors)),
                 if (!is.null(data$tag$special.predictors)) 
                   sapply(data$tag$special.predictors,function(x) x[[3]])) 
  }

  # prepare specification search
  if (is.cv&&(spec.search.treated||spec.search.placebos)) {
    subsets <- expand.grid(rep(list(c(FALSE,TRUE)), 
                           ncol(times.pred)-length(v.special)))
    subsets <- as.matrix(cbind(matrix(TRUE,ncol=length(v.special),
                                      nrow=nrow(subsets)),subsets))
    colnames(subsets) <- colnames(times.pred)
    if (placebo) {
      tmp <- vector("list",length(controls.identifier)+1)
      names(tmp) <- c(treatment.identifier,controls.identifier)
      tmp[[treatment.identifier]] <- if (spec.search.treated) subsets else NULL
      for (id in controls.identifier) 
        tmp[[id]] <- if (spec.search.placebos) subsets else NULL
      subsets <- tmp
    }
  } else subsets <- NULL

  # main workhorse
  synthSingle <- function(treatment.identifier, controls.identifier, 
                          times.dep, times.pred, agg.fns, v=NULL,
                          subsets) {

    if (!is.matrix(times.dep)) stop("times.dep must be a matrix")
    dependent <- colnames(times.dep)
    predictor <- colnames(times.pred)
    time.optimize.ssr <- vector("list",length(dependent))
    for (i in seq_along(dependent)) {
      tmp <- try(seqAQM(times.dep[1,i],times.dep[2,i]),silent=TRUE)
      if (inherits(tmp,"try-error")) 
        stop(paste0("unsupported date format for dependent variable '",
                    dependent[i],"'")) else
        time.optimize.ssr[[i]] <- tmp
    }
  
    if (!is.matrix(times.pred)) stop("times.pred must be a matrix")
    if (is.null(agg.fns)) agg.fns <- rep("id",ncol(times.pred))
    if ((!is.character(agg.fns))||(length(agg.fns)!=ncol(times.pred)))
      stop("agg.fns is not a character vector with ncol(times.pred) elements")

    if (is.dataprep) {
      X  <- cbind(data$X0, data$X1)
      colnames(X) <- unit.names[as.character(colnames(X))]
      if (any(is.na(colnames(X)))) 
        colnames(X) <- c(colnames(data$X0),colnames(data$X1))
      X  <- X/apply(X, 1, sd)                                                 # scale X0,X1
      X0u <- data$X0
      colnames(X0u) <- unit.names[as.character(colnames(data$X0))]
      if (any(is.na(colnames(X0u)))) 
        colnames(X0u) <- colnames(data$X0)
      X1u <- data$X1
      colnames(X1u) <- unit.names[as.character(colnames(data$X1))]
      if (any(is.na(colnames(X1u)))) 
        colnames(X1u) <- colnames(data$X1)
      Z0 <- data$Z0
      colnames(Z0) <- unit.names[as.character(colnames(data$X0))]             # correct potentially wrong column names of Z0
      if (any(is.na(colnames(Z0)))) 
        colnames(Z0) <- colnames(data$X0)
      Z1 <- data$Z1
      colnames(Z1) <- unit.names[as.character(colnames(Z1))]
      if (any(is.na(colnames(Z1)))) 
        colnames(Z1) <- colnames(data$X1)
      Z  <- cbind(Z0,Z1)
      Y  <- rbind(cbind(data$Z0,data$Z1),cbind(data$Y0plot,data$Y1plot))
      Y  <- Y[sort(unique(row.names(Y))),]
      colnames(Y) <- colnames(X)
      Z1 <- Z[,treatment.identifier,drop=FALSE]
      Z0 <- Z[,controls.identifier,drop=FALSE]
      X1 <- X[,treatment.identifier,drop=FALSE]
      X0 <- X[,controls.identifier,drop=FALSE]
    }    
        
    singleSubset <- function(times.pred,agg.fns) {
      names.v    <- colnames(times.pred)
      len.v      <- length(names.v)
      special.predictors <- vector("list",len.v)
      for (i in 1:len.v) {
        tmp <- try(seqAQM(times.pred[1,i],times.pred[2,i]),silent=TRUE)
        if (inherits(tmp,"try-error")) 
          stop(paste0("unsupported date format for predictor variable '",
                      names.v[i],"'")) else
          special.predictors[[i]] <- 
            list(names.v[i],list(tmp),
                 if (is.null(agg.fns)) "id" else agg.fns[i])
      }
      # prepare the data
      if (is.dataprep) {
        dat <- list(X0 = X0, X1 = X1, Z0 = Z0, Z1 = Z1, X0.unscaled = X0u, 
                    X1.unscaled = X1u,
                    trafo.v=genTrafo(n.v=ncol(times.pred),
                                     names.v=colnames(times.pred)),
                    Z.scaled=FALSE)
      } else {
        dat <- prepare(data,predictors=NULL, predictors.op="mean", 
                       special.predictors, dependent, treatment.identifier, 
                       controls.identifier, time.predictors.prior=NULL, 
                       time.optimize.ssr, alpha=alpha, beta=beta, gamma=gamma, 
                       scale.Z=(ncol(times.dep)>1))
      }                 
    
      # run the optimization                           
      res <- multiOpt(X0=dat$X0,X1=dat$X1,Z0=dat$Z0,Z1=dat$Z1,
                      trafo.v=dat$trafo.v,check.global=check.global,
                      inner.optim=inner.optim,inner.opar=inner.opar,
                      starting.values=NULL,outer.par=outer.par,
                      outer.optim=if (is.null(v)) outer.optim else "fixed",
                      outer.opar=if (is.null(v)) outer.opar else list(v=v),
                      std.v=std.v,single.v=single.v,verbose=verbose,debug=debug,
                      seed=seed,cl=if (placebo.on.cluster) NULL else cl)
                      
      list(res=res,dat=dat)
    }
    
    if (is.null(subsets)||!is.matrix(subsets)) {
      if (is.null(subsets)) subset <- rep(TRUE,ncol(times.pred)) else
                            subset <- subsets
      tmp  <- singleSubset(times.pred[,subset,drop=FALSE],agg.fns[subset])
      dat  <- tmp$dat
      res  <- tmp$res
    } else {
      CVcrit <- numeric(nrow(subsets))
      catn("Starting specification search")
      for (i in seq_len(nrow(subsets))) {
        catn("Specification ",i," out of ",nrow(subsets))
        CVcrit[i] <- singleSubset(times.pred[,subsets[i,],drop=FALSE],
                                  agg.fns[subsets[i,]])$res$rmspe
      }
      minCVcrit <- min(CVcrit)
      npreds    <- rowSums(subsets)
      predopt   <- CVcrit <= minCVcrit*(1+1e-8)
      npgood    <- min(ifelse(predopt,npreds,1e8))
      goodsub   <- predopt&(npreds==npgood)
      catn(sum(goodsub)," 'best' model(s)")
      CVcrit[!goodsub] <- Inf
      subset <- subsets[which.min(CVcrit),]
      catn("Predictors included in 'best' specification: ",
           paste0(colnames(times.pred)[subset],collapse=", "))
      tmp  <- singleSubset(times.pred[,subset,drop=FALSE],agg.fns[subset])
      dat  <- tmp$dat
      res  <- tmp$res
    }
                    
    w <- blow(res$w,controls.identifier)
  
    pred.table <- cbind("Treated"     = drop(dat$X1.unscaled), 
                        "Synthetic"   = drop(dat$X0.unscaled %*% w), 
                        "Sample Mean" = apply(dat$X0.unscaled, 1, mean))
  
    if (is.dataprep) {
      ind        <- (w>0)
      data.synth <- list(Y=(Y[,names(w[ind]),drop=FALSE] %*% w[ind])[,1])
      data.treat <- Y[,treatment.identifier,drop=FALSE]
      colnames(data.treat) <- NULL
      data.treat <- list(Y=drop(data.treat))
    } else {
      ind        <- (w>0)
      data.synth <- lapply(data,function(x) if (mode(x)=="numeric")
                            (x[,names(w[ind]),drop=FALSE] %*% w[ind])[,1] else {
                             tmp <- rep(NA,nrow(x))
                             names(tmp) <- rownames(x)
                             tmp
                          })
                                
      data.treat <- lapply(data,function(x) {
                             if (treatment.identifier %in% colnames(x))
                               tmp <- x[,treatment.identifier,drop=FALSE] else {
                                 tmp <- cbind(rep(NA,nrow(x)))
                                 rownames(tmp) <- rownames(x)
                               }  
                             colnames(tmp) <- NULL
                             drop(tmp)
                           })
    }                       
    gaps <- data.treat
    for (i in seq_along(gaps)) 
      if (mode(gaps[[i]])!="numeric") gaps[[i]][] <- NA else 
        gaps[[i]] <- gaps[[i]] - data.synth[[i]]
    
    if (return.ts) {
      data.synth <- lapply(data.synth,AQM2ts)
      data.treat <- lapply(data.treat,AQM2ts)
      gaps       <- lapply(gaps,AQM2ts)
    }
    
    combined <- vector("list",length(gaps))
    names(combined) <- names(gaps)
    for (i in seq_along(gaps)) 
      combined[[i]] <- cbind(treated=data.treat[[i]],synth=data.synth[[i]],
                             gaps=gaps[[i]])
    names(agg.fns)  <- colnames(times.pred)                             
  
    res <- c(res,list(
               dataprep.scaled=dat,data.synth=data.synth,data.treat=data.treat,
               gaps=gaps,combined=combined,treated.unit=treatment.identifier,
               control.units=controls.identifier,dependent=dependent,
               predictor=predictor[subset],agg.fns=agg.fns[subset],
               agg.pred=rownames(dat$X0),times.dep=times.dep,
               times.pred=times.pred[,subset,drop=FALSE],
               std.v=std.v,predictor.table=pred.table,
               subset=subset))                                                  # insert alpha, beta, gamma, ...?
               
    if (dat$Z.scaled) {
      Zu    <- dat$Z0.unscaled - drop(dat$Z1.unscaled)
      mspeu <- lossDep(Zu,w)
      Zl    <- dat$Z.len
      Zw    <- as.numeric(Zu %*% w)^2
      Zsel  <- genA(Zl)
      res   <- c(res,list(mspes=as.numeric(t(Zsel)%*%Zw)/Zl),
                          mspe.unscaled=mspeu)
    }  
    res
  }           

  synthPlacebo <- function(treatment.identifier, controls.identifier, 
                           times.dep, times.pred, agg.fns, v=NULL, subsets) {
    all.units <- c(treatment.identifier,controls.identifier)
    mySynth <- function(treated) {
      if (verbose) catn("Using ",treated," as treated unit now.")
      res <- synthSingle(treated, 
        if (placebo.with.treated) setdiff(all.units,treated) else 
                                  setdiff(controls.identifier,treated),
        times.dep, times.pred, agg.fns,
        v=if (!is.null(v)) v[[treated]] else NULL, 
        subsets=if (!is.null(subsets)) subsets[[treated]] else NULL)
      class(res) <- "mscmt"
      res
    }  
    if (verbose) catn("Starting placebo study, ",
      if (placebo.with.treated) "in" else "ex","cluding original treated ",
      "unit",if (placebo.on.cluster) ", on the cluster. Please hold the line",
      ".")
    if (placebo.on.cluster) clusterExport(cl,c("mySynth","treatment.identifier",
      "controls.identifier","times.dep","times.pred","agg.fns","all.units","v",
      setdiff(unique(agg.fns),"id"),"subsets","multiOpt","atomOpt"),
      envir=environment())
    res <- if (placebo.on.cluster) clusterApplyLB(cl,all.units,mySynth) else
                                   lapply(all.units,mySynth)
                            
    names(res) <- all.units
    combined <- vector("list",length(res[[1]]$combined))
    names(combined) <- names(res[[1]]$combined)
    for (j in names(combined)) {
      combined[[j]] <- list()
      for (i in seq_along(res)) {
        combined[[j]]$data.synth <- cbind(combined[[j]]$data.synth,
                                          res[[i]]$data.synth[[j]])
        combined[[j]]$data.treat <- cbind(combined[[j]]$data.treat,
                                          res[[i]]$data.treat[[j]])
        combined[[j]]$gaps       <- cbind(combined[[j]]$gaps,
                                          res[[i]]$gaps[[j]])
      }
      colnames(combined[[j]]$data.synth)<-colnames(combined[[j]]$data.treat)<-
        colnames(combined[[j]]$gaps) <- names(res)
    }    
    c(res,list(placebo=combined))
  }

  # separate univariate SCM for many dependent variables
  synthUnivariate <- function(treatment.identifier, controls.identifier, 
                              times.dep, times.pred, agg.fns, v=NULL, subsets) {
    if (nrow(subsets)>1) 
      catn("Specification search for univariate SCM not yet supported")
    if (verbose) catn("Starting univariate SCMT for single dependent ",
      "variables, ",if (univariate.with.dependent) "in" else "ex",
      "cluding other dependent variables as predictors.")
    synthFun <- if (placebo) synthPlacebo else synthSingle                               
    dependent <- colnames(times.dep)
    res <- vector("list",length(dependent))
    names(res) <- dependent
    for (i in seq_along(dependent)) {
      if (verbose) catn("Performing SCMT for ",names(res)[i]," now.")
      times.predT <- cbind(times.pred, if (univariate.with.dependent) 
                                         times.dep[,-i,drop=FALSE] else NULL)
      agg.fnsT <- if (is.null(agg.fns)) NULL else c(agg.fns,
        if (univariate.with.dependent) rep("id",length(dependent)-1) else NULL)
      res[[i]] <- synthFun(treatment.identifier, controls.identifier, 
                           times.dep[,i,drop=FALSE], times.predT, agg.fnsT, 
                           if (!is.null(v)) v[[i]] else NULL)
      class(res[[i]]) <- "mscmt"
    }
    res
  }
  
  # what shall we do?
  synthFun <- if (univariate) synthUnivariate else 
                 if (placebo) synthPlacebo else synthSingle
  
  # do it!
  if (is.cv) {
    lb <- if(!is.null(outer.par$lb)) outer.par$lb else 1e-8
    genV <- function(x,xname=NULL) {                                            # cross-validation
      if (is.list(x)) {
        if (is.null(x[["v"]])) {
          tmp <- lapply(seq_along(x),function(i) genV(x[[i]],names(x)[[i]])) 
          names(tmp) <- names(x)
          tmp
        } else {
          if (verbose) catn("Determining cross-validation predictor weights",
            if (!is.null(xname)) paste0(" for ",xname),".")
          tmp_v <- cbind(
            single_v(x$w,x$dataprep.scaled$X0-drop(x$dataprep.scaled$X1),
                     x$dataprep.scaled$Z0-drop(x$dataprep.scaled$Z1),
                     x$trafo.v,cv.alpha=cv.alpha,v.special=v.special,
                     verbose=verbose,debug=debug,lb=lb,backup.v=x$v))
          if (any(is.na(tmp_v))) cbind("backup"=x$v[,ncol(x$v)]) else tmp_v
        }  
      } else NULL
    }  
    if (verbose) catn("Starting optimization for cross-validation period.")
    resCV <- synthFun(treatment.identifier, controls.identifier, 
                      times.dep.validation, times.pred.training, agg.fns, v,
                      subsets)  
    if ("v" %in% names(outer.opar)) v <- outer.opar$v
    if (is.null(v)) {
      v <- genV(resCV)
      if (!(univariate||placebo)) if (!is.null(v)) v=v[,1]
    }  
    CVsubsets <- if (placebo) lapply(resCV,function(x) x$subset) else
                              resCV$subset
    if (verbose) catn("Starting calculations for main period.")
    resM <- synthFun(treatment.identifier, controls.identifier, 
                     times.dep, times.pred, agg.fns, v, CVsubsets)  
    class(resCV) <- class(resM) <- "mscmt"
    res <- list(cv=resCV, main=resM)
  } else {                                                                      # no cross-validation
    res <- synthFun(treatment.identifier, controls.identifier, times.dep, 
                    times.pred, agg.fns, v, subsets)
    class(res) <- "mscmt"
  }
  if (placebo.on.cluster&&verbose) 
    catn("Placebo study on cluster finished.")
  res                  
}

## internal functions and objects
## todo (?): @export multiOpt
## @export multiOpt
#' @importFrom parallel clusterApplyLB
multiOpt <- function(X0,X1=0,Z0,Z1=0,check.global=TRUE,
                     trafo.v,info.v,n.v,names.v,inner.optim="wnnlsOpt",
                     inner.opar=list(),starting.values=NULL,
                     outer.optim="DEoptC",outer.par=list(),outer.opar=list(),
                     std.v=c("sum","mean","min","max"),single.v=FALSE,
                     verbose=TRUE,debug=FALSE,seed=NULL,cl=NULL) {
  outer.optim   <- match.arg(outer.optim,c("DEoptC","DEoptim","GenSA","genoud",
                                           "nlminbOpt","JDEoptim","optimOpt",
                                           "DEopt","PSopt","nloptr",
                                           "cma_es","malschains",
                                           "soma","psoptim","ga","crs","isres",
                                           "fixed","regression","none"),
                                           several.ok=TRUE)
  inner.optim   <- match.arg(inner.optim,c("wnnlsOpt","ipopOpt",
                                           "LowRankQPOpt","benchmarkOpt"))                                           
  do.optimize   <- !(all(outer.optim %in% c("fixed","regression","none")))
  checkPkg <- function(pkg) if(!requireNamespace(pkg,quietly=TRUE))
    stop("package ",pkg," not installed") else TRUE
  getFun <- function(pkg,fun) 
    if(!requireNamespace(pkg,quietly=TRUE))
      stop("package ",pkg," not installed") else 
      get(fun,envir=asNamespace(pkg),mode="function")
  
  solution.type <- "normal"
  std.v <- match.arg(std.v)
  X     <- X0 - drop(X1)                                                        # X0, X1 (Z0, Z1) have to be scaled already!
  Z     <- Z0 - drop(Z1)                                                            
  nW    <- ncol(X)
  cnX   <- colnames(X)
  storage.mode(X) <- "double"
  storage.mode(Z) <- "double"
  comb   <- NULL
  gRmspe <- NA
  X.orig <- X; Z.orig <- Z                                                      # make backup of original X&Z 
  
  if (missing(trafo.v)) {
    if (missing(info.v)) {
      if (missing(n.v))     n.v     <- nrow(X)
      if (missing(names.v)) names.v <- rownames(X)
      if (missing(trafo.v)) trafo.v <- genTrafo(n.v=n.v,names.v=names.v)
    } else {
      if (missing(n.v))     n.v     <- info.v$n.v 
      if (missing(names.v)) names.v <- info.v$names.v 
      if (missing(trafo.v)) trafo.v <- genTrafo(n.v=n.v,names.v=names.v,
                                                len.v=info.v$len.v)
    } 
  } else {
    if (missing(n.v))     n.v     <- trafo.v$n.v
    if (missing(names.v)) names.v <- trafo.v$names.v
  }  
  
  # recombine parameters for inner optimization
  inner.args <- list(debug=FALSE)
  inner.args[names(inner.opar)] <- inner.opar

  # recombine parameters for outer optimization
  outer.args1 <- list(lb=1e-8,opt.separate=TRUE)
  outer.args1[names(outer.par)] <- outer.par
  
  if ((outer.args1$lb<=0)||(outer.args1$lb>=1)) 
    stop("lb must be positive and must be smaller than 1")

  # initialize parameters for outer optimizer (depending on optimizer)
  if (any(outer.optim == "fixed")) {
    if (length(outer.optim)>1) 
      stop("optimizer 'fixed' not supported for multiple optimizations")
    solution.type <- "fixed"  
    if (is.null(outer.opar$v)) {
      warning("element v of outer.opar is missing, using constant weights.")
      outer.opar$v <- rep(1,n.v)
    }  
    v <- cbind("fixed"=outer.opar$v)
  }
  if (any(outer.optim == "regression")) {
    if (length(outer.optim)>1) 
      stop("optimizer 'regression' not supported for multiple optimizations")
    solution.type <- "regression"  
    if (is.null(outer.opar$add.const)) outer.opar$add.const <- TRUE
    if (is.null(outer.opar$solve.tol)) 
      outer.opar$solve.tol <- .Machine$double.eps
    Xr  <- cbind(if (outer.opar$add.const) 1,t(cbind(X1,X0)))
	  Zr  <- t(cbind(Z1,Z0))
	  v   <- cbind("regression" = if (outer.opar$add.const)
                   rowSums(solve(crossprod(Xr),crossprod(Xr,Zr),
                                 tol=outer.opar$solve.tol)[-1,]^2) else 
                   rowSums(solve(crossprod(Xr),crossprod(Xr,Zr),
                                 tol=outer.opar$solve.tol)^2))
  }
  if (any(outer.optim == "none")) {
    if (length(outer.optim)>1) 
      stop("optimizer 'none' not supported for multiple optimizations")
    solution.type <- "artificial"
    if (is.null(outer.opar$w)) 
      stop("outer.opar$w must be provided when outer.optim=='none'")
#    v     <- outer.opar$v                                                           # may be NULL
    v     <- cbind("fixed"=outer.opar$v)
    w     <- blow(outer.opar$w,cnX)
    rmspe <- sqrt(lossDep(Z,w))
    loss.w <- c("fixed"=lossPred(X,w,outer.opar$v,trafo.v))
    res   <- list(w=w,v=v,loss.v=rmspe^2,rmspe=rmspe,loss.w=loss.w,conv=0,
                  single.v=single.v,ncalls.inner=0)
  }

  # initialize variables for benchmarking if applicable
  if (inner.optim=="benchmarkOpt") {                                           
    globals$raise_wnnls <- numeric(2e6)
    globals$raise_ipop  <- numeric(2e6)
    globals$raise_LRQP  <- numeric(2e6)
    globals$vs          <- matrix(0,ncol=n.v,nrow=2e6)
  }

  if (n.v==1) {                                                                 # search space is null space?
    if (verbose) catn("Only one predictor, optimization not required.")
    res <- list(v=cbind("optimizer"=1),conv=0,single.v=TRUE,ncalls.inner=1)
  } else {

    #########################
    # Look for sunny donors #
    #########################
    is.sunny <- logical(nW)
    for (i in seq_len(nW)) is.sunny[i] <- isSunny(e(i,nW),X)
    nS <- sum(is.sunny)

    if (nS<=1&&(!any(outer.optim == "none"))) {                                   # at most one 'sunny' donor, special cases!
      if (nS==0) {                                                                # no 'sunny' donors
        if (verbose) catn("No 'sunny' donors!")
        solution.type <- "nosunny"
        res <- solveNoSunny(X,Z,trafo.v,verbose=verbose)			
      } else {                                                                    # one 'sunny' donor
        if (verbose) catn("Only one 'sunny' donor!")                              # v does not matter and does not need to be optimized
        solution.type <- "onesunny"
        res <- solveOneSunny(X[,is.sunny,drop=FALSE],Z[,is.sunny,drop=FALSE],
                             trafo.v)			
      }
      if ((!do.optimize)&&(!is.null(v))) {                                        # restore custom v if provided 
        v           <- cbind(v)
        colnames(v) <- outer.optim
        res$v       <- v
      }  
    } else {                                                                      # at least two 'sunny' donors, standard case
      if (verbose) catn("Number of 'sunny' donors: ",nS," out of ",nW)

      # restrict optimization to sunny donors
      X <- X[,is.sunny,drop=FALSE]; Z <- Z[,is.sunny,drop=FALSE] 

      ############################
      # Checks for global optima #
      ############################
      if (check.global) {
        res <- checkGlobalOpt(X.orig,Z.orig,trafo.v,outer.args1$lb,               # check for feasibility of 'true' outer optimum
                              single.v=single.v,verbose=verbose,debug=debug)
        gRmspe <- res$rmspe
        if (!is.na(res$conv)) {
          if (verbose) catn("Unrestricted outer optimum (obtained by ignoring all ",
                            "predictors) is FEASIBLE even when respecting the ",
                            "predictors.")
          if (do.optimize) solution.type <- "global"
          do.optimize   <- FALSE
        } else {
          if (verbose) catn("Unrestricted outer optimum (obtained by ignoring all ",
                            "predictors) with RMSPE ",gRmspe,
                            " and MSPE (loss v) ",gRmspe^2," is INFEASIBLE when ",
                            "respecting the predictors.")
        }
        if (do.optimize) {                                                        # check for feasibility of 'sunny' outer optimum 
          res <- checkGlobalOpt(X,Z,trafo.v,outer.args1$lb,single.v=single.v,
                                verbose=verbose,debug=debug)
          if (!is.na(res$conv)) {
            do.optimize <- FALSE
            v <- res$v                                                            # check this !!! XXX changed 2017-04-19
          }
        }  
      } 

      # check availability of packages for inner optimizer
      if ((inner.optim=="ipopOpt")||(inner.optim=="benchmarkOpt"))
        checkPkg("kernlab")
      if ((inner.optim=="LowRankQPOpt")||(inner.optim=="benchmarkOpt"))
        checkPkg("LowRankQP")
                   
      # pre-allocate workspace for inner optimizer wnnlsOpt
      if ((inner.optim=="wnnlsOpt")||(inner.optim=="benchmarkOpt")) {                                           
        ME  <- 1L
        MA  <- sum(trafo.v$len.v)
        N   <- ncol(X)
        N   <- as.integer(nS)
        MDW <- ME+MA
        globals$Ipar  <- as.integer(c(ME=ME,MA=MA,MDW=MDW,N=N)) 
        globals$IWORK <- integer(MDW+N)
        globals$WORK  <- double(MDW+5*N)
        globals$IWORK[1:2] <- c(length(globals$WORK),length(globals$IWORK))
        globals$RNORM <- double(1)
        globals$MODE  <- integer(1)
        globals$X     <- double(N)
      }
                 
      if (do.optimize&&(is.na(gRmspe)||is.na(res$conv))) {                      # outer optimization needed
        if (is.null(seed)) seed <- NA
        comb <- expand.grid(seed=seed,outer.optim=outer.optim,
                            stringsAsFactors=FALSE)
        if ((inner.optim=="benchmarkOpt")&&(nrow(comb)>1)) 
          stop("benchmarkOpt not supported for multiple optimizations")
        if ((length(outer.opar)>0)&&(!isTRUE(names(outer.opar)%in%outer.optim))) 
        {
          if (length(outer.optim)>1) 
            stop("names of outer.opar do not match outer optimizers")
          outer.opar=list(outer.opar)
          names(outer.opar) <- outer.optim
        }
        results <- cases <- vector("list",nrow(comb))
        times <- rep(NA,length(results))
        for (i in seq_along(cases)) {
          cases[[i]] <- list(outer.optim=comb$outer.optim[i],seed=comb$seed[i],
                             outer.opar=outer.opar[[comb$outer.optim[i]]])
        }
        #######################################
        # Do (multiple) outer optimization(s) #
        #######################################
        if (is.null(cl)) {                                                      # without cluster
          if (verbose&&(length(results)>1)) 
            catn("Starting ",length(results)," optimizations with ",
                 length(outer.optim)," different outer optimizers and ",
                 length(seed)," different seeds.")
          for (i in seq_along(results)) {
            if (verbose&&(length(results)>1)) 
              catn("Optimization ",i," out of ",length(results),":")
            results[[i]] <- 
              atomOpt(cases[[i]],X,Z,trafo.v,single.v,inner.optim,
                      inner.args,outer.args1,starting.values,verbose,
                      debug)
          }            
        } else {                                                                # with cluster
          if (verbose) 
            catn("Distributing ",length(results)," optimizations with ",
                 length(outer.optim)," different outer optimizers and ",
                 length(seed)," different seeds to the cluster. ",
                 "Please hold the line!")
  #          clusterEvalQ(cl,library(MSCMT))
  #          clusterExport(cl,c("atomOpt","cases","X","Z","trafo.v","single.v",
  #                             "inner.optim","inner.args","outer.args1",
  #                             "starting.values","verbose","debug"),
  #                        envir=environment())
          results <- clusterApplyLB(cl,cases,atomOpt,X,Z,trafo.v,single.v,
                                    inner.optim,inner.args,outer.args1,
                                    starting.values,verbose,debug)
          if (verbose) catn("Optimization on cluster finished!")
        }              
        times   <- sapply(results,function(x) x$user.self)                      # collect benchmark information
        rmspes  <- sapply(results,function(x) x$rmspe)
        loss.v  <- sapply(results,function(x) x$loss.v)
        n.inner <- sapply(results,function(x) x$ncalls.inner)
        all.w   <- sapply(results,function(x) x$w)
        all.v   <- sapply(results,function(x) x$v[,ncol(x$v)])
        dim(times) <- dim(rmspes) <- dim(loss.v) <- dim(n.inner) <- 
          c(length(seed),length(outer.optim))
        rownames(times) <- rownames(rmspes) <- rownames(loss.v) <- 
          rownames(n.inner) <- seed  
        colnames(times) <- colnames(rmspes) <- colnames(loss.v) <- 
          colnames(n.inner) <- outer.optim
        colnames(all.w) <- colnames(all.v) <- 
          paste0(rep(outer.optim,each=length(seed)),".",
                 rep(seed,times=length(outer.optim)))
        tmploss <- loss.v
        tmploss[tmploss==0] <- Inf
        bestrun <- which.min(tmploss)
        if (verbose&&(length(cases)>1)) {
          catn("Best optimization run(s):")
          for (i in bestrun) catn("Run ",i," with optimizer ",
            cases[[i]]$outer.optim," and seed ",cases[[i]]$seed,".")
          if (isTRUE(any(is.infinite(tmploss)))) {
          catn("Optimization run(s) with problems:")
          for (i in which(is.infinite(tmploss))) catn("Run ",i," with optimizer ",
            cases[[i]]$outer.optim," and seed ",cases[[i]]$seed,".")
          }
        }
        res     <- results[[bestrun[1]]]               
      } else {                                                                    # outer optimization not needed
        if (any(outer.optim == "none")) {
          if (is.null(v)) {
            w     <- blow(outer.opar$w,colnames(X))
            rmspe <- sqrt(lossDep(Z,w))
            if (exists_v(w,X,Z,trafo.v,outer.args1$lb)) {
              v <- if (single.v) single_v(w,X,Z,trafo.v,outer.args1$lb) else 
                                 all_v(w,X,Z,trafo.v,outer.args1$lb)
            } else v <- NA
          } else {
            w     <- blow(outer.opar$w,colnames(X.orig))
            rmspe <- sqrt(lossDep(Z.orig,w))
            v     <- cbind("fixed"=v)
            loss.w <- c("fixed"=lossPred(X.orig,w,outer.opar$v,trafo.v))
          }
          res <- list(w=w,v=v,loss.v=rmspe^2,rmspe=rmspe,loss.w=loss.w,conv=0,
                      single.v=single.v,ncalls.inner=0)
        } else if (solution.type!="global") {
          fn.min.par  <- c(list(X=X,Z=Z,trafo=trafo.v$trafo),inner.args)   
          if (!is.matrix(v)) {                                                    # check this !!! XXX changed 2017-04-19
            v <- cbind(v)
            colnames(v) <- outer.optim
          }  
          v <- v[,apply(v,2,function(x) !any(is.na(x))),drop=FALSE]
          rownames(v) <- names.v
          mspe <- do.call(inner.optim,args=c(list(v=as.double(drop(v))),
                                             fn.min.par))
          w    <- do.call(inner.optim,args=c(list(v=as.double(drop(v))),
                                             fn.min.par,return.w=TRUE))
          names(w) <- colnames(X)
          res  <- list(w=w, v=v, loss.v=mspe, rmspe=sqrt(mspe), conv=0,
                       single.v=TRUE,ncalls.inner=2)
        }               
      } 
    }
  }
  
  res$v <- apply(res$v,2,function(x) x/do.call(std.v,list(x)))
  if (!is.matrix(res$v)) {                                                      # apply oversimplifies if length(v)==1
    res$v <- t(as.matrix(res$v))
    rownames(res$v) <- trafo.v$names.v
  }  

  if (!any(outer.optim == "none")) {                                            # sanitize/fix w only if applicable! (changed 2018-10-19)
    ws    <- apply(res$v[,apply(res$v,2,function(x) !any(is.na(x))),drop=FALSE],2,
                   function(x) wnnlsExt(x,X.orig,NULL,trafo.v,return.w=TRUE))
    ws    <- sanitize_w(ws)
    loss  <- apply(ws,2,function(w) lossDep(Z.orig,w))
    idx   <- which.min(loss)[1]
    res$loss.v   <- loss[idx]
    res$rmspe    <- sqrt(res$loss.v)
    rownames(ws) <- cnX
    res$w        <- ws[,idx]
    res$loss.w   <- apply(res$v,2,function(v) lossPred(X.orig,res$w,v,trafo.v))
  }

  res$trafo.v     <- trafo.v
  res$outer.rmspe <- gRmspe
  
  if (verbose) {
    catf("Final rmspe: ",res$rmspe,", mspe (loss v): ",res$rmspe^2,"\n",                           
         "Optimal weights:")
    print(ess(res$w))
  }
  if (debug) {
    print(res$v)
    print(res$loss.w)
    print(rbind(loss=loss,
                increase=pmin((loss-min(loss))/min(loss),loss-min(loss))),
          digits=16)
  }
  if (verbose||debug) cat("\n")

  if (inner.optim=="benchmarkOpt") {                                           
    c(res,list(solution.type=solution.type,
      raise_wnnls=globals$raise_wnnls[1:globals$NRUNS],
      raise_ipop=globals$raise_ipop[1:globals$NRUNS],
      raise_LRQP=globals$raise_LRQP[1:globals$NRUNS],
      bench_vs=globals$vs[1:globals$NRUNS,],
      bench_X=X,bench_Z=Z))
  } else {
    if ((!is.null(comb))&&(nrow(comb)>1))
      c(res,list(solution.type=solution.type,
                 multi.opt=list(rmspe=rmspes,loss.v=loss.v,time=times,
                                n.inner=n.inner,w=all.w,v=all.v))) else
      c(res,list(solution.type=solution.type))
  }  
}

ggplotBench <- function(results,rmspe=TRUE,time.lim,loss.lim,xlab="",ylab="",
                        title="Benchmark for different optimizers",
                        what=c("Time","Loss")) {
  what <- match.arg(what,several.ok=TRUE)                        
  if (!inherits(results,"mscmt")) 
    stop("results is not an object of class mscmt")
  if(!requireNamespace("reshape",quietly=TRUE))
    stop("package reshape not installed")
  if (is.null(results$placebo)) {                      
    if (is.null(results$multi.opt))
      stop("mscmt results with multiple optimization needed")
    if (missing(time.lim)) time.lim <- range(results$multi.opt$time)
    if (missing(loss.lim)) loss.lim <- range(if (rmspe) results$multi.opt$rmspe 
                                                   else results$multi.opt$loss.v)
    Time <- cbind(reshape::melt(results$multi.opt$time),typ="Time",
                  min=min(time.lim),max=max(time.lim))
    Loss <- cbind(reshape::melt(if (rmspe) results$multi.opt$rmspe else 
                                           results$multi.opt$loss.v),
                  typ=if (rmspe) "RMSPE" else "MSPE",
                  min=min(loss.lim),max=max(loss.lim))
    All <- rbind(if ("Time" %in% what) Time, if ("Loss" %in% what) Loss)  
    All$X2 <- factor(All$X2,levels = rev(levels(All$X2)),ordered = TRUE)
    ggplot(All,aes_string("X2","value")) + ggplot2::geom_boxplot() + 
      ggplot2::coord_flip() + ggplot2::facet_grid(. ~ typ, scales="free") + 
      xlab(xlab) + ylab(ylab) + ggplot2::geom_blank(aes(y=min)) + 
      ggplot2::geom_blank(aes(y=max)) + ggplot2::ggtitle(title)
  } else {
    ind <- sapply(results[1:(length(results)-1)],
                  function(x) !is.null(x$multi.opt))
    if (!any(ind))
      stop("no mscmt results with multiple optimization in placebo results")
    if (!(missing(time.lim)&&missing(loss.lim)))
      warning("time.lim and loss.lim ignored for placebo results")  
    All <- NULL
    for (i in which(ind)) {  
      Time <- cbind(reshape::melt(results[[i]]$multi.opt$time),typ="Time",
                    treated=names(results)[i])
      Loss <- cbind(reshape::melt(if (rmspe) results[[i]]$multi.opt$rmspe else 
                                             results[[i]]$multi.opt$loss.v),
                    typ=if (rmspe) "RMSPE" else "MSPE",
                    treated=names(results)[i])
      All <- rbind(All,if ("Time" %in% what) Time, if ("Loss" %in% what) Loss)  
    }
    All$X2 <- factor(All$X2,levels = rev(levels(All$X2)),ordered = TRUE)
    ggplot(All,aes_string("X2","value")) + ggplot2::geom_boxplot() + 
      ggplot2::coord_flip() + ggplot2::facet_grid(treated~typ, scales="free") + 
      xlab(xlab) + ylab(ylab) + ggplot2::ggtitle(title)  
  }    
}

## internal functions, maybe document and export some of the functions in 
## later release

################################################################################
################################################################################
# Helper functions for TIME SERIES HANDLING                                    #
################################################################################
################################################################################

## Guess frequency of time series data
##
## \code{freq} guesses the frequency of time series data.
##
## \code{freq} guesses the frequency of time series data.
##
## @param datum XX.
## @param only.first XX
## @return character string (one of annually, quarterly, monthly)
## TO DO(?): export?
## @export freq
freq <- function(datum,only.first=FALSE) {
  res <- ifelse(grepl("W",datum),"weekly",
         ifelse(grepl("Q",datum),"quarterly",
         ifelse(nchar(datum)<4,NA,
         ifelse(nchar(datum)==4,"annually",
         ifelse(nchar(datum)>=8,"daily","monthly")))))
  if (only.first) {
    if (length(unique(res))>1) 
      warning("frequency ambiguous, only first entry is used")
    res[1]  
  } else res
}

## sequence-function for data of various granularity (annual, quarterly, 
## monthly, weekly, daily) format for monthly data: YYYY-MM 
## (alternatively, e.g., YYYY/MM), format for quarterly data: YYYYQX 
## (alternatively, e.g., YYYY-QX), format for weekly data: YYYYWXX,
## format for daily-data: YYYYMMDD or YYYY-MM-DD

## Constructing names for time series data of various frequencies
##
## \code{seqAQM} constructs sequences of names for annual, quarterly, monthly,
## weekly and daily time series data.
##
## \code{seqAQM} constructs sequences of names for annual, quarterly, monthly,
## weekly and daily time series data.
##
## @param from XX.
## @param to XX
## @param by XX
## @return vector of character strings XX
## TO DO(?): export?
## @export seqAQM
seqAQM <- function(from,to,by=1) {
  fr   <- switch(freq(from),"quarterly"=4,"monthly"=12,"weekly"=53,"daily"=366,
                 1)
  DtoN <- function(x,fr) {
            year <- as.numeric(substr(x,1,4))
            if (fr==1) year else
            if (fr==366) 366*year + as.numeric(format(as.Date(x),"%j")) - 1 else
            fr*year + as.numeric(substr(x,6,nchar(x))) - 1
          }
  NtoD <- function(x,fr) {
            if (fr==1)   as.character(x) else
            if (fr==4)   paste0(x%/%4,"Q",(x%%4)+1) else
            if (fr==12)  paste0(x%/%12,"-",sprintf("%02d",(x%%12)+1)) else
            if (fr==53)  paste0(x%/%53,"W",sprintf("%02d",(x%%53)+1)) else
            format(as.Date(paste0(x%/%366,"-",(x%%366)+1),format="%Y-%j"),
                   format="%Y-%m-%d")
          }        
  NtoD(seq(DtoN(from,fr),DtoN(to,fr),by=by),fr)        
}

## convert data with character names (annual, quarterly, monthly data)
## to class 'ts'

## Convert (actual) time series data to class 'ts'
##
## \code{AQM2ts} converts annual, quarterly and monthly time series data to 
## class 'ts'.
##
## \code{AQM2ts} converts annual, quarterly and monthly time series data to 
## class 'ts'.
##
## @param data XX.
## @return object of class \code{ts}
#' @importFrom stats ts
#' @importFrom utils head tail
## TO DO(?): export?
## @export AQM2ts
AQM2ts <- function(data) {
  data  <- data[order(names(data))]
  sanitize <- function(nam,fre) 
    switch(as.character(fre),
           "4"   = paste0(substr(nam,1,4),"Q",substr(nam,nchar(nam),nchar(nam))),
           "12"  = paste0(substr(nam,1,4),"-",sprintf("%02d",as.numeric(substr(nam,6,nchar(nam))))),
           "53"  = paste0(substr(nam,1,4),"W",sprintf("%02d",as.numeric(substr(nam,6,nchar(nam))))),
           "366" = format(as.Date(nam),"%Y-%m-%d"),
           substr(nam,1,4))
  mq   <- function(nam,fre) 
    switch(as.character(fre),
           "4"   = as.numeric(substr(nam,nchar(nam),nchar(nam))),
           "12"  = as.numeric(substr(nam,6,nchar(nam))),
           "53"  = as.numeric(substr(nam,6,nchar(nam))),
           "366" = as.numeric(format(as.Date(nam),format="%j")),
           NULL)
  snam  <- head(names(data),1)
  fr    <- switch(freq(snam),"quarterly"=4,"monthly"=12,"weekly"=53,"daily"=366,
                  1)
  syear <- as.numeric(substr(snam,1,4))
  smq   <- mq(snam,fr)
  enam  <- tail(names(data),1)
  eyear <- as.numeric(substr(enam,1,4))
  emq   <- mq(enam,fr)
  AQM   <- seqAQM(snam,enam)
  res   <- rep(NA,length(AQM))
  names(res) <- AQM
  res[sanitize(names(data),fr)] <- if (mode(data)=="numeric") data else 
                                     rep(NA,length(data))
  ts(as.numeric(res),start=c(syear,smq),end=c(eyear,emq),frequency=fr)
}

## @export AQM2Date
AQM2Date <- function(rng,from,to) {
  if (missing(from)) from <- rng[1]
  if (missing(to))   to   <- rng[2]
  from <- switch(freq(from),
            "annually"  = as.Date(paste0(from,"-01-01")),
            "monthly"   = as.Date(paste0(from,"-01")),
            "quarterly" = as.Date(paste0(substr(from,1,4),"-",
                                    as.numeric(substr(from,nchar(from),
                                                      nchar(from)))*3-2,"-01")),
            "weekly"    = as.Date(paste0(substr(from,1,4),"-",
                                         substr(from,6,nchar(from)),"-0"),
                                  format="%Y-%W-%w"),
            "daily"     = as.Date(from))
  to   <- switch(freq(to),
            "annually"  = as.Date(paste0(to,"-12-31")),
            "monthly"   = as.Date(paste0(to,"-28")),
            "quarterly" = as.Date(paste0(substr(to,1,4),"-",
                                    as.numeric(substr(to,nchar(to),
                                                      nchar(to)))*3,"-28")),
            "weekly"    = as.Date(paste0(substr(to,1,4),"-",
                                         substr(to,6,nchar(to)),"-0"),
                                  format="%Y-%W-%w"),
            "daily"     = as.Date(to))
  c(from=from,to=to)
}

#' @importFrom stats window
## @export AQMwindow
AQMwindow <- function(data,range) {
  fr   <- switch(freq(range,only.first=TRUE),"quarterly"=4,"monthly"=12,
                 "weekly"=53,"daily"=366,1)
  year <- as.numeric(substr(range,1,4))
  mq   <- switch(as.character(fr),
                   "4"   = as.numeric(substr(range,nchar(range),nchar(range))),
                   "12"  = as.numeric(substr(range,6,nchar(range))),
                   "53"  = as.numeric(substr(range,6,nchar(range))),
                   "366" = as.numeric(format(as.Date(range),format="%j")),
                 NULL)
  window(data,start=c(year[1],mq[1]),end=c(year[2],mq[2]))
}

#' @importFrom stats window
## @export AQMtail
AQMtail <- function(data,after) {
  fr   <- switch(freq(after,only.first=TRUE),"quarterly"=4,"monthly"=12,
                 "weekly"=53,"daily"=366,1)
  year <- as.numeric(substr(after,1,4))
  mq   <- switch(as.character(fr),
                   "4"   = as.numeric(substr(after,nchar(after),nchar(after))),
                   "12"  = as.numeric(substr(after,6,nchar(after))),
                   "53"  = as.numeric(substr(after,6,nchar(after))),
                   "366" = as.numeric(format(as.Date(after),format="%j")),
                 NULL)
  if (fr==1) year <- year + 1 else {
    if (fr==mq) {  
      mq <- 1; year <- year + 1
    } else mq <- mq + 1               
  }
  window(data,start=c(year,mq))
}

#' @importFrom stats frequency start end
ts2df <- function(tsl,maxfreq) {
  if (missing(maxfreq)) 
    maxfreq <- if (is.list(tsl)) max(sapply(tsl,frequency)) else frequency(tsl)
  if (is.list(tsl)) lapply(tsl,ts2df,maxfreq=maxfreq) else {
    freq  <- frequency(tsl)
    if (freq==1) {
        startts <- as.Date(paste0(start(tsl)[1],"-07-01"))
        endts   <- as.Date(paste0(end(tsl)[1],"-07-01"))
      byts <- "year"
    }
    if (freq==4) {
      startts <- as.Date(paste0(start(tsl)[1],"-",(start(tsl)[2]-1)*3+2,"-15"))
      endts   <- as.Date(paste0(end(tsl)[1],"-",(end(tsl)[2]-1)*3+2,"-15"))
      byts <- "quarter"
    }
    if (freq==12) {
      startts <- as.Date(paste0(start(tsl)[1],"-",start(tsl)[2],"-15"))
      endts   <- as.Date(paste0(end(tsl)[1],"-",end(tsl)[2],"-15"))
      byts <- "month"
    }
    if (freq==53) {
      startts <- as.Date(paste0(start(tsl)[1],"-",start(tsl)[2],"-0"),format="%Y-%W-%w")
      endts   <- as.Date(paste0(end(tsl)[1],"-",end(tsl)[2],"-0"),format="%Y-%W-%w")
      byts <- "week"
    }
    if (freq==366) {
      startts <- as.Date(paste0(start(tsl)[1],"-",start(tsl)[2]),format="%Y-%j")
      endts   <- as.Date(paste0(end(tsl)[1],"-",end(tsl)[2]),format="%Y-%j")
      byts <- "day"
    }
    res <- cbind(seq.Date(startts,endts,byts),as.data.frame(tsl))
    colnames(res) <- c("date",
                       if (is.null(colnames(tsl))) "value" else colnames(tsl))
    res                   
  }
}

################################################################################
################################################################################
# Helper functions for OPTIMIZATION PROCEDURE                                  #
################################################################################
################################################################################

## generates 'trafo'-matrix A
genA <- function(len) 
  matrix(unlist(sapply(len,function(x) c(rep.int(1,x),rep(0,sum(len))))),
         nrow=sum(len))[,seq_along(len)]

## generates a trafo for v
genTrafo <- function(n.v=NULL,names.v=NULL,len.v) {
  has.trafo <- !(missing(len.v)||is.null(len.v)||all(len.v==1))
  if (has.trafo) {                                                              # for SCM*T*, 'real' trafo
    trafo     <- function(v) rep.int(v,len.v) 
    A         <- genA(len.v)
    n.v       <- length(len.v)
  } else {
    if (is.null(n.v)) 
      n.v     <- length(names.v)
    trafo     <- function(v) v
    A         <- diag(n.v)
    len.v     <- rep.int(1,n.v)
  }    
  list(n.v=n.v,names.v=names.v,trafo=trafo,A=A,has.trafo=has.trafo,len.v=len.v)
}

## eliminate and re-impute (near) zero elements of a vector
ess  <- function(x,lb=0) x[x>lb]
blow <- function(x,full.names) {
  stopifnot(isTRUE(all(names(x) %in% full.names)))
  res           <- rep(0,length(full.names))
  names(res)    <- full.names
  res[names(x)] <- x
  res
}

## unit vectors
e <- function(i,n) { tmp <- numeric(n); tmp[i] <- 1/length(i); tmp }

## fast calculation of V'MV
fastVpMV <- function(M,V) .Call(C_fastVpMV,M,V)

## fast calculation of V'M'MV
fastVpMpMV <- function(M,V) .Call(C_fastVpMpMV,M,V)

## fast calculation of M'diag(V)M
fastMpdVM <- function(M,V) .Call(C_fastMpdVM,M,V)

## calculation of loss for dependent variables
lossDep <- function(Z,w) fastVpMpMV(Z,w)/nrow(Z)

## calculation of loss for predictor variables
lossPred <- function(X,w,v,trafo.v) 
  if (missing(trafo.v)) abs(fastVpMV(fastMpdVM(X,v),w)) else 
                        abs(fastVpMV(fastMpdVM(X,trafo.v$trafo(v)),w))

## is vector w0 sunny wrt. X?
#' @importFrom lpSolve lp
isSunny <- function(w0,X) 
  isTRUE(all.equal(suppressWarnings(
    lp(direction="min",
       objective.in=c(rep(0,length(w0)),1),
       const.mat=rbind(c(rep(1,length(w0)),0),
                       cbind(X,-(X %*% w0))),
       const.rhs=c(1,rep(0,nrow(X))),
       const.dir=rep("==",nrow(X)+1)            
      ))$objval,1))

## calculate 'global' optimum and check for feasibility
checkGlobalOpt <- function(X,Z,trafo.v,lb,single.v=FALSE,verbose=FALSE,
                           debug=FALSE) {
  tmp   <- wnnlsGetGlobalOpt(Z)
  w     <- tmp$w
  rmspe <- tmp$rmspe
  if (exists_v(w,X,Z,trafo.v,lb)) {
    list(w=w,
         v=if (is.na(single.v)) 
           cbind("backup"=single_v(w,X,Z,trafo.v,lb,check.exists=TRUE)) else
           if (isTRUE(single.v))
             cbind("max.order"=single_v(w,X,Z,trafo.v,lb,verbose=verbose,
                                        debug=debug)) else 
             all_v(w,X,Z,trafo.v,lb,verbose=verbose,debug=debug),
         loss.v=rmspe^2,rmspe=rmspe,conv=0,single.v=single.v)
  } else list(w=w,v=NA,loss.v=rmspe^2,rmspe=rmspe,conv=NA,single.v=NA)
}

## calculate 'global' optimum with wnnls
wnnlsGetGlobalOpt <- function(X) {
  w <- wnnls(A=X,B=matrix(0,nrow=nrow(X)),E=matrix(rep(1,ncol(X)),nrow=1),
             F=matrix(1),check.input=FALSE)$x
  list(w=w,rmspe=sqrt(lossDep(X,w)))
}    

## try to improve outer target function (if solution of inner opt. is ambig.)   # not used by default
improveZw <- function(Z,X,w,tol=.Machine$double.eps) {
  w <- wnnlsInt(W=.Call(C_prepareW1,Z,X,w),ME=nrow(X)+1,MA=nrow(Z),
                N=length(w))$x
  w[w<tol] <- 0
  w/sum(w)
}

## solve outer optimization when there are no sunny donors
solveNoSunny <- function(X,Z,trafo.v,lb.loss=.Machine$double.eps,
                         lb=sqrt(.Machine$double.eps),verbose=FALSE) {
  n <- ncol(X)
  w <- wnnls(A=Z,B=matrix(0,nrow=nrow(Z)),E=rbind(X,rep(1,n)),
             F=matrix(c(rep(0,nrow(X)),1),ncol=1))$x
  names(w) <- colnames(X)             
  w[w<lb]  <- 0
  w <- w/sum(w)
  v <- rep(1/trafo.v$n.v,trafo.v$n.v)
  names(v) <- trafo.v$names.v
  loss.w <- lossPred(X,w,v,trafo.v)
  if (loss.w>lb.loss) 
    if (verbose) catn("Warning: loss.w is ",loss.w,">0!")
  list(w=w,loss.w=loss.w,v=cbind("max.order"=v),loss.v=lossDep(Z,w),
       rmspe=sqrt(lossDep(Z,w)),conv=0,single.v=TRUE)
}

## 'solve' outer optimization when there is exactly one sunny donor
solveOneSunny <- function(X,Z,trafo.v) {
  w        <- 1
  names(w) <- colnames(X)             
  v        <- rep(1/trafo.v$n.v,trafo.v$n.v)
  names(v) <- trafo.v$names.v
  loss.w   <- lossPred(X,w,v,trafo.v)
  list(w=w,loss.w=loss.w,v=cbind("max.order"=v),loss.v=lossDep(Z,w),
       rmspe=sqrt(lossDep(Z,w)),conv=0,single.v=TRUE)
}

################################################################################
################################################################################
# Helper functions for NICE OUTPUT                                             #
################################################################################
################################################################################

## format output as comma separated list with maximum line width and indentation
csl <- function(title,x,max.width=options()$width,min.indent=0,newline=FALSE,
                sep=", ") {
  n1  <- max(min.indent,nchar(title)+2)
  na  <- max(0,n1 - nchar(title) - 2)
  n2  <- max.width - n1
  nc  <- nchar(x)
  i   <- n1
  out <- paste0(title,if (title!="") ": " else "  ",
                paste0(rep(" ",na),collapse=""))
  for (j in seq_along(x)) {
    if ((i+nc[j]>max.width-1)||((j>1)&&newline)) {
      out <- paste0(out,"\n",paste0(rep(" ",n1),collapse=""),x[j],
                    if (j<length(x)) sep else "\n")
      i   <- n1 + nc[j] + nchar(sep)
    } else {
      out <- paste0(out,x[j],if (j<length(x)) sep else "\n")
      i   <- i + nc[j] + nchar(sep)
    }   
  }
  out
}

## format tab-like
tab <- function(nam,val,sep=": ") {
  nc <- max(nchar(nam))
  paste0(format(nam,width=nc,justify="left"),sep,val)
}

## wrap text with title
wrap <- function(title,x,max.width=options()$width,min.indent=0) {
  x <- unlist(strsplit(x," "))
  csl(title,x,max.width,min.indent,sep=" ")
}

## wrap matrices with titles (as argument names)
wrapM <- function(...,max.width=options()$width,min.indent=0,matrix.width=12) {
  out <- ""
  arg <- list(`...`)
  argnames <- names(arg)
  if (is.null(argnames)) argnames <- rep("",length(arg))
  op  <- options()
  n1  <- max(min.indent,max(nchar(argnames))+2)
  n2  <- max.width - n1
  options(width=n2)
  for (i in seq_along(arg)) 
    if (is.null(rownames(arg[[i]]))) 
      rownames(arg[[i]]) <- rep("",nrow(arg[[i]]))
  lrn <- max(sapply(arg,function(x) max(nchar(rownames(x)))))
  for (i in seq_along(arg)) {
    rownames(arg[[i]]) <- format(rownames(arg[[i]]),width=lrn)
    na  <- max(0,n1 - nchar(argnames[i]) - 2)
    out <- paste0(out,argnames[i],if (argnames[i]!="") ": " else "  ",
                  paste0(rep(" ",na),collapse=""),collapse="")
    if (any(is.character(arg[[i]]))) {
      tmp <- as.data.frame(arg[[i]])
    } else {
      colnames(arg[[i]]) <- format(colnames(arg[[i]]),width=matrix.width,
                                   justify="right")
      tmp <- format(arg[[i]],justify="right",width=matrix.width,na.encode=TRUE)
      tmp[which(tmp==format("NA",justify="right",width=matrix.width))] <- NA
    }                  
    tmp <- capture.output(print(tmp,na.print="",quote=FALSE))
    tmp <- paste0(tmp,sep=paste0("\n",paste0(rep(" ",n1),collapse=""),sep=""))
    out <- paste0(out,paste0(tmp,collapse=""),"\n",collapse="")
  }
  options(op)
  out
}

## wrap text with title (... version)
catw <- function(title,...) {
  x <- paste0(unlist(list(`...`)),collapse="")
  cat0(wrap(title,x))
}

## functions for pretty printing
now  <- function() paste0(format(Sys.time(),"%H:%M:%S"),": ")
now1 <- function() paste0(format(Sys.time(),"%H:%M:%S"),":")
cat0 <- function(...) cat(...,sep="")

## cat without separator and with fill option
catf <- function(...) cat(...,fill=TRUE,sep="")

## cat without separator with fill and label (current time) option
catn <- function(...) 
  cat(unlist(strsplit(paste0(unlist(list(`...`)),collapse="")," ")),
      fill=TRUE,sep=" ",labels=now1())

## align
align <- function(...) {
  obj <- list(`...`)
  nc  <- sapply(obj,function(x) max(nchar(format(x))))
  tmp <- vector("list",length(obj))
  for (i in seq_along(tmp))
    tmp[[i]] <- format(obj[[i]],width=nc[i],justify="left")
  do.call("paste0",tmp)
}


################################################################################
################################################################################
# Helper functions for LINEAR PROGRAMS                                         #
################################################################################
################################################################################

#' @importFrom lpSolveAPI lp.control make.lp set.rhs set.objfn set.column
#' @importFrom lpSolveAPI set.constr.type set.bounds get.objective get.variables
lp3 <- function(direction,objective.in,const.mat,const.rhs,const.dir,
                #vmin=0,vmax=1,
                ...) {
  n <- ncol(const.mat)
#  const.mat <- const.mat[const.dir!="<=",]
#  const.rhs <- const.rhs[const.dir!="<="]
#  const.dir <- const.dir[const.dir!="<="]
  const.dir[const.dir=="=="] <- "="
  mylp <- make.lp(nrow(const.mat),n)
  set.rhs(mylp,const.rhs)
  set.constr.type(mylp,const.dir)
  for (i in 1:n) set.column(mylp,i,const.mat[,i])
#  set.bounds(mylp,lower=rep(vmin,n),upper=rep(vmax,n))
  set.objfn(mylp,objective.in)
  lp.control(mylp,presolve=c("lindep","rows","rowdominate"),
             sense=direction,timeout=10,verbose="neutral")
  status <- solve(mylp)
  list(solution=if (status==0) get.variables(mylp) else NA, status=status,
       objval=if (status==0) get.objective(mylp) else NA)
}

#' @importFrom Rglpk Rglpk_solve_LP
#' @importFrom lpSolve lp
lpsolve <- function(direction,objective.in,const.mat,const.rhs,const.dir,
                    not.fixed=NULL,scale=196L,tol.lp=1e-13,
                    verbose=TRUE,debug=FALSE,check.min=FALSE,...) {
  tol_mul <- 1 + sqrt(.Machine$double.eps)
  rglpk <- Rglpk_solve_LP(obj=objective.in,mat=const.mat,rhs=const.rhs,         # 1st try: solve lp via Rglpk
                          dir=const.dir,max=isTRUE(direction=="max"),
                          control=list(tm_limit=5000,presolve=TRUE))
  rglpk$objval <- rglpk$optimum
  dglpk <- lp_diag(objective.in,const.mat,const.rhs,const.dir,rglpk)
  mglpk <- if (check.min)                                                       # variable defined as minimum indeed minimal?
             isTRUE(rglpk$solution[length(rglpk$solution)] <=
                      min(rglpk$solution[not.fixed]) * tol_mul) else TRUE
  sglpk <- isTRUE(rglpk$status==0)
  score.glpk <- (!mglpk) + (!sglpk) + max(dglpk[2:3])
  if (score.glpk > tol.lp) {                                                    # minimum not ok or status not ok or constraints violated?
    rlpA <- lp3(objective.in,const.mat=const.mat,const.rhs=const.rhs,           # 2nd try: solve lp via lpSolveAPI
                const.dir=const.dir,direction=direction,scale=scale)
    dlpA <- lp_diag(objective.in,const.mat,const.rhs,const.dir,rlpA)
    mlpA <- if (check.min)                                                      # variable defined as minimum indeed minimal?
               isTRUE(rlpA$solution[length(rlpA$solution)] <=
                        min(rlpA$solution[not.fixed]) * tol_mul) else TRUE
    slpA <- isTRUE(rlpA$status==0)
    score.lpA <- (!mlpA) + (!slpA) + max(dlpA[2:3])
    if (score.lpA > tol.lp) {                                                   # minimum not ok or status not ok or constraints violated?
      rlp <- lp(objective.in,const.mat=const.mat,const.rhs=const.rhs,           # 3rd try: solve lp via lpSolve
                const.dir=const.dir,direction=direction,scale=scale)
      dlp <- lp_diag(objective.in,const.mat,const.rhs,const.dir,rlp)
      mlp <- if (check.min)                                                     # variable defined as minimum indeed minimal?
                 isTRUE(rlp$solution[length(rlp$solution)] <=
                          min(rlp$solution[not.fixed]) * tol_mul) else TRUE
      slp <- isTRUE(rlp$status==0)
      score.lp <- (!mlp) + (!slp) + max(dlp[2:3])
      if (score.lp > tol.lp) {                                                  # minimum not ok or status not ok or constraints violated?
        if (min(score.glpk,score.lpA,score.lp)>100*tol.lp) {
          if (debug) 
            catn("lp solvers Rglpk, lpSolveAPI and lpSolve failed severely. score: ", min(score.glpk,score.lpA,score.lp))
        } else if (debug) catn("lp solvers failed.")
        if (debug) print(rbind("glpk"=dglpk,"lpSolveAPI"=dlpA,"lpSolve"=dlp))
      } 
      switch(which.min(c(score.glpk,score.lpA,score.lp)),
              "1" = rglpk, "2" = rlpA, "3" = rlp)
    } else rlpA
  } else rglpk
}

## linear programs for ratios in target function
lprsolve <- function(direction = "min", a0, a, d0,d, const.mat, const.dir, 
                     const.rhs, tol.lp=1e-13, scale=196L,
                     verbose=FALSE,debug=FALSE,...) {
  const.mat    <- cbind(c(-const.rhs,d0),rbind(const.mat,d))
  objective.in <- c(a0,a)
  const.dir    <- c(const.dir,"==")
  const.rhs    <- c(rep(0,length(const.rhs)),1)
  tmp <- lpsolve(objective.in=objective.in,const.mat=const.mat,
                 const.dir=const.dir,const.rhs=const.rhs,direction=direction,
                 tol.lp=tol.lp,check.min=FALSE,scale=scale,verbose=verbose,
                 debug=debug,...)
  if (tmp$status==0) {
    if (tmp$solution[1]>0) tmp$solution[-1]/tmp$solution[1] else 
                           rep(0,length(a))
  } else rep(NA,length(a))
}

# new version of PUFAS which allows for inequality constraints as well as 
#   checking for unique optimal solution (default) and for unique feasible 
#   solution (by setting optimal=FALSE)
pufas <- function(objective.in,const.mat,const.rhs,const.dir,solution,tol=0,
                  do.check=FALSE,tol.check=tol,tol.cond=.Machine$double.eps,
                  optimal=TRUE,scale=196L,verbose=FALSE,debug=FALSE,...) {
  # x is an optimal basic (!) solution of max objective x 
  # s.t. const.mat x = const.rhs, x >= 0
  # Appa (2002): solve linear program max d x s.t. const.mat x = const.rhs, 
  #              objective x = objective solution, x >= 0, with d = 1 for 
  #              zero components of solution and 0 else
  
  if (do.check && !(rcond(const.mat[,solution>tol.check])>tol.cond)) return(NA) # check whether given solution is a basis solution, return NA else

  eqs       <- const.dir %in% c("=","==")
  ineq_less <- const.dir %in% c("<","<=")
  ineq_more <- const.dir %in% c(">",">=")
  mat_slack <- const.mat
  mat_slack[eqs,] <- 0
  mat_slack[ineq_more,] <- -mat_slack[ineq_more,]
  slack_zero      <- rep(NA,nrow(const.mat))
  slack_zero[eqs] <- FALSE
  slack_zero[ineq_less] <- ( const.mat[ineq_less,] %*% 
                                              solution >= const.rhs[ineq_less] )
  slack_zero[ineq_more] <- ( const.mat[ineq_more,] %*% 
                                              solution <= const.rhs[ineq_more] )
  
  obj.new <- 1*(solution<=tol)-colSums(mat_slack[slack_zero,,drop=FALSE])
  
  if (optimal) {
    mat.new <- rbind(const.mat,objective.in)
    new.rhs <- c(const.rhs,sum(objective.in*solution))
    if (length(const.dir)==1) const.dir <- rep(const.dir,nrow(const.mat))
    new.dir <- c(const.dir,"==")
  } else {
    mat.new <- const.mat
    new.rhs <- const.rhs
    new.dir <- const.dir
  }

  new_lp_res <- lpsolve(objective.in=obj.new,const.mat=mat.new,
                        const.rhs=new.rhs,const.dir=new.dir,direction="max",
                        verbose=verbose,debug=debug,scale=scale,...)

# if (new_lp_res$status==3) return(FALSE)                                       # if new_lp is unbounded, then there are many alternative solutions, FIX ME!!!
  if (new_lp_res$status!=0) return(NA)
  new_slack_zero      <- rep(NA,nrow(const.mat))
  new_slack_zero[eqs] <- FALSE
  new_slack_zero[ineq_less] <- ( const.mat[ineq_less,] %*% 
                                   new_lp_res$solution >= const.rhs[ineq_less] )
  new_slack_zero[ineq_more] <- ( const.mat[ineq_more,] %*% 
                                   new_lp_res$solution <= const.rhs[ineq_more] )
  all(new_lp_res$solution[solution<=tol]<=tol) && 
    all( !slack_zero | new_slack_zero)
}

################################################################################
################################################################################
# Helper functions for evaluating and improving solutions of linear programs   #
################################################################################
################################################################################
  
max_error <- function(a,b,subset=NULL,direction="==",correct="none") {
  if (!is.null(subset)) { a <- a[subset]; b <- b[subset] }
  if (length(a)>0) {
    abserr <- switch(direction,
                     "==" = abs(a-b),"<=" = pmax(a-b,0),">=" = pmax(b-a,0))
    base   <- switch(correct,
                     "none" = pmax(abs(a),abs(b)),"a"=abs(a),"b"=abs(b))
    relerr <- ifelse(base==0,Inf,abserr/base)
    max(pmin(abserr,relerr))
  } else 0
}

lp_diag <- function(objective.in,const.mat,const.rhs,const.dir,lpres) {
  if (is.null(lpres$solution)) {
    sapply(lpres, function(x) lp_diag(objective.in,const.mat,const.rhs,
                                      const.dir,x$solution)) 
  } else {
    if ((lpres$status==0)&&(length(lpres$solution)==ncol(const.mat))) {
      obj     <- sum(objective.in*lpres$solution)
      lhs     <- const.mat%*%lpres$solution
      eqmax   <- max_error(lhs,const.rhs,subset=(const.dir=="=="))
      ineqmax <- max(max_error(lhs,const.rhs,subset=(const.dir=="<="),
                               direction="<="),
                     max_error(lhs,const.rhs,subset=(const.dir==">="),
                               direction=">="))
      minimum <- min(lpres$solution[-length(lpres$solution)])
      target  <- lpres$solution[length(lpres$solution)]
    } else obj <- lhs <- eqmax <- ineqmax <- minimum <- target <- Inf  
    c("objective"=obj,"eqmaxinf"=eqmax,"ineqmaxinf"=ineqmax,
      "status"=lpres$status,"minimum"=minimum,"target"=target)
  }
}

sanitize_w <- function(w,tol=1e-14) {
  if (is.matrix(w)) apply(w,2,function(x) sanitize_w(x,tol=tol)) else {
    w[w<tol] <- 0
    w/sum(w)
  }
}

sanitize_v <- function(v,X,Z,trafo.v,lb=1e-8,rel.tol=1e-3,tol.lpq=1e-13,
                       tol.loss=1e-12,verbose=FALSE,debug=FALSE) {
  v <- v/max(v)
  check_lp <- function(v,w) {
    M0   <- M(w,X,trafo.v)
    K    <- ncol(M0)
    J    <- nrow(M0)
    const.mat <- M0
    const.rhs <- rep(0,nrow(M0))
    const.dir <- ifelse(w==0,">=","==")
    lhs     <- const.mat%*%v
    eqmax   <- max(c(0,abs(lhs-const.rhs)[const.dir=="=="]))
    ineqmax <- max(c(0,(lhs-const.rhs)[const.dir=="<="],
                       (const.rhs-lhs)[const.dir==">="]))
    max(eqmax,ineqmax)
  }
  
  if (any(is.na(v))) return(v)
  oldv.w <- wnnlsExt(v,X,NULL,trafo.v,return.w=TRUE)
  oldv.l <- lossDep(Z,oldv.w)
  newv <- v
  newv[v<lb*(1+rel.tol)] <- lb
  newv[v>1*(1-rel.tol)]  <- 1
  newv.w <- wnnlsExt(newv,X,NULL,trafo.v,return.w=TRUE)
  newv.l <- lossDep(Z,newv.w)

  if (debug) catn("sanitize_v: old v: ",oldv.l,", new v: ",newv.l,
                  ", change new vs. old: ",newv.l-oldv.l)

  if ((oldv.l>0)&&(min((newv.l-oldv.l)/oldv.l,newv.l-oldv.l)<tol.loss)&&
      (check_lp(newv,newv.w)-check_lp(v,oldv.w)<tol.lpq)) newv else v
}

sanitize_res <- function(res,X,Z,trafo.v,lb=1e-8,rel.tol=1e-3,tol.lpq=1e-13,
                         tol.loss=1e-12,verbose=FALSE,debug=FALSE) {
  res$v <- res$v/max(res$v)
  check_lp <- function(v,w) {
    M0   <- M(w,X,trafo.v)
    K    <- ncol(M0)
    J    <- nrow(M0)
    const.mat <- M0
    const.rhs <- rep(0,nrow(M0))
    const.dir <- ifelse(w==0,">=","==")
    lhs     <- const.mat%*%v
    eqmax   <- max(c(0,abs(lhs-const.rhs)[const.dir=="=="]))
    ineqmax <- max(c(0,(lhs-const.rhs)[const.dir=="<="],
                       (const.rhs-lhs)[const.dir==">="]))
    max(eqmax,ineqmax)
  }
  
  if (any(is.na(res$v))) return(res)
  wnames <- names(res$w)
  res$w  <- wnnlsExt(res$v,X,NULL,trafo.v,return.w=TRUE)
  res$loss.v <- lossDep(Z,res$w)
  res$rmspe  <- sqrt(res$loss.v)
  newv <- res$v
  newv[res$v<lb*(1+rel.tol)] <- lb
  newv[res$v>1*(1-rel.tol)]  <- 1
  newv.w <- wnnlsExt(newv,X,NULL,trafo.v,return.w=TRUE)
  newv.l <- lossDep(Z,newv.w)

  if (debug) catn("sanitize_res: old v: ",res$loss.v,", new v: ",newv.l,
                  ", change new vs. old: ",newv.l-res$loss.v)

  if ((res$loss.v>0)&&
      (min((newv.l-res$loss.v)/res$loss.v,newv.l-res$loss.v)<tol.loss)&&
      (check_lp(newv,newv.w)-check_lp(res$v,res$w)<tol.lpq)) {
    res$w <- newv.w
    res$v <- newv
    res$loss.v <- newv.l
    res$rmspe  <- sqrt(res$loss.v)
  }
  names(res$w) <- wnames
  res
}

vdiff <- function(v) {
  n <- ncol(v)
  if (n==1) return(0)
  idx   <- expand.grid(1:n,1:n)
  idx   <- idx[idx[,1]<idx[,2],,drop=FALSE]
  diffs <- apply(idx,1,function(x) 
                         max(pmax(abs(v[,x[1]]-v[,x[2]]),
                             abs(v[,x[1]]-v[,x[2]])/(v[,x[1]]+v[,x[2]]))))
  max(diffs)
}

wdiff <- function(v) {
  n <- ncol(v)
  if (n==1) return(0)
  idx   <- expand.grid(1:n,1:n)
  idx   <- idx[idx[,1]<idx[,2],,drop=FALSE]
  diffs <- apply(idx,1,function(x) max(abs(v[,x[1]]-v[,x[2]])))
  max(diffs)
}

bestifnear <- function(v,X,Z,trafo.v,tol=1e-3) {
  if(vdiff(v)<tol) {
    losses <- apply(v,2,function(x) {
      w <- wnnlsExt(x,X,NULL,trafo.v,return.w=TRUE)
      lossDep(Z,w)
    })
    v[,which.min(losses)]
  } else v[,1]
}

#' Compare MSCMT estimation results
#'
#' \code{compare} collects estimation results from \code{\link{mscmt}} for
#' comparison purposes.
#'
#' \code{compare} collects (potentially many) estimation results from 
#' \code{\link{mscmt}} in a special object of class \code{"mscmt"}, which 
#' includes a component \code{"comparison"} where the different estimation 
#' results are aggregated.
#' This aggregated information is used by the \code{\link{ggplot.mscmt}} and 
#' \code{\link{print.mscmt}} methods to present summaries of the different
#' results.
#'
#' @param ... Objects of class \code{"mscmt"} or (a) list(s) containing objects
#' of class \code{"mscmt"}.
#' @param auto.name.prefix A character string (default: "") internally used to 
#' facilitate automatic naming in nested lists of unnamed estimation results.
#' 
#' @return An object of class \code{"mscmt"}, which itself contains the 
#' individual estimation results as well as a component \code{"comparison"} 
#' with aggregated information.
#' @importFrom stats sd
#' @export compare
compare <- function(...,auto.name.prefix="") {
  arg      <- list(`...`)
  argnames <- names(arg)
  if (is.null(argnames)) argnames <- rep("",length(arg))
  res <- list()
  agg <- NULL
  scalars <- c("loss.v","rmspe","loss.w","solution.type","treated.unit","std.v",
               "model.type")  
  XwN <- function(nam) {
    res        <- rep("x",length(nam))
    names(res) <- nam
    res
  }         
  myjoin <- function(a,b) {
    newrn <- unique(c(rownames(a),rownames(b)))
    res   <- matrix(NA,nrow=length(newrn),ncol=ncol(a)+ncol(b))
    colnames(res) <- c(colnames(a),colnames(b))
    rownames(res) <- newrn
    res[rownames(a),colnames(a)] <- a
    res[rownames(b),colnames(b)] <- b
    res
  }
  add <- function(a,b) {
    if (is.null(a)) return(b)
    if (length(unique(c(a$names,b$names)))!=length(a$names)+length(b$names)) 
      stop("names are not unique")
    res <- a
    res$names <- c(a$names,b$names)                                             # merge names
    for (le in names(res$variables)) {                                          # merge variables
      for (i in 1:3) {
        res$variables[[le]][[i]] <- 
          cbind(res$variables[[le]][[i]],b$variables[[le]][[i]])
        if (length(res$names)>1) colnames(res$variables[[le]][[i]]) <- res$names
      }    
    }
    for (le in scalars)                                                         # merge results
      res$results[[le]] <- c(a$results[[le]],b$results[[le]])
    for (le in setdiff(names(res$results),scalars)) 
      res$results[[le]] <- myjoin(a$results[[le]],b$results[[le]])
    res
  }
  if (("comparison" %in% names(arg))&&(!inherits(arg,"mscmt")))
    stop("'comparison' must not be one of the results' names")
  if (("placebo" %in% names(arg))&&(!inherits(arg,"mscmt")))
    stop("'placebo' must not be one of the results' names")
  for (i in seq_along(arg)) {
    if (inherits(arg[[i]],"mscmt")&&(!is.null(arg[[i]]$comparison))) {          # current argument already is a comparison
      res    <- c(res,arg[[i]][setdiff(names(arg[[i]]),"comparison")])
      agg    <- add(agg,arg[[i]]$comparison)
    } else
    if (((!inherits(arg[[i]],"mscmt"))&&is.list(arg[[i]]))||                    # current argument is a list, but not an object of class "mscmt" -> 'recursion'
       (inherits(arg[[i]],"mscmt")&&(is.null(arg[[i]]$placebo)&&
                                     is.null(arg[[i]]$combined))))  {           # current argument is a result of univariate "mscmt" estimation  -> 'recursion'
      tmpres <- do.call("compare",c(arg[[i]],
                        list(auto.name.prefix=paste0(auto.name.prefix,
                                                     LETTERS[i]))))
      res    <- c(res,tmpres[setdiff(names(tmpres),"comparison")])
      agg    <- add(agg,tmpres$comparison)
    } else
    if (inherits(arg[[i]],"mscmt")&&(is.null(arg[[i]]$comparison))) {           # current is object of class "mscmt", but not a comparison
      if (!is.null(arg[[i]]$placebo)) arg[[i]] <- arg[[i]][[1]]                 # placebo study? -> take first component!
      if (argnames[i]=="") argnames[i] <- 
        paste0("Result",auto.name.prefix,LETTERS[i])
      tmpR <- list(arg[[i]])
      names(tmpR) <- argnames[i]
      res  <- c(res,tmpR)                                                       # add whole 'mscmt' object
      tmp1 <- list(w=arg[[i]]$w,                                                # prepare estimation results
                   v=arg[[i]]$v[,ncol(arg[[i]]$v)],
                   loss.v=arg[[i]]$loss.v,                                     
                   rmspe=arg[[i]]$rmspe,
                   loss.w=arg[[i]]$loss.w[[length(arg[[i]]$loss.w)]],
                   solution.type=arg[[i]]$solution.type,
                   treated.unit=arg[[i]]$treated.unit,
                   control.units=XwN(arg[[i]]$control.units),
                   dependent=XwN(arg[[i]]$dependent),
                   predictor=XwN(arg[[i]]$predictor),
                   dependent.start=arg[[i]]$times.dep[1,,drop=FALSE],
                   dependent.end=arg[[i]]$times.dep[2,,drop=FALSE],
                   predictor.start=arg[[i]]$times.pred[1,,drop=FALSE],
                   predictor.end=arg[[i]]$times.pred[2,,drop=FALSE],
                   agg.pred=XwN(arg[[i]]$agg.pred),
                   agg.fns=arg[[i]]$agg.fns,
                   std.v=arg[[i]]$std.v,
                   model.type=paste0(if(length(arg[[i]]$dependent)>1) "M","SCM",
                     if(arg[[i]]$dataprep.scaled$trafo.v$has.trafo) "T")
                   )
      for (le in setdiff(names(tmp1),scalars)) {
        tmp1[[le]] <- if (is.matrix(tmp1[[le]])) t(tmp1[[le]]) else 
                                                 cbind(tmp1[[le]])
        colnames(tmp1[[le]]) <- argnames[i]
      }
      for (le in scalars) names(tmp1[[le]]) <- argnames[i]
      tmp2 <- vector("list",length(arg[[i]]$combined))                          # prepare combined data
      names(tmp2) <- names(arg[[i]]$combined)
      for (j in seq_along(arg[[i]]$combined)) 
        tmp2[[j]] <- list(data.treat=arg[[i]]$combined[[j]][,1],
                          data.synth=arg[[i]]$combined[[j]][,2],
                          gaps=arg[[i]]$combined[[j]][,3])
      tmp3 <- argnames[i]
      agg  <- add(agg,list(names=tmp3,variables=tmp2,results=tmp1))
    } else                                                                      # no valid input
#    if (!is.list(arg[[i]])&&(!inherits(arg[[i]],"mscmt")))                      
      warning("skipping argument which is neither of class 'list' nor 'mscmt'")
  }  
  final <- c(res,list(comparison=agg))
  class(final) <- "mscmt"
  if (!is.null(final$comparison)) final else NULL                               # do we have a valid result?
}

#' P-values for placebo studies
#'
#' \code{pvalue} calculates p-values for placebo studies.
#'
#' \code{pvalue} calculates p-values for placebo studies based
#' on Synthetic Control Methods.
#'
#' @param x An object of class \code{"mscmt"}, usually obtained as
#' the result of a call to function \code{\link{mscmt}}.
#' @param what A character vector. Name of the variable to be considered. If 
#' missing, the (first) dependent variable will be used.
#' @param range.pre A vector of length 2 defining the range of the pre-treatment
#' period with start and end time given as 
#' \itemize{
#' \item annual dates, if the format of start/end time is "dddd", e.g. "2016",
#' \item quarterly dates, if the format of start/end time is "ddddQd", e.g. 
#' "2016Q1",
#' \item monthly dates, if the format of start/end time is "dddd?dd" with "?"
#' different from "W" (see below), e.g. "2016/03" or "2016-10",
#' \item weekly dates, if the format of start/end time is "ddddWdd", e.g. 
#' "2016W23",
#' \item daily dates, if the format of start/end time is "dddd-dd-dd", e.g. 
#' "2016-08-18",
#' }
#' corresponding to the format of the respective column of the \code{times.dep}
#' argument of \code{\link{mscmt}}.
#' If missing, the corresponding column of \code{times.dep} will be used.
#' @param range.post A vector of length 2 defining the range of the 
#' post-treatment period with start and end time given as 
#' \itemize{
#' \item annual dates, if the format of start/end time is "dddd", e.g. "2016",
#' \item quarterly dates, if the format of start/end time is "ddddQd", e.g. 
#' "2016Q1",
#' \item monthly dates, if the format of start/end time is "dddd?dd" with "?"
#' different from "W" (see below), e.g. "2016/03" or "2016-10",
#' \item weekly dates, if the format of start/end time is "ddddWdd", e.g. 
#' "2016W23",
#' \item daily dates, if the format of start/end time is "dddd-dd-dd", e.g. 
#' "2016-08-18",
#' }
#' corresponding to the format of the respective column of the \code{times.dep}
#' argument of \code{\link{mscmt}}. Will be guessed if missing.
#' @param alternative A character string giving the alternative of the test. 
#' Either \code{"two.sided"} (default), \code{"less"}, or \code{"greater"}.
#' @param exclude.ratio A numerical scalar (default: \code{Inf}). Control units
#' with a pre-treatment-(r)mspe of more than \code{exclude.ratio} times the
#' pre-treatment-(r)mspe of the treated unit are excluded from the calculations
#' of the p-value.
#' @param ratio.type A character string. Either \code{rmspe} (default) or 
#' \code{mspe}. 
#' Selects whether root mean squared errors or mean squared errors are 
#' calculated.
#' @return A time series containing the p-values for the 
#' post-treatment periods.
#' @export pvalue
#' @examples
#' \dontrun{
#' ## for an example, see the main package vignette:
#'  vignette("WorkingWithMSCMT",package="MSCMT")
#' }
pvalue <- function(x,what,range.pre,range.post,
                   alternative = c("two.sided", "less", "greater"),
                   exclude.ratio = Inf, ratio.type = c("rmspe","mspe")) {
  ratio.type <- match.arg(ratio.type)
  if (is.null(x$placebo)) stop("results of placebo study are missing")
  alternative <- match.arg(alternative)  
  if (missing(what)) what  <- x[[1]]$dependent[1]
  gaps <- x$placebo[[what]]$gaps
  if (missing(range.pre))  
    if (what %in% x[[1]]$dependent) range.pre <- x[[1]]$times.dep[,what] else
                                    stop("range.pre is missing")
  gaps.pre  <- AQMwindow(gaps,range.pre)
  gaps.post <- if (missing(range.post)) AQMtail(gaps,range.pre[2]) else
                                        AQMwindow(gaps,range.post)
  rmspe.pre <- apply(gaps.pre,2,function(x) mean(x^2))
  if (ratio.type=="rmspe") rmspe.pre <- sqrt(rmspe.pre)
  excl <- (rmspe.pre > exclude.ratio*rmspe.pre[1])
  gaps.post <- gaps.post[,!excl,drop=FALSE]
  res  <- apply(gaps.post,1,switch(alternative,
                  two.sided = function(x) mean(abs(x[1])<=abs(x)),
                  less      = function(x) mean(x[1]>=x),
                  greater   = function(x) mean(x[1]<=x)))
  ts(res,start=start(gaps.post),frequency=frequency(gaps.post))                  
}

#' Check (and Improve) Results of Package Synth
#'
#' \code{improveSynth} checks the results of \code{\link[Synth]{synth}}
#' for feasibility and optimality and tries to find a better solution.
#'
#' Performing SCM means solving a nested optimization problem. Depending on
#' the validity of the results of the inner optimization, SCM may produce
#' \itemize{
#' \item invalid or infeasible results, if the vector \code{w} of donor 
#' weights reported as the result of the inner optimization 
#' is in fact not optimal, ie. produces too large \code{loss.w},
#' \item suboptimal results, if the vector \code{v} of predictor weights 
#' reported as the result of the outer optimization is in fact not
#' optimal (which may be caused by shortcomings of the inner optimization).
#' }
#'
#' \code{improveSynth} first checks \code{synth.out} for feasibility and
#' then tries to find a feasible and optimal solution by applying the
#' optimization methods of package \code{MSCMT} to \code{dataprep.out} 
#' (with default settings, more flexibility will probably be added in a 
#' future release).
#'
#' @param synth.out A result of \code{\link[Synth]{synth}} from package 
#' \code{'Synth'}.
#' @param dataprep.out The input of function \code{\link[Synth]{synth}} which
#' led to \code{synth.out}.
#' @param lb A numerical scalar (default: \code{1e-8}), corresponding to the
#' lower bound for the outer optimization.
#' @param tol A numerical scalar (default: \code{1e-5}). If the relative *and*
#' absolute improvement of \code{loss.v} and \code{loss.w}, respectively,
#' exceed \code{tol}, this is reported 
#' (if \code{verbose} is \code{TRUE}). Better optima are always looked for 
#' (independent of the value of \code{tol}).
#' @param verbose A logical scalar. Should the ouput be verbose (defaults to 
#' \code{TRUE}).
#' @param seed A numerical vector or \code{NULL}. See the corresponding
#' documentation for \code{\link[MSCMT]{mscmt}}. Defaults to 1 in order to
#' provide reproducibility of the results. 
#' @param ... Further arguments to \code{\link[MSCMT]{mscmt}}. Supported 
#' arguments are \code{check.global}, \code{inner.optim}, \code{inner.opar}, 
#' \code{outer.optim}, and \code{outer.opar}.
#' @return An updated version of \code{synth.out}, where \code{solution.v},
#' \code{solution.w}, \code{loss.v}, and \code{loss.w} are replaced by the
#' optimum obtained by package \code{'MSCMT'} and all other components 
#' of \code{synth.out} are removed.
#' @export improveSynth
#' @examples 
#' \dontrun{
#' ## check whether package 'Synth' is available
#' if (require("Synth")) {
#'
#' ## process first example of function "synth" in package 'Synth' 
#' ## (comments are removed):
#'
#'   data(synth.data)
#'   dataprep.out<-
#'     dataprep(
#'      foo = synth.data,
#'      predictors = c("X1", "X2", "X3"),
#'      predictors.op = "mean",
#'      dependent = "Y",
#'      unit.variable = "unit.num",
#'      time.variable = "year",
#'      special.predictors = list(
#'         list("Y", 1991, "mean"),
#'         list("Y", 1985, "mean"),
#'         list("Y", 1980, "mean")
#'                               ),
#'      treatment.identifier = 7,
#'      controls.identifier = c(29, 2, 13, 17, 32, 38),
#'      time.predictors.prior = c(1984:1989),
#'      time.optimize.ssr = c(1984:1990),
#'      unit.names.variable = "name",
#'      time.plot = 1984:1996
#'      )
#' 
#'   synth.out <- synth(dataprep.out)
#'
#' ## check and (try to) improve these results:
#'   synth2.out <- improveSynth(synth.out,dataprep.out)
#' }
#' }
#' @importFrom stats sd
improveSynth <- function(synth.out,dataprep.out,lb=1e-8,tol=1e-5,
                         verbose=TRUE,seed=1,...) {
  storage.mode(dataprep.out$X0) <- storage.mode(dataprep.out$X1) <- 
    storage.mode(dataprep.out$Z0) <- storage.mode(dataprep.out$Z1) <- "double"
  v   <- as.numeric(synth.out$solution.v)
  Xu  <- cbind(dataprep.out$X0, dataprep.out$X1)                                # generate (scaled!) X from dataprep object
  Xs  <- Xu/apply(Xu, 1, sd)
  X   <- Xs[, -ncol(Xs)] - drop(Xs[, ncol(Xs)])
  Z   <- dataprep.out$Z0 - drop(dataprep.out$Z1)
  ME  <- 1L
  MA  <- length(v)
  N   <- ncol(X)
  MDW <- ME+MA
  globals$Ipar  <- as.integer(c(ME=ME,MA=MA,MDW=MDW,N=N)) 
  globals$IWORK <- integer(MDW+N)
  globals$WORK  <- double(MDW+5*N)
  globals$IWORK[1:2] <- c(length(globals$WORK),length(globals$IWORK))
  globals$RNORM <- double(1)
  globals$MODE  <- integer(1)
  globals$X     <- double(N)
  tv  <- as.numeric(v)
  sol <-.Fortran(C_wnnls,W=.Call(C_prepareW4,X,tv),
                 MDW=globals$Ipar[3],ME=globals$Ipar[1],MA=globals$Ipar[2],
                 N=globals$Ipar[4],L=0L,PRGOPT=1.0,X=globals$X,
                 RNORM=globals$RNORM,MODE=globals$MODE,IWORK=globals$IWORK,
                 WORK=globals$WORK)
  if ((any(is.infinite(sol$X))) || (sol$MODE>0)) 
    warning("error in inner optimization (wnnls)")                
  w <- sol$X
  if (verbose) {
    cat0("Results reported by package Synth\n",
         "=================================\n\n")
    catw("Optimal V    ",paste0(synth.out$solution.v,sep=" "))
    catw("Optimal W*(V)",paste0(synth.out$solution.w,sep=" "))
    catf("with corresponding predictor loss ('loss W') of ",
         as.numeric(synth.out$loss.w)," ",
         "and corresponding dependent loss ('loss V') of ",
         as.numeric(synth.out$loss.v),".\n\n")
  }         
  
  w.orig <- as.numeric(synth.out$solution.w)
  if ((sum(w.orig)<1)&&verbose) 
    catf("Components of W*(V) do not sum to 1, dependent loss ('loss V') of ",
         "rescaled W*(V) is ",lossDep(Z,w.orig/sum(w.orig)),".\n\n")

  new.loss.w <- lossPred(X,w,tv)
  new.loss.v <- lossDep(Z,w)
  infeasible.w <- ((synth.out$loss.w-new.loss.w)/new.loss.w > tol) #&&
                    #(synth.out$loss.w-new.loss.w > tol)
  if (verbose) {
    cat0("Feasibility of W*(V)\n",
         "====================\n\n")
    if (infeasible.w) {
      catw("WARNING","W*(V) is NOT optimal and thus infeasible!")
      catw("'True' W*(V)",paste0(w,sep=" "))
      catf("with corresponding predictor loss ('loss W') of ",
           new.loss.w," ",
           "and corresponding dependent loss ('loss V') of ",
           new.loss.v,".\n\n")
    } else {       
      catw("GOOD","W*(V) is (essentially) optimal (new loss W: ",new.loss.w,
           ").\n\n")
    }       
  }     

  tmp    <- multiOpt(X,0,Z,0,outer.par=list(lb=lb),verbose=FALSE,
                     seed=seed,...)
  v      <- favourite_v(tmp$v,tmp$w,X,Z,tmp$trafo.v,lb=lb)
  v      <- v/sum(v)
  loss.w <- lossPred(X,tmp$w,v,tmp$trafo.v)
  loss.v <- tmp$rmspe^2
  suboptimal.v <- ((as.numeric(synth.out$loss.v)-loss.v)/loss.v > tol) #&&
                     #(as.numeric(synth.out$loss.v)-loss.v > tol)
                     
  if (verbose) {
    cat0("Optimality of V\n",
         "===============\n\n")
    
    if (suboptimal.v||infeasible.w) {
      catw("WARNING","'Optimal' V (as reported by package Synth) is not ",
           "optimal",if (infeasible.w) " (W*(V) was infeasible)",", ",
           "(one of potentially many) 'true' optimal V* (with sum(V*)=1):")
      catw("Optimal V*    ",paste0(v,sep=" "))
      catw("Optimal W*(V*)",paste0(tmp$w,sep=" "))
      catf("with corresponding predictor loss ('loss W') of ",
           loss.w," ","and corresponding dependent loss ('loss V') of ",
           loss.v,".\n")
    } else {       
      catw("GOOD","V is (essentially) optimal (new loss V: ",loss.v,").\n")     
    }  
  }    
  
  synth2.out                 <- synth.out
  if (is.numeric(synth2.out$solution.v)) {
    names(v) <- names(synth2.out$solution.v)
    synth2.out$solution.v  <- v 
  } else synth2.out$solution.v[1,]  <- v
  synth2.out$solution.w[,1]  <- tmp$w
  synth2.out$loss.w[1,1]     <- loss.w
  synth2.out$loss.v[1,1]     <- loss.v
  synth2.out$new.loss.v      <- new.loss.v
  synth2.out$new.loss.w      <- new.loss.w
  synth2.out$custom.v        <- NULL
  synth2.out$rgV.optim       <- NULL
       
  invisible(synth2.out)
}

checkInner <- function(X0,X1,v,w,Z0=NULL,Z1=NULL,tol=sqrt(.Machine$double.eps),
                       verbose=TRUE,...) {
  storage.mode(X0) <- storage.mode(X1) <- 
    storage.mode(Z0) <- storage.mode(Z1) <- "double"
  Xu  <- cbind(X0,X1)
  Xs  <- Xu/apply(Xu, 1, sd)
  X   <- Xs[, -ncol(Xs)] - drop(Xs[, ncol(Xs)])
  Z   <- if (!is.null(Z0)&&!is.null(Z1)) Z0 - drop(Z1) else NULL
  ME  <- 1L
  MA  <- length(v)
  N   <- ncol(X)
  MDW <- ME+MA
  globals$Ipar  <- as.integer(c(ME=ME,MA=MA,MDW=MDW,N=N)) 
  globals$IWORK <- integer(MDW+N)
  globals$WORK  <- double(MDW+5*N)
  globals$IWORK[1:2] <- c(length(globals$WORK),length(globals$IWORK))
  globals$RNORM <- double(1)
  globals$MODE  <- integer(1)
  globals$X     <- double(N)
  tv  <- as.numeric(v)
  sol <-.Fortran(C_wnnls,W=.Call(C_prepareW4,X,tv),
                 MDW=globals$Ipar[3],ME=globals$Ipar[1],MA=globals$Ipar[2],
                 N=globals$Ipar[4],L=0L,PRGOPT=1.0,X=globals$X,
                 RNORM=globals$RNORM,MODE=globals$MODE,IWORK=globals$IWORK,
                 WORK=globals$WORK)
  if ((any(is.infinite(sol$X))) || (sol$MODE>0)) 
    warning("error in inner optimization (wnnls)")                
  wnew <- sol$X
  new.loss.w <- lossPred(X,wnew,tv)
  old.loss.w <- lossPred(X,w,tv)
  new.loss.v <- if (is.null(Z)) NULL else lossDep(Z,wnew)
  old.loss.v <- if (is.null(Z)) NULL else lossDep(Z,w)
  abs.diff   <- old.loss.w-new.loss.w
  rel.diff   <- abs.diff/new.loss.w
  changed    <- (abs(rel.diff) > tol) #|| (abs(abs.diff) > tol)
  if (isTRUE(verbose&&changed))  
    cat0("Old: ",old.loss.w,", New: ",new.loss.w,", Relative Difference: ",
         (old.loss.w-new.loss.w)/new.loss.w,"\n")
  c(old.loss.w=old.loss.w, new.loss.w=new.loss.w, old.loss.v=old.loss.v,
    new.loss.v=new.loss.v)
}

rescaleLoss <- function(synth.out,dataprep.out) {
  storage.mode(dataprep.out$X0) <- storage.mode(dataprep.out$X1) <- 
    storage.mode(dataprep.out$Z0) <- storage.mode(dataprep.out$Z1) <- "double"
  Z   <- dataprep.out$Z0 - drop(dataprep.out$Z1)
  w.orig <- as.numeric(synth.out$solution.w)
  lossDep(Z,w.orig/sum(w.orig))
}

is.inner.wrong <- function(synth.out,dataprep.out,lb=1e-8,tol=1e-8) {
  storage.mode(dataprep.out$X0) <- storage.mode(dataprep.out$X1) <- 
    storage.mode(dataprep.out$Z0) <- storage.mode(dataprep.out$Z1) <- "double"
  v   <- as.numeric(synth.out$solution.v)
  Xu  <- cbind(dataprep.out$X0, dataprep.out$X1)                                # generate (scaled!) X from dataprep object
  Xs  <- Xu/apply(Xu, 1, sd)
  X   <- Xs[, -ncol(Xs)] - drop(Xs[, ncol(Xs)])
  Z   <- dataprep.out$Z0 - drop(dataprep.out$Z1)
  ME  <- 1L
  MA  <- length(v)
  N   <- ncol(X)
  MDW <- ME+MA
  globals$Ipar  <- as.integer(c(ME=ME,MA=MA,MDW=MDW,N=N)) 
  globals$IWORK <- integer(MDW+N)
  globals$WORK  <- double(MDW+5*N)
  globals$IWORK[1:2] <- c(length(globals$WORK),length(globals$IWORK))
  globals$RNORM <- double(1)
  globals$MODE  <- integer(1)
  globals$X     <- double(N)
  tv  <- as.numeric(v)
  sol <-.Fortran(C_wnnls,W=.Call(C_prepareW4,X,tv),
                 MDW=globals$Ipar[3],ME=globals$Ipar[1],MA=globals$Ipar[2],
                 N=globals$Ipar[4],L=0L,PRGOPT=1.0,X=globals$X,
                 RNORM=globals$RNORM,MODE=globals$MODE,IWORK=globals$IWORK,
                 WORK=globals$WORK)
  if ((any(is.infinite(sol$X))) || (sol$MODE>0)) 
    warning("error in inner optimization (wnnls)")                
  w <- sol$X
  w.orig <- as.numeric(synth.out$solution.w)

  real.loss.w <- lossPred(X,w.orig/sum(w.orig),tv)
  new.loss.w <- lossPred(X,w,tv)
  ((real.loss.w-new.loss.w)/new.loss.w > tol)
}

updateInner <- function(X0,X1,v,w,Z0=NULL,Z1=NULL,tol=sqrt(.Machine$double.eps),
                        verbose=TRUE,...) {
  storage.mode(X0) <- storage.mode(X1) <- 
    storage.mode(Z0) <- storage.mode(Z1) <- "double"
  Xu  <- cbind(X0,X1)
  Xs  <- Xu/apply(Xu, 1, sd)
  X   <- Xs[, -ncol(Xs)] - drop(Xs[, ncol(Xs)])
  Z   <- if (!is.null(Z0)&&!is.null(Z1)) Z0 - drop(Z1) else NULL
  ME  <- 1L
  MA  <- length(v)
  N   <- ncol(X)
  MDW <- ME+MA
  globals$Ipar  <- as.integer(c(ME=ME,MA=MA,MDW=MDW,N=N)) 
  globals$IWORK <- integer(MDW+N)
  globals$WORK  <- double(MDW+5*N)
  globals$IWORK[1:2] <- c(length(globals$WORK),length(globals$IWORK))
  globals$RNORM <- double(1)
  globals$MODE  <- integer(1)
  globals$X     <- double(N)
  tv  <- as.numeric(v)
  sol <-.Fortran(C_wnnls,W=.Call(C_prepareW4,X,tv),
                 MDW=globals$Ipar[3],ME=globals$Ipar[1],MA=globals$Ipar[2],
                 N=globals$Ipar[4],L=0L,PRGOPT=1.0,X=globals$X,
                 RNORM=globals$RNORM,MODE=globals$MODE,IWORK=globals$IWORK,
                 WORK=globals$WORK)
  if ((any(is.infinite(sol$X))) || (sol$MODE>0)) 
    warning("error in inner optimization (wnnls)")                
  wnew <- sol$X
  names(wnew) <- names(w)
  wnew
}

C two problems with strings solved:
C "INDEX" in xxerrprn and string comparison in ILAENV
C ilaenv was completely removed.

C************************************************************************
C TRIDIAGONAL MATRIX SOLVERS
C************************************************************************
 

      SUBROUTINE tridia(Nmx,au,bu,cu,du,value)

      IMPLICIT NONE

C-------------------------------------------------------------------------
C Solves a tridiagonal system of linear equations A*X=DU by backsubstitution
C where the non-zero diagonal elements are stored in au (below diagonal), 
C bu (on the diagonal), cu (above diagonal)
C and the equations are from first to last
C-------------------------------------------------------------------------

      INTEGER            Nmx                   
      DOUBLE PRECISION   au(Nmx)               
      DOUBLE PRECISION   bu(Nmx)               
      DOUBLE PRECISION   cu(Nmx)               
      DOUBLE PRECISION   du(Nmx)               

      DOUBLE PRECISION   value(Nmx)    
C
      DOUBLE PRECISION   ru(Nmx),qu(Nmx)
      INTEGER            I
C-------------------------------------------------------------------------
C Backsubstitution
      ru(Nmx)=au(Nmx)/bu(Nmx)
      qu(Nmx)=du(Nmx)/bu(Nmx)

      DO 10 I=Nmx-1,2,-1
         ru(I)=au(I)/(bu(I)-cu(I)*ru(I+1))
         qu(I)=(du(I)-cu(I)*qu(I+1))/(bu(I)-cu(I)*ru(I+1))
10    CONTINUE
 
      qu(1)=(du(1)-cu(1)*qu(2))/(bu(1)-cu(1)*ru(2))

C Forward substitution
      value(1)=qu(1)

      DO 20 I=2,Nmx
         value(I)=qu(I)-ru(I)*value(I-1)
20    CONTINUE

      RETURN
      END SUBROUTINE TRIDIA

C************************************************************************
C BANDED MATRIX SOLVER
C************************************************************************


      SUBROUTINE banded(abd,beta,NumABD,NumSvar,BandDown,BandUp,                 &
     &                  indx,info,IsError)
      INTEGER           NumABD,NumSvar,BandDown,BandUp
      INTEGER           indx(NumSvar), info
      DOUBLE PRECISION  ABD(NumABD,NumSVar),beta(NumSvar)
      LOGICAL           isError

        IsError = .FALSE.

        CALL dgbfa(abd,NumABD,NumSvar,BandDown,BandUp,indx,info)

        IF (Info .LT. 0) THEN
C xdgbsl will divide by 0 (singular matrix)
          IsError = .TRUE.
          RETURN
        ENDIF 
        CALL dgbsl(abd,NumABD,NumSvar,BandDown,BandUp,indx,beta,info)
      RETURN
      END SUBROUTINE BANDED   

c********************************************************************
c  tridiagonal solver
c********************************************************************

      SUBROUTINE DGTSV( N, NRHS, DL, D, DU, B, LDB, INFO )
*
*  -- LAPACK routine (version 3.0) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     October 31, 1999
*
*     .. Scalar Arguments ..
      INTEGER            INFO, LDB, N, NRHS
*     ..
*     .. Array Arguments ..
      DOUBLE PRECISION   B( LDB, * ), D( * ), DL( * ), DU( * )
*     ..
*
*  Purpose
*  =======
*
*  DGTSV  solves the equation
*
*     A*X = B,
*
*  where A is an n by n tridiagonal matrix, by Gaussian elimination with
*  partial pivoting.
*
*  Note that the equation  A'*X = B  may be solved by interchanging the
*  order of the arguments DU and DL.
*
*  Arguments
*  =========
*
*  N       (input) INTEGER
*          The order of the matrix A.  N >= 0.
*
*  NRHS    (input) INTEGER
*          The number of right hand sides, i.e., the number of columns
*          of the matrix B.  NRHS >= 0.
*
*  DL      (input/output) DOUBLE PRECISION array, dimension (N-1)
*          On entry, DL must contain the (n-1) sub-diagonal elements of
*          A.
*
*          On exit, DL is overwritten by the (n-2) elements of the
*          second super-diagonal of the upper triangular matrix U from
*          the LU factorization of A, in DL(1), ..., DL(n-2).
*
*  D       (input/output) DOUBLE PRECISION array, dimension (N)
*          On entry, D must contain the diagonal elements of A.
*
*          On exit, D is overwritten by the n diagonal elements of U.
*
*  DU      (input/output) DOUBLE PRECISION array, dimension (N-1)
*          On entry, DU must contain the (n-1) super-diagonal elements
*          of A.
*
*          On exit, DU is overwritten by the (n-1) elements of the first
*          super-diagonal of U.
*
*  B       (input/output) DOUBLE PRECISION array, dimension (LDB,NRHS)
*          On entry, the N by NRHS matrix of right hand side matrix B.
*          On exit, if INFO = 0, the N by NRHS solution matrix X.
*
*  LDB     (input) INTEGER
*          The leading dimension of the array B.  LDB >= max(1,N).
*
*  INFO    (output) INTEGER
*          = 0: successful exit
*          < 0: if INFO = -i, the i-th argument had an illegal value
*          > 0: if INFO = i, U(i,i) is exactly zero, and the solution
*               has not been computed.  The factorization has not been
*               completed unless i = N.
*
*  =====================================================================
*
*     .. Parameters ..
      DOUBLE PRECISION   ZERO
      PARAMETER          ( ZERO = 0.0D+0 )
*     ..
*     .. Local Scalars ..
      INTEGER            I, J
      DOUBLE PRECISION   FACT, TEMP
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          ABS, MAX
*     ..
*     .. External Subroutines ..
      EXTERNAL           XERBLA
*     ..
*     .. Executable Statements ..
*
      INFO = 0
      IF( N.LT.0 ) THEN
         INFO = -1
      ELSE IF( NRHS.LT.0 ) THEN
         INFO = -2
      ELSE IF( LDB.LT.MAX( 1, N ) ) THEN
         INFO = -7
      END IF
      IF( INFO.NE.0 ) THEN
         CALL XERBLA( 'DGTSV ', -INFO )
         RETURN
      END IF
*
      IF( N.EQ.0 )
     $   RETURN
*
      IF( NRHS.EQ.1 ) THEN
         DO 10 I = 1, N - 2
            IF( ABS( D( I ) ).GE.ABS( DL( I ) ) ) THEN
*
*              No row interchange required
*
               IF( D( I ).NE.ZERO ) THEN
                  FACT = DL( I ) / D( I )
                  D( I+1 ) = D( I+1 ) - FACT*DU( I )
                  B( I+1, 1 ) = B( I+1, 1 ) - FACT*B( I, 1 )
               ELSE
                  INFO = I
                  RETURN
               END IF
               DL( I ) = ZERO
            ELSE
*
*              Interchange rows I and I+1
*
               FACT = D( I ) / DL( I )
               D( I ) = DL( I )
               TEMP = D( I+1 )
               D( I+1 ) = DU( I ) - FACT*TEMP
               DL( I ) = DU( I+1 )
               DU( I+1 ) = -FACT*DL( I )
               DU( I ) = TEMP
               TEMP = B( I, 1 )
               B( I, 1 ) = B( I+1, 1 )
               B( I+1, 1 ) = TEMP - FACT*B( I+1, 1 )
            END IF
   10    CONTINUE
         IF( N.GT.1 ) THEN
            I = N - 1
            IF( ABS( D( I ) ).GE.ABS( DL( I ) ) ) THEN
               IF( D( I ).NE.ZERO ) THEN
                  FACT = DL( I ) / D( I )
                  D( I+1 ) = D( I+1 ) - FACT*DU( I )
                  B( I+1, 1 ) = B( I+1, 1 ) - FACT*B( I, 1 )
               ELSE
                  INFO = I
                  RETURN
               END IF
            ELSE
               FACT = D( I ) / DL( I )
               D( I ) = DL( I )
               TEMP = D( I+1 )
               D( I+1 ) = DU( I ) - FACT*TEMP
               DU( I ) = TEMP
               TEMP = B( I, 1 )
               B( I, 1 ) = B( I+1, 1 )
               B( I+1, 1 ) = TEMP - FACT*B( I+1, 1 )
            END IF
         END IF
         IF( D( N ).EQ.ZERO ) THEN
            INFO = N
            RETURN
         END IF
      ELSE
         DO 40 I = 1, N - 2
            IF( ABS( D( I ) ).GE.ABS( DL( I ) ) ) THEN
*
*              No row interchange required
*
               IF( D( I ).NE.ZERO ) THEN
                  FACT = DL( I ) / D( I )
                  D( I+1 ) = D( I+1 ) - FACT*DU( I )
                  DO 20 J = 1, NRHS
                     B( I+1, J ) = B( I+1, J ) - FACT*B( I, J )
   20             CONTINUE
               ELSE
                  INFO = I
                  RETURN
               END IF
               DL( I ) = ZERO
            ELSE
*
*              Interchange rows I and I+1
*
               FACT = D( I ) / DL( I )
               D( I ) = DL( I )
               TEMP = D( I+1 )
               D( I+1 ) = DU( I ) - FACT*TEMP
               DL( I ) = DU( I+1 )
               DU( I+1 ) = -FACT*DL( I )
               DU( I ) = TEMP
               DO 30 J = 1, NRHS
                  TEMP = B( I, J )
                  B( I, J ) = B( I+1, J )
                  B( I+1, J ) = TEMP - FACT*B( I+1, J )
   30          CONTINUE
            END IF
   40    CONTINUE
         IF( N.GT.1 ) THEN
            I = N - 1
            IF( ABS( D( I ) ).GE.ABS( DL( I ) ) ) THEN
               IF( D( I ).NE.ZERO ) THEN
                  FACT = DL( I ) / D( I )
                  D( I+1 ) = D( I+1 ) - FACT*DU( I )
                  DO 50 J = 1, NRHS
                     B( I+1, J ) = B( I+1, J ) - FACT*B( I, J )
   50             CONTINUE
               ELSE
                  INFO = I
                  RETURN
               END IF
            ELSE
               FACT = D( I ) / DL( I )
               D( I ) = DL( I )
               TEMP = D( I+1 )
               D( I+1 ) = DU( I ) - FACT*TEMP
               DU( I ) = TEMP
               DO 60 J = 1, NRHS
                  TEMP = B( I, J )
                  B( I, J ) = B( I+1, J )
                  B( I+1, J ) = TEMP - FACT*B( I+1, J )
   60          CONTINUE
            END IF
         END IF
         IF( D( N ).EQ.ZERO ) THEN
            INFO = N
            RETURN
         END IF
      END IF
*
*     Back solve with the matrix U from the factorization.
*
      IF( NRHS.LE.2 ) THEN
         J = 1
   70    CONTINUE
         B( N, J ) = B( N, J ) / D( N )
         IF( N.GT.1 )
     $      B( N-1, J ) = ( B( N-1, J )-DU( N-1 )*B( N, J ) ) / D( N-1 )
         DO 80 I = N - 2, 1, -1
            B( I, J ) = ( B( I, J )-DU( I )*B( I+1, J )-DL( I )*
     $                  B( I+2, J ) ) / D( I )
   80    CONTINUE
         IF( J.LT.NRHS ) THEN
            J = J + 1
            GO TO 70
         END IF
      ELSE
         DO 100 J = 1, NRHS
            B( N, J ) = B( N, J ) / D( N )
            IF( N.GT.1 )
     $         B( N-1, J ) = ( B( N-1, J )-DU( N-1 )*B( N, J ) ) /
     $                       D( N-1 )
            DO 90 I = N - 2, 1, -1
               B( I, J ) = ( B( I, J )-DU( I )*B( I+1, J )-DL( I )*
     $                     B( I+2, J ) ) / D( I )
   90       CONTINUE
  100    CONTINUE
      END IF
*
      RETURN
*
*     End of DGTSV
*
      END

c********************************************************************
      subroutine dgbsl(abd,lda,n,ml,mu,ipvt,b,job)

c-------------------------------------------------------------------*
c Subroutine obtained from LINPACK (ftp-site netlib.att.com)        *
c-------------------------------------------------------------------*

      integer lda,n,ml,mu,ipvt(*),job
      double precision abd(lda,*),b(*)
c
c     xdgbsl solves the double precision band system
c     a * x = b  or  trans(a) * x = b
c     using the factors computed by dgbco or dgbfa.
c
c     on entry
c
c        abd     double precision(lda, n)
c                the output from dgbco or xdgbfa.
c
c        lda     integer
c                the leading dimension of the array  abd .
c
c        n       integer
c                the order of the original matrix.
c
c        ml      integer
c                number of diagonals below the main diagonal.
c
c        mu      integer
c                number of diagonals above the main diagonal.
c
c        ipvt    integer(n)
c                the pivot vector from dgbco or xdgbfa.
c
c        b       double precision(n)
c                the right hand side vector.
c
c        job     integer
c                = 0         to solve  a*x = b ,
c                = nonzero   to solve  trans(a)*x = b , where
c                            trans(a)  is the transpose.
c
c     on return
c
c        b       the solution vector  x .
c
c     error condition
c
c        a division by zero will occur if the input factor contains a
c        zero on the diagonal.  technically this indicates singularity
c        but it is often caused by improper arguments or improper
c        setting of lda .  it will not occur if the subroutines are
c        called correctly and if dgbco has set rcond .gt. 0.0
c        or xdgbfa has set info .eq. 0 .
c
c     to compute  inverse(a) * c  where  c  is a matrix
c     with  p  columns
c           call dgbco(abd,lda,n,ml,mu,ipvt,rcond,z)
c           if (rcond is too small) go to ...
c           do 10 j = 1, p
c              call xdgbsl(abd,lda,n,ml,mu,ipvt,c(1,j),0)
c        10 continue
c
c     linpack. this version dated 08/14/78 .
c     cleve moler, university of new mexico, argonne national lab.
c
c     subroutines and functions
c
c     xblas daxpy,ddot
c     fortran min0
c
c     internal variables
c
      double precision ddot,t
      integer k,kb,l,la,lb,lm,m,nm1
c
      m = mu + ml + 1
      nm1 = n - 1
      if (job .ne. 0) go to 50
c
c        job = 0 , solve  a * x = b
c        first solve l*y = b
c
         if (ml .eq. 0) go to 30
         if (nm1 .lt. 1) go to 30
            do 20 k = 1, nm1
               lm = min0(ml,n-k)
               l = ipvt(k)
               t = b(l)
               if (l .eq. k) go to 10
                  b(l) = b(k)
                  b(k) = t
   10          continue
               call daxpy(lm,t,abd(m+1,k),1,b(k+1),1)
   20       continue
   30    continue
c
c        now solve  u*x = y
c
         do 40 kb = 1, n
            k = n + 1 - kb
            b(k) = b(k)/abd(m,k)
            lm = min0(k,m) - 1
            la = m - lm
            lb = k - lm
            t = -b(k)
            call daxpy(lm,t,abd(la,k),1,b(lb),1)
   40    continue
      go to 100
   50 continue
c
c        job = nonzero, solve  trans(a) * x = b
c        first solve  trans(u)*y = b
c
         do 60 k = 1, n
            lm = min0(k,m) - 1
            la = m - lm
            lb = k - lm
            t = ddot(lm,abd(la,k),1,b(lb),1)
            b(k) = (b(k) - t)/abd(m,k)
   60    continue
c
c        now solve trans(l)*x = y
c
         if (ml .eq. 0) go to 90
         if (nm1 .lt. 1) go to 90
            do 80 kb = 1, nm1
               k = n - kb
               lm = min0(ml,n-k)
               b(k) = b(k) + ddot(lm,abd(m+1,k),1,b(k+1),1)
               l = ipvt(k)
               if (l .eq. k) go to 70
                  t = b(l)
                  b(l) = b(k)
                  b(k) = t
   70          continue
   80       continue
   90    continue
  100 continue
      return
      end                                        


                       
c********************************************************************
      subroutine dgbfa(abd,lda,n,ml,mu,ipvt,info)

c-------------------------------------------------------------------*
c Subroutine obtained from LINPACK (ftp-site netlib.att.com)        *
c-------------------------------------------------------------------*

      integer lda,n,ml,mu,ipvt(*),info
      double precision abd(lda,*)
c
c     xdgbfa factors a double precision band matrix by elimination.
c
c     xdgbfa is usually called by dgbco, but it can be called
c     directly with a saving in time if  rcond  is not needed.
c
c     on entry
c
c        abd     double precision(lda, n)
c                contains the matrix in band storage.  the columns
c                of the matrix are stored in the columns of  abd  and
c                the diagonals of the matrix are stored in rows
c                ml+1 through 2*ml+mu+1 of  abd .
c                see the comments below for details.
c
c        lda     integer
c                the leading dimension of the array  abd .
c                lda must be .ge. 2*ml + mu + 1 .
c
c        n       integer
c                the order of the original matrix.
c
c        ml      integer
c                number of diagonals below the main diagonal.
c                0 .le. ml .lt. n .
c
c        mu      integer
c                number of diagonals above the main diagonal.
c                0 .le. mu .lt. n .
c                more efficient if  ml .le. mu .
c     on return
c
c        abd     an upper triangular matrix in band storage and
c                the multipliers which were used to obtain it.
c                the factorization can be written  a = l*u  where
c                l  is a product of permutation and unit lower
c                triangular matrices and  u  is upper triangular.
c
c        ipvt    integer(n)
c                an integer vector of pivot indices.
c
c        info    integer
c                = 0  normal value.
c                = k  if  u(k,k) .eq. 0.0 .  this is not an error
c                     condition for this subroutine, but it does
c                     indicate that xdgbsl will divide by zero if
c                     called.  use  rcond  in dgbco for a reliable
c                     indication of singularity.
c
c     band storage
c
c           if  a  is a band matrix, the following program segment
c           will set up the input.
c
c                   ml = (band width below the diagonal)
c                   mu = (band width above the diagonal)
c                   m = ml + mu + 1
c                   do 20 j = 1, n
c                      i1 = max0(1, j-mu)
c                      i2 = min0(n, j+ml)
c                      do 10 i = i1, i2
c                         k = i - j + m
c                         abd(k,j) = a(i,j)
c                10    continue
c                20 continue
c
c           this uses rows  ml+1  through  2*ml+mu+1  of  abd .
c           in addition, the first  ml  rows in  abd  are used for
c           elements generated during the triangularization.
c           the total number of rows needed in  abd  is  2*ml+mu+1 .
c           the  ml+mu by ml+mu  upper left triangle and the
c           ml by ml  lower right triangle are not referenced.
c
c     linpack. this version dated 08/14/78 .
c     cleve moler, university of new mexico, argonne national lab.
c
c     subroutines and functions
c
c     blas daxpy,dscal,idamax
c     fortran max0,min0
c
c     internal variables
c
      double precision t
      integer i,idamax,i0,j,ju,jz,j0,j1,k,kp1,l,lm,m,mm,nm1
c
c
      m = ml + mu + 1
      info = 0
c
c     zero initial fill-in columns
c
      j0 = mu + 2
      j1 = min0(n,m) - 1
      if (j1 .lt. j0) go to 30
      do 20 jz = j0, j1
         i0 = m + 1 - jz
         do 10 i = i0, ml
            abd(i,jz) = 0.0d0
   10    continue
   20 continue
   30 continue
      jz = j1
      ju = 0
c
c     gaussian elimination with partial pivoting
c
      nm1 = n - 1
      if (nm1 .lt. 1) go to 130
      do 120 k = 1, nm1
         kp1 = k + 1
c
c        zero next fill-in column
c
         jz = jz + 1
         if (jz .gt. n) go to 50
         if (ml .lt. 1) go to 50
            do 40 i = 1, ml
               abd(i,jz) = 0.0d0
   40       continue
   50    continue
c
c        find l = pivot index
c
         lm = min0(ml,n-k)
         l = idamax(lm+1,abd(m,k),1) + m - 1
         ipvt(k) = l + k - m
c
c        zero pivot implies this column already triangularized
c
         if (abd(l,k) .eq. 0.0d0) go to 100
c
c           interchange if necessary
c
            if (l .eq. m) go to 60
               t = abd(l,k)
               abd(l,k) = abd(m,k)
               abd(m,k) = t
   60       continue
c
c           compute multipliers
c
            t = -1.0d0/abd(m,k)
            call dscal(lm,t,abd(m+1,k),1)
c
c           row elimination with column indexing
c
            ju = min0(max0(ju,mu+ipvt(k)),n)
            mm = m
            if (ju .lt. kp1) go to 90
            do 80 j = kp1, ju
               l = l - 1
               mm = mm - 1
               t = abd(l,j)
               if (l .eq. mm) go to 70
                  abd(l,j) = abd(mm,j)
                  abd(mm,j) = t
   70          continue
               call daxpy(lm,t,abd(m+1,k),1,abd(mm+1,j),1)
   80       continue
   90       continue
         go to 110
  100    continue
            info = k
  110    continue
  120 continue
  130 continue
      ipvt(n) = n
      if (abd(m,n) .eq. 0.0d0) info = n
      return
      end                                        


      subroutine dgefa(a,lda,n,ipvt,info)
      integer lda,n,ipvt(*),info
      double precision a(lda,*)
c
c     dgefa factors a double precision matrix by gaussian elimination.
c
c     dgefa is usually called by dgeco, but it can be called
c     directly with a saving in time if  rcond  is not needed.
c     (time for dgeco) = (1 + 9/n)*(time for dgefa) .
c
c     on entry
c
c        a       double precision(lda, n)
c                the matrix to be factored.
c
c        lda     integer
c                the leading dimension of the array  a .
c
c        n       integer
c                the order of the matrix  a .
c
c     on return
c
c        a       an upper triangular matrix and the multipliers
c                which were used to obtain it.
c                the factorization can be written  a = l*u  where
c                l  is a product of permutation and unit lower
c                triangular matrices and  u  is upper triangular.
c
c        ipvt    integer(n)
c                an integer vector of pivot indices.
c
c        info    integer
c                = 0  normal value.
c                = k  if  u(k,k) .eq. 0.0 .  this is not an error
c                     condition for this subroutine, but it does
c                     indicate that dgesl or dgedi will divide by zero
c                     if called.  use  rcond  in dgeco for a reliable
c                     indication of singularity.
c
c     linpack. this version dated 08/14/78 .
c     cleve moler, university of new mexico, argonne national lab.
c
c     subroutines and functions
c
c     blas daxpy,dscal,idamax
c
c     internal variables
c
      double precision t
      integer idamax,j,k,kp1,l,nm1
c
c
c     gaussian elimination with partial pivoting
c
      info = 0
      nm1 = n - 1
      if (nm1 .lt. 1) go to 70
      do 60 k = 1, nm1
         kp1 = k + 1
c
c        find l = pivot index
c
         l = idamax(n-k+1,a(k,k),1) + k - 1
         ipvt(k) = l
c
c        zero pivot implies this column already triangularized
c
         if (a(l,k) .eq. 0.0d0) go to 40
c
c           interchange if necessary
c
            if (l .eq. k) go to 10
               t = a(l,k)
               a(l,k) = a(k,k)
               a(k,k) = t
   10       continue
c
c           compute multipliers
c
            t = -1.0d0/a(k,k)
            call dscal(n-k,t,a(k+1,k),1)
c
c           row elimination with column indexing
c
            do 30 j = kp1, n
               t = a(l,j)
               if (l .eq. k) go to 20
                  a(l,j) = a(k,j)
                  a(k,j) = t
   20          continue
               call daxpy(n-k,t,a(k+1,k),1,a(k+1,j),1)
   30       continue
         go to 50
   40    continue
            info = k
   50    continue
   60 continue
   70 continue
      ipvt(n) = n
      if (a(n,n) .eq. 0.0d0) info = n
      return
      end
      subroutine dgesl(a,lda,n,ipvt,b,job)
      integer lda,n,ipvt(*),job
      double precision a(lda,*),b(*)
c
c     dgesl solves the double precision system
c     a * x = b  or  trans(a) * x = b
c     using the factors computed by dgeco or dgefa.
c
c     on entry
c
c        a       double precision(lda, n)
c                the output from dgeco or dgefa.
c
c        lda     integer
c                the leading dimension of the array  a .
c
c        n       integer
c                the order of the matrix  a .
c
c        ipvt    integer(n)
c                the pivot vector from dgeco or dgefa.
c
c        b       double precision(n)
c                the right hand side vector.
c
c        job     integer
c                = 0         to solve  a*x = b ,
c                = nonzero   to solve  trans(a)*x = b  where
c                            trans(a)  is the transpose.
c
c     on return
c
c        b       the solution vector  x .
c
c     error condition
c
c        a division by zero will occur if the input factor contains a
c        zero on the diagonal.  technically this indicates singularity
c        but it is often caused by improper arguments or improper
c        setting of lda .  it will not occur if the subroutines are
c        called correctly and if dgeco has set rcond .gt. 0.0
c        or dgefa has set info .eq. 0 .
c
c     to compute  inverse(a) * c  where  c  is a matrix
c     with  p  columns
c           call dgeco(a,lda,n,ipvt,rcond,z)
c           if (rcond is too small) go to ...
c           do 10 j = 1, p
c              call dgesl(a,lda,n,ipvt,c(1,j),0)
c        10 continue
c
c     linpack. this version dated 08/14/78 .
c     cleve moler, university of new mexico, argonne national lab.
c
c     subroutines and functions
c
c     blas daxpy,ddot
c
c     internal variables
c
      double precision ddot,t
      integer k,kb,l,nm1
c
      nm1 = n - 1
      if (job .ne. 0) go to 50
c
c        job = 0 , solve  a * x = b
c        first solve  l*y = b
c
         if (nm1 .lt. 1) go to 30
         do 20 k = 1, nm1
            l = ipvt(k)
            t = b(l)
            if (l .eq. k) go to 10
               b(l) = b(k)
               b(k) = t
   10       continue
            call daxpy(n-k,t,a(k+1,k),1,b(k+1),1)
   20    continue
   30    continue
c
c        now solve  u*x = y
c
         do 40 kb = 1, n
            k = n + 1 - kb
            b(k) = b(k)/a(k,k)
            t = -b(k)
            call daxpy(k-1,t,a(1,k),1,b(1),1)
   40    continue
      go to 100
   50 continue
c
c        job = nonzero, solve  trans(a) * x = b
c        first solve  trans(u)*y = b
c
         do 60 k = 1, n
            t = ddot(k-1,a(1,k),1,b(1),1)
            b(k) = (b(k) - t)/a(k,k)
   60    continue
c
c        now solve trans(l)*x = y
c
         if (nm1 .lt. 1) go to 90
         do 80 kb = 1, nm1
            k = n - kb
            b(k) = b(k) + ddot(n-k,a(k+1,k),1,b(k+1),1)
            l = ipvt(k)
            if (l .eq. k) go to 70
               t = b(l)
               b(l) = b(k)
               b(k) = t
   70       continue
   80    continue
   90    continue
  100 continue
      return
      end
c********************************************************************



      double precision function Xdasum(n,dx,incx)
C
C     takes the sum of the absolute values.
C     jack dongarra, linpack, 3/11/78.
C     modified 3/93 to return if incx .le. 0.
C     modified 12/3/93, array(1) declarations changed to array(*)
C
      IMPLICIT NONE 
      double precision dx(*),dtemp
      integer i,incx,m,mp1,n,nincx
C
      Xdasum = 0.0d0
      dtemp = 0.0d0
      if( n.le.0 .or. incx.le.0 )return
      if(incx.eq.1)go to 20
C
C        code for increment not equal to 1
C
      nincx = n*incx
      do 10 i = 1,nincx,incx
        dtemp = dtemp + dabs(dx(i))
   10 continue
      Xdasum = dtemp
      return
C
C        code for increment equal to 1
C
C
C        clean-up loop
C
   20 m = mod(n,6)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dtemp = dtemp + dabs(dx(i))
   30 continue
      if( n .lt. 6 ) go to 60
   40 mp1 = m + 1
      do 50 i = mp1,n,6
        dtemp = dtemp + dabs(dx(i)) + dabs(dx(i + 1)) + dabs(dx(i + 2))  &
     &  + dabs(dx(i + 3)) + dabs(dx(i + 4)) + dabs(dx(i + 5))
   50 continue
   60 Xdasum = dtemp
      return
      end




      SUBROUTINE xXERMSG (LIBRAR, SUBROU, MESSG, NERR, LEVEL)

C***PURPOSE  Process error messages for SLATEC and other libraries.

      IMPLICIT NONE
      CHARACTER(LEN=*)   LIBRAR, SUBROU, MESSG
      CHARACTER(LEN=8)   XLIBR, XSUBR
      CHARACTER(LEN=72)  TEMP
      CHARACTER(LEN=20)  LFIRST
      INTEGER :: LKNTRL,MAXMES,NERR,Lerr,LEVEL,LLEVEL,LTEMP,Kount
      INTEGER :: MKNTRL,I

CKARLINE:
       IF (Level .GE. 2) THEN 
C         WRITE(*,*) " Unrecoverable error C- program stopped"
       CALL XMESSAGE (" Unrecoverable error C- LINPACK routine stopped")
       RETURN
       ENDIF


C***FIRST EXECUTABLE STATEMENT  xXERMSG
C KARLINE: CHANGED -ERRMSG NOT WRITTEN TO FILE
C      LKNTRL = J4SAVE (2, 0, .FALSE.)
C      MAXMES = J4SAVE (4, 0, .FALSE.)
       LKNTRL = 1
       MAXMES = 1 
C
C       LKNTRL IS A LOCAL COPY OF THE CONTROL FLAG KONTRL.
C       MAXMES IS THE MAXIMUM NUMBER OF TIMES ANY PARTICULAR MESSAGE
C          SHOULD BE PRINTED.
C
C       WE PRINT A FATAL ERROR MESSAGE AND TERMINATE FOR AN ERROR IN
C          CALLING xXERMSG.  THE ERROR NUMBER SHOULD BE POSITIVE,
C          AND THE LEVEL SHOULD BE BETWEEN 0 AND 2.
C
      IF (NERR.LT.-9999999 .OR. NERR.GT.99999999 .OR. NERR.EQ.0 .OR.     &
     &   LEVEL.LT.-1 .OR. LEVEL.GT.2) THEN
         CALL xXERPRN (' ***', -1, 'FATAL ERROR IN...$$ ' //             &
     &      'xXERMSG -- INVALID ERROR NUMBER OR LEVEL$$ '//               &
     &      'JOB ABORT DUE TO FATAL ERROR.', 72)
C KS        CALL XERSVE (' ', ' ', ' ', 0, 0, 0, KDUMMY)

C KS        CALL XERHLT (' ***xXERMSG -- INVALID INPUT')
         
C         WRITE(*,*) ' ***xXERMSG -- INVALID INPUT'
         CALL XMESSAGE                                                    &
     & (" ***xXERMSG -- INVALID INPUT- LINPACK routine stopped")
         RETURN
C
         RETURN
      ENDIF
C
C       RECORD THE MESSAGE.
C
C KS      I = J4SAVE (1, NERR, .TRUE.)
C KS      CALL XERSVE (LIBRAR, SUBROU, MESSG, 1, NERR, LEVEL, KOUNT)
C
C       HANDLE PRINT-ONCE WARNING MESSAGES.
C KS: 
      KOUNT = 1 
      IF (LEVEL.EQ.-1 .AND. KOUNT.GT.1) RETURN
C
C       ALLOW TEMPORARY USER OVERRIDE OF THE CONTROL FLAG.
C
      XLIBR  = LIBRAR
      XSUBR  = SUBROU
      LFIRST = MESSG
      LERR   = NERR
      LLEVEL = LEVEL
C KS      CALL XERCNT (XLIBR, XSUBR, LFIRST, LERR, LLEVEL, LKNTRL)
C
      LKNTRL = MAX(-2, MIN(2,LKNTRL))
      MKNTRL = ABS(LKNTRL)
C
C       SKIP PRINTING IF THE CONTROL FLAG VALUE AS RESET IN XERCNT IS
C       ZERO AND THE ERROR IS NOT FATAL.
C
      IF (LEVEL.LT.2 .AND. LKNTRL.EQ.0) GO TO 30
      IF (LEVEL.EQ.0 .AND. KOUNT.GT.MAXMES) GO TO 30
      IF (LEVEL.EQ.1 .AND. KOUNT.GT.MAXMES .AND. MKNTRL.EQ.1) GO TO 30
      IF (LEVEL.EQ.2 .AND. KOUNT.GT.MAX(1,MAXMES)) GO TO 30
C
C       ANNOUNCE THE NAMES OF THE LIBRARY AND SUBROUTINE BY BUILDING A
C       MESSAGE IN CHARACTER VARIABLE TEMP (NOT EXCEEDING 66 CHARACTERS)
C       AND SENDING IT OUT VIA XERPRN.  PRINT ONLY IF CONTROL FLAG
C       IS NOT ZERO.
C
      IF (LKNTRL .NE. 0) THEN
         TEMP(1:21) = 'MESSAGE FROM ROUTINE '
         I = MIN(LEN(SUBROU), 16)
         TEMP(22:21+I) = SUBROU(1:I)
         TEMP(22+I:33+I) = ' IN LIBRARY '
         LTEMP = 33 + I
         I = MIN(LEN(LIBRAR), 16)
         TEMP(LTEMP+1:LTEMP+I) = LIBRAR (1:I)
         TEMP(LTEMP+I+1:LTEMP+I+1) = '.'
         LTEMP = LTEMP + I + 1
         CALL XxERPRN (' ***', -1, TEMP(1:LTEMP), 72)
      ENDIF
C
C       IF LKNTRL IS POSITIVE, PRINT AN INTRODUCTORY LINE BEFORE
C       PRINTING THE MESSAGE.  THE INTRODUCTORY LINE TELLS THE CHOICE
C       FROM EACH OF THE FOLLOWING THREE OPTIONS.
C       1.  LEVEL OF THE MESSAGE
C              'INFORMATIVE MESSAGE'
C              'POTENTIALLY RECOVERABLE ERROR'
C              'FATAL ERROR'
C       2.  WHETHER CONTROL FLAG WILL ALLOW PROGRAM TO CONTINUE
C              'PROG CONTINUES'
C              'PROG ABORTED'
C       3.  WHETHER OR NOT A TRACEBACK WAS REQUESTED.  (THE TRACEBACK
C           MAY NOT BE IMPLEMENTED AT SOME SITES, SO THIS ONLY TELLS
C           WHAT WAS REQUESTED, NOT WHAT WAS DELIVERED.)
C              'TRACEBACK REQUESTED'
C              'TRACEBACK NOT REQUESTED'
C       NOTICE THAT THE LINE INCLUDING FOUR PREFIX CHARACTERS WILL NOT
C       EXCEED 74 CHARACTERS.
C       WE SKIP THE NEXT BLOCK IF THE INTRODUCTORY LINE IS NOT NEEDED.
C
      IF (LKNTRL .GT. 0) THEN
C
C       THE FIRST PART OF THE MESSAGE TELLS ABOUT THE LEVEL.
C
         IF (LEVEL .LE. 0) THEN
            TEMP(1:20) = 'INFORMATIVE MESSAGE,'
            LTEMP = 20
         ELSEIF (LEVEL .EQ. 1) THEN
            TEMP(1:30) = 'POTENTIALLY RECOVERABLE ERROR,'
            LTEMP = 30
         ELSE
            TEMP(1:12) = 'FATAL ERROR,'
            LTEMP = 12
         ENDIF
C
C       THEN WHETHER THE PROGRAM WILL CONTINUE.
C
         IF ((MKNTRL.EQ.2 .AND. LEVEL.GE.1) .OR.                         &
     &       (MKNTRL.EQ.1 .AND. LEVEL.EQ.2)) THEN
            TEMP(LTEMP+1:LTEMP+14) = ' PROG ABORTED,'
            LTEMP = LTEMP + 14
         ELSE
            TEMP(LTEMP+1:LTEMP+16) = ' PROG CONTINUES,'
            LTEMP = LTEMP + 16
         ENDIF
C
C       FINALLY TELL WHETHER THERE SHOULD BE A TRACEBACK.
C
         IF (LKNTRL .GT. 0) THEN
            TEMP(LTEMP+1:LTEMP+20) = ' TRACEBACK REQUESTED'
            LTEMP = LTEMP + 20
         ELSE
            TEMP(LTEMP+1:LTEMP+24) = ' TRACEBACK NOT REQUESTED'
            LTEMP = LTEMP + 24
         ENDIF
         CALL xXERPRN (' ***', -1, TEMP(1:LTEMP), 72)
      ENDIF
C
C       NOW SEND OUT THE MESSAGE.
C
      CALL xXERPRN (' *  ', -1, MESSG, 72)
C
C       IF LKNTRL IS POSITIVE, WRITE THE ERROR NUMBER AND REQUEST A
C          TRACEBACK.
C
      IF (LKNTRL .GT. 0) THEN
C KARLINE: REMOVED WRITE		 
         CALL rwarn ('An error occurred')

C   WRITE (TEMP, '(''ERROR NUMBER = '', I8)') NERR
C         DO 10 I=16,22
C            IF (TEMP(I:I) .NE. ' ') GO TO 20
C   10    CONTINUE
C
C   20    CALL xXERPRN (' *  ', -1, TEMP(1:15) // TEMP(I:23), 72)
C KS         CALL FDUMP
      ENDIF
C
C       IF LKNTRL IS NOT ZERO, PRINT A BLANK LINE AND AN END OF MESSAGE.
C
      IF (LKNTRL .NE. 0) THEN
         CALL xXERPRN (' *  ', -1, ' ', 72)
         CALL xXERPRN (' ***', -1, 'END OF MESSAGE', 72)
         CALL xXERPRN ('    ',  0, ' ', 72)
      ENDIF
C
C       IF THE ERROR IS NOT FATAL OR THE ERROR IS RECOVERABLE AND THE
C       CONTROL FLAG IS SET FOR RECOVERY, THEN RETURN.
C
   30 IF (LEVEL.LE.0 .OR. (LEVEL.EQ.1 .AND. MKNTRL.LE.1)) RETURN
C
C       THE PROGRAM WILL BE STOPPED DUE TO AN UNRECOVERED ERROR OR A
C       FATAL ERROR.  PRINT THE REASON FOR THE ABORT AND THE ERROR
C       SUMMARY IF THE CONTROL FLAG AND THE MAXIMUM ERROR COUNT PERMIT.
C
      IF (LKNTRL.GT.0 .AND. KOUNT.LT.MAX(1,MAXMES)) THEN
         IF (LEVEL .EQ. 1) THEN
            CALL xXERPRN                                                  &
     &         (' ***', -1, 'JOB ABORT DUE TO UNRECOVERED ERROR.', 72)
         ELSE
         CALL xXERPRN(' ***', -1, 'JOB ABORT DUE TO FATAL ERROR.', 72)
         ENDIF
C KS      CALL XERSVE (' ', ' ', ' ', -1, 0, 0, KDUMMY)
C         CALL XERHLT (' ')
c         STOP
C
          call rexit("STOPPED")
      ELSE
C         CALL XERHLT (MESSG)
C         WRITE(*,*) MESSG
          CALL rexit ("STOPPED")

C         STOP
C
      ENDIF
      RETURN
      END SUBROUTINE xXERMSG

C*********************************************************************

      SUBROUTINE XXERPRN (PREFIX, NPREF, MESSG, NWRAP)

CC***PURPOSE  Print error messages processed by xXERMSG.
C KARLINE: DO NOT PRINT TO SCREEN ->          CALL XMESSAGE 

      IMPLICIT NONE
      CHARACTER(LEN=*)   PREFIX, MESSG
      INTEGER            NPREF, NWRAP
      CHARACTER(LEN=148) CBUFF
      CHARACTER(LEN=2)   NEWLIN
      PARAMETER (NEWLIN = '$$')
      INTEGER :: LPREF,LWRAP,LENMSG,NEXTC,LPIECE,IDELTA,I,N

C KARLINE: RETURN
C      RETURN

C       LPREF IS THE LENGTH OF THE PREFIX.  THE PREFIX IS PLACED AT THE
C       BEGINNING OF CBUFF, THE CHARACTER BUFFER, AND KEPT THERE DURING
C       THE REST OF THIS ROUTINE.
C
      IF ( NPREF .LT. 0 ) THEN
         LPREF = LEN(PREFIX)
      ELSE
         LPREF = NPREF
      ENDIF
      LPREF = MIN(16, LPREF)
      IF (LPREF .NE. 0) CBUFF(1:LPREF) = PREFIX
C
C       LWRAP IS THE MAXIMUM NUMBER OF CHARACTERS WE WANT TO TAKE AT ONE
C       TIME FROM MESSG TO PRINT ON ONE LINE.
C
      LWRAP = MAX(16, MIN(132, NWRAP))
C
C       SET LENMSG TO THE LENGTH OF MESSG, IGNORE ANY TRAILING BLANKS.
C
      LENMSG = LEN(MESSG)
      N = LENMSG
      DO 20 I=1,N
         IF (MESSG(LENMSG:LENMSG) .NE. ' ') GO TO 30
         LENMSG = LENMSG - 1
   20 CONTINUE
   30 CONTINUE
C
C       IF THE MESSAGE IS ALL BLANKS, THEN PRINT ONE BLANK LINE.
C
      IF (LENMSG .EQ. 0) THEN
         CBUFF(LPREF+1:LPREF+1) = ' '
C            WRITE(*, '(A)') CBUFF(1:LPREF+1)
         CALL XMESSAGE (CBUFF(1:LPREF+1))

         RETURN
      ENDIF
C
C       SET NEXTC TO THE POSITION IN MESSG WHERE THE NEXT SUBSTRING
C       STARTS.  FROM THIS POSITION WE SCAN FOR THE NEW LINE SENTINEL.
C       WHEN NEXTC EXCEEDS LENMSG, THERE IS NO MORE TO PRINT.
C       WE LOOP BACK TO LABEL 50 UNTIL ALL PIECES HAVE BEEN PRINTED.
C
C       WE LOOK FOR THE NEXT OCCURRENCE OF THE NEW LINE SENTINEL.  THE
C       INDEX INTRINSIC FUNCTION RETURNS ZERO IF THERE IS NO OCCURRENCE
C       OR IF THE LENGTH OF THE FIRST ARGUMENT IS LESS THAN THE LENGTH
C       OF THE SECOND ARGUMENT.
C
C       THERE ARE SEVERAL CASES WHICH SHOULD BE CHECKED FOR IN THE
C       FOLLOWING ORDER.  WE ARE ATTEMPTING TO SET LPIECE TO THE NUMBER
C       OF CHARACTERS THAT SHOULD BE TAKEN FROM MESSG STARTING AT
C       POSITION NEXTC.
C
C       LPIECE .EQ. 0   THE NEW LINE SENTINEL DOES NOT OCCUR IN THE
C                       REMAINDER OF THE CHARACTER STRING.  LPIECE
C                       SHOULD BE SET TO LWRAP OR LENMSG+1-NEXTC,
C                       WHICHEVER IS LESS.
C
C       LPIECE .EQ. 1   THE NEW LINE SENTINEL STARTS AT MESSG(NEXTC:
C                       NEXTC).  LPIECE IS EFFECTIVELY ZERO, AND WE
C                       PRINT NOTHING TO AVOID PRODUCING UNNECESSARY
C                       BLANK LINES.  THIS TAKES CARE OF THE SITUATION
C                       WHERE THE LIBRARY ROUTINE HAS A MESSAGE OF
C                       EXACTLY 72 CHARACTERS FOLLOWED BY A NEW LINE
C                       SENTINEL FOLLOWED BY MORE CHARACTERS.  NEXTC
C                       SHOULD BE INCREMENTED BY 2.
C
C       LPIECE .GT. LWRAP+1  REDUCE LPIECE TO LWRAP.
C
C       ELSE            THIS LAST CASE MEANS 2 .LE. LPIECE .LE. LWRAP+1
C                       RESET LPIECE = LPIECE-1.  NOTE THAT THIS
C                       PROPERLY HANDLES THE END CASE WHERE LPIECE .EQ.
C                       LWRAP+1.  THAT IS, THE SENTINEL FALLS EXACTLY
C                       AT THE END OF A LINE.
C
      NEXTC = 1

C KARLINE: The INDEX function triggers a "note" in gfortran, so it is removed
   50 LPIECE = 1 
C   50 LPIECE = INDEX(MESSG(NEXTC:LENMSG), NEWLIN)
      IF (LPIECE .EQ. 0) THEN
C
C       THERE WAS NO NEW LINE SENTINEL FOUND.
C
         IDELTA = 0
         LPIECE = MIN(LWRAP, LENMSG+1-NEXTC)
         IF (LPIECE .LT. LENMSG+1-NEXTC) THEN
            DO 52 I=LPIECE+1,2,-1
               IF (MESSG(NEXTC+I-1:NEXTC+I-1) .EQ. ' ') THEN
                  LPIECE = I-1
                  IDELTA = 1
                  GOTO 54
               ENDIF
   52       CONTINUE
         ENDIF
   54    CBUFF(LPREF+1:LPREF+LPIECE) = MESSG(NEXTC:NEXTC+LPIECE-1)
         NEXTC = NEXTC + LPIECE + IDELTA
      ELSEIF (LPIECE .EQ. 1) THEN
C
C       WE HAVE A NEW LINE SENTINEL AT MESSG(NEXTC:NEXTC+1).
C       DON'T PRINT A BLANK LINE.
C
         NEXTC = NEXTC + 2
         GO TO 50
      ELSEIF (LPIECE .GT. LWRAP+1) THEN
C
C       LPIECE SHOULD BE SET DOWN TO LWRAP.
C
         IDELTA = 0
         LPIECE = LWRAP
         DO 56 I=LPIECE+1,2,-1
            IF (MESSG(NEXTC+I-1:NEXTC+I-1) .EQ. ' ') THEN
               LPIECE = I-1
               IDELTA = 1
               GOTO 58
            ENDIF
   56    CONTINUE
   58    CBUFF(LPREF+1:LPREF+LPIECE) = MESSG(NEXTC:NEXTC+LPIECE-1)
         NEXTC = NEXTC + LPIECE + IDELTA
      ELSE
C
C       IF WE ARRIVE HERE, IT MEANS 2 .LE. LPIECE .LE. LWRAP+1.
C       WE SHOULD DECREMENT LPIECE BY ONE.
C
         LPIECE = LPIECE - 1
         CBUFF(LPREF+1:LPREF+LPIECE) = MESSG(NEXTC:NEXTC+LPIECE-1)
         NEXTC  = NEXTC + LPIECE + 2
      ENDIF
C
C       PRINT
C
C         WRITE(*, '(A)') CBUFF(1:LPREF+LPIECE)
         CALL XMESSAGE (CBUFF(1:LPREF+LPIECE))
C
      IF (NEXTC .LE. LENMSG) GO TO 50
      RETURN

      END SUBROUTINE XXERPRN



CCCCCCCCCCCCCCCCCCCCC  BANDED SYSTEMS

      SUBROUTINE DGBSV( N, KL, KU, NRHS, AB, LDAB, IPIV, B, LDB, INFO )
*
*  -- LAPACK driver routine (version 3.0) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     March 31, 1993
*
*     .. Scalar Arguments ..
      INTEGER            INFO, KL, KU, LDAB, LDB, N, NRHS
*     ..
*     .. Array Arguments ..
      INTEGER            IPIV( * )
      DOUBLE PRECISION   AB( LDAB, * ), B( LDB, * )
*     ..
*
*  Purpose
*  =======
*
*  DGBSV computes the solution to a real system of linear equations
*  A * X = B, where A is a band matrix of order N with KL subdiagonals
*  and KU superdiagonals, and X and B are N-by-NRHS matrices.
*
*  The LU decomposition with partial pivoting and row interchanges is
*  used to factor A as A = L * U, where L is a product of permutation
*  and unit lower triangular matrices with KL subdiagonals, and U is
*  upper triangular with KL+KU superdiagonals.  The factored form of A
*  is then used to solve the system of equations A * X = B.
*
*  Arguments
*  =========
*
*  N       (input) INTEGER
*          The number of linear equations, i.e., the order of the
*          matrix A.  N >= 0.
*
*  KL      (input) INTEGER
*          The number of subdiagonals within the band of A.  KL >= 0.
*
*  KU      (input) INTEGER
*          The number of superdiagonals within the band of A.  KU >= 0.
*
*  NRHS    (input) INTEGER
*          The number of right hand sides, i.e., the number of columns
*          of the matrix B.  NRHS >= 0.
*
*  AB      (input/output) DOUBLE PRECISION array, dimension (LDAB,N)
*          On entry, the matrix A in band storage, in rows KL+1 to
*          2*KL+KU+1; rows 1 to KL of the array need not be set.
*          The j-th column of A is stored in the j-th column of the
*          array AB as follows:
*          AB(KL+KU+1+i-j,j) = A(i,j) for max(1,j-KU)<=i<=min(N,j+KL)
*          On exit, details of the factorization: U is stored as an
*          upper triangular band matrix with KL+KU superdiagonals in
*          rows 1 to KL+KU+1, and the multipliers used during the
*          factorization are stored in rows KL+KU+2 to 2*KL+KU+1.
*          See below for further details.
*
*  LDAB    (input) INTEGER
*          The leading dimension of the array AB.  LDAB >= 2*KL+KU+1.
*
*  IPIV    (output) INTEGER array, dimension (N)
*          The pivot indices that define the permutation matrix P;
*          row i of the matrix was interchanged with row IPIV(i).
*
*  B       (input/output) DOUBLE PRECISION array, dimension (LDB,NRHS)
*          On entry, the N-by-NRHS right hand side matrix B.
*          On exit, if INFO = 0, the N-by-NRHS solution matrix X.
*
*  LDB     (input) INTEGER
*          The leading dimension of the array B.  LDB >= max(1,N).
*
*  INFO    (output) INTEGER
*          = 0:  successful exit
*          < 0:  if INFO = -i, the i-th argument had an illegal value
*          > 0:  if INFO = i, U(i,i) is exactly zero.  The factorization
*                has been completed, but the factor U is exactly
*                singular, and the solution has not been computed.
*
*  Further Details
*  ===============
*
*  The band storage scheme is illustrated by the following example, when
*  M = N = 6, KL = 2, KU = 1:
*
*  On entry:                       On exit:
*
*      *    *    *    +    +    +       *    *    *   u14  u25  u36
*      *    *    +    +    +    +       *    *   u13  u24  u35  u46
*      *   a12  a23  a34  a45  a56      *   u12  u23  u34  u45  u56
*     a11  a22  a33  a44  a55  a66     u11  u22  u33  u44  u55  u66
*     a21  a32  a43  a54  a65   *      m21  m32  m43  m54  m65   *
*     a31  a42  a53  a64   *    *      m31  m42  m53  m64   *    *
*
*  Array elements marked * are not used by the routine; elements marked
*  + need not be set on entry, but are required by the routine to store
*  elements of U because of fill-in resulting from the row interchanges.
*
*  =====================================================================
*
*     .. External Subroutines ..
      EXTERNAL           DGBTRF, DGBTRS, XERBLA
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          MAX
*     ..
*     .. Executable Statements ..
*
*     Test the input parameters.
*
      INFO = 0
      IF( N.LT.0 ) THEN
         INFO = -1
      ELSE IF( KL.LT.0 ) THEN
         INFO = -2
      ELSE IF( KU.LT.0 ) THEN
         INFO = -3
      ELSE IF( NRHS.LT.0 ) THEN
         INFO = -4
      ELSE IF( LDAB.LT.2*KL+KU+1 ) THEN
         INFO = -6
      ELSE IF( LDB.LT.MAX( N, 1 ) ) THEN
         INFO = -9
      END IF
      IF( INFO.NE.0 ) THEN
         CALL XERBLA( 'DGBSV ', -INFO )
         RETURN
      END IF
*
*     Compute the LU factorization of the band matrix A.
*
      CALL DGBTRF( N, N, KL, KU, AB, LDAB, IPIV, INFO )
      IF( INFO.EQ.0 ) THEN
*
*        Solve the system A*X = B, overwriting B with X.
*
         CALL DGBTRS( 'N', N, KL, KU, NRHS, AB, LDAB, IPIV,
     $                B, LDB, INFO )
      END IF
      RETURN
*
*     End of DGBSV
*
      END

      SUBROUTINE DGBTRS( TRANS, N, KL, KU, NRHS, AB, LDAB, IPIV, B, LDB,
     $                   INFO )
*
*  -- LAPACK routine (version 3.0) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     March 31, 1993
*
*     .. Scalar Arguments ..
      CHARACTER          TRANS
      INTEGER            INFO, KL, KU, LDAB, LDB, N, NRHS
*     ..
*     .. Array Arguments ..
      INTEGER            IPIV( * )
      DOUBLE PRECISION   AB( LDAB, * ), B( LDB, * )
*     ..
*
*  Purpose
*  =======
*
*  DGBTRS solves a system of linear equations
*     A * X = B  or  A' * X = B
*  with a general band matrix A using the LU factorization computed
*  by DGBTRF.
*
*  Arguments
*  =========
*
*  TRANS   (input) CHARACTER*1
*          Specifies the form of the system of equations.
*          = 'N':  A * X = B  (No transpose)
*          = 'T':  A'* X = B  (Transpose)
*          = 'C':  A'* X = B  (Conjugate transpose = Transpose)
*
*  N       (input) INTEGER
*          The order of the matrix A.  N >= 0.
*
*  KL      (input) INTEGER
*          The number of subdiagonals within the band of A.  KL >= 0.
*
*  KU      (input) INTEGER
*          The number of superdiagonals within the band of A.  KU >= 0.
*
*  NRHS    (input) INTEGER
*          The number of right hand sides, i.e., the number of columns
*          of the matrix B.  NRHS >= 0.
*
*  AB      (input) DOUBLE PRECISION array, dimension (LDAB,N)
*          Details of the LU factorization of the band matrix A, as
*          computed by DGBTRF.  U is stored as an upper triangular band
*          matrix with KL+KU superdiagonals in rows 1 to KL+KU+1, and
*          the multipliers used during the factorization are stored in
*          rows KL+KU+2 to 2*KL+KU+1.
*
*  LDAB    (input) INTEGER
*          The leading dimension of the array AB.  LDAB >= 2*KL+KU+1.
*
*  IPIV    (input) INTEGER array, dimension (N)
*          The pivot indices; for 1 <= i <= N, row i of the matrix was
*          interchanged with row IPIV(i).
*
*  B       (input/output) DOUBLE PRECISION array, dimension (LDB,NRHS)
*          On entry, the right hand side matrix B.
*          On exit, the solution matrix X.
*
*  LDB     (input) INTEGER
*          The leading dimension of the array B.  LDB >= max(1,N).
*
*  INFO    (output) INTEGER
*          = 0:  successful exit
*          < 0: if INFO = -i, the i-th argument had an illegal value
*
*  =====================================================================
*
*     .. Parameters ..
      DOUBLE PRECISION   ONE
      PARAMETER          ( ONE = 1.0D+0 )
*     ..
*     .. Local Scalars ..
      LOGICAL            LNOTI, NOTRAN
      INTEGER            I, J, KD, L, LM
*     ..
*     .. External Functions ..
      LOGICAL            LSAME
      EXTERNAL           LSAME
*     ..
*     .. External Subroutines ..
      EXTERNAL           DGEMV, DGER, DSWAP, DTBSV, XERBLA
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          MAX, MIN
*     ..
*     .. Executable Statements ..
*
*     Test the input parameters.
*
      INFO = 0
      NOTRAN = LSAME( TRANS, 'N' )
      IF( .NOT.NOTRAN .AND. .NOT.LSAME( TRANS, 'T' ) .AND. .NOT.
     $    LSAME( TRANS, 'C' ) ) THEN
         INFO = -1
      ELSE IF( N.LT.0 ) THEN
         INFO = -2
      ELSE IF( KL.LT.0 ) THEN
         INFO = -3
      ELSE IF( KU.LT.0 ) THEN
         INFO = -4
      ELSE IF( NRHS.LT.0 ) THEN
         INFO = -5
      ELSE IF( LDAB.LT.( 2*KL+KU+1 ) ) THEN
         INFO = -7
      ELSE IF( LDB.LT.MAX( 1, N ) ) THEN
         INFO = -10
      END IF
      IF( INFO.NE.0 ) THEN
         CALL XERBLA( 'DGBTRS', -INFO )
         RETURN
      END IF
*
*     Quick return if possible
*
      IF( N.EQ.0 .OR. NRHS.EQ.0 )
     $   RETURN
*
      KD = KU + KL + 1
      LNOTI = KL.GT.0
*
      IF( NOTRAN ) THEN
*
*        Solve  A*X = B.
*
*        Solve L*X = B, overwriting B with X.
*
*        L is represented as a product of permutations and unit lower
*        triangular matrices L = P(1) * L(1) * ... * P(n-1) * L(n-1),
*        where each transformation L(i) is a rank-one modification of
*        the identity matrix.
*
         IF( LNOTI ) THEN
            DO 10 J = 1, N - 1
               LM = MIN( KL, N-J )
               L = IPIV( J )
               IF( L.NE.J )
     $            CALL DSWAP( NRHS, B( L, 1 ), LDB, B( J, 1 ), LDB )
               CALL DGER( LM, NRHS, -ONE, AB( KD+1, J ), 1, B( J, 1 ),
     $                    LDB, B( J+1, 1 ), LDB )
   10       CONTINUE
         END IF
*
         DO 20 I = 1, NRHS
*
*           Solve U*X = B, overwriting B with X.
*
            CALL DTBSV( 'Upper', 'No transpose', 'Non-unit', N, KL+KU,
     $                  AB, LDAB, B( 1, I ), 1 )
   20    CONTINUE
*
      ELSE
*
*        Solve A'*X = B.
*
         DO 30 I = 1, NRHS
*
*           Solve U'*X = B, overwriting B with X.
*
            CALL DTBSV( 'Upper', 'Transpose', 'Non-unit', N, KL+KU, AB,
     $                  LDAB, B( 1, I ), 1 )
   30    CONTINUE
*
*        Solve L'*X = B, overwriting B with X.
*
         IF( LNOTI ) THEN
            DO 40 J = N - 1, 1, -1
               LM = MIN( KL, N-J )
               CALL DGEMV( 'Transpose', LM, NRHS, -ONE, B( J+1, 1 ),
     $                     LDB, AB( KD+1, J ), 1, ONE, B( J, 1 ), LDB )
               L = IPIV( J )
               IF( L.NE.J )
     $            CALL DSWAP( NRHS, B( L, 1 ), LDB, B( J, 1 ), LDB )
   40       CONTINUE
         END IF
      END IF
      RETURN
*
*     End of DGBTRS
*
      END

      SUBROUTINE DGBTRF( M, N, KL, KU, AB, LDAB, IPIV, INFO )
*
*  -- LAPACK routine (version 3.0) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     February 29, 1992
*
*     .. Scalar Arguments ..
      INTEGER            INFO, KL, KU, LDAB, M, N
*     ..
*     .. Array Arguments ..
      INTEGER            IPIV( * )
      DOUBLE PRECISION   AB( LDAB, * )
*     ..
*
*  Purpose
*  =======
*
*  DGBTRF computes an LU factorization of a real m-by-n band matrix A
*  using partial pivoting with row interchanges.
*
*  This is the blocked version of the algorithm, calling Level 3 BLAS.
*
*  Arguments
*  =========
*
*  M       (input) INTEGER
*          The number of rows of the matrix A.  M >= 0.
*
*  N       (input) INTEGER
*          The number of columns of the matrix A.  N >= 0.
*
*  KL      (input) INTEGER
*          The number of subdiagonals within the band of A.  KL >= 0.
*
*  KU      (input) INTEGER
*          The number of superdiagonals within the band of A.  KU >= 0.
*
*  AB      (input/output) DOUBLE PRECISION array, dimension (LDAB,N)
*          On entry, the matrix A in band storage, in rows KL+1 to
*          2*KL+KU+1; rows 1 to KL of the array need not be set.
*          The j-th column of A is stored in the j-th column of the
*          array AB as follows:
*          AB(kl+ku+1+i-j,j) = A(i,j) for max(1,j-ku)<=i<=min(m,j+kl)
*
*          On exit, details of the factorization: U is stored as an
*          upper triangular band matrix with KL+KU superdiagonals in
*          rows 1 to KL+KU+1, and the multipliers used during the
*          factorization are stored in rows KL+KU+2 to 2*KL+KU+1.
*          See below for further details.
*
*  LDAB    (input) INTEGER
*          The leading dimension of the array AB.  LDAB >= 2*KL+KU+1.
*
*  IPIV    (output) INTEGER array, dimension (min(M,N))
*          The pivot indices; for 1 <= i <= min(M,N), row i of the
*          matrix was interchanged with row IPIV(i).
*
*  INFO    (output) INTEGER
*          = 0: successful exit
*          < 0: if INFO = -i, the i-th argument had an illegal value
*          > 0: if INFO = +i, U(i,i) is exactly zero. The factorization
*               has been completed, but the factor U is exactly
*               singular, and division by zero will occur if it is used
*               to solve a system of equations.
*
*  Further Details
*  ===============
*
*  The band storage scheme is illustrated by the following example, when
*  M = N = 6, KL = 2, KU = 1:
*
*  On entry:                       On exit:
*
*      *    *    *    +    +    +       *    *    *   u14  u25  u36
*      *    *    +    +    +    +       *    *   u13  u24  u35  u46
*      *   a12  a23  a34  a45  a56      *   u12  u23  u34  u45  u56
*     a11  a22  a33  a44  a55  a66     u11  u22  u33  u44  u55  u66
*     a21  a32  a43  a54  a65   *      m21  m32  m43  m54  m65   *
*     a31  a42  a53  a64   *    *      m31  m42  m53  m64   *    *
*
*  Array elements marked * are not used by the routine; elements marked
*  + need not be set on entry, but are required by the routine to store
*  elements of U because of fill-in resulting from the row interchanges.
*
*  =====================================================================
*
*     .. Parameters ..
      DOUBLE PRECISION   ONE, ZERO
      PARAMETER          ( ONE = 1.0D+0, ZERO = 0.0D+0 )
      INTEGER            NBMAX, LDWORK
      PARAMETER          ( NBMAX = 64, LDWORK = NBMAX+1 )
*     ..
*     .. Local Scalars ..
      INTEGER            I, I2, I3, II, IP, J, J2, J3, JB, JJ, JM, JP,
     $                   JU, K2, KM, KV, NB, NW
      DOUBLE PRECISION   TEMP
*     ..
*     .. Local Arrays ..
      DOUBLE PRECISION   WORK13( LDWORK, NBMAX ),
     $                   WORK31( LDWORK, NBMAX )
*     ..
*     .. External Functions ..
      INTEGER            IDAMAX
      EXTERNAL           IDAMAX
*     ..
*     .. External Subroutines ..
      EXTERNAL           DCOPY, DGBTF2, DGEMM, DGER, DLASWP, DSCAL,
     $                   DSWAP, DTRSM, XERBLA
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          MAX, MIN
*     ..
*     .. Executable Statements ..
*
*     KV is the number of superdiagonals in the factor U, allowing for
*     fill-in
*
      KV = KU + KL
*
*     Test the input parameters.
*
      INFO = 0
      IF( M.LT.0 ) THEN
         INFO = -1
      ELSE IF( N.LT.0 ) THEN
         INFO = -2
      ELSE IF( KL.LT.0 ) THEN
         INFO = -3
      ELSE IF( KU.LT.0 ) THEN
         INFO = -4
      ELSE IF( LDAB.LT.KL+KV+1 ) THEN
         INFO = -6
      END IF
      IF( INFO.NE.0 ) THEN
         CALL XERBLA( 'DGBTRF', -INFO )
         RETURN
      END IF
*
*     Quick return if possible
*
      IF( M.EQ.0 .OR. N.EQ.0 )
     $   RETURN
*
*     Determine the block size for this environment
*
C      NB = ILAENV( 1, 'DGBTRF', ' ', M, N, KL, KU )
C KARLINE: HARD-CODED THIS TO AVOID CALL TO ILAENV
      IF (KU > 64) THEN
         NB = 32
      ELSE
         NB = 1
      ENDIF
C KARLINE-END            
*
*     The block size must not exceed the limit set by the size of the
*     local arrays WORK13 and WORK31.
*
      NB = MIN( NB, NBMAX )
*
      IF( NB.LE.1 .OR. NB.GT.KL ) THEN
*
*        Use unblocked code
*
         CALL DGBTF2( M, N, KL, KU, AB, LDAB, IPIV, INFO )
      ELSE
*
*        Use blocked code
*
*        Zero the superdiagonal elements of the work array WORK13
*
         DO 20 J = 1, NB
            DO 10 I = 1, J - 1
               WORK13( I, J ) = ZERO
   10       CONTINUE
   20    CONTINUE
*
*        Zero the subdiagonal elements of the work array WORK31
*
         DO 40 J = 1, NB
            DO 30 I = J + 1, NB
               WORK31( I, J ) = ZERO
   30       CONTINUE
   40    CONTINUE
*
*        Gaussian elimination with partial pivoting
*
*        Set fill-in elements in columns KU+2 to KV to zero
*
         DO 60 J = KU + 2, MIN( KV, N )
            DO 50 I = KV - J + 2, KL
               AB( I, J ) = ZERO
   50       CONTINUE
   60    CONTINUE
*
*        JU is the index of the last column affected by the current
*        stage of the factorization
*
         JU = 1
*
         DO 180 J = 1, MIN( M, N ), NB
            JB = MIN( NB, MIN( M, N )-J+1 )
*
*           The active part of the matrix is partitioned
*
*              A11   A12   A13
*              A21   A22   A23
*              A31   A32   A33
*
*           Here A11, A21 and A31 denote the current block of JB columns
*           which is about to be factorized. The number of rows in the
*           partitioning are JB, I2, I3 respectively, and the numbers
*           of columns are JB, J2, J3. The superdiagonal elements of A13
*           and the subdiagonal elements of A31 lie outside the band.
*
            I2 = MIN( KL-JB, M-J-JB+1 )
            I3 = MIN( JB, M-J-KL+1 )
*
*           J2 and J3 are computed after JU has been updated.
*
*           Factorize the current block of JB columns
*
            DO 80 JJ = J, J + JB - 1
*
*              Set fill-in elements in column JJ+KV to zero
*
               IF( JJ+KV.LE.N ) THEN
                  DO 70 I = 1, KL
                     AB( I, JJ+KV ) = ZERO
   70             CONTINUE
               END IF
*
*              Find pivot and test for singularity. KM is the number of
*              subdiagonal elements in the current column.
*
               KM = MIN( KL, M-JJ )
               JP = IDAMAX( KM+1, AB( KV+1, JJ ), 1 )
               IPIV( JJ ) = JP + JJ - J
               IF( AB( KV+JP, JJ ).NE.ZERO ) THEN
                  JU = MAX( JU, MIN( JJ+KU+JP-1, N ) )
                  IF( JP.NE.1 ) THEN
*
*                    Apply interchange to columns J to J+JB-1
*
                     IF( JP+JJ-1.LT.J+KL ) THEN
*
                        CALL DSWAP( JB, AB( KV+1+JJ-J, J ), LDAB-1,
     $                              AB( KV+JP+JJ-J, J ), LDAB-1 )
                     ELSE
*
*                       The interchange affects columns J to JJ-1 of A31
*                       which are stored in the work array WORK31
*
                        CALL DSWAP( JJ-J, AB( KV+1+JJ-J, J ), LDAB-1,
     $                              WORK31( JP+JJ-J-KL, 1 ), LDWORK )
                        CALL DSWAP( J+JB-JJ, AB( KV+1, JJ ), LDAB-1,
     $                              AB( KV+JP, JJ ), LDAB-1 )
                     END IF
                  END IF
*
*                 Compute multipliers
*
                  CALL DSCAL( KM, ONE / AB( KV+1, JJ ), AB( KV+2, JJ ),
     $                        1 )
*
*                 Update trailing submatrix within the band and within
*                 the current block. JM is the index of the last column
*                 which needs to be updated.
*
                  JM = MIN( JU, J+JB-1 )
                  IF( JM.GT.JJ )
     $               CALL DGER( KM, JM-JJ, -ONE, AB( KV+2, JJ ), 1,
     $                          AB( KV, JJ+1 ), LDAB-1,
     $                          AB( KV+1, JJ+1 ), LDAB-1 )
               ELSE
*
*                 If pivot is zero, set INFO to the index of the pivot
*                 unless a zero pivot has already been found.
*
                  IF( INFO.EQ.0 )
     $               INFO = JJ
               END IF
*
*              Copy current column of A31 into the work array WORK31
*
               NW = MIN( JJ-J+1, I3 )
               IF( NW.GT.0 )
     $            CALL DCOPY( NW, AB( KV+KL+1-JJ+J, JJ ), 1,
     $                        WORK31( 1, JJ-J+1 ), 1 )
   80       CONTINUE
            IF( J+JB.LE.N ) THEN
*
*              Apply the row interchanges to the other blocks.
*
               J2 = MIN( JU-J+1, KV ) - JB
               J3 = MAX( 0, JU-J-KV+1 )
*
*              Use DLASWP to apply the row interchanges to A12, A22, and
*              A32.
*
               CALL DLASWP( J2, AB( KV+1-JB, J+JB ), LDAB-1, 1, JB,
     $                      IPIV( J ), 1 )
*
*              Adjust the pivot indices.
*
               DO 90 I = J, J + JB - 1
                  IPIV( I ) = IPIV( I ) + J - 1
   90          CONTINUE
*
*              Apply the row interchanges to A13, A23, and A33
*              columnwise.
*
               K2 = J - 1 + JB + J2
               DO 110 I = 1, J3
                  JJ = K2 + I
                  DO 100 II = J + I - 1, J + JB - 1
                     IP = IPIV( II )
                     IF( IP.NE.II ) THEN
                        TEMP = AB( KV+1+II-JJ, JJ )
                        AB( KV+1+II-JJ, JJ ) = AB( KV+1+IP-JJ, JJ )
                        AB( KV+1+IP-JJ, JJ ) = TEMP
                     END IF
  100             CONTINUE
  110          CONTINUE
*
*              Update the relevant part of the trailing submatrix
*
               IF( J2.GT.0 ) THEN
*
*                 Update A12
*
                  CALL DTRSM( 'Left', 'Lower', 'No transpose', 'Unit',
     $                        JB, J2, ONE, AB( KV+1, J ), LDAB-1,
     $                        AB( KV+1-JB, J+JB ), LDAB-1 )
*
                  IF( I2.GT.0 ) THEN
*
*                    Update A22
*
                     CALL DGEMM( 'No transpose', 'No transpose', I2, J2,
     $                           JB, -ONE, AB( KV+1+JB, J ), LDAB-1,
     $                           AB( KV+1-JB, J+JB ), LDAB-1, ONE,
     $                           AB( KV+1, J+JB ), LDAB-1 )
                  END IF
*
                  IF( I3.GT.0 ) THEN
*
*                    Update A32
*
                     CALL DGEMM( 'No transpose', 'No transpose', I3, J2,
     $                           JB, -ONE, WORK31, LDWORK,
     $                           AB( KV+1-JB, J+JB ), LDAB-1, ONE,
     $                           AB( KV+KL+1-JB, J+JB ), LDAB-1 )
                  END IF
               END IF
*
               IF( J3.GT.0 ) THEN
*
*                 Copy the lower triangle of A13 into the work array
*                 WORK13
*
                  DO 130 JJ = 1, J3
                     DO 120 II = JJ, JB
                        WORK13( II, JJ ) = AB( II-JJ+1, JJ+J+KV-1 )
  120                CONTINUE
  130             CONTINUE
*
*                 Update A13 in the work array
*
                  CALL DTRSM( 'Left', 'Lower', 'No transpose', 'Unit',
     $                        JB, J3, ONE, AB( KV+1, J ), LDAB-1,
     $                        WORK13, LDWORK )
*
                  IF( I2.GT.0 ) THEN
*
*                    Update A23
*
                     CALL DGEMM( 'No transpose', 'No transpose', I2, J3,
     $                           JB, -ONE, AB( KV+1+JB, J ), LDAB-1,
     $                           WORK13, LDWORK, ONE, AB( 1+JB, J+KV ),
     $                           LDAB-1 )
                  END IF
*
                  IF( I3.GT.0 ) THEN
*
*                    Update A33
*
                     CALL DGEMM( 'No transpose', 'No transpose', I3, J3,
     $                           JB, -ONE, WORK31, LDWORK, WORK13,
     $                           LDWORK, ONE, AB( 1+KL, J+KV ), LDAB-1 )
                  END IF
*
*                 Copy the lower triangle of A13 back into place
*
                  DO 150 JJ = 1, J3
                     DO 140 II = JJ, JB
                        AB( II-JJ+1, JJ+J+KV-1 ) = WORK13( II, JJ )
  140                CONTINUE
  150             CONTINUE
               END IF
            ELSE
*
*              Adjust the pivot indices.
*
               DO 160 I = J, J + JB - 1
                  IPIV( I ) = IPIV( I ) + J - 1
  160          CONTINUE
            END IF
*
*           Partially undo the interchanges in the current block to
*           restore the upper triangular form of A31 and copy the upper
*           triangle of A31 back into place
*
            DO 170 JJ = J + JB - 1, J, -1
               JP = IPIV( JJ ) - JJ + 1
               IF( JP.NE.1 ) THEN
*
*                 Apply interchange to columns J to JJ-1
*
                  IF( JP+JJ-1.LT.J+KL ) THEN
*
*                    The interchange does not affect A31
*
                     CALL DSWAP( JJ-J, AB( KV+1+JJ-J, J ), LDAB-1,
     $                           AB( KV+JP+JJ-J, J ), LDAB-1 )
                  ELSE
*
*                    The interchange does affect A31
*
                     CALL DSWAP( JJ-J, AB( KV+1+JJ-J, J ), LDAB-1,
     $                           WORK31( JP+JJ-J-KL, 1 ), LDWORK )
                  END IF
               END IF
*
*              Copy the current column of A31 back into place
*
               NW = MIN( I3, JJ-J+1 )
               IF( NW.GT.0 )
     $            CALL DCOPY( NW, WORK31( 1, JJ-J+1 ), 1,
     $                        AB( KV+KL+1-JJ+J, JJ ), 1 )
  170       CONTINUE
  180    CONTINUE
      END IF
*
      RETURN
*
*     End of DGBTRF
*
      END


      SUBROUTINE DGBTF2( M, N, KL, KU, AB, LDAB, IPIV, INFO )
*
*  -- LAPACK routine (version 3.1) --
*     Univ. of Tennessee, Univ. of California Berkeley and NAG Ltd..
*     November 2006
*
*     .. Scalar Arguments ..
      INTEGER            INFO, KL, KU, LDAB, M, N
*     ..
*     .. Array Arguments ..
      INTEGER            IPIV( * )
      DOUBLE PRECISION   AB( LDAB, * )
*     ..
*
*  Purpose
*  =======
*
*  DGBTF2 computes an LU factorization of a real m-by-n band matrix A
*  using partial pivoting with row interchanges.
*
*  This is the unblocked version of the algorithm, calling Level 2 BLAS.
*
*  Arguments
*  =========
*
*  M       (input) INTEGER
*          The number of rows of the matrix A.  M >= 0.
*
*  N       (input) INTEGER
*          The number of columns of the matrix A.  N >= 0.
*
*  KL      (input) INTEGER
*          The number of subdiagonals within the band of A.  KL >= 0.
*
*  KU      (input) INTEGER
*          The number of superdiagonals within the band of A.  KU >= 0.
*
*  AB      (input/output) DOUBLE PRECISION array, dimension (LDAB,N)
*          On entry, the matrix A in band storage, in rows KL+1 to
*          2*KL+KU+1; rows 1 to KL of the array need not be set.
*          The j-th column of A is stored in the j-th column of the
*          array AB as follows:
*          AB(kl+ku+1+i-j,j) = A(i,j) for max(1,j-ku)<=i<=min(m,j+kl)
*
*          On exit, details of the factorization: U is stored as an
*          upper triangular band matrix with KL+KU superdiagonals in
*          rows 1 to KL+KU+1, and the multipliers used during the
*          factorization are stored in rows KL+KU+2 to 2*KL+KU+1.
*          See below for further details.
*
*  LDAB    (input) INTEGER
*          The leading dimension of the array AB.  LDAB >= 2*KL+KU+1.
*
*  IPIV    (output) INTEGER array, dimension (min(M,N))
*          The pivot indices; for 1 <= i <= min(M,N), row i of the
*          matrix was interchanged with row IPIV(i).
*
*  INFO    (output) INTEGER
*          = 0: successful exit
*          < 0: if INFO = -i, the i-th argument had an illegal value
*          > 0: if INFO = +i, U(i,i) is exactly zero. The factorization
*               has been completed, but the factor U is exactly
*               singular, and division by zero will occur if it is used
*               to solve a system of equations.
*
*  Further Details
*  ===============
*
*  The band storage scheme is illustrated by the following example, when
*  M = N = 6, KL = 2, KU = 1:
*
*  On entry:                       On exit:
*
*      *    *    *    +    +    +       *    *    *   u14  u25  u36
*      *    *    +    +    +    +       *    *   u13  u24  u35  u46
*      *   a12  a23  a34  a45  a56      *   u12  u23  u34  u45  u56
*     a11  a22  a33  a44  a55  a66     u11  u22  u33  u44  u55  u66
*     a21  a32  a43  a54  a65   *      m21  m32  m43  m54  m65   *
*     a31  a42  a53  a64   *    *      m31  m42  m53  m64   *    *
*
*  Array elements marked * are not used by the routine; elements marked
*  + need not be set on entry, but are required by the routine to store
*  elements of U, because of fill-in resulting from the row
*  interchanges.
*
*  =====================================================================
*
*     .. Parameters ..
      DOUBLE PRECISION   ONE, ZERO
      PARAMETER          ( ONE = 1.0D+0, ZERO = 0.0D+0 )
*     ..
*     .. Local Scalars ..
      INTEGER            I, J, JP, JU, KM, KV
*     ..
*     .. External Functions ..
      INTEGER            IDAMAX
      EXTERNAL           IDAMAX
*     ..
*     .. External Subroutines ..
      EXTERNAL           DGER, DSCAL, DSWAP, XERBLA
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          MAX, MIN
*     ..
*     .. Executable Statements ..
*
*     KV is the number of superdiagonals in the factor U, allowing for
*     fill-in.
*
      KV = KU + KL
*
*     Test the input parameters.
*
      INFO = 0
      IF( M.LT.0 ) THEN
         INFO = -1
      ELSE IF( N.LT.0 ) THEN
         INFO = -2
      ELSE IF( KL.LT.0 ) THEN
         INFO = -3
      ELSE IF( KU.LT.0 ) THEN
         INFO = -4
      ELSE IF( LDAB.LT.KL+KV+1 ) THEN
         INFO = -6
      END IF
      IF( INFO.NE.0 ) THEN
         CALL XERBLA( 'DGBTF2', -INFO )
         RETURN
      END IF
*
*     Quick return if possible
*
      IF( M.EQ.0 .OR. N.EQ.0 )
     $   RETURN
*
*     Gaussian elimination with partial pivoting
*
*     Set fill-in elements in columns KU+2 to KV to zero.
*
      DO 20 J = KU + 2, MIN( KV, N )
         DO 10 I = KV - J + 2, KL
            AB( I, J ) = ZERO
   10    CONTINUE
   20 CONTINUE
*
*     JU is the index of the last column affected by the current stage
*     of the factorization.
*
      JU = 1
*
      DO 40 J = 1, MIN( M, N )
*
*        Set fill-in elements in column J+KV to zero.
*
         IF( J+KV.LE.N ) THEN
            DO 30 I = 1, KL
               AB( I, J+KV ) = ZERO
   30       CONTINUE
         END IF
*
*        Find pivot and test for singularity. KM is the number of
*        subdiagonal elements in the current column.
*
         KM = MIN( KL, M-J )
         JP = IDAMAX( KM+1, AB( KV+1, J ), 1 )
         IPIV( J ) = JP + J - 1
         IF( AB( KV+JP, J ).NE.ZERO ) THEN
            JU = MAX( JU, MIN( J+KU+JP-1, N ) )
*
*           Apply interchange to columns J to JU.
*
            IF( JP.NE.1 )
     $         CALL DSWAP( JU-J+1, AB( KV+JP, J ), LDAB-1,
     $                     AB( KV+1, J ), LDAB-1 )
*
            IF( KM.GT.0 ) THEN
*
*              Compute multipliers.
*
               CALL DSCAL( KM, ONE / AB( KV+1, J ), AB( KV+2, J ), 1 )
*
*              Update trailing submatrix within the band.
*
               IF( JU.GT.J )
     $            CALL DGER( KM, JU-J, -ONE, AB( KV+2, J ), 1,
     $                       AB( KV, J+1 ), LDAB-1, AB( KV+1, J+1 ),
     $                       LDAB-1 )
            END IF
         ELSE
*
*           If pivot is zero, set INFO to the index of the pivot
*           unless a zero pivot has already been found.
*
            IF( INFO.EQ.0 )
     $         INFO = J
         END IF
   40 CONTINUE
      RETURN
*
*     End of DGBTF2
*
      END








cccccccccccc banded matrices

      SUBROUTINE DLASWP( N, A, LDA, K1, K2, IPIV, INCX )
*
*  -- LAPACK auxiliary routine (version 3.1) --
*     Univ. of Tennessee, Univ. of California Berkeley and NAG Ltd..
*     November 2006
*
*     .. Scalar Arguments ..
      INTEGER            INCX, K1, K2, LDA, N
*     ..
*     .. Array Arguments ..
      INTEGER            IPIV( * )
      DOUBLE PRECISION   A( LDA, * )
*     ..
*
*  Purpose
*  =======
*
*  DLASWP performs a series of row interchanges on the matrix A.
*  One row interchange is initiated for each of rows K1 through K2 of A.
*
*  Arguments
*  =========
*
*  N       (input) INTEGER
*          The number of columns of the matrix A.
*
*  A       (input/output) DOUBLE PRECISION array, dimension (LDA,N)
*          On entry, the matrix of column dimension N to which the row
*          interchanges will be applied.
*          On exit, the permuted matrix.
*
*  LDA     (input) INTEGER
*          The leading dimension of the array A.
*
*  K1      (input) INTEGER
*          The first element of IPIV for which a row interchange will
*          be done.
*
*  K2      (input) INTEGER
*          The last element of IPIV for which a row interchange will
*          be done.
*
*  IPIV    (input) INTEGER array, dimension (K2*abs(INCX))
*          The vector of pivot indices.  Only the elements in positions
*          K1 through K2 of IPIV are accessed.
*          IPIV(K) = L implies rows K and L are to be interchanged.
*
*  INCX    (input) INTEGER
*          The increment between successive values of IPIV.  If IPIV
*          is negative, the pivots are applied in reverse order.
*
*  Further Details
*  ===============
*
*  Modified by
*   R. C. Whaley, Computer Science Dept., Univ. of Tenn., Knoxville, USA
*
* =====================================================================
*
*     .. Local Scalars ..
      INTEGER            I, I1, I2, INC, IP, IX, IX0, J, K, N32
      DOUBLE PRECISION   TEMP
*     ..
*     .. Executable Statements ..
*
*     Interchange row I with row IPIV(I) for each of rows K1 through K2.
*
      IF( INCX.GT.0 ) THEN
         IX0 = K1
         I1 = K1
         I2 = K2
         INC = 1
      ELSE IF( INCX.LT.0 ) THEN
         IX0 = 1 + ( 1-K2 )*INCX
         I1 = K2
         I2 = K1
         INC = -1
      ELSE
         RETURN
      END IF
*
      N32 = ( N / 32 )*32
      IF( N32.NE.0 ) THEN
         DO 30 J = 1, N32, 32
            IX = IX0
            DO 20 I = I1, I2, INC
               IP = IPIV( IX )
               IF( IP.NE.I ) THEN
                  DO 10 K = J, J + 31
                     TEMP = A( I, K )
                     A( I, K ) = A( IP, K )
                     A( IP, K ) = TEMP
   10             CONTINUE
               END IF
               IX = IX + INCX
   20       CONTINUE
   30    CONTINUE
      END IF
      IF( N32.NE.N ) THEN
         N32 = N32 + 1
         IX = IX0
         DO 50 I = I1, I2, INC
            IP = IPIV( IX )
            IF( IP.NE.I ) THEN
               DO 40 K = N32, N
                  TEMP = A( I, K )
                  A( I, K ) = A( IP, K )
                  A( IP, K ) = TEMP
   40          CONTINUE
            END IF
            IX = IX + INCX
   50    CONTINUE
      END IF
*
      RETURN
*
*     End of DLASWP
*
      END


      SUBROUTINE DGETRS( TRANS, N, NRHS, A, LDA, IPIV, B, LDB, INFO )
*
*  -- LAPACK routine (version 3.1) --
*     Univ. of Tennessee, Univ. of California Berkeley and NAG Ltd..
*     November 2006
*
*     .. Scalar Arguments ..
      CHARACTER          TRANS
      INTEGER            INFO, LDA, LDB, N, NRHS
*     ..
*     .. Array Arguments ..
      INTEGER            IPIV( * )
      DOUBLE PRECISION   A( LDA, * ), B( LDB, * )
*     ..
*
*  Purpose
*  =======
*
*  DGETRS solves a system of linear equations
*     A * X = B  or  A' * X = B
*  with a general N-by-N matrix A using the LU factorization computed
*  by DGETRF.
*
*  Arguments
*  =========
*
*  TRANS   (input) CHARACTER*1
*          Specifies the form of the system of equations:
*          = 'N':  A * X = B  (No transpose)
*          = 'T':  A'* X = B  (Transpose)
*          = 'C':  A'* X = B  (Conjugate transpose = Transpose)
*
*  N       (input) INTEGER
*          The order of the matrix A.  N >= 0.
*
*  NRHS    (input) INTEGER
*          The number of right hand sides, i.e., the number of columns
*          of the matrix B.  NRHS >= 0.
*
*  A       (input) DOUBLE PRECISION array, dimension (LDA,N)
*          The factors L and U from the factorization A = P*L*U
*          as computed by DGETRF.
*
*  LDA     (input) INTEGER
*          The leading dimension of the array A.  LDA >= max(1,N).
*
*  IPIV    (input) INTEGER array, dimension (N)
*          The pivot indices from DGETRF; for 1<=i<=N, row i of the
*          matrix was interchanged with row IPIV(i).
*
*  B       (input/output) DOUBLE PRECISION array, dimension (LDB,NRHS)
*          On entry, the right hand side matrix B.
*          On exit, the solution matrix X.
*
*  LDB     (input) INTEGER
*          The leading dimension of the array B.  LDB >= max(1,N).
*
*  INFO    (output) INTEGER
*          = 0:  successful exit
*          < 0:  if INFO = -i, the i-th argument had an illegal value
*
*  =====================================================================
*
*     .. Parameters ..
      DOUBLE PRECISION   ONE
      PARAMETER          ( ONE = 1.0D+0 )
*     ..
*     .. Local Scalars ..
      LOGICAL            NOTRAN
*     ..
*     .. External Functions ..
      LOGICAL            LSAME
      EXTERNAL           LSAME
*     ..
*     .. External Subroutines ..
      EXTERNAL           DLASWP, DTRSM, XERBLA
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          MAX
*     ..
*     .. Executable Statements ..
*
*     Test the input parameters.
*
      INFO = 0
      NOTRAN = LSAME( TRANS, 'N' )
      IF( .NOT.NOTRAN .AND. .NOT.LSAME( TRANS, 'T' ) .AND. .NOT.
     $    LSAME( TRANS, 'C' ) ) THEN
         INFO = -1
      ELSE IF( N.LT.0 ) THEN
         INFO = -2
      ELSE IF( NRHS.LT.0 ) THEN
         INFO = -3
      ELSE IF( LDA.LT.MAX( 1, N ) ) THEN
         INFO = -5
      ELSE IF( LDB.LT.MAX( 1, N ) ) THEN
         INFO = -8
      END IF
      IF( INFO.NE.0 ) THEN
         CALL XERBLA( 'DGETRS', -INFO )
         RETURN
      END IF
*
*     Quick return if possible
*
      IF( N.EQ.0 .OR. NRHS.EQ.0 )
     $   RETURN
*
      IF( NOTRAN ) THEN
*
*        Solve A * X = B.
*
*        Apply row interchanges to the right hand sides.
*
         CALL DLASWP( NRHS, B, LDB, 1, N, IPIV, 1 )
*
*        Solve L*X = B, overwriting B with X.
*
         CALL DTRSM( 'Left', 'Lower', 'No transpose', 'Unit', N, NRHS,
     $               ONE, A, LDA, B, LDB )
*
*        Solve U*X = B, overwriting B with X.
*
         CALL DTRSM( 'Left', 'Upper', 'No transpose', 'Non-unit', N,
     $               NRHS, ONE, A, LDA, B, LDB )
      ELSE
*
*        Solve A' * X = B.
*
*        Solve U'*X = B, overwriting B with X.
*
         CALL DTRSM( 'Left', 'Upper', 'Transpose', 'Non-unit', N, NRHS,
     $               ONE, A, LDA, B, LDB )
*
*        Solve L'*X = B, overwriting B with X.
*
         CALL DTRSM( 'Left', 'Lower', 'Transpose', 'Unit', N, NRHS, ONE,
     $               A, LDA, B, LDB )
*
*        Apply row interchanges to the solution vectors.
*
         CALL DLASWP( NRHS, B, LDB, 1, N, IPIV, -1 )
      END IF
*
      RETURN
*
*     End of DGETRS
*
      END
      SUBROUTINE DGETRF( M, N, A, LDA, IPIV, INFO )
*
*  -- LAPACK routine (version 3.1) --
*     Univ. of Tennessee, Univ. of California Berkeley and NAG Ltd..
*     November 2006
*
*     .. Scalar Arguments ..
      INTEGER            INFO, LDA, M, N
*     ..
*     .. Array Arguments ..
      INTEGER            IPIV( * )
      DOUBLE PRECISION   A( LDA, * )
*     ..
*
*  Purpose
*  =======
*
*  DGETRF computes an LU factorization of a general M-by-N matrix A
*  using partial pivoting with row interchanges.
*
*  The factorization has the form
*     A = P * L * U
*  where P is a permutation matrix, L is lower triangular with unit
*  diagonal elements (lower trapezoidal if m > n), and U is upper
*  triangular (upper trapezoidal if m < n).
*
*  This is the right-looking Level 3 BLAS version of the algorithm.
*
*  Arguments
*  =========
*
*  M       (input) INTEGER
*          The number of rows of the matrix A.  M >= 0.
*
*  N       (input) INTEGER
*          The number of columns of the matrix A.  N >= 0.
*
*  A       (input/output) DOUBLE PRECISION array, dimension (LDA,N)
*          On entry, the M-by-N matrix to be factored.
*          On exit, the factors L and U from the factorization
*          A = P*L*U; the unit diagonal elements of L are not stored.
*
*  LDA     (input) INTEGER
*          The leading dimension of the array A.  LDA >= max(1,M).
*
*  IPIV    (output) INTEGER array, dimension (min(M,N))
*          The pivot indices; for 1 <= i <= min(M,N), row i of the
*          matrix was interchanged with row IPIV(i).
*
*  INFO    (output) INTEGER
*          = 0:  successful exit
*          < 0:  if INFO = -i, the i-th argument had an illegal value
*          > 0:  if INFO = i, U(i,i) is exactly zero. The factorization
*                has been completed, but the factor U is exactly
*                singular, and division by zero will occur if it is used
*                to solve a system of equations.
*
*  =====================================================================
*
*     .. Parameters ..
      DOUBLE PRECISION   ONE
      PARAMETER          ( ONE = 1.0D+0 )
*     ..
*     .. Local Scalars ..
      INTEGER            I, IINFO, J, JB, NB
*     ..
*     .. External Subroutines ..
      EXTERNAL           DGEMM, DGETF2, DLASWP, DTRSM, XERBLA
*     ..
*     .. External Functions ..
C     INTEGER            ILAENV
C      EXTERNAL           ILAENV
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          MAX, MIN
*     ..
*     .. Executable Statements ..
*
*     Test the input parameters.
*
      INFO = 0
      IF( M.LT.0 ) THEN
         INFO = -1
      ELSE IF( N.LT.0 ) THEN
         INFO = -2
      ELSE IF( LDA.LT.MAX( 1, M ) ) THEN
         INFO = -4
      END IF
      IF( INFO.NE.0 ) THEN
         CALL XERBLA( 'DGETRF', -INFO )
         RETURN
      END IF
*
*     Quick return if possible
*
      IF( M.EQ.0 .OR. N.EQ.0 )
     $   RETURN
*
*     Determine the block size for this environment.
*
C KARLINE: HARD-CODED THIS TO AVOID ILAAENV
C      NB = ILAENV( 1, 'DGETRF', ' ', M, N, -1, -1 )
      NB = 64
      IF( NB.LE.1 .OR. NB.GE.MIN( M, N ) ) THEN
*
*        Use unblocked code.
*
         CALL DGETF2( M, N, A, LDA, IPIV, INFO )
      ELSE
*
*        Use blocked code.
*
         DO 20 J = 1, MIN( M, N ), NB
            JB = MIN( MIN( M, N )-J+1, NB )
*
*           Factor diagonal and subdiagonal blocks and test for exact
*           singularity.
*
            CALL DGETF2( M-J+1, JB, A( J, J ), LDA, IPIV( J ), IINFO )
*
*           Adjust INFO and the pivot indices.
*
            IF( INFO.EQ.0 .AND. IINFO.GT.0 )
     $         INFO = IINFO + J - 1
            DO 10 I = J, MIN( M, J+JB-1 )
               IPIV( I ) = J - 1 + IPIV( I )
   10       CONTINUE
*
*           Apply interchanges to columns 1:J-1.
*
            CALL DLASWP( J-1, A, LDA, J, J+JB-1, IPIV, 1 )
*
            IF( J+JB.LE.N ) THEN
*
*              Apply interchanges to columns J+JB:N.
*
               CALL DLASWP( N-J-JB+1, A( 1, J+JB ), LDA, J, J+JB-1,
     $                      IPIV, 1 )
*
*              Compute block row of U.
*
               CALL DTRSM( 'Left', 'Lower', 'No transpose', 'Unit', JB,
     $                     N-J-JB+1, ONE, A( J, J ), LDA, A( J, J+JB ),
     $                     LDA )
               IF( J+JB.LE.M ) THEN
*
*                 Update trailing submatrix.
*
                  CALL DGEMM( 'No transpose', 'No transpose', M-J-JB+1,
     $                        N-J-JB+1, JB, -ONE, A( J+JB, J ), LDA,
     $                        A( J, J+JB ), LDA, ONE, A( J+JB, J+JB ),
     $                        LDA )
               END IF
            END IF
   20    CONTINUE
      END IF
      RETURN
*
*     End of DGETRF
*
      END
      SUBROUTINE DGETF2( M, N, A, LDA, IPIV, INFO )
*
*  -- LAPACK routine (version 3.1) --
*     Univ. of Tennessee, Univ. of California Berkeley and NAG Ltd..
*     November 2006
*
*     .. Scalar Arguments ..
      INTEGER            INFO, LDA, M, N
*     ..
*     .. Array Arguments ..
      INTEGER            IPIV( * )
      DOUBLE PRECISION   A( LDA, * )
*     ..
*
*  Purpose
*  =======
*
*  DGETF2 computes an LU factorization of a general m-by-n matrix A
*  using partial pivoting with row interchanges.
*
*  The factorization has the form
*     A = P * L * U
*  where P is a permutation matrix, L is lower triangular with unit
*  diagonal elements (lower trapezoidal if m > n), and U is upper
*  triangular (upper trapezoidal if m < n).
*
*  This is the right-looking Level 2 BLAS version of the algorithm.
*
*  Arguments
*  =========
*
*  M       (input) INTEGER
*          The number of rows of the matrix A.  M >= 0.
*
*  N       (input) INTEGER
*          The number of columns of the matrix A.  N >= 0.
*
*  A       (input/output) DOUBLE PRECISION array, dimension (LDA,N)
*          On entry, the m by n matrix to be factored.
*          On exit, the factors L and U from the factorization
*          A = P*L*U; the unit diagonal elements of L are not stored.
*
*  LDA     (input) INTEGER
*          The leading dimension of the array A.  LDA >= max(1,M).
*
*  IPIV    (output) INTEGER array, dimension (min(M,N))
*          The pivot indices; for 1 <= i <= min(M,N), row i of the
*          matrix was interchanged with row IPIV(i).
*
*  INFO    (output) INTEGER
*          = 0: successful exit
*          < 0: if INFO = -k, the k-th argument had an illegal value
*          > 0: if INFO = k, U(k,k) is exactly zero. The factorization
*               has been completed, but the factor U is exactly
*               singular, and division by zero will occur if it is used
*               to solve a system of equations.
*
*  =====================================================================
*
*     .. Parameters ..
      DOUBLE PRECISION   ONE, ZERO
      PARAMETER          ( ONE = 1.0D+0, ZERO = 0.0D+0 )
*     ..
*     .. Local Scalars ..
      DOUBLE PRECISION   SFMIN
      INTEGER            I, J, JP
*     ..
*     .. External Functions ..
      DOUBLE PRECISION   DLAMCH
      INTEGER            IDAMAX
      EXTERNAL           DLAMCH, IDAMAX
*     ..
*     .. External Subroutines ..
      EXTERNAL           DGER, DSCAL, DSWAP, XERBLA
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          MAX, MIN
*     ..
*     .. Executable Statements ..
*
*     Test the input parameters.
*
      INFO = 0
      IF( M.LT.0 ) THEN
         INFO = -1
      ELSE IF( N.LT.0 ) THEN
         INFO = -2
      ELSE IF( LDA.LT.MAX( 1, M ) ) THEN
         INFO = -4
      END IF
      IF( INFO.NE.0 ) THEN
         CALL XERBLA( 'DGETF2', -INFO )
         RETURN
      END IF
*
*     Quick return if possible
*
      IF( M.EQ.0 .OR. N.EQ.0 )
     $   RETURN
*
*     Compute machine safe minimum
*
      SFMIN = DLAMCH('S')
*
      DO 10 J = 1, MIN( M, N )
*
*        Find pivot and test for singularity.
*
         JP = J - 1 + IDAMAX( M-J+1, A( J, J ), 1 )
         IPIV( J ) = JP
         IF( A( JP, J ).NE.ZERO ) THEN
*
*           Apply the interchange to columns 1:N.
*
            IF( JP.NE.J )
     $         CALL DSWAP( N, A( J, 1 ), LDA, A( JP, 1 ), LDA )
*
*           Compute elements J+1:M of J-th column.
*
            IF( J.LT.M ) THEN
               IF( ABS(A( J, J )) .GE. SFMIN ) THEN
                  CALL DSCAL( M-J, ONE / A( J, J ), A( J+1, J ), 1 )
               ELSE
                 DO 20 I = 1, M-J
                    A( J+I, J ) = A( J+I, J ) / A( J, J )
   20            CONTINUE
               END IF
            END IF
*
         ELSE IF( INFO.EQ.0 ) THEN
*
            INFO = J
         END IF
*
         IF( J.LT.MIN( M, N ) ) THEN
*
*           Update trailing submatrix.
*
            CALL DGER( M-J, N-J, -ONE, A( J+1, J ), 1, A( J, J+1 ), LDA,
     $                 A( J+1, J+1 ), LDA )
         END IF
   10 CONTINUE
      RETURN
*
*     End of DGETF2
*
      END
      SUBROUTINE DGESV( N, NRHS, A, LDA, IPIV, B, LDB, INFO )
*
*  -- LAPACK driver routine (version 3.1) --
*     Univ. of Tennessee, Univ. of California Berkeley and NAG Ltd..
*     November 2006
*
*     .. Scalar Arguments ..
      INTEGER            INFO, LDA, LDB, N, NRHS
*     ..
*     .. Array Arguments ..
      INTEGER            IPIV( * )
      DOUBLE PRECISION   A( LDA, * ), B( LDB, * )
*     ..
*
*  Purpose
*  =======
*
*  DGESV computes the solution to a real system of linear equations
*     A * X = B,
*  where A is an N-by-N matrix and X and B are N-by-NRHS matrices.
*
*  The LU decomposition with partial pivoting and row interchanges is
*  used to factor A as
*     A = P * L * U,
*  where P is a permutation matrix, L is unit lower triangular, and U is
*  upper triangular.  The factored form of A is then used to solve the
*  system of equations A * X = B.
*
*  Arguments
*  =========
*
*  N       (input) INTEGER
*          The number of linear equations, i.e., the order of the
*          matrix A.  N >= 0.
*
*  NRHS    (input) INTEGER
*          The number of right hand sides, i.e., the number of columns
*          of the matrix B.  NRHS >= 0.
*
*  A       (input/output) DOUBLE PRECISION array, dimension (LDA,N)
*          On entry, the N-by-N coefficient matrix A.
*          On exit, the factors L and U from the factorization
*          A = P*L*U; the unit diagonal elements of L are not stored.
*
*  LDA     (input) INTEGER
*          The leading dimension of the array A.  LDA >= max(1,N).
*
*  IPIV    (output) INTEGER array, dimension (N)
*          The pivot indices that define the permutation matrix P;
*          row i of the matrix was interchanged with row IPIV(i).
*
*  B       (input/output) DOUBLE PRECISION array, dimension (LDB,NRHS)
*          On entry, the N-by-NRHS matrix of right hand side matrix B.
*          On exit, if INFO = 0, the N-by-NRHS solution matrix X.
*
*  LDB     (input) INTEGER
*          The leading dimension of the array B.  LDB >= max(1,N).
*
*  INFO    (output) INTEGER
*          = 0:  successful exit
*          < 0:  if INFO = -i, the i-th argument had an illegal value
*          > 0:  if INFO = i, U(i,i) is exactly zero.  The factorization
*                has been completed, but the factor U is exactly
*                singular, so the solution could not be computed.
*
*  =====================================================================
*
*     .. External Subroutines ..
      EXTERNAL           DGETRF, DGETRS, XERBLA
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          MAX
*     ..
*     .. Executable Statements ..
*
*     Test the input parameters.
*
      INFO = 0
      IF( N.LT.0 ) THEN
         INFO = -1
      ELSE IF( NRHS.LT.0 ) THEN
         INFO = -2
      ELSE IF( LDA.LT.MAX( 1, N ) ) THEN
         INFO = -4
      ELSE IF( LDB.LT.MAX( 1, N ) ) THEN
         INFO = -7
      END IF
      IF( INFO.NE.0 ) THEN
         CALL XERBLA( 'DGESV ', -INFO )
         RETURN
      END IF
*
*     Compute the LU factorization of A.
*
      CALL DGETRF( N, N, A, LDA, IPIV, INFO )
      IF( INFO.EQ.0 ) THEN
*
*        Solve the system A*X = B, overwriting B with X.
*
         CALL DGETRS( 'N', N, NRHS, A, LDA, IPIV, B, LDB,
     $                INFO )
      END IF
      RETURN
*
*     End of DGESV
*
      END








ccccccccccccccccc auxilliary functions ccccccccccccc


      DOUBLE PRECISION FUNCTION DLAMCH( CMACH )
*
*  -- LAPACK auxiliary routine (version 3.1) --
*     Univ. of Tennessee, Univ. of California Berkeley and NAG Ltd..
*     November 2006
*
*     .. Scalar Arguments ..
      CHARACTER          CMACH
*     ..
*
*  Purpose
*  =======
*
*  DLAMCH determines double precision machine parameters.
*
*  Arguments
*  =========
*
*  CMACH   (input) CHARACTER*1
*          Specifies the value to be returned by DLAMCH:
*          = 'E' or 'e',   DLAMCH := eps
*          = 'S' or 's ,   DLAMCH := sfmin
*          = 'B' or 'b',   DLAMCH := base
*          = 'P' or 'p',   DLAMCH := eps*base
*          = 'N' or 'n',   DLAMCH := t
*          = 'R' or 'r',   DLAMCH := rnd
*          = 'M' or 'm',   DLAMCH := emin
*          = 'U' or 'u',   DLAMCH := rmin
*          = 'L' or 'l',   DLAMCH := emax
*          = 'O' or 'o',   DLAMCH := rmax
*
*          where
*
*          eps   = relative machine precision
*          sfmin = safe minimum, such that 1/sfmin does not overflow
*          base  = base of the machine
*          prec  = eps*base
*          t     = number of (base) digits in the mantissa
*          rnd   = 1.0 when rounding occurs in addition, 0.0 otherwise
*          emin  = minimum exponent before (gradual) underflow
*          rmin  = underflow threshold - base**(emin-1)
*          emax  = largest exponent before overflow
*          rmax  = overflow threshold  - (base**emax)*(1-eps)
*
* =====================================================================
*
*     .. Parameters ..
      DOUBLE PRECISION   ONE, ZERO
      PARAMETER          ( ONE = 1.0D+0, ZERO = 0.0D+0 )
*     ..
*     .. Local Scalars ..
      LOGICAL            FIRST, LRND
      INTEGER            BETA, IMAX, IMIN, IT
      DOUBLE PRECISION   BASE, EMAX, EMIN, EPS, PREC, RMACH, RMAX, RMIN,
     $                   RND, SFMIN, SMALL, T
*     ..
*     .. External Functions ..
      LOGICAL            LSAME
      EXTERNAL           LSAME
*     ..
*     .. External Subroutines ..
      EXTERNAL           DLAMC2
*     ..
*     .. Save statement ..
      SAVE               FIRST, EPS, SFMIN, BASE, T, RND, EMIN, RMIN,
     $                   EMAX, RMAX, PREC
*     ..
*     .. Data statements ..
      DATA               FIRST / .TRUE. /
*     ..
*     .. Executable Statements ..
*
      IF( FIRST ) THEN
         CALL DLAMC2( BETA, IT, LRND, EPS, IMIN, RMIN, IMAX, RMAX )
         BASE = BETA
         T = IT
         IF( LRND ) THEN
            RND = ONE
            EPS = ( BASE**( 1-IT ) ) / 2
         ELSE
            RND = ZERO
            EPS = BASE**( 1-IT )
         END IF
         PREC = EPS*BASE
         EMIN = IMIN
         EMAX = IMAX
         SFMIN = RMIN
         SMALL = ONE / RMAX
         IF( SMALL.GE.SFMIN ) THEN
*
*           Use SMALL plus a bit, to avoid the possibility of rounding
*           causing overflow when computing  1/sfmin.
*
            SFMIN = SMALL*( ONE+EPS )
         END IF
      END IF
*
      IF( LSAME( CMACH, 'E' ) ) THEN
         RMACH = EPS
      ELSE IF( LSAME( CMACH, 'S' ) ) THEN
         RMACH = SFMIN
      ELSE IF( LSAME( CMACH, 'B' ) ) THEN
         RMACH = BASE
      ELSE IF( LSAME( CMACH, 'P' ) ) THEN
         RMACH = PREC
      ELSE IF( LSAME( CMACH, 'N' ) ) THEN
         RMACH = T
      ELSE IF( LSAME( CMACH, 'R' ) ) THEN
         RMACH = RND
      ELSE IF( LSAME( CMACH, 'M' ) ) THEN
         RMACH = EMIN
      ELSE IF( LSAME( CMACH, 'U' ) ) THEN
         RMACH = RMIN
      ELSE IF( LSAME( CMACH, 'L' ) ) THEN
         RMACH = EMAX
      ELSE IF( LSAME( CMACH, 'O' ) ) THEN
         RMACH = RMAX
      END IF
*
      DLAMCH = RMACH
      FIRST  = .FALSE.
      RETURN
*
*     End of DLAMCH
*
      END
*




      SUBROUTINE DLAMC2( BETA, T, RND, EPS, EMIN, RMIN, EMAX, RMAX )
*
*  -- LAPACK auxiliary routine (version 3.0) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     October 31, 1992
*
*     .. Scalar Arguments ..
      LOGICAL            RND
      INTEGER            BETA, EMAX, EMIN, T
      DOUBLE PRECISION   EPS, RMAX, RMIN
*     ..
*
*  Purpose
*  =======
*
*  DLAMC2 determines the machine parameters specified in its argument
*  list.
*
*  Arguments
*  =========
*
*  BETA    (output) INTEGER
*          The base of the machine.
*
*  T       (output) INTEGER
*          The number of ( BETA ) digits in the mantissa.
*
*  RND     (output) LOGICAL
*          Specifies whether proper rounding  ( RND = .TRUE. )  or
*          chopping  ( RND = .FALSE. )  occurs in addition. This may not
*          be a reliable guide to the way in which the machine performs
*          its arithmetic.
*
*  EPS     (output) DOUBLE PRECISION
*          The smallest positive number such that
*
*             fl( 1.0 - EPS ) .LT. 1.0,
*
*          where fl denotes the computed value.
*
*  EMIN    (output) INTEGER
*          The minimum exponent before (gradual) underflow occurs.
*
*  RMIN    (output) DOUBLE PRECISION
*          The smallest normalized number for the machine, given by
*          BASE**( EMIN - 1 ), where  BASE  is the floating point value
*          of BETA.
*
*  EMAX    (output) INTEGER
*          The maximum exponent before overflow occurs.
*
*  RMAX    (output) DOUBLE PRECISION
*          The largest positive number for the machine, given by
*          BASE**EMAX * ( 1 - EPS ), where  BASE  is the floating point
*          value of BETA.
*
*  Further Details
*  ===============
*
*  The computation of  EPS  is based on a routine PARANOIA by
*  W. Kahan of the University of California at Berkeley.
*
* =====================================================================
*
*     .. Local Scalars ..
      LOGICAL            FIRST, IEEE, IWARN, LIEEE1, LRND
      INTEGER            GNMIN, GPMIN, I, LBETA, LEMAX, LEMIN, LT,
     $                   NGNMIN, NGPMIN
      DOUBLE PRECISION   A, B, C, HALF, LEPS, LRMAX, LRMIN, ONE, RBASE,
     $                   SIXTH, SMALL, THIRD, TWO, ZERO
*     ..
*     .. External Functions ..
      DOUBLE PRECISION   DLAMC3
      EXTERNAL           DLAMC3
*     ..
*     .. External Subroutines ..
      EXTERNAL           DLAMC1, DLAMC4, DLAMC5
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          ABS, MAX, MIN
*     ..
*     .. Save statement ..
      SAVE               FIRST, IWARN, LBETA, LEMAX, LEMIN, LEPS, LRMAX,
     $                   LRMIN, LT
*     ..
*     .. Data statements ..
      DATA               FIRST / .TRUE. / , IWARN / .FALSE. /
*     ..
*     .. Executable Statements ..
*
      IF( FIRST ) THEN
         FIRST = .FALSE.
         ZERO = 0
         ONE = 1
         TWO = 2
*
*        LBETA, LT, LRND, LEPS, LEMIN and LRMIN  are the local values of
*        BETA, T, RND, EPS, EMIN and RMIN.
*
*        Throughout this routine  we use the function  DLAMC3  to ensure
*        that relevant values are stored  and not held in registers,  or
*        are not affected by optimizers.
*
*        DLAMC1 returns the parameters  LBETA, LT, LRND and LIEEE1.
*
         CALL DLAMC1( LBETA, LT, LRND, LIEEE1 )
*
*        Start to find EPS.
*
         B = LBETA
         A = B**( -LT )
         LEPS = A
*
*        Try some tricks to see whether or not this is the correct  EPS.
*
         B = TWO / 3
         HALF = ONE / 2
         SIXTH = DLAMC3( B, -HALF )
         THIRD = DLAMC3( SIXTH, SIXTH )
         B = DLAMC3( THIRD, -HALF )
         B = DLAMC3( B, SIXTH )
         B = ABS( B )
         IF( B.LT.LEPS )
     $      B = LEPS
*
         LEPS = 1
*
*+       WHILE( ( LEPS.GT.B ).AND.( B.GT.ZERO ) )LOOP
   10    CONTINUE
         IF( ( LEPS.GT.B ) .AND. ( B.GT.ZERO ) ) THEN
            LEPS = B
            C = DLAMC3( HALF*LEPS, ( TWO**5 )*( LEPS**2 ) )
            C = DLAMC3( HALF, -C )
            B = DLAMC3( HALF, C )
            C = DLAMC3( HALF, -B )
            B = DLAMC3( HALF, C )
            GO TO 10
         END IF
*+       END WHILE
*
         IF( A.LT.LEPS )
     $      LEPS = A
*
*        Computation of EPS complete.
*
*        Now find  EMIN.  Let A = + or - 1, and + or - (1 + BASE**(-3)).
*        Keep dividing  A by BETA until (gradual) underflow occurs. This
*        is detected when we cannot recover the previous A.
*
         RBASE = ONE / LBETA
         SMALL = ONE
         DO 20 I = 1, 3
            SMALL = DLAMC3( SMALL*RBASE, ZERO )
   20    CONTINUE
         A = DLAMC3( ONE, SMALL )
         CALL DLAMC4( NGPMIN, ONE, LBETA )
         CALL DLAMC4( NGNMIN, -ONE, LBETA )
         CALL DLAMC4( GPMIN, A, LBETA )
         CALL DLAMC4( GNMIN, -A, LBETA )
         IEEE = .FALSE.
*
         IF( ( NGPMIN.EQ.NGNMIN ) .AND. ( GPMIN.EQ.GNMIN ) ) THEN
            IF( NGPMIN.EQ.GPMIN ) THEN
               LEMIN = NGPMIN
*            ( Non twos-complement machines, no gradual underflow;
*              e.g.,  VAX )
            ELSE IF( ( GPMIN-NGPMIN ).EQ.3 ) THEN
               LEMIN = NGPMIN - 1 + LT
               IEEE = .TRUE.
*            ( Non twos-complement machines, with gradual underflow;
*              e.g., IEEE standard followers )
            ELSE
               LEMIN = MIN( NGPMIN, GPMIN )
*            ( A guess; no known machine )
               IWARN = .TRUE.
            END IF
*
         ELSE IF( ( NGPMIN.EQ.GPMIN ) .AND. ( NGNMIN.EQ.GNMIN ) ) THEN
            IF( ABS( NGPMIN-NGNMIN ).EQ.1 ) THEN
               LEMIN = MAX( NGPMIN, NGNMIN )
*            ( Twos-complement machines, no gradual underflow;
*              e.g., CYBER 205 )
            ELSE
               LEMIN = MIN( NGPMIN, NGNMIN )
*            ( A guess; no known machine )
               IWARN = .TRUE.
            END IF
*
         ELSE IF( ( ABS( NGPMIN-NGNMIN ).EQ.1 ) .AND.
     $            ( GPMIN.EQ.GNMIN ) ) THEN
            IF( ( GPMIN-MIN( NGPMIN, NGNMIN ) ).EQ.3 ) THEN
               LEMIN = MAX( NGPMIN, NGNMIN ) - 1 + LT
*            ( Twos-complement machines with gradual underflow;
*              no known machine )
            ELSE
               LEMIN = MIN( NGPMIN, NGNMIN )
*            ( A guess; no known machine )
               IWARN = .TRUE.
            END IF
*
         ELSE
            LEMIN = MIN( NGPMIN, NGNMIN, GPMIN, GNMIN )
*         ( A guess; no known machine )
            IWARN = .TRUE.
         END IF
***
* Comment out this if block if EMIN is ok
C         IF( IWARN ) THEN
C            FIRST = .TRUE.
C            WRITE( 6, FMT = 9999 )LEMIN
C         END IF
***
*
*        Assume IEEE arithmetic if we found denormalised  numbers above,
*        or if arithmetic seems to round in the  IEEE style,  determined
*        in routine DLAMC1. A true IEEE machine should have both  things
*        true; however, faulty machines may have one or the other.
*
         IEEE = IEEE .OR. LIEEE1
*
*        Compute  RMIN by successive division by  BETA. We could compute
*        RMIN as BASE**( EMIN - 1 ),  but some machines underflow during
*        this computation.
*
         LRMIN = 1
         DO 30 I = 1, 1 - LEMIN
            LRMIN = DLAMC3( LRMIN*RBASE, ZERO )
   30    CONTINUE
*
*        Finally, call DLAMC5 to compute EMAX and RMAX.
*
         CALL DLAMC5( LBETA, LT, LEMIN, IEEE, LEMAX, LRMAX )
      END IF
*
      BETA = LBETA
      T = LT
      RND = LRND
      EPS = LEPS
      EMIN = LEMIN
      RMIN = LRMIN
      EMAX = LEMAX
      RMAX = LRMAX
*
      RETURN
*
* 9999 FORMAT( / / ' WARNING. The value EMIN may be incorrect:-',
*     $      '  EMIN = ', I8, /
*     $      ' If, after inspection, the value EMIN looks',
*     $      ' acceptable please comment out ',
*     $      / ' the IF block as marked within the code of routine',
*     $      ' DLAMC2,', / ' otherwise supply EMIN explicitly.', / )
*
*     End of DLAMC2
*
      END

      SUBROUTINE DLAMC1( BETA, T, RND, IEEE1 )
*
*  -- LAPACK auxiliary routine (version 3.0) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     October 31, 1992
*
*     .. Scalar Arguments ..
      LOGICAL            IEEE1, RND
      INTEGER            BETA, T
*     ..
*
*  Purpose
*  =======
*
*  DLAMC1 determines the machine parameters given by BETA, T, RND, and
*  IEEE1.
*
*  Arguments
*  =========
*
*  BETA    (output) INTEGER
*          The base of the machine.
*
*  T       (output) INTEGER
*          The number of ( BETA ) digits in the mantissa.
*
*  RND     (output) LOGICAL
*          Specifies whether proper rounding  ( RND = .TRUE. )  or
*          chopping  ( RND = .FALSE. )  occurs in addition. This may not
*          be a reliable guide to the way in which the machine performs
*          its arithmetic.
*
*  IEEE1   (output) LOGICAL
*          Specifies whether rounding appears to be done in the IEEE
*          'round to nearest' style.
*
*  Further Details
*  ===============
*
*  The routine is based on the routine  ENVRON  by Malcolm and
*  incorporates suggestions by Gentleman and Marovich. See
*
*     Malcolm M. A. (1972) Algorithms to reveal properties of
*        floating-point arithmetic. Comms. of the ACM, 15, 949-951.
*
*     Gentleman W. M. and Marovich S. B. (1974) More on algorithms
*        that reveal properties of floating point arithmetic units.
*        Comms. of the ACM, 17, 276-277.
*
* =====================================================================
*
*     .. Local Scalars ..
      LOGICAL            FIRST, LIEEE1, LRND
      INTEGER            LBETA, LT
      DOUBLE PRECISION   A, B, C, F, ONE, QTR, SAVEC, T1, T2
*     ..
*     .. External Functions ..
      DOUBLE PRECISION   DLAMC3
      EXTERNAL           DLAMC3
*     ..
*     .. Save statement ..
      SAVE               FIRST, LIEEE1, LBETA, LRND, LT
*     ..
*     .. Data statements ..
      DATA               FIRST / .TRUE. /
*     ..
*     .. Executable Statements ..
*
      IF( FIRST ) THEN
         FIRST = .FALSE.
         ONE = 1
*
*        LBETA,  LIEEE1,  LT and  LRND  are the  local values  of  BETA,
*        IEEE1, T and RND.
*
*        Throughout this routine  we use the function  DLAMC3  to ensure
*        that relevant values are  stored and not held in registers,  or
*        are not affected by optimizers.
*
*        Compute  a = 2.0**m  with the  smallest positive integer m such
*        that
*
*           fl( a + 1.0 ) = a.
*
         A = 1
         C = 1
*
*+       WHILE( C.EQ.ONE )LOOP
   10    CONTINUE
         IF( C.EQ.ONE ) THEN
            A = 2*A
            C = DLAMC3( A, ONE )
            C = DLAMC3( C, -A )
            GO TO 10
         END IF
*+       END WHILE
*
*        Now compute  b = 2.0**m  with the smallest positive integer m
*        such that
*
*           fl( a + b ) .gt. a.
*
         B = 1
         C = DLAMC3( A, B )
*
*+       WHILE( C.EQ.A )LOOP
   20    CONTINUE
         IF( C.EQ.A ) THEN
            B = 2*B
            C = DLAMC3( A, B )
            GO TO 20
         END IF
*+       END WHILE
*
*        Now compute the base.  a and c  are neighbouring floating point
*        numbers  in the  interval  ( beta**t, beta**( t + 1 ) )  and so
*        their difference is beta. Adding 0.25 to c is to ensure that it
*        is truncated to beta and not ( beta - 1 ).
*
         QTR = ONE / 4
         SAVEC = C
         C = DLAMC3( C, -A )
         LBETA = INT(C + QTR,4)
*
*        Now determine whether rounding or chopping occurs,  by adding a
*        bit  less  than  beta/2  and a  bit  more  than  beta/2  to  a.
*
         B = LBETA
         F = DLAMC3( B / 2, -B / 100 )
         C = DLAMC3( F, A )
         IF( C.EQ.A ) THEN
            LRND = .TRUE.
         ELSE
            LRND = .FALSE.
         END IF
         F = DLAMC3( B / 2, B / 100 )
         C = DLAMC3( F, A )
         IF( ( LRND ) .AND. ( C.EQ.A ) )
     $      LRND = .FALSE.
*
*        Try and decide whether rounding is done in the  IEEE  'round to
*        nearest' style. B/2 is half a unit in the last place of the two
*        numbers A and SAVEC. Furthermore, A is even, i.e. has last  bit
*        zero, and SAVEC is odd. Thus adding B/2 to A should not  change
*        A, but adding B/2 to SAVEC should change SAVEC.
*
         T1 = DLAMC3( B / 2, A )
         T2 = DLAMC3( B / 2, SAVEC )
         LIEEE1 = ( T1.EQ.A ) .AND. ( T2.GT.SAVEC ) .AND. LRND
*
*        Now find  the  mantissa, t.  It should  be the  integer part of
*        log to the base beta of a,  however it is safer to determine  t
*        by powering.  So we find t as the smallest positive integer for
*        which
*
*           fl( beta**t + 1.0 ) = 1.0.
*
         LT = 0
         A = 1
         C = 1
*
*+       WHILE( C.EQ.ONE )LOOP
   30    CONTINUE
         IF( C.EQ.ONE ) THEN
            LT = LT + 1
            A = A*LBETA
            C = DLAMC3( A, ONE )
            C = DLAMC3( C, -A )
            GO TO 30
         END IF
*+       END WHILE
*
      END IF
*
      BETA = LBETA
      T = LT
      RND = LRND
      IEEE1 = LIEEE1
      RETURN
*
*     End of DLAMC1
*
      END

      DOUBLE PRECISION FUNCTION DLAMC3( A, B )
*
*  -- LAPACK auxiliary routine (version 3.0) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     October 31, 1992
*
*     .. Scalar Arguments ..
      DOUBLE PRECISION   A, B
*     ..
*
*  Purpose
*  =======
*
*  DLAMC3  is intended to force  A  and  B  to be stored prior to doing
*  the addition of  A  and  B ,  for use in situations where optimizers
*  might hold one of these in a register.
*
*  Arguments
*  =========
*
*  A, B    (input) DOUBLE PRECISION
*          The values A and B.
*
* =====================================================================
*
*     .. Executable Statements ..
*
      DLAMC3 = A + B
*
      RETURN
*
*     End of DLAMC3
*
      END
      SUBROUTINE DLAMC4( EMIN, START, BASE )
*
*  -- LAPACK auxiliary routine (version 3.0) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     October 31, 1992
*
*     .. Scalar Arguments ..
      INTEGER            BASE, EMIN
      DOUBLE PRECISION   START
*     ..
*
*  Purpose
*  =======
*
*  DLAMC4 is a service routine for DLAMC2.
*
*  Arguments
*  =========
*
*  EMIN    (output) EMIN
*          The minimum exponent before (gradual) underflow, computed by
*          setting A = START and dividing by BASE until the previous A
*          can not be recovered.
*
*  START   (input) DOUBLE PRECISION
*          The starting point for determining EMIN.
*
*  BASE    (input) INTEGER
*          The base of the machine.
*
* =====================================================================
*
*     .. Local Scalars ..
      INTEGER            I
      DOUBLE PRECISION   A, B1, B2, C1, C2, D1, D2, ONE, RBASE, ZERO
*     ..
*     .. External Functions ..
      DOUBLE PRECISION   DLAMC3
      EXTERNAL           DLAMC3
*     ..
*     .. Executable Statements ..
*
      A = START
      ONE = 1
      RBASE = ONE / BASE
      ZERO = 0
      EMIN = 1
      B1 = DLAMC3( A*RBASE, ZERO )
      C1 = A
      C2 = A
      D1 = A
      D2 = A
*+    WHILE( ( C1.EQ.A ).AND.( C2.EQ.A ).AND.
*    $       ( D1.EQ.A ).AND.( D2.EQ.A )      )LOOP
   10 CONTINUE
      IF( ( C1.EQ.A ) .AND. ( C2.EQ.A ) .AND. ( D1.EQ.A ) .AND.
     $    ( D2.EQ.A ) ) THEN
         EMIN = EMIN - 1
         A = B1
         B1 = DLAMC3( A / BASE, ZERO )
         C1 = DLAMC3( B1*BASE, ZERO )
         D1 = ZERO
         DO 20 I = 1, BASE
            D1 = D1 + B1
   20    CONTINUE
         B2 = DLAMC3( A*RBASE, ZERO )
         C2 = DLAMC3( B2 / RBASE, ZERO )
         D2 = ZERO
         DO 30 I = 1, BASE
            D2 = D2 + B2
   30    CONTINUE
         GO TO 10
      END IF
*+    END WHILE
*
      RETURN
*
*     End of DLAMC4
*
      END
      SUBROUTINE DLAMC5( BETA, P, EMIN, IEEE, EMAX, RMAX )
*
*  -- LAPACK auxiliary routine (version 3.0) --
*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
*     Courant Institute, Argonne National Lab, and Rice University
*     October 31, 1992
*
*     .. Scalar Arguments ..
      LOGICAL            IEEE
      INTEGER            BETA, EMAX, EMIN, P
      DOUBLE PRECISION   RMAX
*     ..
*
*  Purpose
*  =======
*
*  DLAMC5 attempts to compute RMAX, the largest machine floating-point
*  number, without overflow.  It assumes that EMAX + abs(EMIN) sum
*  approximately to a power of 2.  It will fail on machines where this
*  assumption does not hold, for example, the Cyber 205 (EMIN = -28625,
*  EMAX = 28718).  It will also fail if the value supplied for EMIN is
*  too large (i.e. too close to zero), probably with overflow.
*
*  Arguments
*  =========
*
*  BETA    (input) INTEGER
*          The base of floating-point arithmetic.
*
*  P       (input) INTEGER
*          The number of base BETA digits in the mantissa of a
*          floating-point value.
*
*  EMIN    (input) INTEGER
*          The minimum exponent before (gradual) underflow.
*
*  IEEE    (input) LOGICAL
*          A logical flag specifying whether or not the arithmetic
*          system is thought to comply with the IEEE standard.
*
*  EMAX    (output) INTEGER
*          The largest exponent before overflow
*
*  RMAX    (output) DOUBLE PRECISION
*          The largest machine floating-point number.
*
* =====================================================================
*
*     .. Parameters ..
      DOUBLE PRECISION   ZERO, ONE
      PARAMETER          ( ZERO = 0.0D0, ONE = 1.0D0 )
*     ..
*     .. Local Scalars ..
      INTEGER            EXBITS, EXPSUM, I, LEXP, NBITS, TRY, UEXP
      DOUBLE PRECISION   OLDY, RECBAS, Y, Z
*     ..
*     .. External Functions ..
      DOUBLE PRECISION   DLAMC3
      EXTERNAL           DLAMC3
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC          MOD
*     ..
*     .. Executable Statements ..
*
*     First compute LEXP and UEXP, two powers of 2 that bound
*     abs(EMIN). We then assume that EMAX + abs(EMIN) will sum
*     approximately to the bound that is closest to abs(EMIN).
*     (EMAX is the exponent of the required number RMAX).
*
      LEXP = 1
      EXBITS = 1
   10 CONTINUE
      TRY = LEXP*2
      IF( TRY.LE.( -EMIN ) ) THEN
         LEXP = TRY
         EXBITS = EXBITS + 1
         GO TO 10
      END IF
      IF( LEXP.EQ.-EMIN ) THEN
         UEXP = LEXP
      ELSE
         UEXP = TRY
         EXBITS = EXBITS + 1
      END IF
*
*     Now -LEXP is less than or equal to EMIN, and -UEXP is greater
*     than or equal to EMIN. EXBITS is the number of bits needed to
*     store the exponent.
*
      IF( ( UEXP+EMIN ).GT.( -LEXP-EMIN ) ) THEN
         EXPSUM = 2*LEXP
      ELSE
         EXPSUM = 2*UEXP
      END IF
*
*     EXPSUM is the exponent range, approximately equal to
*     EMAX - EMIN + 1 .
*
      EMAX = EXPSUM + EMIN - 1
      NBITS = 1 + EXBITS + P
*
*     NBITS is the total number of bits needed to store a
*     floating-point number.
*
      IF( ( MOD( NBITS, 2 ).EQ.1 ) .AND. ( BETA.EQ.2 ) ) THEN
*
*        Either there are an odd number of bits used to store a
*        floating-point number, which is unlikely, or some bits are
*        not used in the representation of numbers, which is possible,
*        (e.g. Cray machines) or the mantissa has an implicit bit,
*        (e.g. IEEE machines, Dec Vax machines), which is perhaps the
*        most likely. We have to assume the last alternative.
*        If this is true, then we need to reduce EMAX by one because
*        there must be some way of representing zero in an implicit-bit
*        system. On machines like Cray, we are reducing EMAX by one
*        unnecessarily.
*
         EMAX = EMAX - 1
      END IF
*
      IF( IEEE ) THEN
*
*        Assume we are on an IEEE machine which reserves one exponent
*        for infinity and NaN.
*
         EMAX = EMAX - 1
      END IF
*
*     Now create RMAX, the largest machine number, which should
*     be equal to (1.0 - BETA**(-P)) * BETA**EMAX .
*
*     First compute 1.0 - BETA**(-P), being careful that the
*     result is less than 1.0 .
*
      RECBAS = ONE / BETA
      Z = BETA - ONE
      Y = ZERO
      DO 20 I = 1, P
         Z = Z*RECBAS
         IF( Y.LT.ONE )
     $      OLDY = Y
         Y = DLAMC3( Y, Z )
   20 CONTINUE
      IF( Y.GE.ONE )
     $   Y = OLDY
*
*     Now multiply by BETA**EMAX to get RMAX.
*
      DO 30 I = 1, EMAX
         Y = DLAMC3( Y*BETA, ZERO )
   30 CONTINUE
*
      RMAX = Y
      RETURN
*
*     End of DLAMC5
*
      END

## ----echo = FALSE-------------------------------------------------------------
knitr::opts_chunk$set(
  fig.width  = 7,
  fig.height = 4,
  fig.align  = "center",
#  cache      = TRUE,
  autodep    = TRUE
)

## -----------------------------------------------------------------------------
library(Synth)
data(basque)
library(MSCMT)
Basque <- listFromLong(basque, unit.variable="regionno", time.variable="year", unit.names.variable="regionname")
school.sum <- with(Basque,colSums(school.illit + school.prim + school.med + school.high  + school.post.high))
Basque$school.higher <- Basque$school.high + Basque$school.post.high
for (item in c("school.illit", "school.prim", "school.med", "school.higher"))      
  Basque[[item]] <- 6 * 100 * t(t(Basque[[item]]) / school.sum)

## -----------------------------------------------------------------------------
treatment.identifier <- "Basque Country (Pais Vasco)"
controls.identifier  <- setdiff(colnames(Basque[[1]]),
                                c(treatment.identifier, "Spain (Espana)"))
times.dep  <- cbind("gdpcap"                = c(1960,1969))
times.pred <- cbind("school.illit"          = c(1964,1969),
                    "school.prim"           = c(1964,1969),
                    "school.med"            = c(1964,1969),
                    "school.higher"         = c(1964,1969),
                    "invest"                = c(1964,1969),
                    "gdpcap"                = c(1960,1969),
                    "sec.agriculture"       = c(1961,1969),
                    "sec.energy"            = c(1961,1969),
                    "sec.industry"          = c(1961,1969),
                    "sec.construction"      = c(1961,1969),
                    "sec.services.venta"    = c(1961,1969),
                    "sec.services.nonventa" = c(1961,1969),
                    "popdens"               = c(1969,1969))
agg.fns <- rep("mean", ncol(times.pred))                       

## -----------------------------------------------------------------------------
res <- mscmt(Basque, treatment.identifier, controls.identifier, times.dep, times.pred, agg.fns, seed=1, single.v=TRUE, verbose=FALSE)
res

## -----------------------------------------------------------------------------
times.pred <- times.pred[,-6]
agg.fns <- rep("mean", ncol(times.pred))                       
res2 <- mscmt(Basque, treatment.identifier, controls.identifier, times.dep, times.pred, agg.fns, seed=1, single.v=TRUE, verbose=FALSE)
res2

## -----------------------------------------------------------------------------
agg.fns <- rep("id", ncol(times.pred))   # Omitting agg.fns has the same effect (as "id" is the default)
res3 <- mscmt(Basque, treatment.identifier, controls.identifier, times.dep, times.pred, agg.fns, seed=1, single.v=TRUE, verbose=FALSE)
res3

## internal functions

M <- function(w,X,trafo.v) {
  M <- (diag(1,ncol(X))-rep(1,ncol(X))%*%t(w)) %*% t(X) %*% 
    (if (nrow(X)>1) diag(drop(X%*%w)) else drop(X%*%w))
  if (trafo.v$has.trafo) M %*% trafo.v$A else M 
}

all_v <- function(w,X,Z,trafo.v,lb=1e-8,verbose=FALSE,debug=FALSE) 
  cbind("min.loss.w" = loss_v(w,X,Z,trafo.v,lb,verbose=verbose,debug=debug),
#        if (debug) "max.min"    = single_v(w,X,Z,trafo.v,lb,check.exists=TRUE,
#                                           verbose=verbose),
        "max.order"  = single_v(w,X,Z,trafo.v,lb,verbose=verbose,debug=debug))

exists_v <- function(w,X,Z,trafo.v,lb=1e-8,all=FALSE,...) 
  check_v(single_v(w,X,Z,trafo.v,lb,check.exists=TRUE,...),
          w,X,Z,trafo.v,lb,...) || 
  if (all) check_v(loss_v(w,X,Z,trafo.v,lb,...),
                   w,X,Z,trafo.v,lb,...) else FALSE

## used in improveSynth
favourite_v <- function(v,w,X,Z,trafo.v,lb=1e-8,...) {
  tmp <- loss_v(w,X,Z,trafo.v,lb,...)
  if (check_v(tmp,w,X,Z,trafo.v,lb,...)) tmp else {
    tmp <- single_v(w,X,Z,trafo.v,lb,check.exists=TRUE,...)
    if (check_v(tmp,w,X,Z,trafo.v,lb,...)) tmp else v
  }
}

check_v <- function(v,w,X,Z,trafo.v,lb=1e-8,tol.lp=1e-12,tol.loss=1e-8,
                    verbose=FALSE,debug=FALSE) {
  if (any(is.na(v))) {
    if (verbose||debug) catn("checking v: v contains NAs!") 
    invisible(FALSE)
  } else if ((min(v/max(v))<lb)||(max(v)>1)) {
    if (verbose||debug) {
      if (min(v/max(v))<lb) catn("smallest predictor weight too small with ",
                                 min(v)," (lb=",lb,")")
      if (max(v)>1)  catn("biggest predictor weight too big with ",max(v))
    }  
    invisible(FALSE)
  } else {
    M0        <- M(w,X,trafo.v)
    K         <- ncol(M0)
    J         <- nrow(M0)
    const.mat <- M0
    const.rhs <- rep(0,nrow(M0))
    const.dir <- ifelse(w==0,">=","==")
    lhs       <- const.mat%*%v
    eqmax     <- max(abs(lhs-const.rhs)[const.dir=="=="])
    ineqmax   <- max(c(0,(lhs-const.rhs)[const.dir=="<="],
                         (const.rhs-lhs)[const.dir==">="]))
    wnew      <- wnnlsExt(v,X,NULL,trafo.v,return.w=TRUE)
    maxwdev   <- max(pmax(abs(wnew-w),ifelse(w>0,abs((wnew-w)/w),0)))
    
    if (!is.null(Z)) {
      loss.old <- lossDep(Z,w)
      loss.new <- lossDep(Z,wnew)
      loss.d <- if (loss.old==0) loss.new else 
                  min((loss.new-loss.old)/loss.old,loss.new-loss.old)
    } else loss.d <- 0

    if (debug) 
      catn("max. eq. residuum: ",eqmax,", max. ineq. residuum: ",ineqmax,
           ", w deviation: ",maxwdev,", loss difference: ",loss.d)

    invisible((max(eqmax,ineqmax)<=tol.lp)&&(loss.d<=tol.loss))
  }
}

calc_lpq <- function(v,w,X,trafo.v) {
  if (any(is.na(v))) NA else {
    M0        <- M(w,X,trafo.v)
    K         <- ncol(M0)
    J         <- nrow(M0)
    const.mat <- M0
    const.rhs <- rep(0,nrow(M0))
    const.dir <- ifelse(w==0,">=","==")
    lhs       <- const.mat%*%v
    eqmax     <- max(abs(lhs-const.rhs)[const.dir=="=="])
    ineqmax   <- max(c(0,(lhs-const.rhs)[const.dir=="<="],
                         (const.rhs-lhs)[const.dir==">="]))
    max(eqmax,ineqmax)
  }
}

# function delivers "min.loss.w"-v
loss_v <- function(w,X,Z,trafo.v,lb=1e-8,verbose=FALSE,debug=FALSE) {
  M0   <- M(w,X,trafo.v)
  K    <- ncol(M0)

  # linear program written in terms of x_k := v_k - lb
  # conditions that x_k <= 1-lb for all k
  const.mat <- diag(K)
  const.rhs <- rep(1-lb,K)
  const.dir <- rep("<=",K)

  # conditions concerning derivatives w.r.t. w 
  const.mat <- rbind(const.mat,M0)
  const.rhs <- c(const.rhs,-lb*rowSums(M0))
  const.dir <- c(const.dir,ifelse(w==0,">=","=="))

  obj <- if (trafo.v$has.trafo) (X%*%w)[,1]^2 %*% trafo.v$A else (X%*%w)[,1]^2
  lp_res <- lprsolve(direction="min",a0=lb*sum(obj),a=obj,d0=lb*K,d=rep(1,K),   # solve lp via various solvers
                 const.mat=const.mat,const.rhs=const.rhs,const.dir=const.dir,
                 verbose=verbose,debug=debug)
  if (!any(is.na(lp_res))) res <- pmax(lb,pmin(1,lp_res[1:K] + lb)) else 
                           res <- lp_res
  names(res) <- trafo.v$names.v
  res <- sanitize_v(res,X,Z,trafo.v,lb,verbose=verbose,debug=debug)
  if (debug&&!check_v(res,w,X,Z,trafo.v,lb,verbose=TRUE,debug=TRUE))
    catn("min.loss.w v probably unreliable: ",calc_lpq(res,w,X,trafo.v))
  res
}

# function delivers "max.order"-v
single_v <- function(w,X,Z,trafo.v,lb=0,check.exists=FALSE,tol_PUFAS=0.5*lb,
                     tol_min=1e-6,tol_lp=1e-13,scale=196L,                      #,scale=4L+8L+64L+256L
                     cv.alpha=0,v.special=integer(),backup.v=NULL,verbose=FALSE,
                     debug=FALSE) {

  M0   <- M(w,X,trafo.v)                                                        # matrix for calculating derivatives w.r.t. donor weights
  K    <- ncol(M0)                                                              # number of predictors
  J    <- nrow(M0)                                                              # number of donors

  # linear program written in terms of v_k
  # conditions concerning derivatives w.r.t. w 
  const.mat <- M0
  const.rhs <- rep(0,J)
  const.dir <- ifelse(w==0,">=","==")
  if (sum(w>0)>1) {
    J <- J - 1
    const.mat <- const.mat[-which.max(w>0),,drop=FALSE]
    const.rhs <- const.rhs[-which.max(w>0)]
    const.dir <- const.dir[-which.max(w>0)]
  }
  
  # conditions that v_k <= 1 for all k
  const.mat <- rbind(const.mat,diag(K)) # 
  const.rhs <- c(const.rhs,rep(1,K))
  const.dir <- c(const.dir,rep("<=",K))

  if ((cv.alpha>0) && (length(v.special)>0)) {                                  # cross-validation?
    obj.special <- rep(0,K)
    obj.special[v.special] <- 1
    
    tmp <- lpsolve(direction="max",objective.in=obj.special,
                   const.mat=rbind(const.mat,diag(K)),
                   const.rhs=c(const.rhs,rep(lb,K)),
                   const.dir=c(const.dir,rep(">=",K)),scale=scale,tol.lp=tol_lp,
                   verbose=verbose,debug=debug)

    dtmp <- lp_diag(obj.special,const.mat=rbind(const.mat,diag(K)),
                    const.rhs=c(const.rhs,rep(lb,K)),
                    const.dir=c(const.dir,rep(">=",K)),tmp)

    if (max(dtmp[2:3])>tol_lp) {
      tmpv <- backup.v[,ncol(backup.v)]
      tmpv <- tmpv/max(tmpv)
      tmp <- sum(tmpv[v.special])
#      if (debug) {
        catn("maximum of special.v unreliable (inf:",max(dtmp[2:3]),"), using ",tmp,
             " determined by backup v.")
        print(dtmp)
#      }
    } else tmp <- tmp$objval    
    
    if (tmp*cv.alpha>lb*length(v.special)) {
      const.mat <- rbind(const.mat,obj.special)                                 # add constraint for cross-validation
      const.rhs <- c(const.rhs,cv.alpha*tmp)
      const.dir <- c(const.dir,">=")
      CV <- 1
    } else CV <- 0
  } else CV <- 0
  
  # add new variable (minimum of v_1,...,v_K)
  const.mat <- rbind(cbind(const.mat,0),cbind(diag(K),-1))
  const.rhs <- c(const.rhs,rep(0,K))
  const.dir <- c(const.dir,rep(">=",K))
  
  fixed      <- c()                                                             # components of v which are fixed
  not.fixed  <- 1:K
  res        <- numeric(K)
  names(res) <- trafo.v$names.v

  solve_fixed <- function(fixed=c(),PUFAS=FALSE) {                              # helper function
    not.fixed <- setdiff(1:K,fixed)
    n.fixed   <- length(fixed)
    n.free    <- K - n.fixed
    if (n.fixed>0) {                                                            # remove currently fixed variables
      eq_change <- J+fixed                                                      # change equations that currently fixed variables do not exceed upper bound 
      const.rhs[eq_change] <- res[fixed]
      const.dir[eq_change] <- "=="
      eq_remove <- J+K+CV+fixed
      const.mat <- const.mat[-c(eq_remove),,drop=FALSE]                         
      const.dir <- const.dir[-c(eq_remove)]                                     
      const.rhs <- const.rhs[-c(eq_remove)]
    }
    obj.in <- c(rep(0,K),1)
    lp_res <- lpsolve(obj.in,const.mat=const.mat,const.rhs=const.rhs,           # try to solve lp with various solvers
                      const.dir=const.dir,direction="max",tol.lp=tol_lp,
                      not.fixed=not.fixed,check.min=TRUE,verbose=verbose,
                      debug=debug)
                  
    sol <- if (lp_res$status!=0) rep(NA,K) else                                 # check whether lp was solved properly
                                 lp_res$solution[-length(lp_res$solution)]
    sol[(sol>0)&(sol<lb)] <- lb
    if (!PUFAS) return(sol) else return(
       list(solution = sol,
            pufas    = if (n.free < 3) TRUE else 
                         isTRUE(pufas(objective.in=c(rep(0,K),1),               # change this to FALSE if sol contains NA ???
                                const.mat=const.mat,const.rhs=const.rhs,
                                const.dir=const.dir,solution=lp_res$solution,
                                scale=scale,tol=tol_PUFAS,verbose=verbose,
                                debug=debug)))) 
  }
  
  maxmin <- solve_fixed()
  if (min(maxmin)>0) maxmin <- pmax(lb,pmin(1,maxmin))
  maxmin <- sanitize_v(maxmin,X,Z,trafo.v,lb,verbose=verbose,debug=debug)
  if (check.exists) return(maxmin)

  for (i in 1:max(1,(K-2))) {                                                   # loop through orders 

    # checking for uniqueness via PUFAS
    lptmp     <- solve_fixed(fixed,PUFAS=TRUE)
    if (any(is.na(lptmp$solution))||any(lptmp$solution<=0)) {
      if (debug) catn("max.order v possibly unreliable!")
      break
    }
    res[1:K] <- lptmp$solution                                                  # update free (non-fixed) components
    res <- sanitize_v(res,X,Z,trafo.v,lb,verbose=verbose,debug=debug)
    if (lptmp$pufas) break                                                      # PUFAS condition fulfilled (or only 2 free variables left)? => DONE!

    ind <- intersect(which(res<=min(res[not.fixed])*(1+tol_min)),not.fixed)     # determine components which - up to some tolerance - produce the current order statistics
    if (length(ind) == 1) {                                                     # if there is only one candidate, we use this one.
      best_ind <- ind
    }	else {                                                                    # if there is more than one candidate:
      best_value <- 0                                                           # initialize best_value
      best_ind   <- NA
      for (index in ind) {                                                      # loop through all candidates to find the one which maximizes the next order statistics
        cur_fixed <- c(fixed,index)
        res_tmp   <- solve_fixed(cur_fixed,PUFAS=FALSE)
        if (!any(is.na(res_tmp)))
          if (min(res_tmp[-c(cur_fixed)])>best_value) {                         # if currently found next order statistics is larger than current best
            best_value <- min(res_tmp[-c(cur_fixed)])                           # update best value ... 
            best_ind   <- index                                                 # ... and index
        }
      }
    }
    if (is.na(best_ind)) break
    fixed     <- c(fixed,best_ind)                                              # update set of fixed indices
    not.fixed <- setdiff(1:K,fixed)
  }
  res <- res/max(res)
  if (!isTRUE(min(res)>0)) res[1:K] <- NA else res <- pmax(lb,pmin(1,res))      # if lb <= v[i], return res (vector of NAs else)
  res <- sanitize_v(res,X,Z,trafo.v,lb,verbose=verbose,debug=debug)
  maxmin.ok   <- check_v(maxmin,w,X,Z,trafo.v,lb,verbose=debug,debug=debug)
  maxorder.ok <- check_v(res,w,X,Z,trafo.v,lb,verbose=debug,debug=debug)
  if (maxorder.ok) {
    if (maxmin.ok) bestifnear(cbind(res,maxmin),X,Z,trafo.v) else res
  } else {
    if (maxmin.ok) maxmin else res
  }
  if (debug&&!maxorder.ok)
    catn("max.order v probably unreliable: ",calc_lpq(res,w,X,trafo.v))
  res
}

#' Plotting Results of MSCMT
#'
#' \code{plot.mscmt} plots results of \code{mscmt}.
#'
#' A unified basic plot function for gaps plots, comparison of treated and 
#' synthetic values, as well as plots for placebo studies.
#' Consider using \code{\link{ggplot.mscmt}} instead, which is the preferred
#' plot method and has more functionality than \code{\link{plot.mscmt}}.
#'
#' @param x An object of class \code{"mscmt"}, usually obtained as
#' the result of a call to function \code{\link{mscmt}}.
#' @param what A character scalar. Name of the variable to be plotted. If 
#' missing, the (first) dependent variable will be used.
#' @param type A character scalar denoting the type of the plot containing 
#' either \code{"gaps"}, \code{"comparison"}, \code{"placebo.gaps"}, 
#' or \code{"placebo.data"}. 
#' Partial matching allowed, defaults to \code{"placebo.gaps"}, if results
#' of a placebo study are present, and to \code{"gaps"}, else.
#' @param treatment.time An optional numerical scalar. If not missing, a 
#' vertical dotted line at the given point in time is included in the plot. 
#' \code{treatment.time} is measured in years, but may as well be a decimal 
#' number to reflect treatment times different from January 1st.
#' @param zero.line A logical scalar. If \code{TRUE} (default), a horizontal
#' dotted line (at zero level) is plotted for \code{"gaps"} and \code{"placebo"}
#' plots.
#' @param ylab Optional label for the y-axis, automatically generated if 
#' missing.
#' @param xlab Optional label for the x-axis, defaults to \code{"Date"}.
#' @param main Optional main title for the plot, automatically generated 
#' if missing.
#' @param sub Optional subtitle for the plot. If missing, the subtitle is 
#' generated automatically for \code{"comparison"} and \code{"gaps"} plots.
#' @param col Optional character vector with length corresponding to the number
#' of units. Contains the colours for the different units, automatically 
#' generated if missing. 
#' @param lty Optional numerical vector with length corresponding to the number
#' of units. Contains the line types for the different units, automatically 
#' generated if missing.
#' @param lwd Optional numerical vector with length corresponding to the number
#' of units. Contains the line widths for the different units, automatically 
#' generated if missing.
#' @param legend A logical scalar. If \code{TRUE} (default), a legend is 
#' included in the plot.
#' @param bw A logical scalar. If \code{FALSE} (default), the automatically 
#' generated colours and line types are optimized for a colour plot, 
#' if \code{TRUE}, the automatic colours and line types are set for a black and
#' white plot.
#' @param ... Further optional parameters for the underlying 
#' \code{\link[graphics]{plot}} function.
#' @return Nothing useful (function is called for its side effects).
#' @importFrom graphics plot abline
#' @method plot mscmt
#' @export
plot.mscmt <- function(x,what,type=c("gaps","comparison","placebo.gaps",
                       "placebo.data"),
                       treatment.time,zero.line=TRUE,ylab,xlab="Date",main,sub,
                       col,lty,lwd,legend=TRUE,bw=FALSE,...) {
  if (missing(type)&&(!is.null(x$placebo))) type <- "placebo.gaps"
  type <- match.arg(type)                      
  if (missing(what)) {
    if (is.null(x$placebo)) what <- x$dependent[1] else
                            what <- x[[1]]$dependent[1]
  }  
  if ((type=="placebo.gaps")||(type=="placebo.data")) {
    if (is.null(x$placebo)) stop("results of placebo study missing")
    if (!(what %in% names(x$placebo)))
      stop(paste("variable",what,"missing in results"))
    unames <- colnames(x$placebo[[what]]$gaps)
    if (missing(col)&&bw) col=rep("black",length(unames))
    if (missing(col)&&(!bw)) col=c("red",rep("black",length(unames)-1))
    if (missing(lty)&&bw) lty=c(1,rep(2,length(unames)-1))
    if (missing(lty)&&(!bw)) lty=rep(1,length(unames))
    if (missing(lwd)) lwd=c(2,rep(1,length(unames)-1))
    if (missing(ylab)) ylab=paste(if (type=="placebo.gaps") "Gaps" else "Data",
                                  "for",what)
    if (missing(main)) main=paste("Placebo plot for",what)
    if (missing(sub)) sub=""
    plot(if (type=="placebo.gaps") x$placebo[[what]]$gaps else 
                                   x$placebo[[what]]$data.treat,
         xlab=xlab,ylab=ylab,lty=lty,col=col,
         plot.type="single",main=main,sub=sub,lwd=lwd,...)
    if (!missing(treatment.time)) abline(v=treatment.time,lty=3)
    if (zero.line) abline(h=0,lty=3)         
    if (legend) legend("topleft",legend=c(unames[1],"other"),col=col[1:2],
                       lty=lty[1:2],lwd=lwd[1:2])
  }
  if (type=="comparison") {
    if (is.null(x$combined)) stop("input is not an individual mscmt result")
    if (!(what %in% names(x$combined)))
      stop(paste("variable",what,"missing in results"))
    if (missing(col)&&bw) col=rep("black",2)
    if (missing(col)&&(!bw)) col=c("black","red")
    if (missing(lty)&&bw) lty=c(1,2)
    if (missing(lty)&&(!bw)) lty=rep(1,2)
    if (missing(lwd)) lwd=c(2,2)
    if (missing(ylab)) ylab=what
    if (missing(main)) main=paste("Comparison of",what)
    if (missing(sub)) sub=paste("treated unit: ",x$treated.unit)
    plot(x$combined[[what]][,1:2],xlab=xlab,ylab=ylab,lty=lty,col=col,
         plot.type="single",main=main,sub=sub,lwd=lwd,...)
    if (!missing(treatment.time)) abline(v=treatment.time,lty=3)
    if (legend) legend("topleft",legend=c("actual data","synthesized data"),
                       col=col[1:2],lty=lty[1:2],lwd=lwd[1:2])
  }
  if (type=="gaps") {
    if (is.null(x$combined)) stop("input is not an individual mscmt result")
    if (!(what %in% names(x$combined)))
      stop(paste("variable",what,"missing in results"))
    if (missing(col)&&bw) col="black"
    if (missing(col)&&(!bw)) col="black"
    if (missing(lty)&&bw) lty=1
    if (missing(lty)&&(!bw)) lty=1
    if (missing(lwd)) lwd=2
    if (missing(ylab)) ylab=what
    if (missing(main)) main=paste("Gaps for",what)
    if (missing(sub)) sub=paste("treated unit: ",x$treated.unit)
    plot(x$combined[[what]][,3],xlab=xlab,ylab=ylab,lty=lty,col=col,
         plot.type="single",main=main,sub=sub,lwd=lwd,...)
    if (!missing(treatment.time)) abline(v=treatment.time,lty=3)
    if (zero.line) abline(h=0,lty=3)         
  }
}

#' Plotting Results of mscmt with ggplot2
#'
#' \code{ggplot.mscmt} plots results of \code{\link{mscmt}} based on 
#' \code{\link[ggplot2]{ggplot}}.
#'
#' A unified plot method for gaps plots, comparison of treated and synthetic
#' values, as well as plots for placebo studies, based on 
#' \code{\link[ggplot2]{ggplot}}. \code{\link{ggplot.mscmt}} is the preferred
#' plot method and has more functionality than \code{\link{plot.mscmt}}.
#'
#' @param data An object of class \code{"mscmt"}, usually obtained as
#' the result of a call to function \code{\link{mscmt}}.
#' @param mapping An object necessary to match the definition of the
#' \code{ggplot} generic (passed to \code{ggplot} as is). Defaults to 
#' \code{aes()}.
#' @param what A character vector. Name(s) of the variables to be plotted. If 
#' missing, the (first) dependent variable will be used.
#' @param type A character scalar denoting the type of the plot containing 
#' either \code{"gaps"}, \code{"comparison"}, \code{"placebo.gaps"}, 
#' \code{"placebo.data"}, or \code{"p.value"}. 
#' Partial matching allowed, defaults to \code{"placebo.gaps"}, if results
#' of a placebo study are present, and to \code{"gaps"}, else.
#' @param treatment.time An optional scalar (numeric, character, or 
#' \code{\link[base]{Date}}) giving the treatment time. 
#' If \code{treatment.time} is numeric, Jan 01 of that year will be used. If 
#' \code{treatment.time} is a character string, it will be converted to a 
#' \code{\link[base]{Date}} and must thus be in an unambiguous format.
#' A vertical dotted line at the given point in time is included in the plot. 
#' @param zero.line A logical scalar. If \code{TRUE} (default), a horizontal
#' dotted line (at zero level) is plotted for \code{"gaps"} and 
#' \code{"placebo.gaps"} plots.
#' @param ylab Optional label for the y-axis, automatically generated if 
#' missing.
#' @param xlab Optional label for the x-axis, defaults to \code{"Date"}.
#' @param main Optional main title for the plot, automatically generated 
#' if missing.
#' @param col Optional character vector with length 1 (for gaps plots) or
#' 2 (for all other plot types). For comparison plots, \code{col} contains the
#' colours for the actual and synthesized data, for placebo.plots (with 
#' \code{full.legend==FALSE}), \code{col} contains the colours for 
#' the treated unit and the control units. Automatically generated 
#' if missing. 
#' @param lty Optional numerical vector with length 1 (for gaps plots) or
#' 2 (for all other plot types). For comparison plots, \code{lty} contains the
#' linetypes for the actual and synthesized data, for placebo.plots (with 
#' \code{full.legend==FALSE}), \code{col} contains the linetypes for 
#' the treated unit and the control units. Automatically generated 
#' if missing. 
#' @param lwd Optional numerical vector with length 1 (for gaps plots) or
#' 2 (for all other plot types). For comparison plots, \code{lty} contains the
#' linewidths for the actual and synthesized data, for placebo.plots (with 
#' \code{full.legend==FALSE}), \code{col} contains the linewidths for 
#' the treated unit and the control units. Automatically generated 
#' if missing. 
#' @param legend A logical scalar. If \code{TRUE} (default), a legend is 
#' included in the plot.
#' @param bw A logical scalar. If \code{FALSE} (default), the automatically 
#' generated colours and line types are optimized for a colour plot, 
#' if \code{TRUE}, the automatic colours and line types are set for a black and
#' white plot.
#' @param date.format A character string giving the format for the tick labels
#' of the x axis as documented in \code{\link[base]{strptime}}. Defaults to
#' \code{"\%b \%y"} or \code{"\%Y"}, depending on the granularity of the data.
#' @param unit.name A character string with the title of the legend 
#' for comparison and placebo plots. Defaults to "Estimation" for comparison 
#' and "Unit" for placebo plots.
#' @param full.legend A logical scalar. If \code{TRUE} (default), a full legend
#' of all units (donors) is constructed. If \code{FALSE}, only the treated and
#' the control units are distinguished.
#' @param include.smooth A logical scalar. If \code{TRUE}, a geometric smoother
#' based on the control units is added to placebo plots. Default: \code{FALSE}.
#' @param include.mean A logical scalar. If \code{TRUE}, the arithmetic mean
#' of all control units is added to placebo plots. Default: \code{FALSE}.
#' @param include.synth A logical scalar. If \code{TRUE}, the synthesized data
#' for the treated unit are added to plots of type \code{"placebo.data"}.
#' Defaults to \code{FALSE}.
#' @param draw.estwindow A logical scalar. If \code{TRUE} (default), the time
#' range containing all optimization periods is shaded in the corresponding
#' plots.
#' @param what.set An optional character string for a convenient selection of 
#' multiple variables. Accepted values are \code{"dependents"}, 
#' \code{"predictors"}, and \code{"all"}, which collects all dependent, all
#' predictor, or all variables of both types, respectively. Overrides parameter
#' \code{what} (if the latter is present).
#' @param limits An optional vector of length 2 giving the range of the plot or
#' \code{NULL}.
#' If \code{limits} is numeric, Jan 01 of the corresponding years will be used.
#' If \code{limits} is of type character, both strings will be converted to 
#' Dates (via \code{\link[base]{as.Date}}) and must thus be in an unambiguous 
#' format.
#' @param alpha Either a numerical scalar, a numerical vector of length 
#' corresponding to the number of units, or the character string \code{"auto"}.
#' If \code{alpha} is a numerical scalar (default with value \code{1}), a fixed
#' value for the alpha channel (transparency) is included for all units in 
#' placebo plots. If \code{alpha} is numeric and has length corresponding to the
#' number of units, these values are assigned as alpha channel to the individual 
#' units. If \code{"auto"}, the alpha channel information is obtained from the
#' w weights of the control units.
#' @param alpha.min A numerical scalar (default: \code{0.1}). If \code{alpha} is
#' set to \code{"auto"}, the individual alpha channel information for control 
#' unit \code{i} is set to \code{alpha.min + (1-alpha.min) * w[i]}.
#' @param exclude.units An optional (default: \code{NULL}) character vector with 
#' names for control units which shall be excluded from placebo plots and
#' p-value calculations.
#' @param exclude.ratio A numeric scalar (default: \code{Inf}). Control units 
#' with a pre-treatment (r)mspe of more than \code{exclude.ratio} times
#' the pre-treatment (r)mspe of the treated unit are excluded from placebo 
#' plots and p-value calculations.
#' @param ratio.type A character string. Either \code{rmspe} (default) or 
#' \code{mspe}. Selects whether root mean squared errors or mean squared errors 
#' are considered for the exclusion of control units (see \code{exclude.ratio}).
#' @param alternative A character string giving the alternative of the test for 
#' plots of type \code{"p.value"}. Either \code{"two.sided"} (default), 
#' \code{"less"}, or \code{"greater"}.
#' @param draw.points A logical scalar. If \code{TRUE} (default), points are
#' added to the line plots to enhance visibility.
#' @param control.name A character string for the naming of the non-treated
#' units in placebo plots. Defaults to \code{"control units"}.
#' @param size A numerical scalar (default: \code{1}). If \code{draw.points} is
#' \code{TRUE} (default), \code{size} specifies the size of the points.
#' @param treated.name A character string giving the label for the treated
#' unit. Defaults to \code{"treated unit"}.
#' @param labels A character vector of length 2 giving the labels for the actual 
#' and synthesized data. Defaults to \code{c("actual data","synthsized data")}.
#' @param ... Necessary to match the definition of the
#' \code{"ggplot"} generic (passed to \code{ggplot} as is).
#' @param environment An object necessary to match the definition of the
#' \code{"ggplot"} generic (passed to \code{ggplot} as is). Defaults to 
#' \code{parent.frame()}.
#' @return An object of class \code{\link[ggplot2]{ggplot}}.
#' @importFrom ggplot2 ggplot aes_string geom_line labs scale_x_date geom_hline
#' @importFrom ggplot2 scale_colour_manual scale_linetype_manual 
#' @importFrom ggplot2 scale_size_manual scale_alpha_manual geom_rect 
#' @importFrom ggplot2 geom_vline aes facet_wrap geom_smooth guides
#' @importFrom ggplot2 guide_legend stat_summary geom_point scale_y_continuous
#' @importFrom stats frequency
#' @importFrom utils stack
#' @method ggplot mscmt
#' @export 
ggplot.mscmt <- function(data,mapping=aes(),what,
                         type=c("gaps","comparison","placebo.gaps",
                                "placebo.data","p.value"),
                         treatment.time,zero.line=TRUE,ylab,xlab="Date",main,
                         col,lty,lwd,legend=TRUE,bw=FALSE,date.format,
                         unit.name,full.legend=TRUE,include.smooth=FALSE,
                         include.mean=FALSE,include.synth=FALSE,
                         draw.estwindow=TRUE,what.set,limits=NULL,
                         alpha=1,alpha.min=0.1,exclude.units=NULL,
                         exclude.ratio=Inf,ratio.type=c("rmspe","mspe"),
                         alternative=c("two.sided", "less", "greater"),
                         draw.points=TRUE,control.name="control units",
                         size=1,treated.name="treated unit",
                         labels=c("actual data","synthsized data"),
                         ...,environment=parent.frame()) {
  ratio.type  <- match.arg(ratio.type)                         
  alternative <- match.arg(alternative) 
  x           <- data
  if (!missing(what.set)) 
    what.set <- match.arg(what.set,c("dependents","predictors","all")) 
  if (missing(type)&&(!is.null(x$placebo))) type <- "placebo.gaps"
  type        <- match.arg(type) 
  if (exclude.ratio<1) { 
    warning("exclude.ratio too small, using 1")
    exclude.ratio <- 1
  }  
  fixTimes <- function(a,b) {                                                   # helper function for results of compare()
    resA <- as.vector(a)
    names(resA) <- rep(rownames(a),times=ncol(a))
    resB <- as.vector(b)
    names(resB) <- rep(rownames(b),times=ncol(b))
    res <- rbind(Start=resA,End=resB)
    res[,apply(res,2,function(x) !all(is.na(x))),drop=FALSE]
  }
  if (is.null(x$placebo)) {                                                     # special treatment for placebo studies and comparions
    if (is.null(x$comparison)) tmp  <- x else {
      tmp <- x$comparison$results
      tmp$times.dep  <- fixTimes(tmp$dependent.start,tmp$dependent.end)
      tmp$times.pred <- fixTimes(tmp$predictor.start,tmp$predictor.end)
      tmp$dependent <- rownames(tmp$dependent)
    }
  } else tmp  <- x[[1]]
  if (missing(what)) what <- tmp$dependent[1]                                   # try to guess good defaults...
  if (!missing(what.set)) what <- switch(what.set,
    "dependents" = colnames(tmp$times.dep),
    "predictors" = colnames(tmp$times.pred),
    "all"        = c(colnames(tmp$times.dep),colnames(tmp$times.pred)))         # make this unique??? CHECK!!!
  if (!missing(treatment.time)) if (!inherits(treatment.time,"Date")) {         # convert treatment.time to Date
    if (is.numeric(treatment.time)) 
      treatment.time <- as.Date(paste0(treatment.time,"-01-01")) else
      treatment.time <- as.Date(treatment.time)
  }  
  if (!is.null(limits)) if (!inherits(limits,"Date")) {                         # convert limits to Date
    if (is.numeric(limits)) 
      limits <- as.Date(paste0(limits,"-01-01")) else
      limits <- as.Date(limits)
  }
  estwindow <- list()
  for (wh in what)
    estwindow[[wh]] <- cbind(
      apply(tmp$times.dep[,which(colnames(tmp$times.dep)==wh),drop=FALSE],2,
            AQM2Date),
      apply(tmp$times.pred[,which(colnames(tmp$times.pred)==wh),drop=FALSE],2,
            AQM2Date))
  estwindow <- lapply(estwindow,function(x) 
                        range(as.Date(as.vector(x),origin="1970-01-01")))
  res <- ggplot(mapping=mapping,...,environment=environment)                    # initialize the ggplot object
  if (!missing(treatment.time))                                                 # add line for treatment time
    res <- res + geom_vline(xintercept=as.numeric(treatment.time),
                            color="grey50")
  if ((type=="placebo.gaps")||(type=="placebo.data")||(type=="p.value")) {      # placebo-based plots
    if (is.null(x$placebo)) stop("results of placebo study missing")
    what.missing <- !(what %in% names(x$placebo))
    if (any(what.missing))
      stop(paste("variable(s)",paste0(what[what.missing],collapse=", "),
                 "missing in results"))
    unames <- colnames(x$placebo[[what[1]]]$gaps)
    mspe   <- sapply(x[-length(x)],function(x) x$loss.v)
    if (ratio.type=="rmspe") mspe <- sqrt(mspe)
    cunits <- mspe/mspe[1]<=exclude.ratio
    unames <- unames[cunits]
    cunits[exclude.units] <- FALSE
    unames <- setdiff(unames,exclude.units)
    ncontr <- length(unames)-1
    if (missing(col)&&bw) col=c("black","grey20")
    if (missing(col)&&(!bw)) col=c("red","grey20")
    if (missing(lty)&&bw) lty=c(1,2)
    if (missing(lty)&&(!bw)) lty=rep(1,2)
    if (missing(lwd)) lwd=c(2,1)
    if (missing(main)) main=if (length(what)==1) what[[1]] else ""
    if (missing(date.format)) 
      date.format <- if (frequency(x$placebo[[what[[1]]]]$gaps)>1) "%b %y" else 
                                                                   "%Y" 
  }  
  if (type=="p.value") {
    draw.estwindow <- FALSE
    dat  <- NULL
    for (wh in what) {
      gaps.post <- AQMtail(x$placebo[[wh]]$gaps[,cunits,drop=FALSE],
                           x[[1]]$times.dep[2,wh])
      pval  <- apply(gaps.post,1,switch(alternative,
                       two.sided = function(x) mean(abs(x[1])<=abs(x)),
                       less      = function(x) mean(x[1]>=x),
                       greater   = function(x) mean(x[1]<=x)))
      tmpdf <- ts2df(ts(pval,start=start(gaps.post),
                        frequency=frequency(gaps.post)))
      das   <- data.frame(date=tmpdf[,1],pvalue=as.numeric(tmpdf[,2]),
                          which.data=wh)
      dat   <- rbind(dat,das)
    }
    dat  <- dat[!is.na(dat$pvalue),]
    if (length(what)>1) res <- res + facet_wrap(~which.data,scales="free")
    if (missing(ylab)) ylab <- "p-value"
    res <- res + geom_point(data=dat,aes_string("date","pvalue"),size=size) +
                 scale_y_continuous(limits=c(0,1))
    if (zero.line) res <- res + geom_hline(yintercept=0,colour="grey50")
  }
  if ((type=="placebo.gaps")||(type=="placebo.data")) {
    if (missing(ylab)) ylab=paste(if (type=="placebo.gaps") "Gaps" else "Data",
                                  if (length(what)==1) 
                                    paste0(" for ",what[[1]],collapse=""))
    if (missing(unit.name)) unit.name <- "Unit"
    dat  <- NULL
    dat2 <- NULL
    for (wh in what) {
      tmpdf <- ts2df(if (type=="placebo.gaps") 
        x$placebo[[wh]]$gaps[,cunits,drop=FALSE] else 
        x$placebo[[wh]]$data.treat[,cunits,drop=FALSE])
      das <- cbind(date=tmpdf[,1],stack(tmpdf,select=-1))
      das <- cbind(das,treated=factor(
                                ifelse(das$ind==das$ind[1],"treated","control"),
                                levels=c("treated","control")),
                   start.estwindow=c(estwindow[[wh]][[1]],
                                   rep(NA,nrow(das)-!is.null(estwindow[[wh]]))),
                   end.estwindow=c(estwindow[[wh]][[2]],
                                   rep(NA,nrow(das)-!is.null(estwindow[[wh]]))),
                   which.data=wh)
      das$ind <- factor(das$ind,levels=c(as.character(das$ind[1]),
                                         setdiff(levels(das$ind),
                                                 as.character(das$ind[1]))))
      estwindow[[wh]] <- NULL                                                   
      dat  <- rbind(dat,das)
      dat2 <- rbind(dat2,cbind(ts2df(x$placebo[[wh]]$data.synth)[,1:2],
                               which.data=wh))
    }
    names(dat)[names(dat)=="ind"] <- unit.name
    names(dat2)[2] <- "value"
    dat  <- dat[!is.na(dat$values),]
    dat2 <- dat2[!is.na(dat2$value),]
    if (length(what)>1) res <- res + facet_wrap(~which.data,scales="free")
    if (alpha=="auto") {
      alpha <- rep(alpha.min,ncontr+1)
      names(alpha) <- levels(dat[[unit.name]])
      alpha[dat[[unit.name]][1]] <- 1
      alpha[names(x[[1]]$w)] <- alpha.min+(1-alpha.min)*x[[1]]$w
    } else if (length(alpha)==1) alpha <- rep(alpha,ncontr+1)
    if (full.legend) {
      lvals <- (1:6)[((seq_along(unames)-1)%%5)+1]                           
      res <- res + 
        geom_line(data=dat,aes_string("date","values",colour=unit.name,
                  linetype=unit.name,size=unit.name,alpha=unit.name),
                  na.rm=TRUE) +
        geom_line(data=dat[dat$treated=="treated",],aes_string("date","values",
                  colour=unit.name,linetype=unit.name,size=unit.name,
                  alpha=unit.name),na.rm=TRUE)
      if (draw.points) res <- res + 
        geom_point(data=dat,size=size,aes_string("date","values",
                   colour=unit.name,size=unit.name,alpha=unit.name),na.rm=TRUE) +
        geom_point(data=dat[dat$treated=="treated",],size=size,
                   aes_string("date","values",colour=unit.name,size=unit.name,
                   alpha=unit.name),na.rm=TRUE)
      res <- res + 
        scale_linetype_manual(values=lvals) +
        scale_size_manual(values=c(lwd[1],rep(lwd[2],length(unames)-1))) +
        scale_alpha_manual(values=alpha) +                          
        if (legend) guides(colour=guide_legend(override.aes=list(alpha=1))) else
                    guides(colour="none",linetype="none",size="none",
                           alpha="none")
    } else {
      res <- res + 
        geom_line(data=dat,aes_string("date","values",colour="treated",
                  linetype="treated",size="treated",group=unit.name,
                  alpha=unit.name),na.rm=TRUE) +
        geom_line(data=dat[dat$treated=="treated",],aes_string("date","values",
                  colour="treated",linetype="treated",size="treated",
                  alpha=unit.name,group=unit.name),na.rm=TRUE)
      if (draw.points) res <- res + 
        geom_point(data=dat,size=size,aes_string("date","values",
                   colour="treated",size="treated",group=unit.name,
                   alpha=unit.name),na.rm=TRUE) + 
        geom_point(data=dat[dat$treated=="treated",],size=size,
                   aes_string("date","values",colour="treated",size="treated",
                   alpha=unit.name,group=unit.name),na.rm=TRUE)
      res <- res + 
        scale_colour_manual("",values=col,
                            labels=c(treated.name,control.name)) +
        scale_linetype_manual("",values=lty,
                              labels=c(treated.name,control.name)) +
        scale_size_manual("",values=lwd,
                          labels=c(treated.name,control.name)) +
        scale_alpha_manual("",values=alpha) +                          
        if (legend) {
          if (missing(col))
            guides(colour=guide_legend(override.aes=list(alpha=1)),
                   alpha="none") else
            guides(colour=guide_legend(override.aes=list(alpha=1,colour=col)),
                   alpha="none") 
        } else guides(colour="none",linetype="none",size="none",
                      alpha="none")
    }  
    if (include.smooth) 
      res <- res + geom_smooth(data=dat[dat$treated=="control",],
                               mapping=aes_string("date","values"),
                               na.rm=TRUE) 
    if (include.mean) 
      res <- res + stat_summary(data=dat[dat$treated=="control",],
                                mapping=aes_string("date","values"),
                                fun.y="mean",
                                colour="black",geom="line",size=2,alpha=0.5,
                                na.rm=TRUE)
    if (include.synth&&(type=="placebo.data"))
      res <- res + geom_line(data=dat2,aes_string("date","value"),size=lwd[1],
                             linetype=if (full.legend) lvals[1] else lwd[1],
                             col="grey20",alpha=0.5)
    if (zero.line&&(type=="placebo.gaps")) 
      res <- res + geom_hline(yintercept=0,colour="grey50")
  }
  if ((type=="comparison")||(type=="gaps")) {                                   # non-placebo based plots
    if (is.null(x$combined)) {                                                  # input is a comparison
      if (missing(unit.name)) unit.name <- "Estimation"
      if (is.null(x$comparison)) stop("input is not an individual mscmt result")
      what.missing <- !(what %in% names(x$comparison$variables))
      if (any(what.missing))
        stop(paste("variable(s)",paste0(what[what.missing],collapse=", "),
                   "missing in results"))
      unames <- colnames(x$comparison$variables[[what[1]]]$gaps)
      nunits <- length(unames)
#      if (missing(col)&&bw) col=c("black","grey20")
#      if (missing(col)&&(!bw)) col=c("red","grey20")
      if (missing(col)) col=rep("black",nunits)
      if (missing(lty)) lty=(1:6)[((seq_len(nunits)-1)%%5)+1]                   # changed 2017-02-08
      if (missing(lwd)) lwd=rep(1,nunits)
      if (missing(main)) main=if (length(what)==1) what[[1]] else ""
      if (missing(date.format)) date.format <- 
        if (frequency(x$comparison$variables[[what[[1]]]]$gaps)>1) "%b %y" else 
                                                                   "%Y" 
      if (missing(ylab)) ylab=paste(if (type=="gaps") "Gaps" else "Data",
                                    if (length(what)==1) 
                                      paste0(" for ",what[[1]],collapse=""))
      dat  <- NULL
      dat2 <- NULL
      for (wh in what) {
        tmpdf <- ts2df(if (type=="gaps") 
          x$comparison$variables[[wh]]$gaps else 
          x$comparison$variables[[wh]]$data.synth)
        das <- cbind(date=tmpdf[,1],stack(tmpdf,select=-1))
        das <- cbind(das,
                     start.estwindow=c(estwindow[[wh]][[1]],
                                   rep(NA,nrow(das)-!is.null(estwindow[[wh]]))),
                     end.estwindow=c(estwindow[[wh]][[2]],
                                   rep(NA,nrow(das)-!is.null(estwindow[[wh]]))),
                     which.data=wh)
#        das$ind <- factor(das$ind,levels=c(as.character(das$ind[1]),
#                                           setdiff(levels(das$ind),
#                                                   as.character(das$ind[1]))))
        estwindow[[wh]] <- NULL                                                   
        dat  <- rbind(dat,das)
        dat2 <- rbind(dat2,
                      cbind(ts2df(x$comparison$variables[[wh]]$data.treat)[,1:2],
                            which.data=wh))
      }
      names(dat)[names(dat)=="ind"] <- unit.name
      names(dat2)[2] <- "value"
      dat  <- dat[!is.na(dat$values),]
      dat2 <- dat2[!is.na(dat2$value),]
      if (length(what)>1) res <- res + facet_wrap(~which.data,scales="free")
      if (length(alpha)==1) alpha <- rep(alpha,nunits)
      res <- res + 
        geom_line(data=dat,aes_string("date","values",colour=unit.name,
                  linetype=unit.name,size=unit.name,alpha=unit.name),
                  na.rm=TRUE) +
        geom_line(data=dat[dat$treated=="treated",],aes_string("date","values",
                  colour=unit.name,linetype=unit.name,size=unit.name,
                  alpha=unit.name),na.rm=TRUE)
      if (draw.points) res <- res + 
        geom_point(data=dat,aes_string("date","values",colour=unit.name,
                   size=unit.name,shape=unit.name,alpha=unit.name),size=size,
                   na.rm=TRUE) +
        geom_point(data=dat[dat$treated=="treated",],
                   aes_string("date","values",colour=unit.name,size=unit.name,
                   shape=unit.name,alpha=unit.name),size=size,na.rm=TRUE) 
      if (!missing(col)) res <- res + scale_colour_manual(values=col)
      res <- res + 
        scale_linetype_manual(values=lty) +
        scale_size_manual(values=lwd) +
        scale_alpha_manual(values=alpha) +                          
        if (legend) {
          if (missing(col))
            guides(colour=guide_legend(override.aes=list(alpha=1)),
                   alpha="none") else
            guides(colour=guide_legend(override.aes=list(alpha=1,colour=col)),
                   alpha="none") 
        } else guides(colour="none",linetype="none",size="none",
                      alpha="none")
#        if (legend) guides(colour=guide_legend(override.aes=list(alpha=1))) else
#                    guides(colour="none",linetype="none",size="none",
#                           shape="none",alpha="none")    
        if ((type=="comparison")&&
            (length(unique(x$comparison$results$treated.unit))==1))
          res <- res + geom_line(data=dat2,aes_string("date","value"),size=2,
                                 linetype=1,col="black",alpha=0.5)
      if (zero.line&&(type=="gaps")) 
        res <- res + geom_hline(yintercept=0,colour="grey50")
                            
    } else {                                                                    # input is not a comparison
      what.missing <- !(what %in% names(x$combined))
      if (any(what.missing))
        stop(paste("variable(s)",paste0(what[what.missing],collapse=", "),
                   "missing in results"))
      dat <- NULL
      for (wh in what) {
        das <- ts2df(x$combined[[wh]])
        if (type=="comparison") das <- cbind(date=das[,1],stack(das,select=2:3))
        das <- cbind(das,
                    start.estwindow=c(estwindow[[wh]][[1]],
                                   rep(NA,nrow(das)-!is.null(estwindow[[wh]]))),
                    end.estwindow=c(estwindow[[wh]][[2]],
                                   rep(NA,nrow(das)-!is.null(estwindow[[wh]]))),
                    which.data=wh)
        estwindow[[wh]] <- NULL                                                   
        if (type=="comparison") 
          das$ind <- factor(das$ind,levels=c(as.character(das$ind[1]),
                                             setdiff(levels(das$ind),
                                                     as.character(das$ind[1]))))
        dat <- rbind(dat,das)
      }  
      dat  <- if (type=="comparison") dat[!is.na(dat$values),] else 
                                      dat[!is.na(dat$gaps),]
      if (length(what)>1) res <- res + facet_wrap(~which.data,scales="free")
      if (missing(date.format)) 
        date.format <- if (frequency(x$combined[[what[[1]]]])>1) "%b %y" else 
                                                                 "%Y" 
      if (type=="comparison") {
        if (missing(col)&&bw) col=rep("black",2)
        if (missing(col)&&(!bw)) col=c("black","red")
        if (missing(lty)&&bw) lty=c(1,2)
        if (missing(lty)&&(!bw)) lty=rep(1,2)
        if (missing(lwd)) lwd=c(2,2)
        if (missing(ylab)) ylab=if (length(what)==1) what else ""
        if (missing(main)) main=if (length(what)==1) 
                                  paste("Comparison of",what) else ""
        res <- res + 
          geom_line(data=dat,aes_string("date","values",colour="ind",size="ind",
                    linetype="ind"),na.rm=TRUE)
        if (draw.points) res <- res + 
#          geom_point(data=dat,aes_string("date","values",
#                     colour="ind",size="ind"),na.rm=TRUE)
          geom_point(data=dat,aes_string("date","values",
                     colour="ind"),size=size,na.rm=TRUE)
        res <- res + 
          scale_colour_manual("",values=col,
                              labels=labels) +
          scale_linetype_manual("",values=lty,
                                labels=labels) +
          scale_size_manual("",values=lwd,
                            labels=labels) +
          if (legend) guides(colour=guide_legend(override.aes=list(alpha=1))) else
                      guides(colour="none",linetype="none",size="none")
      }
      if (type=="gaps") {
        if (missing(col)&&bw) col="black"
        if (missing(col)&&(!bw)) col="black"
        if (missing(lty)&&bw) lty=1
        if (missing(lty)&&(!bw)) lty=1
        if (missing(lwd)) lwd=2
        if (missing(ylab)) ylab="gap"
        if (missing(main)) main=if (length(what)==1) 
                                  paste("Gap for",what) else ""
        res <- res + geom_line(data=dat,
                               aes_string("date","gaps"),colour=col[1],
                               size=lwd[1],linetype=lty[1],na.rm=TRUE)
        if (draw.points) res <- res + geom_point(data=dat,
                                aes_string("date","gaps"),colour=col[1],
#                                size=lwd[1],na.rm=TRUE) 
                                size=size,na.rm=TRUE) 
        if (zero.line) res <- res + geom_hline(yintercept=0,colour="grey50")
      }  
    }
  }  
  res <- res + labs(title=main,x=xlab,y=ylab) +
               scale_x_date(date_labels=date.format,limits=limits)
  if (draw.estwindow)
    res <- res + geom_rect(data=dat,
                           aes_string(xmin="start.estwindow",
                                      xmax="end.estwindow"),
                           ymin=-Inf,ymax=+Inf,alpha=0.3,fill="grey70",
                           na.rm=TRUE)
  res  
}

## internal functions and objects
## todo (?): @export atomOpt
## @export atomOpt
atomOpt <- function(arglist,X,Z,trafo.v,single.v,inner.optim,inner.args,
                    outer.args1,starting.values,verbose=TRUE,debug=FALSE) {
  timing <- system.time({                    
    checkPkg <- function(pkg) if(!requireNamespace(pkg,quietly=TRUE))
      stop("package ",pkg," not installed") else TRUE
    getFun <- function(pkg,fun) 
      if(!requireNamespace(pkg,quietly=TRUE))
        stop("package ",pkg," not installed") else 
        get(fun,envir=asNamespace(pkg),mode="function")
    
    # obtain parameters from arglist
    outer.optim  <- arglist$outer.optim
    outer.opar   <- arglist$outer.opar
    seed         <- arglist$seed
    if (!is.null(seed)) if (is.na(seed)) seed <- NULL
    
    # initialize further parameters
    opt.separate <- outer.args1$opt.separate
    n.v          <- trafo.v$n.v
    true.n.v     <- if (opt.separate) n.v-1 else n.v
    lb           <- log(outer.args1$lb,base=10)
    idx          <- if (opt.separate) seq_len(n.v) else 1
    
    # starting values for outer optimization?                 
    if (!is.null(starting.values)) 
      if (is.na(starting.values)) starting.values <- rep(1/(true.n.v),true.n.v)
    if (!is.null(starting.values)) starting.values <- log(starting.values,base=10)
    
    # initialize parameters for outer optimizer (depending on optimizer)
    if (outer.optim=="DEoptC") {
      optfn   <- "DEoptC"
      outer.args <- list(min=lb, max=0, nG=500, nP=20*true.n.v, waitgen=100, 
                         minimpr=1e-14, F=0.5, CR=0.9, 
                         check.ambiguity = FALSE,
                         width=if (verbose&&interactive()) getOption("width")-1 
                                 else 0)
      outer.args[names(outer.opar)] <- outer.opar
    }
    if (outer.optim=="genoud") {
      optfn <- getFun("rgenoud","genoud")
      outer.args <- list(print.level=0,
                         max.generations=70,
                         solution.tolerance=1e-12,pop.size=20*true.n.v,
                         wait.generations=true.n.v,boundary.enforcement=2,
                         gradient.check=FALSE,MemoryMatrix=FALSE,
                         cluster=FALSE,transform=FALSE)
      outer.args <- c(outer.args,
                      list(Domains=cbind(rep(lb,true.n.v),rep(0,true.n.v)),
                           nvars=true.n.v,starting.values=starting.values))
      if (!is.null(seed)) 
        outer.args <- c(outer.args,list(unif.seed=seed,int.seed=seed))
      outer.args[names(outer.opar)] <- outer.opar
    }
    if (outer.optim=="DEoptim") {
      optfn <- getFun("DEoptim","DEoptim")
      control <- list(trace=FALSE,reltol=1e-14,steptol=500)
      control[names(outer.opar)] <- outer.opar
      outer.args <- list(lower=rep(lb,true.n.v),upper=rep(0,true.n.v),
                         control=control)  
    } 
    if (outer.optim=="JDEoptim") {
      optfn <- getFun("DEoptimR","JDEoptim")
      outer.args <- list(lower=rep(lb,true.n.v),upper=rep(0,true.n.v))  
      outer.args[names(outer.opar)] <- outer.opar
    } 
    if (outer.optim=="DEopt") {
      optfn <- getFun("NMOF","DEopt")
      algo <- list(min=rep(lb,true.n.v), max=rep(0,true.n.v), nG=100, 
	               nP=20*true.n.v,
                   minmaxConstr=TRUE, printBar=verbose, printDetail=verbose)
      algo[names(outer.opar)] <- outer.opar
      outer.args <- list(algo=algo)  
    }
    if (outer.optim=="PSopt") {
      optfn <- getFun("NMOF","PSopt")
      algo <- list(min=rep(lb,true.n.v), max=rep(0,true.n.v), nG=100, 
	               nP=20*true.n.v, 
                   minmaxConstr=TRUE, printBar=verbose, printDetail=verbose)
      algo[names(outer.opar)] <- outer.opar
      outer.args <- list(algo=algo)  
    }
    if (outer.optim=="GenSA"){
      optfn <- getFun("GenSA","GenSA")
      control <- list(max.call=1e7,max.time=25/true.n.v,trace.mat=FALSE)
      control[names(outer.opar)] <- outer.opar
      outer.args <- list(lower=rep(lb,true.n.v),upper=rep(0,true.n.v),
                         control=control)  
    }
    if (outer.optim=="ga") {
      optfn <- getFun("GA","ga")
      outer.args <- list(type="real-valued",min=rep(lb,true.n.v),
                         max=rep(0,true.n.v),maxiter=50,monitor=FALSE,
                         popSize=20*true.n.v)
      outer.args[names(outer.opar)] <- outer.opar
    }
    if (outer.optim=="soma") {
      optfn <- getFun("soma","soma")
      opts <- list(nMigrations=100)
      opts[names(outer.opar)] <- outer.opar
      outer.args <- list(bounds=list(min=rep(lb,true.n.v),max=rep(0,true.n.v)),
                         options=opts)
    }
    if (outer.optim=="cma_es"){
      optfn <- getFun("cmaes","cma_es")
      control <- list(maxit=2500)
      control[names(outer.opar)] <- outer.opar
      outer.args <- list(par=if(!is.null(starting.values)) starting.values else
                                log(rep(1/true.n.v,true.n.v),base=10),
                         lower=rep(lb,true.n.v),upper=rep(0,true.n.v),
                         control=control)  
    }
    if (outer.optim=="psoptim"){
      optfn <- getFun("pso","psoptim")
      control <- list(maxit=700)
      control[names(outer.opar)] <- outer.opar
      outer.args <- list(par=if(!is.null(starting.values)) starting.values else
                                log(rep(1/true.n.v,true.n.v),base=10),
                         lower=rep(lb,true.n.v),upper=rep(0,true.n.v),
                         control=control)  
    }
    if (outer.optim=="malschains"){
      optfn <- getFun("Rmalschains","malschains")
      control <- list(popsize=20*true.n.v)
      if (!is.null(outer.opar$maxEvals)) {
        maxEvals <- outer.opar$maxEvals
        outer.opar$maxEvals <- NULL
      } else maxEvals <- 25000
      control[names(outer.opar)] <- outer.opar
      outer.args <- list(maxEvals=maxEvals, 
                         lower=rep(lb,true.n.v),upper=rep(0,true.n.v),
                         control=do.call(getFun("Rmalschains",
                                                "malschains.control"),control),
                         verbosity=0)  
    }
    if ((outer.optim=="nloptr")||(outer.optim=="crs")){
      optfn <- getFun("nloptr","nloptr")
      opts <- list(maxeval=2.5e4,xtol_rel=1e-14,population=20*true.n.v,
                   ranseed=if(!is.null(seed)) as.integer(seed) else NULL,
                   algorithm="NLOPT_GN_CRS2_LM")
      opts[names(outer.opar)] <- outer.opar
      outer.args <- list(x0=if(!is.null(starting.values)) starting.values else
                              log(rep(1/true.n.v,true.n.v),base=10),
                         lb=rep(lb,true.n.v),ub=rep(0,true.n.v),
                         opts=opts)  
    }
    if (outer.optim=="isres"){
      optfn <- getFun("nloptr","nloptr")
      opts <- list(maxeval=2e4,xtol_rel=1e-14,population=20*true.n.v,
                   ranseed=if(!is.null(seed)) as.integer(seed) else NULL,
                   algorithm="NLOPT_GN_ISRES")
      opts[names(outer.opar)] <- outer.opar
      outer.args <- list(x0=if(!is.null(starting.values)) starting.values else
                              log(rep(1/true.n.v,true.n.v),base=10),
                         lb=rep(lb,true.n.v),ub=rep(0,true.n.v),
                         opts=opts)  
    }
    if (outer.optim %in% c("nlminbOpt","optimOpt")) {
      checkPkg("stats")
      optfn <- outer.optim
      outer.args <- list(lower=rep(lb,true.n.v),upper=rep(0,true.n.v),
	    nrandom=if (outer.optim=="nlminbOpt") 30 else 25)  
      outer.args[names(outer.opar)] <- outer.opar
    }
  
    # initialize best target function value
    best <- Inf
  
    # count calls to inner optimizer
    globals$NRUNS <- 0
    
    # check availability of packages for inner optimizer
    if ((inner.optim=="ipopOpt")||(inner.optim=="benchmarkOpt"))
      checkPkg("kernlab")
    if ((inner.optim=="LowRankQPOpt")||(inner.optim=="benchmarkOpt"))
      checkPkg("LowRankQP")
                 
    # pre-allocate workspace for inner optimizer wnnlsOpt
    if ((inner.optim=="wnnlsOpt")||(inner.optim=="benchmarkOpt")) {                                           
      ME  <- 1L
      MA  <- sum(trafo.v$len.v)
      N   <- ncol(X)
      MDW <- ME+MA
      globals$Ipar  <- as.integer(c(ME=ME,MA=MA,MDW=MDW,N=N)) 
      globals$IWORK <- integer(MDW+N)
      globals$WORK  <- double(MDW+5*N)
      globals$IWORK[1:2] <- c(length(globals$WORK),length(globals$IWORK))
      globals$RNORM <- double(1)
      globals$MODE  <- integer(1)
      globals$X     <- double(N)
    }
             
    #########################
    # outer optimization    #
    #########################
    if (verbose) catn("Starting optimization via ",outer.optim,
                      if (!is.null(seed)) paste0(", random seed ",seed),".") 
                                            
    if (outer.optim=="DEoptC") {
      if (!is.null(seed)) set.seed(seed) 
      trafo.inner <- function(v) 10^v
      trafo.fn    <- function(v) trafo.v$trafo(10^v)
      fn.min.par <- c(list(X=X,Z=Z,trafo=trafo.fn),inner.args) 
      rgV.optim <- 
        .Call(C_DE,X,Z,as.integer(trafo.v$len),as.integer(outer.args$nP),
              as.integer(outer.args$nG),as.double(outer.args$F),
              as.double(outer.args$CR),as.double(outer.args$min),
              as.double(outer.args$max),as.double(outer.args$minimpr),
              as.integer(outer.args$waitgen),
              as.logical(opt.separate),
              as.logical(outer.args$check.ambiguity),
              as.integer(outer.args$width))
      rmspe <- sqrt(rgV.optim$value)
      opt.v <- as.numeric(rgV.optim$par)
      conv  <- rgV.optim$counts
      globals$NRUNS <- rgV.optim$nruns
      w <- do.call(inner.optim,args=c(list(v=opt.v),fn.min.par,
                   return.w=TRUE))
      names(w) <- colnames(X)
      res  <- list(w=w, v=trafo.inner(opt.v), loss.v=rmspe^2, rmspe=rmspe, 
                   conv=conv)
    } else {
      for (i in idx) {
        if (opt.separate) {
          trafo.inner <- function(v) 
            c(if (i>1) 10^v[1:(i-1)] else NULL,
              1,if (i<n.v) 10^v[i:(n.v-1)] else NULL)
          trafo.fn    <- function(v) 
            trafo.v$trafo(c(if (i>1) 10^v[1:(i-1)] else NULL,
                            1,if (i<n.v) 10^v[i:(n.v-1)] else NULL))
        } else {
          trafo.inner <- function(v) 10^v
          trafo.fn    <- function(v) trafo.v$trafo(10^v)
        }                    
        
        if (!is.null(seed)) set.seed(seed) 
        fn.min.par <- c(list(X=X,Z=Z,trafo=trafo.fn),inner.args) 
        obj.fun <- if (outer.optim=="ga") 
                     function(v) 
                       (-1)*do.call(inner.optim,args=c(list(v=v),fn.min.par))
                   else 
                     function(v) 
                       do.call(inner.optim,args=c(list(v=v),fn.min.par))
        obj.fun.name <- if (outer.optim %in% c("DEopt","PSopt")) "OF" 
                        else if (outer.optim %in% c("nloptr","isres","crs")) 
                                                        "eval_f" 
                        else if (outer.optim=="ga")     "fitness"
                        else if (outer.optim=="soma")   "costFunction"
                        else "fn"                      
        obj.fun.list <- list(obj.fun)
        names(obj.fun.list) <- obj.fun.name                
  
        # this is the actual call of the outer optimizer
        rgV.optim  <- do.call(optfn,args=c(obj.fun.list,outer.args))
  
        if (outer.optim=="genoud") {
          rmspe <- sqrt(rgV.optim$value)
          opt.v <- as.numeric(rgV.optim$par)
          conv  <- rgV.optim$generations
        }
        if (outer.optim=="DEoptim") {
          rmspe <- sqrt(rgV.optim$optim$bestval)
          opt.v <- as.numeric(rgV.optim$optim$bestmem)
          conv  <- rgV.optim$optim$iter
        }
        if (outer.optim=="JDEoptim") {
          rmspe <- sqrt(rgV.optim$value)
          opt.v <- as.numeric(rgV.optim$par)
          conv  <- rgV.optim$iter
        }
        if (outer.optim %in% c("DEopt","PSopt")) {
          rmspe <- sqrt(rgV.optim$OFvalue)
          opt.v <- as.numeric(rgV.optim$xbest)
          conv  <- algo$nG
        }
        if (outer.optim=="GenSA") {
          rmspe <- sqrt(rgV.optim$value)
          opt.v <- as.numeric(rgV.optim$par)
          conv  <- rgV.optim$counts
        }
        if (outer.optim %in% c("nloptr","isres","crs")) {
          rmspe <- sqrt(rgV.optim$objective)
          opt.v <- as.numeric(rgV.optim$solution)
          conv  <- rgV.optim$iterations
        }
        if (outer.optim=="ga") {
          rmspe <- sqrt((-1)*rgV.optim@fitnessValue)
          opt.v <- as.numeric(rgV.optim@solution)
          conv  <- rgV.optim@iter
        }
        if (outer.optim=="soma") {
          rmspe <- sqrt(rgV.optim$cost[rgV.optim$leader])
          opt.v <- as.numeric(rgV.optim$population[,rgV.optim$leader])
          conv  <- rgV.optim$migrations
        }
        if (outer.optim=="cma_es") {
          rmspe <- sqrt(rgV.optim$value)
          opt.v <- as.numeric(rgV.optim$par)
          conv  <- rgV.optim$counts[1]
        }
        if (outer.optim=="malschains") {
          rmspe <- sqrt(rgV.optim$fitness)
          opt.v <- as.numeric(rgV.optim$sol)
          conv  <- rgV.optim$numEvalEA + rgV.optim$numEvalLS
        }
        if (outer.optim=="psoptim") {
          rmspe <- sqrt(rgV.optim$value)
          opt.v <- as.numeric(rgV.optim$par)
          conv  <- rgV.optim$counts[2]
        }
        if (outer.optim=="nlminbOpt") {
          rmspe <- sqrt(rgV.optim$objective)
          opt.v <- as.numeric(rgV.optim$par)
          conv  <- rgV.optim$iterations
        }
        if (outer.optim=="optimOpt") {
          rmspe <- sqrt(rgV.optim$value)
          opt.v <- as.numeric(rgV.optim$par)
          conv  <- rgV.optim$counts[1]
        }
        
        # call the inner optimizer again to obtain w
        w <- do.call(inner.optim,args=c(list(v=opt.v),fn.min.par,
                     return.w=TRUE))
        names(w) <- colnames(X)
  
        if (verbose&&opt.separate)
          catn("Fixing v[",i,"] (as maximum) yields rmspe ",rmspe,
               " and mspe (loss v) ",rmspe^2," after ",conv,
               " generations/iterations.")
        if (rmspe < best) {
          best <- rmspe
          res  <- list(w=w, v=trafo.inner(opt.v), loss.v=rmspe^2, rmspe=rmspe, 
                       conv=conv)
        }
      }
    }
    res <- sanitize_res(res,X,Z,trafo.v,outer.args1$lb,verbose=verbose,
                        debug=debug)
    w <- res$w
    v <- res$v
    if (!check_v(v,w,X,Z,trafo.v,lb=outer.args1$lb,verbose=verbose,
                 debug=debug)) {
      if(verbose) catn("optimizer's v violates optimality constraints!") else 
                  warning("optimizer's v violates optimality constraints")
    }
    if (!is.na(single.v)) {
      v_minlossw  <- loss_v(w,X,Z,trafo.v,outer.args1$lb,verbose=verbose,
                            debug=debug)
      minlossw_ok <- check_v(v_minlossw,w,X,Z,trafo.v,outer.args1$lb,
                             verbose=verbose,debug=debug)
      v_maxorder  <- single_v(w,X,Z,trafo.v,outer.args1$lb,verbose=verbose,
                              debug=debug)
      maxorder_ok <- check_v(v_maxorder,w,X,Z,trafo.v,outer.args1$lb,
                             verbose=verbose,debug=debug)
      if (minlossw_ok) {
        v_minlossw <- bestifnear(cbind(v_minlossw,v),X,Z,trafo.v)
        if (maxorder_ok) 
          v_maxorder <- bestifnear(cbind(v_maxorder,v_minlossw),X,Z,trafo.v) else
          v_maxorder <- bestifnear(cbind(v,v_minlossw),X,Z,trafo.v)
      } else {
        v_minlossw <- v
        if (maxorder_ok) 
          v_maxorder <- bestifnear(cbind(v_maxorder,v),X,Z,trafo.v) else
          v_maxorder <- v
      }
      res$v <- if (single.v) cbind("max.order"=v_maxorder) else
                             cbind(if (debug) "optimizer"=v,
                                   "min.loss.w"=v_minlossw,
                                   "max.order"=v_maxorder)
    } else res$v <- cbind("optimizer"=res$v)
    res$single.v <- single.v
    rownames(res$v) <- trafo.v$names.v
  
    if (verbose) catn("Optimization finished (",globals$NRUNS," calls to ",
                      "inner optimizer), rmspe: ",res$rmspe,", mspe: ",
                      res$loss.v,".")
  })[["user.self"]]  
  c(res,list(ncalls.inner=globals$NRUNS,user.self=timing))
}

## internal functions and objects
globals         <- new.env()

timingBench <- function(vs,X,Z,max.iter=1000L,debug=FALSE,sigf=5,margin=5e-04,
                        bound=10,...) {
  ME  <- 1L
  MA  <- ncol(vs)
  n <- N <- ncol(X)
  MDW <- ME+MA
  globals$Ipar  <- as.integer(c(ME=ME,MA=MA,MDW=MDW,N=N)) 
  globals$IWORK <- integer(MDW+N)
  globals$WORK  <- double(MDW+5*N)
  globals$IWORK[1:2] <- c(length(globals$WORK),length(globals$IWORK))
  globals$RNORM <- double(1)
  globals$MODE  <- integer(1)
  globals$X     <- double(N)  
  
  wnnls.time <- system.time(
    for (i in 1:nrow(vs)) sol <- .Fortran(C_wnnls,W=.Call(C_prepareW4,X,vs[i,]),
                 MDW=globals$Ipar[3],ME=globals$Ipar[1],MA=globals$Ipar[2],
                 N=globals$Ipar[4],L=0L,PRGOPT=1.0,X=globals$X,
                 RNORM=globals$RNORM,MODE=globals$MODE,IWORK=globals$IWORK,
                 WORK=globals$WORK))
  ipop.time <- system.time(
    for (i in 1:nrow(vs)) w.ipop <- try(as.numeric(kernlab::ipop(c=rep(0,n),
                           H=fastMpdVM(X,vs[i,]),A=matrix(rep(1,n),nrow=1),b=1,
                           l=rep(0,n),u=rep(1,n),r=0,sigf=sigf,margin=margin,
                           maxiter=max.iter,bound=bound,verb=debug,...)$primal),
                           silent=TRUE))
  LRQP.time <- system.time(
    for (i in 1:nrow(vs)) w <- LowRankQP::LowRankQP(Vmat=t(sqrt(vs[i,])*X), 
                           dvec=rep(0,n), Amat=matrix(rep(1,n),nrow=1), bvec=1, 
                           uvec=rep(1,n),niter=max.iter,verbose=debug,
                           ...)$alpha)
  list(wnnls=wnnls.time,ipop=ipop.time,LRQP=LRQP.time)
}

## inner optimization methods
## currently, only wnnlsOpt is 'officially' supported
wnnlsOpt <- function(v,X,Z,trafo,debug=FALSE,return.w=FALSE,
                     check.ambiguity=FALSE,...) {
  if (any(is.na(v))) return(Inf) else tv  <- trafo(v)                           # needed for nlminb to work. Better drop nlminb-support for speed reasons?
  sol <-.Fortran(C_wnnls,W=.Call(C_prepareW4,X,tv),
                 MDW=globals$Ipar[3],ME=globals$Ipar[1],MA=globals$Ipar[2],
                 N=globals$Ipar[4],L=0L,PRGOPT=1.0,X=globals$X,
                 RNORM=globals$RNORM,MODE=globals$MODE,IWORK=globals$IWORK,
                 WORK=globals$WORK)
  globals$NRUNS <- globals$NRUNS + 1                 
  if ((any(is.infinite(sol$X))) || (sol$MODE>0)) warning("error in wnnls")                
  w   <- if (check.ambiguity) improveZw(Z,X,sol$X) else sol$X
  if (return.w) w else lossDep(Z,w)
}

wnnlsExt <- function(v,X,Z,trafo.v,debug=FALSE,return.w=FALSE,
                     check.ambiguity=FALSE,...) {
  if (any(is.na(v))) return(Inf) else tv  <- trafo.v$trafo(v)                   # needed for nlminb to work. Better drop nlminb-support for speed reasons?
  ME  <- as.integer(1L)
  MA  <- as.integer(sum(trafo.v$len.v))
  N   <- as.integer(ncol(X))
  MDW <- as.integer(ME+MA)
  IWORK <- integer(MDW+N)
  IWORK[1:2] <- as.integer(c(MDW+5*N,MDW+N))
  sol <-.Fortran(C_wnnls,W=.Call(C_prepareW4,X,tv),
                 MDW=MDW,ME=ME,MA=MA,N=N,L=0L,PRGOPT=1.0,X=double(N),
                 RNORM=double(1),MODE=integer(1),IWORK=IWORK,
                 WORK=double(MDW+5*N))
  if ((any(is.infinite(sol$X))) || (sol$MODE>0)) warning("error in wnnls")                
  w   <- if (check.ambiguity) improveZw(Z,X,sol$X) else sol$X
  if (return.w) w else lossDep(Z,w)
}

LowRankQPOpt <- function(v,X,Z,trafo,debug=FALSE,return.w=FALSE,max.iter=1000L,
                         ...) {
  if (any(is.na(v))) return(Inf) else tv  <- trafo(v)                           # needed for nlminb to work. Better drop nlminb-support for speed reasons?
  n <- ncol(X)
  w <- LowRankQP::LowRankQP(Vmat=t(sqrt(tv)*X), dvec=rep(0,n), 
                            Amat=matrix(rep(1,n),nrow=1), bvec=1, uvec=rep(1,n),
                            niter=max.iter,verbose=debug,...)$alpha
  if (return.w) w else lossDep(Z,w)
}

ipopOpt <- function(v,X,Z,trafo,debug=FALSE,return.w=FALSE,max.iter=1000L,
                    sigf=5,margin=5e-04,bound=10,...) {
  if (any(is.na(v))) return(Inf) else tv  <- trafo(v)                           # needed for nlminb to work. Better drop nlminb-support for speed reasons?
  n <- ncol(X)
  w <- try(as.numeric(kernlab::ipop(c=rep(0,n),H=fastMpdVM(X,tv),
                                    A=matrix(rep(1,n),nrow=1),b=1,l=rep(0,n),
                                    u=rep(1,n),r=0,sigf=sigf,margin=margin,
                                    maxiter=max.iter,bound=bound,verb=debug,
                                    ...)$primal),silent=TRUE)
  if (inherits(w,"try-error")) { 
    if (return.w) return(NA) else return(Inf)
  } else {
    if (return.w) w else lossDep(Z,w)
  }  
}

benchmarkOpt <- function(v,X,Z,trafo,debug=FALSE,return.w=FALSE,max.iter=1000L,
                         sigf=5,margin=5e-04,bound=10,...) {
  if (any(is.na(v))) return(Inf) else tv  <- trafo(v)                           # needed for nlminb to work. Better drop nlminb-support for speed reasons?
  globals$NRUNS <- globals$NRUNS + 1                 
  globals$vs[globals$NRUNS,] <- tv
  
  # wnnls
  sol <-.Fortran(C_wnnls,W=.Call(C_prepareW4,X,tv),
                 MDW=globals$Ipar[3],ME=globals$Ipar[1],MA=globals$Ipar[2],
                 N=globals$Ipar[4],L=0L,PRGOPT=1.0,X=globals$X,
                 RNORM=globals$RNORM,MODE=globals$MODE,IWORK=globals$IWORK,
                 WORK=globals$WORK)
  if ((any(is.infinite(sol$X))) || (sol$MODE>0)) warning("error in wnnls")                
  w.wnnls <- sol$X

  # LowRankQP
  n <- ncol(X)
  w.LRQP <- LowRankQP::LowRankQP(Vmat=t(sqrt(tv)*X), dvec=rep(0,n), 
                            Amat=matrix(rep(1,n),nrow=1), bvec=1, uvec=rep(1,n),
                            niter=max.iter,verbose=debug,...)$alpha

  # ipop                      
  w.ipop  <- try(as.numeric(kernlab::ipop(c=rep(0,n),H=fastMpdVM(X,tv),
                                    A=matrix(rep(1,n),nrow=1),b=1,l=rep(0,n),
                                    u=rep(1,n),r=0,sigf=sigf,margin=margin,
                                    maxiter=max.iter,bound=bound,verb=debug,
                                    ...)$primal),silent=TRUE)
  if (inherits(w.ipop,"try-error")) { 
    trouble <- TRUE; w.ipop <- w.wnnls 
  } else trouble <- FALSE
  
  w.wnnls <- abs(w.wnnls)/sum(abs(w.wnnls))
  w.LRQP  <- abs(w.LRQP)/sum(abs(w.LRQP))
  w.ipop  <- abs(w.ipop)/sum(abs(w.ipop))
  
  Ws  <- cbind("wnnls"=w.wnnls,"LRQP"=w.LRQP,"ipop"=w.ipop)
  TFs <- sqrt(diag(t(Ws)%*%t(Z)%*%Z%*%Ws)/nrow(Z))
  LWs <- diag(t(Ws)%*%t(X)%*%diag(tv)%*%X%*%Ws)
  mLW <- min(LWs)
  
  globals$raise_wnnls[globals$NRUNS] <- (LWs[1]-mLW)/mLW
  globals$raise_LRQP[globals$NRUNS]  <- (LWs[2]-mLW)/mLW
  globals$raise_ipop[globals$NRUNS]  <- if (trouble) NA else (LWs[3]-mLW)/mLW
  
  if (return.w) w.wnnls else lossDep(Z,w.wnnls)
}

## outer optimization with nlminb
nlminbOpt <- function(fn,lower,upper,nrandom=1,debug=FALSE,
                      frandom=function(n,lb,ub) stats::runif(n,lb,ub),
                      distance=1,iter.max=1000L,rel.tol=1e-10,trace=FALSE,...) {
  n       <- length(lower)
  best    <- Inf
  res     <- NULL
  control <- list(trace=trace,rel.tol=rel.tol,iter.max=iter.max)
  control[names(list(`...`))] <- list(`...`)
  if (nrandom==0) {
    if (n>16) stop("dimension of v too big for grid of starting values")        # this is quite arbitrary...
    mask <- bitwShiftL(1, 0:(n-1))
  }    
  nruns   <- if (nrandom>0) nrandom else 2^n
  for (i in 1:nruns) {
    s.v <- if (nrandom>0) frandom(n,lower,upper) else 
             ifelse(bitwAnd(i-1,mask),upper-distance,lower+distance)
    tmp <- stats::nlminb(s.v,fn,lower=lower,upper=upper,control=control)
    if (debug) cat("sv ",i," out of ",nruns," has objective ",tmp$objective, 
                   "(best: ",best,") after ",tmp$iterations," Iterations.\n")
    if (tmp$objective<best) {
      res  <- tmp
      best <- res$objective
    }
  }
  res
}

## outer optimization with optim
optimOpt <- function(fn,lower,upper,nrandom=1,debug=FALSE,
                     frandom=function(n,lb,ub) stats::runif(n,lb,ub),
                     distance=1,maxit=1000L,rel.tol=1e-10,trace=FALSE,...) {
  n       <- length(lower)
  best    <- Inf
  res     <- NULL
  control <- list(trace=trace,factr=rel.tol/.Machine$double.eps,maxit=maxit)
  control[names(list(`...`))] <- list(`...`)
  if (nrandom==0) {
    if (n>16) stop("dimension of v too big for grid of starting values")        # this is quite arbitrary...
    mask <- bitwShiftL(1, 0:(n-1))
  }    
  nruns   <- if (nrandom>0) nrandom else 2^n
  for (i in 1:nruns) {
    s.v <- if (nrandom>0) frandom(n,lower,upper) else 
             ifelse(bitwAnd(i-1,mask),upper-distance,lower+distance)
    tmp <- stats::optim(par=s.v,fn=fn,lower=lower,upper=upper,method="L-BFGS-B",
                        control=control)
    if (debug) cat("sv ",i," out of ",nruns," has objective ",tmp$value, 
                   "(best: ",best,") with conv. code ",tmp$convergence,".\n")
    if (tmp$value<best) {
      res  <- tmp
      best <- res$value
    }
  }
  res
}

#' Convert Long Format to List Format
#'
#' \code{listFromLong} converts long to list format.
#'
#' \code{listFromLong} is a convenience function to convert long format
#' (in a \code{\link[base]{data.frame}}, as used by package 'Synth') to list 
#' format, where data is stored as a list of matrices.
#'
#' Most parameter names are named after their equivalents in the 
#' \code{\link[Synth]{dataprep}} function of package 'Synth'.
#'
#' @param foo A \code{data.frame} containing the data in "long" format.
#' @param unit.variable Either a numeric scalar with the column number (in 
#' \code{foo}) containing the units or a character scalar with the corresponding
#' column name in \code{foo}.
#' @param time.variable Either a numeric scalar with the column number (in 
#' \code{foo}) containing the times or a character scalar with the corresponding
#' column name in \code{foo}.
#' @param unit.names.variable Optional. If not \code{NULL}, either a numeric 
#' scalar with the column number (in \code{foo}) containing the unit names or 
#' a character scalar with  the corresponding column name in \code{foo}. Must
#' match with the units defined by \code{unit.variable} (if not \code{NULL}).
#' @param exclude.columns Optional (defaults to \code{NULL}). Numeric vector 
#' with column numbers of \code{foo} to be excluded from the conversion.
#' @return A list of matrices with rows corresponding to the times and columns
#' corresponding to the unit (or unit names, respectively) for all columns of
#' \code{foo} which are neither excluded nor have a special role as time, unit,
#' or unit names variable.
#' @importFrom stats na.omit
#' @export listFromLong
#' @examples
#' if (require("Synth")) {
#'   data(basque)
#'   Basque <- listFromLong(basque, unit.variable="regionno", 
#'                          time.variable="year", 
#'                          unit.names.variable="regionname")
#'   names(Basque)
#'   head(Basque$gdpcap)
#' }
listFromLong <- function(foo, unit.variable, time.variable, 
                         unit.names.variable=NULL,exclude.columns=NULL) {
  if(!is.data.frame(foo)) stop("foo must be a data.frame")

  # main helper function
  DFtoList <- function(input,rowcol,colcol,colnamecol=NULL,exclude=NULL) {
    stopifnot(length(dim(input))==2)
    datcols    <- setdiff(seq_len(ncol(input)), 
                          c(rowcol,colcol,colnamecol,exclude))
    res        <- vector("list",length(datcols))
    names(res) <- if (!is.null(colnames(input))) colnames(input)[datcols] else 
                                                 as.character(datcols)
    if (!is.null(colnamecol)) {
      c2n        <- na.omit(unique(input[,colnamecol]))
      names(c2n) <- na.omit(unique(input[,colcol]))
    }  
                                                 
    for (i in seq_along(res)) {
      idx  <- !is.na(input[,datcols[i]])
      rown <- unique(input[idx,rowcol])
      coln <- unique(input[idx,colcol])
      res[[i]] <- matrix(NA,nrow=length(rown),ncol=length(coln))
      rownames(res[[i]]) <- rown
      colnames(res[[i]]) <- coln
      for (j in which(idx)) 
        res[[i]][as.character(input[j,rowcol]),as.character(input[j,colcol])] <- 
          input[j,datcols[i]]
      if (!is.null(colnamecol)) colnames(res[[i]]) <- c2n[as.character(coln)]
      res[[i]] <- res[[i]][order(rownames(res[[i]])),,drop=FALSE]
    }
    res
  }
    
  # helper function for data check and column lookup
  nam2idx <- function(...,id,type="numeric") {
    if (missing(id)) id <- as.character(match.call())[2]
    obj <- list(`...`)[[1]]
    if (length(obj)!=1) stop(paste0("\n Please specify only one ",id,"\n"))
    if (!(mode(foo[,obj]) %in% type)) stop("\n ",id," not found as ",type,
                                           " variable in foo.\n")
    if (mode(obj) == "character") which(names(foo)==obj) else obj
  }

  # check and look for unit.variable and time.variable
  unit.variable <- nam2idx(unit.variable)
  time.variable <- nam2idx(time.variable,type=c("numeric","character"))

  # check and look for unit.name.variable (if present)
  if (!is.null(unit.names.variable)) {
    idx <- !(is.na(foo[,unit.variable])|is.na(foo[,unit.names.variable]))
    unit.names.variable <- nam2idx(unit.names.variable,type="character")
    if (length(unique(foo[idx,unit.names.variable])) !=
          length(unique(foo[idx,unit.variable])))
        stop("lengths of unit.names and unit.names.variable do not match")
    if (length(unique(paste(foo[idx,unit.variable],foo[idx,unit.names.variable],
                            sep="----------"))) != 
          length(unique(foo[idx,unit.variable])))
        stop("unit.names and unit.names.variable do not match")
  }
  
  # do the actual conversion
  DFtoList(foo,rowcol=time.variable,colcol=unit.variable,
           colnamecol=unit.names.variable,exclude=exclude.columns)
}

#' Difference-in-difference estimator based on SCM
#'
#' \code{did} calculates difference-in-difference estimators based on SCM.
#'
#' \code{did} calculates difference-in-difference estimators with corresponding
#' p-values (if results of a placebo study are present) based on the Synthetic 
#' Control Method.
#'
#' @param x An object of class \code{"mscmt"}, usually obtained as
#' the result of a call to function \code{\link{mscmt}}.
#' @param what A character vector. Name of the variable to be considered. If 
#' missing, the (first) dependent variable will be used.
#' @param range.pre A vector of length 2 defining the range of the pre-treatment
#' period with start and end time given as 
#' \itemize{
#' \item annual dates, if the format of start/end time is "dddd", e.g. "2016",
#' \item quarterly dates, if the format of start/end time is "ddddQd", e.g. 
#' "2016Q1",
#' \item monthly dates, if the format of start/end time is "dddd?dd" with "?"
#' different from "W" (see below), e.g. "2016/03" or "2016-10",
#' \item weekly dates, if the format of start/end time is "ddddWdd", e.g. 
#' "2016W23",
#' \item daily dates, if the format of start/end time is "dddd-dd-dd", e.g. 
#' "2016-08-18",
#' }
#' corresponding to the format of the respective column of the \code{times.dep}
#' argument of \code{\link{mscmt}}.
#' If missing, the corresponding column of \code{times.dep} will be used.
#' @param range.post A vector of length 2 defining the range of the 
#' post-treatment period with start and end time given as 
#' \itemize{
#' \item annual dates, if the format of start/end time is "dddd", e.g. "2016",
#' \item quarterly dates, if the format of start/end time is "ddddQd", e.g. 
#' "2016Q1",
#' \item monthly dates, if the format of start/end time is "dddd?dd" with "?"
#' different from "W" (see below), e.g. "2016/03" or "2016-10",
#' \item weekly dates, if the format of start/end time is "ddddWdd", e.g. 
#' "2016W23",
#' \item daily dates, if the format of start/end time is "dddd-dd-dd", e.g. 
#' "2016-08-18",
#' }
#' corresponding to the format of the respective column of the \code{times.dep}
#' argument of \code{\link{mscmt}}. Will be guessed if missing.
#' @param alternative A character string giving the alternative of the test. 
#' Either \code{"two.sided"} (default), \code{"less"}, or \code{"greater"}.
#' @param exclude.ratio A numerical scalar (default: \code{Inf}). When 
#' calculating the p-value, control units with an average pre-treatment gap
#' of more then \code{exclude.ratio} times the average pre-treatment gap of
#' the treated unit are excluded from the analysis.
#' @return A list with components \code{effect.size}, \code{average.pre} and
#' \code{average.post}. If \code{x} contains the results of a placebo study,
#' three components \code{p.value}, \code{rank}, and \code{excluded} (with the 
#' names of the excluded units) are included additionally.
#' @export did
#' @examples
#' \dontrun{
#' ## for an example, see the main package vignette:
#'  vignette("WorkingWithMSCMT",package="MSCMT")
#' }
did <- function(x,what,range.pre,range.post,
                alternative = c("two.sided", "less", "greater"),
                exclude.ratio = Inf) {
  alternative <- match.arg(alternative)  
  if (has.placebo <- !is.null(x$placebo)) {
    if (missing(what)) what  <- x[[1]]$dependent[1]
    gaps <- x$placebo[[what]]$gaps
    if (missing(range.pre))  
      if (what %in% x[[1]]$dependent) range.pre <- x[[1]]$times.dep[,what] else
                                      stop("range.pre is missing")
  } else {
    if (missing(what)) what  <- x$dependent[1]
    gaps <- x$gaps[[what]]
    if (missing(range.pre))  
      if (what %in% x$dependent) range.pre <- x$times.dep[,what] else
                                 stop("range.pre is missing")
  }
  gaps.pre  <- AQMwindow(gaps,range.pre)
  gaps.post <- if (missing(range.post)) AQMtail(gaps,range.pre[2]) else
                                        AQMwindow(gaps,range.post)
  if (has.placebo) {
    D.pre     <- apply(gaps.pre,2,function(x) mean(x))
    D.post    <- apply(gaps.post,2,function(x) mean(x))
    excl      <- (abs(D.pre) > exclude.ratio*abs(D.pre[1]))
    effect    <- as.numeric(D.post - D.pre)[!excl]
    p.value   <- switch(alternative,
                   two.sided = mean(abs(effect[1])<=abs(effect)),
                   less      = mean(effect[1]>=effect),
                   greater   = mean(effect[1]<=effect))
    list(effect.size=effect[1],average.pre=as.numeric(D.pre[1]),
         average.post=as.numeric(D.post[1]),p.value=p.value,
         rank=sum(effect[1]>=effect),excluded=names(excl)[excl])
  } else list(effect.size=as.numeric(mean(gaps.post) - mean(gaps.pre)),
              average.pre=as.numeric(mean(gaps.pre)),
              average.post=as.numeric(mean(gaps.post)))
}

---
title: "Working with package MSCMT"
vignette: >
  %\VignetteIndexEntry{Working with package MSCMT}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
output: 
  html_vignette:
    toc: true
bibliography: ../inst/REFERENCES.bib
---

```{r, echo = FALSE}
knitr::opts_chunk$set(
  fig.width  = 7,
  fig.height = 4.5,
  fig.align  = "center",
#  cache      = TRUE,
  autodep    = TRUE
)
```

## Introduction

Based on the `basque` dataset and the corresponding example of package `Synth`, this vignette illustrates the usage of package `MSCMT`.
Thereby, parts of @Abadie2003 are reproduced in this example.

## Preparing the Data

The `MSCMT` package expects its input data to be a `list` of matrices, where each matrix corresponds to one variable of interest.
The column names of the matrices correspond to the names of all units (treated unit and control units) and have to be identical across all elements of the list. 
The row names correspond to the instants of time where data is available, they may well be different between variables.

Currently, all variables must either contain annual, quarterly, monthly, weekly or daily data.
The row names must therefore match one of the following patterns (where `d` corresponds to a single digit):

1.  `dddd` for annual dates, e.g. `2016`,
2.  `ddddQd` for quarterly dates, e.g. `2016Q1` for the first quarter of 2016,
3.  `dddd-dd` for monthly dates, e.g. `2016-03` for March 2016.
4.  `ddddWdd` for weekly dates, e.g. `2016W03` for the third week of 2016.
5.  `dddd-dd-dd` for daily dates, e.g. `2016-03-17`.

A list of matrices is a very useful, but quite unusual data format.
Therefore, a helper function called `listFromLong`, which generates a list of matrices from 
a more common data representation, the so-called 'long' format, has been included in package `MSCMT`. 
With `listFromLong`, migrating from the `Synth` package to `MSCMT` is simple, because package `Synth` (more precisely, its `dataprep` function) expects its input data to be in
'long' format.

### Example

The `basque` dataset in package `Synth` in 'long' format has the following structure:
```{r}
library(Synth)
data(basque)
colnames(basque)
basque[703,]
```

With `listFromLong` from package `MSCMT`, the conversion to 'list' format is a simple task. 
The names of the parameters correspond to the names of the parameters for function `dataprep` of package `Synth`:
```{r}
library(MSCMT)
Basque <- listFromLong(basque, unit.variable="regionno", time.variable="year", unit.names.variable="regionname")
names(Basque)
head(Basque$gdpcap)
```

The 'list' representation allows for simple, reproducible data preparations.
To reproduce the analysis of @Abadie2003, the following preparations are sufficient:
```{r}
# define the sum of all cases
school.sum <- with(Basque,colSums(school.illit + school.prim + school.med + school.high  + school.post.high))
# combine school.high and school.post.high in a single class
Basque$school.higher <- Basque$school.high + Basque$school.post.high
# calculate ratios and multiply by number of observations to obtain percentages from totals 
for (item in c("school.illit", "school.prim", "school.med", "school.higher"))      
  Basque[[item]] <- 6 * 100 * t(t(Basque[[item]]) / school.sum)
```

## Defining the Model

In package `Synth`, data preparation and model formulation are combined in function `dataprep`, whereas the estimation is separated from the model formulation.
The approach of package `MSCMT` is different: in order to facilitate the estimation of (potentially many) different models based on the same dataset, the model formulation has been
moved to function `mscmt`, which also does the model estimation.

To allow for the methodological extensions (**Multivariate** Synthetic Control Methods using **Time-Series**), the model syntax had to be made more flexible. 
A model formulation now consists of

*   defining the (single)^[Changing the treated unit for placebo studies is done automatically.] treated unit (single character string, parameter `treatment.identifier`),
*   defining the (multiple) control units (vector of character strings, parameter `controls.identifier`),
*   defining one or more dependent variables and the corresponding optimization periods as a `2 x L`-matrix (parameter `times.dep`),  where 
    -   the column names consist of the names of the `L` dependent variables,
    -   the first row contains the starting times of the optimization periods in the appropriate (annual, quarterly, or monthly) format,
    -   the second row contains the end times of the optimization periods in the appropriate format,
*   defining several predictor variables and the corresponding optimization periods as a `2 x K`-matrix (parameter `times.pred`), where
    -   the column names consist of the names of the `K` predictor variables,
    -   the first row contains the starting times of the optimization periods in the appropriate (annual, quarterly, or monthly) format,
    -   the second row contains the end times of the optimization periods in the appropriate format,
*   an (optional) vector of the names of `K` aggregation functions (parameter `agg.fns`) for the predictor variables. 
    If missing, all predictor variables are considered to be time series, which corresponds to the function name `"id"`.
    Whenever the result of an aggregation function has length exceeding 1, the resulting data is considered as a time series, too.

Note that `times.dep` and `times.pred` may contain duplicate column names for further increased flexibility of the model specification.

### Example

The following model specification reproduces the model in @Abadie2003:
```{r}
treatment.identifier <- "Basque Country (Pais Vasco)"
controls.identifier  <- setdiff(colnames(Basque[[1]]),
                                c(treatment.identifier, "Spain (Espana)"))
times.dep  <- cbind("gdpcap"                = c(1960,1969))
times.pred <- cbind("school.illit"          = c(1964,1969),
                    "school.prim"           = c(1964,1969),
                    "school.med"            = c(1964,1969),
                    "school.higher"         = c(1964,1969),
                    "invest"                = c(1964,1969),
                    "gdpcap"                = c(1960,1969),
                    "sec.agriculture"       = c(1961,1969),
                    "sec.energy"            = c(1961,1969),
                    "sec.industry"          = c(1961,1969),
                    "sec.construction"      = c(1961,1969),
                    "sec.services.venta"    = c(1961,1969),
                    "sec.services.nonventa" = c(1961,1969),
                    "popdens"               = c(1969,1969))
agg.fns <- rep("mean", ncol(times.pred))                       
```

## Single Estimation

After preparing the data and formulating the model, the model estimation is done with function `mscmt`. 
Apart from the function parameters concerning data and model specification, there are further parameters which, e.g.,

*   control for several aspects of the optimization procedure, including weights for the inner and outer target functions, 
*   choose and tune the optimizers, 
*   verbose output,
*   initiate placebo studies 
*   allow to use the computational power of a multi-core cpu or cluster for placebo studies.

### Example 

A simple estimation (without a placebo study) produces the following console output (if parameter `verbose` is `TRUE`, which is the default):
```{r}
res <- mscmt(Basque, treatment.identifier, controls.identifier, times.dep, times.pred, agg.fns, seed=1)
```

Package `MSCMT` ships with an `S3` method for `print`, which gives a nice human-readable summary of the estimation results:
```{r}
res
```

While there is a basic `S3` method for `plot`, it is strongly recommended to use package `ggplot2` and the corresponding `S3` method for `ggplot` contained in package `MSCMT`.
With the results of a simple estimation, two types of plots are available, the first being a comparison of original and synthesized data for the treated unit. 
The variable to be plotted can be selected with parameter `what`, by default the (first) dependent variable is being chosen:
```{r}
library(ggplot2)
ggplot(res, type="comparison")
```

The second type of plot is a plot of the gaps, ie. the differences between original and synthesized data:
```{r}
ggplot(res, type="gaps")
```

It is possible to plot several variables by providing a vector for argument `what`. Pre-defined sets of variables named `"dependents"`, `"predictors"`, and `"all"` can be selected with parameter `what.set`.
```{r}
ggplot(res, what=c("gdpcap","invest","school.higher","sec.energy"), type="comparison")
```

## Placebo Study

Placebo studies are performed by simply setting the function argument `placebo` to `TRUE`. By default, the original treated unit is not added to the donor pool of the (original) control units, but this can be changed with parameter `placebo.with.treated`. 

### Making Use of a Cluster

A remarkable speed-up (depending on the number of control units) can be achieved for placebo studies by making use of a cluster, which can be set up with the help of package `parallel`. 
The simplest form of a cluster is a local cluster, which makes the power of multi-core cpus available for the (lengthy) computations involved with placebo studies. 
Setting up a local cluster is very easy, see the example below.

The argument `cl` of function `mscmt` can be used to specify the cluster to be used for placebo studies.
The only drawback of using a cluster for placebo studies is losing the verbose output from the individual (one for each unit) SCM estimations, which
includes a lack of progress information for the whole placebo study. 
Nevertheless, the speed-up should compensate for this drawback in all applications where a placebo study is meaningful.

Although clusters should be shut down automatically when the corresponding (master) `R` process is finished, function `stopCluster` can (and should) be used to shut down the cluster manually.

### Example

In the following example, a (local) cluster with two nodes is used for the estimation.

```{r}
library(parallel)
cl <- makeCluster(2)
resplacebo <- mscmt(Basque, treatment.identifier, controls.identifier, times.dep, times.pred, agg.fns, cl=cl, placebo=TRUE, seed=1)
stopCluster(cl)
```

Object `resplacebo` now contains single SCM estimations for each unit as well as aggregated information concerning original data, synthesized data, and gaps for all units.
The individual SCM estimations can be accessed separately (as list elements with names corresponding to the units' names). 
With the following plot, one can inspect whether there is some effect for Catalonia:

```{r}
ggplot(resplacebo[["Cataluna"]], type="comparison")
```

Several functions in package `MSCMT` are able to make use of the results of the placebo study as a whole. 
One example are so-called placebo plots, by setting the plot type to `"placebo.gaps"` (the default for results of a placebo study):
```{r}
ggplot(resplacebo)
```

## Statistical Inference

For statistical inference based on the results of a placebo study, the literature has developed so-called 'placebo tests', which have similarities to permutation tests. Two of these are

*   tests for the (per-period) treatment effect based on the per-period gaps,
*   tests for the aggregated treatment effect based on a difference-in-difference approach for aggregated (average) gaps.

Most often, not all control units can be synthesized with an acceptable fit in a placebo study, resulting in large pre-treatment gaps. 
Of course, large post-treatment gaps are expected for these units, but since these gaps are rather caused from lack of fit than from an existing treatment effect, 
excluding such units is strongly advisable while investigating the effect for the (original) treated unit.

Excluding control units with large pre-treatment errors is usually done by limiting the ratio of a control unit's pre-treatment (r)mspe to the treated unit's (r)mspe. 

### Example

Control units with large pre-treatment errors can easily be excluded from placebo plots:
```{r}
ggplot(resplacebo, exclude.ratio=5, ratio.type="mspe")
```

The p-values of per-period placebo tests can be calculated via function `pvalue` or plotted via `ggplot` with `plot.type="p.values"`:
```{r}
pvalue(resplacebo, exclude.ratio=5, ratio.type="mspe", alternative="less")
ggplot(resplacebo, exclude.ratio=5, ratio.type="mspe", type="p.value", alternative="less")
```

Calculating the aggregated treatment effect and testing its significance can be done with function `did` of package `MSCMT`:
```{r}
did(resplacebo, range.post=c(1970,1990), exclude.ratio=5, alternative="less")
```


## Summary

In this vignette, the basic workflow of preparing the data, defining and estimating the model, and evaluating the results of a (simple) SCM application with package `MSCMT` has been illustrated.

Many features and options of package `MSCMT` remained untouched, because this would have led far beyond the scope of this simple example. 
There are more specialized vignettes which illustrate

*   [how to check the feasibility of (and potentially improve) the results obtained by function `synth` of package `Synth`,](CheckingSynth.html)
*   [an application of SCM using time series (SCM**T**).](UsingTimeSeries.html)

In future releases of this package, additional vignettes will (probably) be added to illustrate

*   applications of **M**SCM and **M**SCM**T**,
*   alternatives for the optimizers and its parameters,
*   more functionality of the `ggplot` method.
*   the use of cross-validation, see @CV.

## References

---
title: "Checking and Improving Results of package Synth"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Checking and Improving Results of package Synth}
  %\VignetteEncoding{UTF-8}
output: 
  html_vignette:
    toc: true
bibliography: ../inst/REFERENCES.bib
---

```{r, echo = FALSE}
knitr::opts_chunk$set(
  fig.width  = 7,
  fig.height = 4,
  fig.align  = "center",
#  cache      = TRUE,
  autodep    = TRUE
)
```

## Introduction

This vignette illustrates the usage of `improveSynth`. For a more general introduction to package `MSCMT` see its [main vignette](WorkingWithMSCMT.html).

Estimating an SCM model involves searching for an approximate solution of a nested optimization problem. 
Although the formulation of the optimization problem is quite simple, finding a (good approximate) solution can be hard for several reasons, see @Mafia and @FastReliable.
While implementing package `MSCMT` we put a lot of effort into the design of a smart and robust (but still fast) optimization procedure.

Apart from function `mscmt` for the estimation of SCM models based on our model syntax, we also included the convenience function `improveSynth`,
which implements checks for feasibility and optimality of results delivered by package `Synth`. 
Below, we illustrate how to use `improveSynth`.


## First Example

We exemplify the usage of `improveSynth` based on the first example of function `synth` in package `Synth`.

### Generating the result of package `Synth`

The following code is thus essentially borrowed from the `example` section of the corresponding help page (all comments have been removed):

```{r}
library(Synth)
data(synth.data)
dataprep.out <-
  dataprep(
    foo = synth.data,
    predictors = c("X1", "X2", "X3"),
    predictors.op = "mean",
    dependent = "Y",
    unit.variable = "unit.num",
    time.variable = "year",
    special.predictors = list(
      list("Y", 1991, "mean"),
      list("Y", 1985, "mean"),
      list("Y", 1980, "mean")
    ),
    treatment.identifier = 7,
    controls.identifier = c(29, 2, 13, 17, 32, 38),
    time.predictors.prior = c(1984:1989),
    time.optimize.ssr = c(1984:1990),
    unit.names.variable = "name",
    time.plot = 1984:1996
  )

synth.out <- synth(dataprep.out)
```

### Checking the result

We check the result by applying function `improveSynth` to `synth.out` and `dataprep.out`:

```{r}
library(MSCMT)
synth2.out <- improveSynth(synth.out,dataprep.out)
```

Package `Synth` generated a (slightly) infeasible solution, returning a (slightly)
suboptimal weight vector `w` for the control units.
However, the predictor weights `v` are (considerably) suboptimal anyway, 
because the original dependent loss of `r round(synth.out$loss.v,6)` (as well as
the dependent loss for the corrected `w` `r round(synth2.out$new.loss.v,6)`)
is considerably larger than the dependent loss `r round(synth2.out$loss.v,6)` 
for the optimal predictor weights obtained by `improveSynth`.

## Second Example

In the second example, we modify the first example by allowing package `Synth` to use `genoud` as (outer) optimization algorithm.

### Generating the result of package `Synth`

`genoud` is switched on by the corresponding function argument. We capture the output with `capture.output` because it is **very** verbose. Furthermore, the calculation is quite lengthy, therefore the results have been cached.^[To reproduce from scratch, please delete `"synth3.out.RData"` from the `vignettes` folder.]

```{r}
if (file.exists("synth3.out.RData")) load ("synth3.out.RData") else {
  set.seed(42)
  out <- capture.output(synth3.out <- synth(dataprep.out,genoud=TRUE))
}  
```

### Checking the result

We again check the result by applying function `improveSynth` to `synth3.out` and `dataprep.out`:

```{r}
synth4.out <- improveSynth(synth3.out,dataprep.out)
```

Now, package `Synth` generated a solution with a dependent loss of `r round(synth3.out$loss.v,6)` which is even smaller than the dependent loss `r round(synth2.out$loss.v,6)` obtained by `improveSynth`.
However, the solution generated by `Synth` is severely **infeasible**: the inner optimization failed, returning a suboptimal weight vector `w` for the control units, which itself lead to a wrong calculation of the dependent loss (which, of course, depends on `w`). 
Implanting the true optimal `w` (depending on `v`) leads to a large increase of the dependent loss, which uncovers the suboptimality of `v`.

`improveSynth` is able to detect this severe problem and calculates an improved *and feasible* solution 
(the improved solution `r if(!isTRUE(all.equal(round(synth4.out$loss.v,6),round(synth2.out$loss.v,6)))) "essentially" else ""` matches the solution obtained from the first call to `improveSynth` above, 
with a dependent loss of `r round(synth4.out$loss.v,6)``r if(!isTRUE(all.equal(round(synth4.out$loss.v,6),round(synth2.out$loss.v,6)))) paste0("as compared to ",round(synth2.out$loss.v,6)," above") else ""`).

## Summary

Issues with the inner and outer optimizers used in `synth` from package `Synth` may lead to infeasible or suboptimal solutions.
This vignette illustrated the usage of the convenience function `improveSynth` from package `MSCMT` for checking and potentially improving results obtained from `synth`. 

## References

---
title: "SCM Using Time Series"
vignette: >
  %\VignetteIndexEntry{SCM Using Time Series}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
output: 
  html_vignette:
    toc: true
bibliography: ../inst/REFERENCES.bib
---

```{r, echo = FALSE}
knitr::opts_chunk$set(
  fig.width  = 7,
  fig.height = 4,
  fig.align  = "center",
#  cache      = TRUE,
  autodep    = TRUE
)
```

## Introduction

This vignette illustrates the syntax of SCM**T** models. For a more general introduction to package `MSCMT` see its [main vignette](WorkingWithMSCMT.html).

Although SCM models are usually based on time series data of predictor variables, standard SCM estimation does not exploit this particular characteristic. 
Instead, time series data of predictors are either aggregated, mostly by calculating (a bunch of) means, or every instant of time is considered as a separate input variable with individual predictor weight.
With package `MSCMT`, a time series of a predictor variable can be considered as single input variable without the need of aggregation, an extension of SCM called SCM**T**, see @KP16.

This vignette illustrates the syntax of SCM**T** models and how SCM**T** models may lead to more meaningful predictor weights without drawbacks concerning the model fit. 


## Definition of the Standard Model

We use the `basque` dataset in package `Synth` as an example and replicate the preparation of the data from the [main vignette](WorkingWithMSCMT.html) of this package:

```{r}
library(Synth)
data(basque)
library(MSCMT)
Basque <- listFromLong(basque, unit.variable="regionno", time.variable="year", unit.names.variable="regionname")
school.sum <- with(Basque,colSums(school.illit + school.prim + school.med + school.high  + school.post.high))
Basque$school.higher <- Basque$school.high + Basque$school.post.high
for (item in c("school.illit", "school.prim", "school.med", "school.higher"))      
  Basque[[item]] <- 6 * 100 * t(t(Basque[[item]]) / school.sum)
```

We also replicate model specification of the [main vignette](WorkingWithMSCMT.html) which reproduces the model in @Abadie2003:

```{r}
treatment.identifier <- "Basque Country (Pais Vasco)"
controls.identifier  <- setdiff(colnames(Basque[[1]]),
                                c(treatment.identifier, "Spain (Espana)"))
times.dep  <- cbind("gdpcap"                = c(1960,1969))
times.pred <- cbind("school.illit"          = c(1964,1969),
                    "school.prim"           = c(1964,1969),
                    "school.med"            = c(1964,1969),
                    "school.higher"         = c(1964,1969),
                    "invest"                = c(1964,1969),
                    "gdpcap"                = c(1960,1969),
                    "sec.agriculture"       = c(1961,1969),
                    "sec.energy"            = c(1961,1969),
                    "sec.industry"          = c(1961,1969),
                    "sec.construction"      = c(1961,1969),
                    "sec.services.venta"    = c(1961,1969),
                    "sec.services.nonventa" = c(1961,1969),
                    "popdens"               = c(1969,1969))
agg.fns <- rep("mean", ncol(times.pred))                       
```

Estimation of the model gives:
```{r}
res <- mscmt(Basque, treatment.identifier, controls.identifier, times.dep, times.pred, agg.fns, seed=1, single.v=TRUE, verbose=FALSE)
res
```

It is remarkable that the mean of the lagged dependent variable `gdpcap.mean.1960.1969` is by far the most important predictor with a weight of `r res$v[6,"max.order"]`, 
all other predictors are only marginally relevant due to their tiny 
(at most `r format(max(res$v[-6,"max.order"]),scientific=FALSE)`) 
weights.^[Notice that the weight vector `v` is obtained by maximizing the order statistics of `v` (while fixing the sum of `v` to 1). 
This choice of 'v' attributes weights as large as possible to even the least relevant predictor(s).]

## Removing the Lagged Dependent Variable

Omitting the lagged dependent variable `gdpcap.mean.1960.1969` from the model definition, however, leads to a significant increase of the dependent loss:
```{r}
times.pred <- times.pred[,-6]
agg.fns <- rep("mean", ncol(times.pred))                       
res2 <- mscmt(Basque, treatment.identifier, controls.identifier, times.dep, times.pred, agg.fns, seed=1, single.v=TRUE, verbose=FALSE)
res2
```

The dependent loss (MSPE) increased considerably from `r res$loss.v` to `r res2$loss.v`. Trying to give more meaning to the economic predictors in this way obviously has the drawback of worsening the fit
of the dependent variable.

## SCMT without the Lagged Dependent Variable

Leaving the lagged dependent variable `gdpcap.mean.1960.1969` aside, but considering all other predictor variables as **time series** instead of aggregating their values leads to the following results:
```{r}
agg.fns <- rep("id", ncol(times.pred))   # Omitting agg.fns has the same effect (as "id" is the default)
res3 <- mscmt(Basque, treatment.identifier, controls.identifier, times.dep, times.pred, agg.fns, seed=1, single.v=TRUE, verbose=FALSE)
res3
```

Notice that this specification's model type is 'SCMT', in contrast to the previous models which were 'SCM' models. 
By using the 'SCMT' model, the dependent loss (`r res3$loss.v`) is even smaller than that of the original model (`r res$loss.v`) which used the dependent variable's mean as an extra economic predictor. 
``r rownames(res3$v)[which.max(res3$v[,"max.order"])]`` has now become the most important predictor with weight `r max(res3$v[,"max.order"])`, 
all other predictor weights are at least `r format(min(res3$v[,"max.order"]),scientific=FALSE)`.

## Summary

This vignette illustrated that considering predictors as *true* time series (without intermediate aggregation) may have various benefits. 
In this example, by excluding the mean of the lagged dependent variable from the set of economic predictors and considering all other predictors as time series,
more meaningful predictor weights could be obtained and the dependent variable's fit could be slightly improved, too.

## References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.r
\name{plot.mscmt}
\alias{plot.mscmt}
\title{Plotting Results of MSCMT}
\usage{
\method{plot}{mscmt}(
  x,
  what,
  type = c("gaps", "comparison", "placebo.gaps", "placebo.data"),
  treatment.time,
  zero.line = TRUE,
  ylab,
  xlab = "Date",
  main,
  sub,
  col,
  lty,
  lwd,
  legend = TRUE,
  bw = FALSE,
  ...
)
}
\arguments{
\item{x}{An object of class \code{"mscmt"}, usually obtained as
the result of a call to function \code{\link{mscmt}}.}

\item{what}{A character scalar. Name of the variable to be plotted. If 
missing, the (first) dependent variable will be used.}

\item{type}{A character scalar denoting the type of the plot containing 
either \code{"gaps"}, \code{"comparison"}, \code{"placebo.gaps"}, 
or \code{"placebo.data"}. 
Partial matching allowed, defaults to \code{"placebo.gaps"}, if results
of a placebo study are present, and to \code{"gaps"}, else.}

\item{treatment.time}{An optional numerical scalar. If not missing, a 
vertical dotted line at the given point in time is included in the plot. 
\code{treatment.time} is measured in years, but may as well be a decimal 
number to reflect treatment times different from January 1st.}

\item{zero.line}{A logical scalar. If \code{TRUE} (default), a horizontal
dotted line (at zero level) is plotted for \code{"gaps"} and \code{"placebo"}
plots.}

\item{ylab}{Optional label for the y-axis, automatically generated if 
missing.}

\item{xlab}{Optional label for the x-axis, defaults to \code{"Date"}.}

\item{main}{Optional main title for the plot, automatically generated 
if missing.}

\item{sub}{Optional subtitle for the plot. If missing, the subtitle is 
generated automatically for \code{"comparison"} and \code{"gaps"} plots.}

\item{col}{Optional character vector with length corresponding to the number
of units. Contains the colours for the different units, automatically 
generated if missing.}

\item{lty}{Optional numerical vector with length corresponding to the number
of units. Contains the line types for the different units, automatically 
generated if missing.}

\item{lwd}{Optional numerical vector with length corresponding to the number
of units. Contains the line widths for the different units, automatically 
generated if missing.}

\item{legend}{A logical scalar. If \code{TRUE} (default), a legend is 
included in the plot.}

\item{bw}{A logical scalar. If \code{FALSE} (default), the automatically 
generated colours and line types are optimized for a colour plot, 
if \code{TRUE}, the automatic colours and line types are set for a black and
white plot.}

\item{...}{Further optional parameters for the underlying 
\code{\link[graphics]{plot}} function.}
}
\value{
Nothing useful (function is called for its side effects).
}
\description{
\code{plot.mscmt} plots results of \code{mscmt}.
}
\details{
A unified basic plot function for gaps plots, comparison of treated and 
synthetic values, as well as plots for placebo studies.
Consider using \code{\link{ggplot.mscmt}} instead, which is the preferred
plot method and has more functionality than \code{\link{plot.mscmt}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mscmt.r
\name{mscmt}
\alias{mscmt}
\title{Multivariate SCM Using Time Series}
\usage{
mscmt(
  data,
  treatment.identifier = NULL,
  controls.identifier = NULL,
  times.dep = NULL,
  times.pred = NULL,
  agg.fns = NULL,
  placebo = FALSE,
  placebo.with.treated = FALSE,
  univariate = FALSE,
  univariate.with.dependent = FALSE,
  check.global = TRUE,
  inner.optim = "wnnlsOpt",
  inner.opar = list(),
  outer.optim = "DEoptC",
  outer.par = list(),
  outer.opar = list(),
  std.v = c("sum", "mean", "min", "max"),
  alpha = NULL,
  beta = NULL,
  gamma = NULL,
  return.ts = TRUE,
  single.v = FALSE,
  verbose = TRUE,
  debug = FALSE,
  seed = NULL,
  cl = NULL,
  times.pred.training = NULL,
  times.dep.validation = NULL,
  v.special = integer(),
  cv.alpha = 0,
  spec.search.treated = FALSE,
  spec.search.placebos = FALSE
)
}
\arguments{
\item{data}{Typically, a list of matrices with rows corresponding to times 
and columns corresponding to units for all relevant features (dependent as
well as predictor variables, identified by the list elements' names).
This might be the result of converting from a 
\code{\link[base]{data.frame}}
by using function \code{\link{listFromLong}}. 

For convenience, \code{data} may alternatively be the 
result of function \code{\link[Synth]{dataprep}} of package 
\code{'Synth'}. In this case, the parameters \code{treatment.identifier},
\code{controls.identifier}, \code{times.dep}, \code{times.pred}, 
and \code{agg.fns} are ignored, as these input parameters are generated
automatically from \code{data}. The parameters \code{univariate}, 
\code{alpha}, \code{beta}, and \code{gamma} are ignored by fixing them to 
their defaults.
Using results of \code{\link[Synth]{dataprep}} is experimental, because
the automatic generation of input parameters may fail due to lack of 
information contained in results of \code{\link[Synth]{dataprep}}.}

\item{treatment.identifier}{A character scalar containing the name of the 
treated unit. 
Must be contained in the column names of the matrices in \code{data}.}

\item{controls.identifier}{A character vector containing the names of at 
least two control units.
Entries must be contained in the column names of the matrices in \code{data}.}

\item{times.dep}{A matrix with two rows (containing start times in
the first and end times in the second row) and one column for each dependent
variable, where the column names must exactly match the names of the
corresponding dependent variables. 
A sequence of dates with the given start and end times of
\itemize{
\item annual dates, if the format of start/end time is "dddd", e.g. "2016",
\item quarterly dates, if the format of start/end time is "ddddQd", e.g. 
"2016Q1",
\item monthly dates, if the format of start/end time is "dddd?dd" with "?"
different from "W" (see below), e.g. "2016/03" or "2016-10",
\item weekly dates, if the format of start/end time is "ddddWdd", e.g. 
"2016W23",
\item daily dates, if the format of start/end time is "dddd-dd-dd", e.g. 
"2016-08-18",
}
will be constructed; these dates are looked for in the row names of
the respective matrices in \code{data}. In applications with 
cross-validation, \code{times.dep} belongs to the main period.}

\item{times.pred}{A matrix with two rows (containing start times in
the first and end times in the second row) and one column for each predictor
variable, where the column names must exactly match the names of the
corresponding predictor variables.
A sequence of dates with the given start and end times of
\itemize{
\item annual dates, if the format of start/end time is "dddd", e.g. "2016",
\item quarterly dates, if the format of start/end time is "ddddQd", e.g. 
"2016Q1",
\item monthly dates, if the format of start/end time is "dddd?dd" with "?"
different from "W" (see below), e.g. "2016/03" or "2016-10",
\item weekly dates, if the format of start/end time is "ddddWdd", e.g. 
"2016W23",
\item daily dates, if the format of start/end time is "dddd-dd-dd", e.g. 
"2016-08-18",
}
will be constructed; these dates are looked for in the row names of
the respective matrices in \code{data}. In applications with 
cross-validation, \code{times.pred} belongs to the main period.}

\item{agg.fns}{Either \code{NULL} (default) or a character vector containing
one name of an aggregation function for each predictor variable (i.e., each
column of \code{times.pred}). The character string "id" may be used as a
"no-op" aggregation. Each aggregation function must accept a numeric vector
and return either a numeric scalar ("classical" MSCM) or a numeric vector 
(leading to MSCM*T* if length of vector is at least two).}

\item{placebo}{A logical scalar. If \code{TRUE}, a placebo study is 
performed where, apart from the treated unit, each control unit is considered
as treated unit in separate optimizations. Defaults to \code{FALSE}. 
Depending on the number of control units and the complexity of the problem, 
placebo studies may take a long time to finish.}

\item{placebo.with.treated}{A logical scalar. If \code{TRUE}, the treated
unit is included as control unit (for other treated units in placebo 
studies). Defaults to \code{FALSE}.}

\item{univariate}{A logical scalar. If \code{TRUE}, a series of univariate
SCMT optimizations is done (instead of one MSCMT optimization) even if
there is more than one dependent variable. Defaults to \code{FALSE}.}

\item{univariate.with.dependent}{A logical scalar. If \code{TRUE} (and if
\code{univariate} is also \code{TRUE}), all dependent variables (contained
in the column names of \code{times.dep}) apart from the current (real) 
dependent variable are included as predictors in the series of univariate
SCMT optimizations. Defaults to \code{FALSE}.}

\item{check.global}{A logical scalar. If \code{TRUE} (default), a check for
the feasibility of the unrestricted outer optimum (where actually no 
restrictions are imposed by the predictor variables) is made before 
starting the actual optimization procedure.}

\item{inner.optim}{A character scalar containing the name of the optimization
method for the inner optimization. Defaults to \code{"wnnlsOpt"}, which
(currently) is the only supported implementation, because it outperforms
all other inner optimizers we are aware of. 
\code{"ipopOpt"}, which uses \code{\link[kernlab]{ipop}}, and 
\code{LowRankQPOpt}, which uses \code{\link[LowRankQP]{LowRankQP}} as inner
optimizer, have experimental support for benchmark purposes.}

\item{inner.opar}{A list containing further parameters for the inner 
optimizer. Defaults to the empty list. (For \code{"wnnlsOpt"}, there are no
meaningful further parameters.)}

\item{outer.optim}{A character vector containing the name(s) of the 
optimization method(s) for the outer optimization. Defaults to 
\code{"DEoptC"}, which (currently) is the recommended global optimizer. 
The optimizers currently supported can be found in the documentation of
parameter \code{outer.opar}, where the default control parameters for
the various optimizers are listed.
If \code{outer.optim} has length greater
than 1, one optimization is invoked for each outer optimizer (and, 
potentially, each random seed, see below), and the best result is used.}

\item{outer.par}{A list containing further parameters for the outer 
optimization procedure. Defaults to the empty list. Entries in this list may 
override the following hard-coded general defaults:
\itemize{
\item \code{lb=1e-8}, corresponding to the lower bound for the ratio of
predictor weights,
\item \code{opt.separate=TRUE}, corresponding
to an improved outer optimization where each predictor is treated as the 
(potentially) most important predictor (i.e. with maximal weight) in 
separate optimizations (one for each predictor), see [1].
}}

\item{outer.opar}{A list (or a list of lists, if \code{outer.optim} has
length greater than 1) containing further parameters for the outer 
optimizer(s). Defaults to the empty list. Entries in this list may override
the following hard-coded defaults for the individual optimizers, which
are quite modest concerning the computing time. 
\code{dim} is a variable holding the problem dimension, 
typically the number of predictors minus one.
\tabular{lll}{
\bold{Optimizer}  \tab \bold{Package}     \tab \bold{Default parameters} \cr 
\code{DEoptC}     \tab \code{MSCMT}       \tab \code{nG=500}, \code{nP=20*dim}, \code{waitgen=100}, \cr
                  \tab                    \tab \code{minimpr=1e-14}, \code{F=0.5}, \code{CR=0.9} \cr
\code{cma_es}     \tab \code{cmaes}       \tab \code{maxit=2500} \cr
\code{crs}        \tab \code{nloptr}      \tab \code{maxeval=2.5e4}, \code{xtol_rel=1e-14}, \cr
                  \tab                    \tab \code{population=20*dim}, \code{algorithm="NLOPT_GN_CRS2_LM"} \cr
\code{DEopt}      \tab \code{NMOF}        \tab \code{nG=100}, \code{nP=20*dim} \cr
\code{DEoptim}    \tab \code{DEoptim}     \tab \code{nP=20*dim} \cr
\code{ga}         \tab \code{GA}          \tab \code{maxiter=50}, \code{monitor=FALSE}, \cr
                  \tab                    \tab \code{popSize=20*dim} \cr
\code{genoud}     \tab \code{rgenoud}     \tab \code{print.level=0}, \code{max.generations=70}, \cr
                  \tab                    \tab \code{solution.tolerance=1e-12}, \code{pop.size=20*dim}, \cr
                  \tab                    \tab \code{wait.generations=dim}, \code{boundary.enforcement=2}, \cr
                  \tab                    \tab \code{gradient.check=FALSE}, \code{MemoryMatrix=FALSE} \cr
\code{GenSA}      \tab \code{GenSA}       \tab \code{max.call=1e7}, \code{max.time=25/dim},  \cr
                  \tab                    \tab \code{trace.mat=FALSE} \cr
\code{isres}      \tab \code{nloptr}      \tab \code{maxeval=2e4}, \code{xtol_rel=1e-14}, \cr
                  \tab                    \tab \code{population=20*dim}, \code{algorithm="NLOPT_GN_ISRES"} \cr
\code{malschains} \tab \code{Rmalschains} \tab \code{popsize=20*dim}, \code{maxEvals=25000} \cr
\code{nlminbOpt}  \tab \code{MSCMT/stats} \tab \code{nrandom=30} \cr
\code{optimOpt}   \tab \code{MSCMT/stats} \tab \code{nrandom=25} \cr
\code{PSopt}      \tab \code{NMOF}        \tab \code{nG=100}, \code{nP=20*dim} \cr
\code{psoptim}    \tab \code{pso}         \tab \code{maxit=700} \cr
\code{soma}       \tab \code{soma}        \tab \code{nMigrations=100} 
}
If \code{outer.opar} is a list of lists, its names must correspond to (a 
subset of) the outer optimizers chosen in \code{outer.optim}.}

\item{std.v}{A character scalar containing one of the function names
"sum", "mean", "min", or "max" for the standardization of the predictor 
weights (weights are divided by \code{std.v(weights)} before reporting). 
Defaults to "sum", partial matching allowed.}

\item{alpha}{A numerical vector with weights for the dependent variables
in an MSCMT optimization or \code{NULL} (default). If not \code{NULL},
the length of \code{alpha} must agree with the number of dependent
variables, \code{NULL} is equivalent to weight 1 for all dependent 
variables.}

\item{beta}{Either \code{NULL} (default), a numerical vector, or a list.
If \code{beta} is a numerical vector or a list, its length must agree
with the number of dependent variables. 
\itemize{
\item If \code{beta} is a numerical vector,
the \code{i}th dependent variable is discounted with discount factor 
\code{beta[i]} (the observations of the dependent variables must thus be 
in chronological order!). 
\item If \code{beta} is a list, the components of \code{beta} must be 
numerical vectors with lengths corresponding to the numbers of observations 
for the individual dependent variables. These observations are then 
multiplied with the corresponding component of \code{beta}.
}}

\item{gamma}{Either \code{NULL} (default), a numerical vector, or a list.
If \code{gamma} is a numerical vector or a list, its length must agree
with the number of predictor variables. 
\itemize{
\item If \code{gamma} is a numerical vector,
the output of \code{agg.fns[i]} applied to the \code{i}th predictor variable
is discounted with discount factor \code{gamma[i]} (the output of 
\code{agg.fns[i]} must therefore be in chronological order!). 
\item If \code{gamma} is a list, the components of \code{gamma} must be 
numerical vectors with lengths corresponding to the lengths of the output of 
\code{agg.fns} for the individual predictor variables. The output of 
\code{agg.fns} is then multiplied with the corresponding component of 
\code{gamma}.
}}

\item{return.ts}{A logical scalar. If \code{TRUE} (default), most results are
converted to time series.}

\item{single.v}{A logical scalar. If \code{FALSE} (default), a selection
of feasible (optimal!) predictor weight vectors is generated. If \code{TRUE}, 
the one optimal weight vector which has maximal order statistics is generated 
to facilitate cross validation studies.}

\item{verbose}{A logical scalar. If \code{TRUE} (default), output is verbose.}

\item{debug}{A logical scalar. If \code{TRUE}, output is very verbose. 
Defaults to \code{FALSE}.}

\item{seed}{A numerical vector or \code{NULL}. If not \code{NULL}, the
random number generator is initialized with the elements of \code{seed} via
\code{set.seed(seed)} (see \link[base]{Random}) before
calling the optimizer, performing repeated optimizations (and staying with 
the best) if \code{seed} has length greater than 1. Defaults to \code{NULL}. 
If not \code{NULL}, the seeds \code{int.seed} (default: 53058) and 
\code{unif.seed} (default: 812821) for \code{\link[rgenoud]{genoud}} are 
also initialized to the corresponding element of \code{seed}, but this can 
be overridden with the list elements \code{int.seed} and \code{unif.seed} 
of (the corresponding element of) \code{outer.opar}.}

\item{cl}{\code{NULL} (default) or an object of class \code{cluster}
obtained by \code{\link[parallel]{makeCluster}} of package \code{parallel}. 
Repeated estimations (see \code{outer.optim} and \code{seed}) and
placebo studies will make use of the cluster \code{cl} (if not \code{NULL}).}

\item{times.pred.training}{A matrix with two rows (containing start times in
the first and end times in the second row) and one column for each predictor
variable, where the column names must exactly match the names of the
corresponding predictor variables (or \code{NULL} by default).
If not \code{NULL}, \code{times.pred.training} defines training periods
for cross-validation applications. For the format of the start and end times,
see the documentation of parameter \code{times.pred}.}

\item{times.dep.validation}{A matrix with two rows (containing start times in
the first and end times in the second row) and one column for each dependent
variable, where the column names must exactly match the names of the
corresponding dependent variables (or \code{NULL} by default). 
If not \code{NULL}, \code{times.dep.validation} defines validation period(s)
for cross-validation applications. For the format of the start and end times,
see the documentation of parameter \code{times.dep}.}

\item{v.special}{integer vector containing indices of important predictors
with special treatment (see below). Defaults to the empty set.}

\item{cv.alpha}{numeric scalar containing the minimal proportion (of the
maximal feasible weight) for the weights of the predictors selected by 
\code{v.special}. Defaults to \code{0}.}

\item{spec.search.treated}{A logical scalar. If \code{TRUE}, a specification 
search (for the optimal set of included predictors) is done for the treated unit. Defaults to \code{FALSE}.}

\item{spec.search.placebos}{A logical scalar. If \code{TRUE}, a specification 
search (for the optimal set of included predictors) is done for the control unit. Defaults to \code{FALSE}.}
}
\value{
An object of class \code{"mscmt"}, which is essentially a list
containing the results of the estimation and, if applicable, the placebo
study.
The most important list elements are 
\itemize{
\item the weight vector \code{w} for the control units,
\item a matrix \code{v} with weight vectors for the predictors in its 
columns,
\item scalars \code{loss.v} and \code{rmspe} with the dependent loss and its 
square root,
\item a vector \code{loss.w} with the predictor losses corresponding to the
various weight vectors in the columns of \code{v},
\item a matrix \code{predictor.table} containing aggregated statistics of
predictor values (similar to list element \code{tab.pred} of 
function \code{\link[Synth]{synth.tab}} of package \code{'Synth'}),
\item a list of multivariate time series \code{combined} containing, 
for each dependent and predictor variable, a multivariate time series 
with elements \code{treated} for the actual values of the treated unit,
\code{synth} for the synthesized values, and \code{gaps} for the differences.
}
Placebo studies produce a list containing individual results for each 
unit (as treated unit), starting with the original treated unit, as well
as a list element named \code{placebo} with aggregated results for each
dependent and predictor variable.

If \code{times.pred.training} and \code{times.dep.validation} are not 
\code{NULL}, a cross-validation is done and a list of elements \code{cv}
with the results of the cross-validation period and \code{main} with
the results of the main period is returned.
}
\description{
\code{mscmt} performs the Multivariate Synthetic Control Method Using Time 
Series.
}
\details{
\code{mscmt} combines, if necessary, the preparation of the raw data (which 
is expected to be in "list" format, possibly after conversion from a 
\code{\link[base]{data.frame}}
with function \code{\link{listFromLong}}) and the call to the appropriate
MSCMT optimization procedures (depending on the input parameters).
For details on the input parameters \code{alpha}, \code{beta}, and 
\code{gamma}, see [1]. For details on cross-validation, see [2].
}
\examples{
\dontrun{
## for examples, see the package vignettes:
browseVignettes(package="MSCMT")
}
}
\references{
[1] \insertRef{FastReliable}{MSCMT} 

[2] \insertRef{CV}{MSCMT}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pvalue.r
\name{pvalue}
\alias{pvalue}
\title{P-values for placebo studies}
\usage{
pvalue(
  x,
  what,
  range.pre,
  range.post,
  alternative = c("two.sided", "less", "greater"),
  exclude.ratio = Inf,
  ratio.type = c("rmspe", "mspe")
)
}
\arguments{
\item{x}{An object of class \code{"mscmt"}, usually obtained as
the result of a call to function \code{\link{mscmt}}.}

\item{what}{A character vector. Name of the variable to be considered. If 
missing, the (first) dependent variable will be used.}

\item{range.pre}{A vector of length 2 defining the range of the pre-treatment
period with start and end time given as 
\itemize{
\item annual dates, if the format of start/end time is "dddd", e.g. "2016",
\item quarterly dates, if the format of start/end time is "ddddQd", e.g. 
"2016Q1",
\item monthly dates, if the format of start/end time is "dddd?dd" with "?"
different from "W" (see below), e.g. "2016/03" or "2016-10",
\item weekly dates, if the format of start/end time is "ddddWdd", e.g. 
"2016W23",
\item daily dates, if the format of start/end time is "dddd-dd-dd", e.g. 
"2016-08-18",
}
corresponding to the format of the respective column of the \code{times.dep}
argument of \code{\link{mscmt}}.
If missing, the corresponding column of \code{times.dep} will be used.}

\item{range.post}{A vector of length 2 defining the range of the 
post-treatment period with start and end time given as 
\itemize{
\item annual dates, if the format of start/end time is "dddd", e.g. "2016",
\item quarterly dates, if the format of start/end time is "ddddQd", e.g. 
"2016Q1",
\item monthly dates, if the format of start/end time is "dddd?dd" with "?"
different from "W" (see below), e.g. "2016/03" or "2016-10",
\item weekly dates, if the format of start/end time is "ddddWdd", e.g. 
"2016W23",
\item daily dates, if the format of start/end time is "dddd-dd-dd", e.g. 
"2016-08-18",
}
corresponding to the format of the respective column of the \code{times.dep}
argument of \code{\link{mscmt}}. Will be guessed if missing.}

\item{alternative}{A character string giving the alternative of the test. 
Either \code{"two.sided"} (default), \code{"less"}, or \code{"greater"}.}

\item{exclude.ratio}{A numerical scalar (default: \code{Inf}). Control units
with a pre-treatment-(r)mspe of more than \code{exclude.ratio} times the
pre-treatment-(r)mspe of the treated unit are excluded from the calculations
of the p-value.}

\item{ratio.type}{A character string. Either \code{rmspe} (default) or 
\code{mspe}. 
Selects whether root mean squared errors or mean squared errors are 
calculated.}
}
\value{
A time series containing the p-values for the 
post-treatment periods.
}
\description{
\code{pvalue} calculates p-values for placebo studies.
}
\details{
\code{pvalue} calculates p-values for placebo studies based
on Synthetic Control Methods.
}
\examples{
\dontrun{
## for an example, see the main package vignette:
 vignette("WorkingWithMSCMT",package="MSCMT")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MSCMT-package.r
\docType{package}
\name{MSCMT}
\alias{MSCMT}
\alias{MSCMT-package}
\title{Multivariate Synthetic Control Method Using Time Series}
\description{
MSCMT implements the Multivariate Synthetic Control Method Using
Time Series.
}
\details{
MSCMT implements three generalizations of the synthetic control method (which 
has already an implementation in package 'Synth'): \enumerate{
\item it allows for using multiple outcome variables, 
\item time series can be supplied as economic predictors,
\item a well-defined cross-validation approach can be used.
} 
Much effort has been taken to make the implementation as stable as possible 
(including edge cases) without losing computational efficiency.
}
\examples{
\dontrun{
## for examples, see the package vignettes:
browseVignettes(package="MSCMT")
}
}
\references{
\insertRef{Abadie2003}{MSCMT}

\insertRef{Abadie2010}{MSCMT}

\insertRef{FastReliable}{MSCMT}

\insertRef{CV}{MSCMT}

\insertRef{KP16}{MSCMT}
}
\author{
\strong{Maintainer}: Martin Becker \email{martin.becker@mx.uni-saarland.de} (\href{https://orcid.org/0000-0003-2336-9751}{ORCID})

Authors:
\itemize{
  \item Stefan Klößner \email{S.Kloessner@mx.uni-saarland.de}
}

Other contributors:
\itemize{
  \item Karline Soetaert \email{karline.soetaert@nioz.nl} [compiler]
  \item Jack Dongarra \email{dongarra@icl.utk.edu} [copyright holder]
  \item R.J. Hanson [copyright holder]
  \item K.H. Haskell [copyright holder]
  \item Cleve Moler [copyright holder]
  \item LAPACK authors [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/listFromLong.r
\name{listFromLong}
\alias{listFromLong}
\title{Convert Long Format to List Format}
\usage{
listFromLong(
  foo,
  unit.variable,
  time.variable,
  unit.names.variable = NULL,
  exclude.columns = NULL
)
}
\arguments{
\item{foo}{A \code{data.frame} containing the data in "long" format.}

\item{unit.variable}{Either a numeric scalar with the column number (in 
\code{foo}) containing the units or a character scalar with the corresponding
column name in \code{foo}.}

\item{time.variable}{Either a numeric scalar with the column number (in 
\code{foo}) containing the times or a character scalar with the corresponding
column name in \code{foo}.}

\item{unit.names.variable}{Optional. If not \code{NULL}, either a numeric 
scalar with the column number (in \code{foo}) containing the unit names or 
a character scalar with  the corresponding column name in \code{foo}. Must
match with the units defined by \code{unit.variable} (if not \code{NULL}).}

\item{exclude.columns}{Optional (defaults to \code{NULL}). Numeric vector 
with column numbers of \code{foo} to be excluded from the conversion.}
}
\value{
A list of matrices with rows corresponding to the times and columns
corresponding to the unit (or unit names, respectively) for all columns of
\code{foo} which are neither excluded nor have a special role as time, unit,
or unit names variable.
}
\description{
\code{listFromLong} converts long to list format.
}
\details{
\code{listFromLong} is a convenience function to convert long format
(in a \code{\link[base]{data.frame}}, as used by package 'Synth') to list 
format, where data is stored as a list of matrices.

Most parameter names are named after their equivalents in the 
\code{\link[Synth]{dataprep}} function of package 'Synth'.
}
\examples{
if (require("Synth")) {
  data(basque)
  Basque <- listFromLong(basque, unit.variable="regionno", 
                         time.variable="year", 
                         unit.names.variable="regionname")
  names(Basque)
  head(Basque$gdpcap)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.r
\name{compare}
\alias{compare}
\title{Compare MSCMT estimation results}
\usage{
compare(..., auto.name.prefix = "")
}
\arguments{
\item{...}{Objects of class \code{"mscmt"} or (a) list(s) containing objects
of class \code{"mscmt"}.}

\item{auto.name.prefix}{A character string (default: "") internally used to 
facilitate automatic naming in nested lists of unnamed estimation results.}
}
\value{
An object of class \code{"mscmt"}, which itself contains the 
individual estimation results as well as a component \code{"comparison"} 
with aggregated information.
}
\description{
\code{compare} collects estimation results from \code{\link{mscmt}} for
comparison purposes.
}
\details{
\code{compare} collects (potentially many) estimation results from 
\code{\link{mscmt}} in a special object of class \code{"mscmt"}, which 
includes a component \code{"comparison"} where the different estimation 
results are aggregated.
This aggregated information is used by the \code{\link{ggplot.mscmt}} and 
\code{\link{print.mscmt}} methods to present summaries of the different
results.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.r
\name{print.mscmt}
\alias{print.mscmt}
\title{Printing Results of MSCMT}
\usage{
\method{print}{mscmt}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"mscmt"}, usually obtained as
the result of a call to function \code{\link{mscmt}}.}

\item{...}{Further arguments to be passed to or from other methods. 
They are ignored in this function.}
}
\value{
Nothing useful (function is called for its side effects).
}
\description{
\code{print.mscmt} prints results of \code{mscmt}.
}
\details{
A human-readable summary of \code{mscmt}'s results.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/improveSynth.r
\name{improveSynth}
\alias{improveSynth}
\title{Check (and Improve) Results of Package Synth}
\usage{
improveSynth(
  synth.out,
  dataprep.out,
  lb = 1e-08,
  tol = 1e-05,
  verbose = TRUE,
  seed = 1,
  ...
)
}
\arguments{
\item{synth.out}{A result of \code{\link[Synth]{synth}} from package 
\code{'Synth'}.}

\item{dataprep.out}{The input of function \code{\link[Synth]{synth}} which
led to \code{synth.out}.}

\item{lb}{A numerical scalar (default: \code{1e-8}), corresponding to the
lower bound for the outer optimization.}

\item{tol}{A numerical scalar (default: \code{1e-5}). If the relative *and*
absolute improvement of \code{loss.v} and \code{loss.w}, respectively,
exceed \code{tol}, this is reported 
(if \code{verbose} is \code{TRUE}). Better optima are always looked for 
(independent of the value of \code{tol}).}

\item{verbose}{A logical scalar. Should the ouput be verbose (defaults to 
\code{TRUE}).}

\item{seed}{A numerical vector or \code{NULL}. See the corresponding
documentation for \code{\link[MSCMT]{mscmt}}. Defaults to 1 in order to
provide reproducibility of the results.}

\item{...}{Further arguments to \code{\link[MSCMT]{mscmt}}. Supported 
arguments are \code{check.global}, \code{inner.optim}, \code{inner.opar}, 
\code{outer.optim}, and \code{outer.opar}.}
}
\value{
An updated version of \code{synth.out}, where \code{solution.v},
\code{solution.w}, \code{loss.v}, and \code{loss.w} are replaced by the
optimum obtained by package \code{'MSCMT'} and all other components 
of \code{synth.out} are removed.
}
\description{
\code{improveSynth} checks the results of \code{\link[Synth]{synth}}
for feasibility and optimality and tries to find a better solution.
}
\details{
Performing SCM means solving a nested optimization problem. Depending on
the validity of the results of the inner optimization, SCM may produce
\itemize{
\item invalid or infeasible results, if the vector \code{w} of donor 
weights reported as the result of the inner optimization 
is in fact not optimal, ie. produces too large \code{loss.w},
\item suboptimal results, if the vector \code{v} of predictor weights 
reported as the result of the outer optimization is in fact not
optimal (which may be caused by shortcomings of the inner optimization).
}

\code{improveSynth} first checks \code{synth.out} for feasibility and
then tries to find a feasible and optimal solution by applying the
optimization methods of package \code{MSCMT} to \code{dataprep.out} 
(with default settings, more flexibility will probably be added in a 
future release).
}
\examples{
\dontrun{
## check whether package 'Synth' is available
if (require("Synth")) {

## process first example of function "synth" in package 'Synth' 
## (comments are removed):

  data(synth.data)
  dataprep.out<-
    dataprep(
     foo = synth.data,
     predictors = c("X1", "X2", "X3"),
     predictors.op = "mean",
     dependent = "Y",
     unit.variable = "unit.num",
     time.variable = "year",
     special.predictors = list(
        list("Y", 1991, "mean"),
        list("Y", 1985, "mean"),
        list("Y", 1980, "mean")
                              ),
     treatment.identifier = 7,
     controls.identifier = c(29, 2, 13, 17, 32, 38),
     time.predictors.prior = c(1984:1989),
     time.optimize.ssr = c(1984:1990),
     unit.names.variable = "name",
     time.plot = 1984:1996
     )

  synth.out <- synth(dataprep.out)

## check and (try to) improve these results:
  synth2.out <- improveSynth(synth.out,dataprep.out)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggplot.r
\name{ggplot.mscmt}
\alias{ggplot.mscmt}
\title{Plotting Results of mscmt with ggplot2}
\usage{
\method{ggplot}{mscmt}(
  data,
  mapping = aes(),
  what,
  type = c("gaps", "comparison", "placebo.gaps", "placebo.data", "p.value"),
  treatment.time,
  zero.line = TRUE,
  ylab,
  xlab = "Date",
  main,
  col,
  lty,
  lwd,
  legend = TRUE,
  bw = FALSE,
  date.format,
  unit.name,
  full.legend = TRUE,
  include.smooth = FALSE,
  include.mean = FALSE,
  include.synth = FALSE,
  draw.estwindow = TRUE,
  what.set,
  limits = NULL,
  alpha = 1,
  alpha.min = 0.1,
  exclude.units = NULL,
  exclude.ratio = Inf,
  ratio.type = c("rmspe", "mspe"),
  alternative = c("two.sided", "less", "greater"),
  draw.points = TRUE,
  control.name = "control units",
  size = 1,
  treated.name = "treated unit",
  labels = c("actual data", "synthsized data"),
  ...,
  environment = parent.frame()
)
}
\arguments{
\item{data}{An object of class \code{"mscmt"}, usually obtained as
the result of a call to function \code{\link{mscmt}}.}

\item{mapping}{An object necessary to match the definition of the
\code{ggplot} generic (passed to \code{ggplot} as is). Defaults to 
\code{aes()}.}

\item{what}{A character vector. Name(s) of the variables to be plotted. If 
missing, the (first) dependent variable will be used.}

\item{type}{A character scalar denoting the type of the plot containing 
either \code{"gaps"}, \code{"comparison"}, \code{"placebo.gaps"}, 
\code{"placebo.data"}, or \code{"p.value"}. 
Partial matching allowed, defaults to \code{"placebo.gaps"}, if results
of a placebo study are present, and to \code{"gaps"}, else.}

\item{treatment.time}{An optional scalar (numeric, character, or 
\code{\link[base]{Date}}) giving the treatment time. 
If \code{treatment.time} is numeric, Jan 01 of that year will be used. If 
\code{treatment.time} is a character string, it will be converted to a 
\code{\link[base]{Date}} and must thus be in an unambiguous format.
A vertical dotted line at the given point in time is included in the plot.}

\item{zero.line}{A logical scalar. If \code{TRUE} (default), a horizontal
dotted line (at zero level) is plotted for \code{"gaps"} and 
\code{"placebo.gaps"} plots.}

\item{ylab}{Optional label for the y-axis, automatically generated if 
missing.}

\item{xlab}{Optional label for the x-axis, defaults to \code{"Date"}.}

\item{main}{Optional main title for the plot, automatically generated 
if missing.}

\item{col}{Optional character vector with length 1 (for gaps plots) or
2 (for all other plot types). For comparison plots, \code{col} contains the
colours for the actual and synthesized data, for placebo.plots (with 
\code{full.legend==FALSE}), \code{col} contains the colours for 
the treated unit and the control units. Automatically generated 
if missing.}

\item{lty}{Optional numerical vector with length 1 (for gaps plots) or
2 (for all other plot types). For comparison plots, \code{lty} contains the
linetypes for the actual and synthesized data, for placebo.plots (with 
\code{full.legend==FALSE}), \code{col} contains the linetypes for 
the treated unit and the control units. Automatically generated 
if missing.}

\item{lwd}{Optional numerical vector with length 1 (for gaps plots) or
2 (for all other plot types). For comparison plots, \code{lty} contains the
linewidths for the actual and synthesized data, for placebo.plots (with 
\code{full.legend==FALSE}), \code{col} contains the linewidths for 
the treated unit and the control units. Automatically generated 
if missing.}

\item{legend}{A logical scalar. If \code{TRUE} (default), a legend is 
included in the plot.}

\item{bw}{A logical scalar. If \code{FALSE} (default), the automatically 
generated colours and line types are optimized for a colour plot, 
if \code{TRUE}, the automatic colours and line types are set for a black and
white plot.}

\item{date.format}{A character string giving the format for the tick labels
of the x axis as documented in \code{\link[base]{strptime}}. Defaults to
\code{"\%b \%y"} or \code{"\%Y"}, depending on the granularity of the data.}

\item{unit.name}{A character string with the title of the legend 
for comparison and placebo plots. Defaults to "Estimation" for comparison 
and "Unit" for placebo plots.}

\item{full.legend}{A logical scalar. If \code{TRUE} (default), a full legend
of all units (donors) is constructed. If \code{FALSE}, only the treated and
the control units are distinguished.}

\item{include.smooth}{A logical scalar. If \code{TRUE}, a geometric smoother
based on the control units is added to placebo plots. Default: \code{FALSE}.}

\item{include.mean}{A logical scalar. If \code{TRUE}, the arithmetic mean
of all control units is added to placebo plots. Default: \code{FALSE}.}

\item{include.synth}{A logical scalar. If \code{TRUE}, the synthesized data
for the treated unit are added to plots of type \code{"placebo.data"}.
Defaults to \code{FALSE}.}

\item{draw.estwindow}{A logical scalar. If \code{TRUE} (default), the time
range containing all optimization periods is shaded in the corresponding
plots.}

\item{what.set}{An optional character string for a convenient selection of 
multiple variables. Accepted values are \code{"dependents"}, 
\code{"predictors"}, and \code{"all"}, which collects all dependent, all
predictor, or all variables of both types, respectively. Overrides parameter
\code{what} (if the latter is present).}

\item{limits}{An optional vector of length 2 giving the range of the plot or
\code{NULL}.
If \code{limits} is numeric, Jan 01 of the corresponding years will be used.
If \code{limits} is of type character, both strings will be converted to 
Dates (via \code{\link[base]{as.Date}}) and must thus be in an unambiguous 
format.}

\item{alpha}{Either a numerical scalar, a numerical vector of length 
corresponding to the number of units, or the character string \code{"auto"}.
If \code{alpha} is a numerical scalar (default with value \code{1}), a fixed
value for the alpha channel (transparency) is included for all units in 
placebo plots. If \code{alpha} is numeric and has length corresponding to the
number of units, these values are assigned as alpha channel to the individual 
units. If \code{"auto"}, the alpha channel information is obtained from the
w weights of the control units.}

\item{alpha.min}{A numerical scalar (default: \code{0.1}). If \code{alpha} is
set to \code{"auto"}, the individual alpha channel information for control 
unit \code{i} is set to \code{alpha.min + (1-alpha.min) * w[i]}.}

\item{exclude.units}{An optional (default: \code{NULL}) character vector with 
names for control units which shall be excluded from placebo plots and
p-value calculations.}

\item{exclude.ratio}{A numeric scalar (default: \code{Inf}). Control units 
with a pre-treatment (r)mspe of more than \code{exclude.ratio} times
the pre-treatment (r)mspe of the treated unit are excluded from placebo 
plots and p-value calculations.}

\item{ratio.type}{A character string. Either \code{rmspe} (default) or 
\code{mspe}. Selects whether root mean squared errors or mean squared errors 
are considered for the exclusion of control units (see \code{exclude.ratio}).}

\item{alternative}{A character string giving the alternative of the test for 
plots of type \code{"p.value"}. Either \code{"two.sided"} (default), 
\code{"less"}, or \code{"greater"}.}

\item{draw.points}{A logical scalar. If \code{TRUE} (default), points are
added to the line plots to enhance visibility.}

\item{control.name}{A character string for the naming of the non-treated
units in placebo plots. Defaults to \code{"control units"}.}

\item{size}{A numerical scalar (default: \code{1}). If \code{draw.points} is
\code{TRUE} (default), \code{size} specifies the size of the points.}

\item{treated.name}{A character string giving the label for the treated
unit. Defaults to \code{"treated unit"}.}

\item{labels}{A character vector of length 2 giving the labels for the actual 
and synthesized data. Defaults to \code{c("actual data","synthsized data")}.}

\item{...}{Necessary to match the definition of the
\code{"ggplot"} generic (passed to \code{ggplot} as is).}

\item{environment}{An object necessary to match the definition of the
\code{"ggplot"} generic (passed to \code{ggplot} as is). Defaults to 
\code{parent.frame()}.}
}
\value{
An object of class \code{\link[ggplot2]{ggplot}}.
}
\description{
\code{ggplot.mscmt} plots results of \code{\link{mscmt}} based on 
\code{\link[ggplot2]{ggplot}}.
}
\details{
A unified plot method for gaps plots, comparison of treated and synthetic
values, as well as plots for placebo studies, based on 
\code{\link[ggplot2]{ggplot}}. \code{\link{ggplot.mscmt}} is the preferred
plot method and has more functionality than \code{\link{plot.mscmt}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppratio.r
\name{ppratio}
\alias{ppratio}
\title{Post-pre-(r)mspe-ratios for placebo studies}
\usage{
ppratio(
  x,
  what,
  range.pre,
  range.post,
  type = c("rmspe", "mspe"),
  return.all = FALSE
)
}
\arguments{
\item{x}{An object of class \code{"mscmt"}, usually obtained as
the result of a call to function \code{\link{mscmt}}.}

\item{what}{A character vector. Name of the variable to be considered. If 
missing, the (first) dependent variable will be used.}

\item{range.pre}{A vector of length 2 defining the range of the pre-treatment
period with start and end time given as 
\itemize{
\item annual dates, if the format of start/end time is "dddd", e.g. "2016",
\item quarterly dates, if the format of start/end time is "ddddQd", e.g. 
"2016Q1",
\item monthly dates, if the format of start/end time is "dddd?dd" with "?"
different from "W" (see below), e.g. "2016/03" or "2016-10",
\item weekly dates, if the format of start/end time is "ddddWdd", e.g. 
"2016W23",
\item daily dates, if the format of start/end time is "dddd-dd-dd", e.g. 
"2016-08-18",
}
corresponding to the format of the respective column of the \code{times.dep}
argument of \code{\link{mscmt}}.
If missing, the corresponding column of \code{times.dep} will be used.}

\item{range.post}{A vector of length 2 defining the range of the 
post-treatment period with start and end time given as 
\itemize{
\item annual dates, if the format of start/end time is "dddd", e.g. "2016",
\item quarterly dates, if the format of start/end time is "ddddQd", e.g. 
"2016Q1",
\item monthly dates, if the format of start/end time is "dddd?dd" with "?"
different from "W" (see below), e.g. "2016/03" or "2016-10",
\item weekly dates, if the format of start/end time is "ddddWdd", e.g. 
"2016W23",
\item daily dates, if the format of start/end time is "dddd-dd-dd", e.g. 
"2016-08-18",
}
corresponding to the format of the respective column of the \code{times.dep}
argument of \code{\link{mscmt}}. Will be guessed if missing.}

\item{type}{A character string. Either \code{rmspe} (default) or \code{mspe}.
Selects whether root mean squared errors or mean squared errors are 
calculated.}

\item{return.all}{A logical scalar. If \code{FALSE} (default), only the
(named) vector of post-pre-(r)mspe-ratios is returned, if \code{TRUE},
a three-column matrix with pre- and post-treatment (r)mspe's as well as the 
post-pre-ratios will be returned.}
}
\value{
If \code{return.all} is \code{FALSE}, a (named) vector of 
post-pre-(r)mspe-ratios. If \code{return.all} is \code{TRUE}, a matrix with
three columns containing the pre-treatment (r)mspe, the post-treatment 
(r)mspe, and the post-pre-ratio.
}
\description{
\code{ppratio} calculates post-to-pre-(r)mspe-ratios for placebo studies.
}
\details{
\code{ppratio} calculates post-to-pre-(r)mspe-ratios for placebo studies based
on Synthetic Control Methods.
}
